Require Export InfrastructureOperations.
Require Export InfrastructureLemmasTerms.
Require Export DeclarationTyping.
Set Implicit Arguments.

(******************************************************************************)
(* Context lookups.                                                           *)
(******************************************************************************)

Lemma lookup_etvar_functional {Γ X K K'} :
  lookup_etvar Γ X K → lookup_etvar Γ X K' → K = K'.
Proof.
  intros gt gt'; revert K' gt'.
  induction gt; intros; dependent destruction gt'; simpl; f_equal; auto.
Qed.

(******************************************************************************)
(* Context insertions.                                                        *)
(******************************************************************************)

Inductive shift_etvar : nat → Env → Env → Prop :=
  | it_here {Γ K} :
      shift_etvar 0 Γ (etvar Γ K)
  | it_var {Γ1 Γ2 c T} :
      shift_etvar c Γ1 Γ2 →
      shift_etvar c (evar Γ1 T) (evar Γ2 (tshiftTy c T))
  | it_tvar {Γ1 Γ2 c K} :
      shift_etvar c Γ1 Γ2 →
      shift_etvar (S c) (etvar Γ1 K) (etvar Γ2 K).
Hint Constructors shift_etvar.

Inductive shift_evar : nat → Env → Env → Prop :=
  | iv_here {Γ T} :
      shift_evar 0 Γ (evar Γ T)
  | iv_var {Γ1 Γ2 : Env} {c T} :
      shift_evar c Γ1 Γ2 →
      shift_evar (S c) (evar Γ1 T) (evar Γ2 T)
  | iv_tvar {Γ1 Γ2 c K} :
      shift_evar c Γ1 Γ2 →
      shift_evar c (etvar Γ1 K) (etvar Γ2 K).
Hint Constructors shift_evar.

(******************************************************************************)
(* Lemmas about shifting and context lookups.                                 *)
(******************************************************************************)

Lemma shift_etvar_lookup_etvar {Γ1 Γ2 c} (ib : shift_etvar c Γ1 Γ2) :
  ∀ X K, lookup_etvar Γ1 X K → lookup_etvar Γ2 (shiftIndex c X) K.
Proof.
  induction ib; inversion 1; subst; simpl; try rewrite tshiftTy_comm; eauto.
Qed.
Hint Resolve shift_etvar_lookup_etvar.

Lemma shift_etvar_lookup_evar {Γ1 Γ2 c} (ib : shift_etvar c Γ1 Γ2) :
  ∀ x T, lookup_evar Γ1 x T → lookup_evar Γ2 x (tshiftTy c T).
Proof.
  induction ib; inversion 1; subst; simpl; try rewrite tshiftTy_comm; eauto.
Qed.
Hint Resolve shift_etvar_lookup_evar.

Lemma shift_evar_lookup_etvar {Γ1 Γ2 c} (iv : shift_evar c Γ1 Γ2) :
  ∀ X K, lookup_etvar Γ1 X K → lookup_etvar Γ2 X K.
Proof.
  induction iv; inversion 1; subst; simpl; eauto.
Qed.
Hint Resolve shift_evar_lookup_etvar.

Lemma shift_evar_lookup_evar {Γ1 Γ2 c} (iv : shift_evar c Γ1 Γ2) :
  ∀ x T, lookup_evar Γ1 x T → lookup_evar Γ2 (shiftIndex c x) T.
Proof.
  induction iv; inversion 1; subst; simpl; eauto.
Qed.
Hint Resolve shift_evar_lookup_evar.

(******************************************************************************)
(* Strengthening lemmas                                                       *)
(******************************************************************************)

Lemma shiftIndex_inj {c} :
  ∀ X Y, shiftIndex c X = shiftIndex c Y → X = Y.
Proof.
  induction c; simpl; intros.
  dependent destruction H; auto.
  destruct X, Y; simpl; try discriminate; auto.
Qed.

Lemma tshift_inj {c T U} :
  tshiftTy c T = tshiftTy c U → T = U.
Proof.
  revert c U; induction T; simpl; intros c U eq; destruct U; simpl in *;
    try discriminate; inversion eq; f_equal; eauto using shiftIndex_inj.
Qed.

Lemma remove_var_lookup_etvar {Γ1 Γ2 c} (iv : shift_evar c Γ2 Γ1) :
  ∀ X K, lookup_etvar Γ1 X K → lookup_etvar Γ2 X K.
Proof.
  induction iv; inversion 1; subst; simpl; eauto.
Qed.

(******************************************************************************)
(* Context etvar substitution                                                 *)
(******************************************************************************)

Inductive subst_etvar : nat → Ty → Env → Env → Prop :=
  | subst_etvar_here {Γ T K} :
      Kinding Γ T K →
      subst_etvar 0 T (etvar Γ K) Γ
  | subst_etvar_there_evar {Γ1 Γ2 T T' X} :
      subst_etvar X T Γ1 Γ2 →
      subst_etvar X T (evar Γ1 T') (evar Γ2 (tsubstTy X T T'))
  | subst_etvar_there_etvar {Γ1 Γ2 X T K} :
      subst_etvar X T Γ1 Γ2 →
      subst_etvar (S X) T (etvar Γ1 K) (etvar Γ2 K).
Hint Constructors subst_etvar.

Lemma subst_etvar_lookup_evar {X T Γ1 Γ2} (esub: subst_etvar X T Γ1 Γ2) :
  ∀ {x T'}, lookup_evar Γ1 x T' → lookup_evar Γ2 x (tsubstTy X T T').
Proof.
  induction esub; inversion 1; subst; simpl;
    try rewrite tsubstTy_tshiftTy_reflection;
    try rewrite tsubstTy_tshiftTy_comm; eauto.
Qed.
Hint Resolve subst_etvar_lookup_evar.

(******************************************************************************)
(* Context evar substitution                                                  *)
(******************************************************************************)

Inductive subst_evar : Trace → Tm → Env → Env → Prop :=
  | subst_evar_here {Γ t T} :
      Typing Γ t T →
      subst_evar I0 t (evar Γ T) Γ
  | subst_evar_there_evar {Γ1 Γ2 T t x} :
      subst_evar x t Γ1 Γ2 →
      subst_evar (IV x) t (evar Γ1 T) (evar Γ2 T)
  | subst_evar_there_etvar {Γ1 Γ2 K x t} :
      subst_evar x t Γ1 Γ2 →
      subst_evar (IT x) t (etvar Γ1 K) (etvar Γ2 K).
Hint Constructors subst_evar.

Lemma subst_evar_lookup_etvar {x t Γ1 Γ2} (esub: subst_evar x t Γ1 Γ2) :
  ∀ {X K}, lookup_etvar Γ1 X K → lookup_etvar Γ2 X K.
Proof.
  induction esub; inversion 1; subst; simpl; eauto.
Qed.
Hint Resolve subst_evar_lookup_etvar.
