Require Export DeclarationEvaluation.
Require Export DeclarationTyping.
Require Export InfrastructureLemmasTerms.
Require Export InfrastructureLemmasContext.
Set Implicit Arguments.

(******************************************************************************)
(* Weakening lemmas                                                           *)
(******************************************************************************)

Lemma DTyping_bind_lengthExt {Γ ds Δ} (wd : DTyping Γ ds Δ) :
  bind ds = lengthExt Δ.
Proof.
  induction wd; simpl; auto.
  - rewrite IHwd, lengthExt_append; simpl.
    reflexivity.
Qed.

Lemma shift_etvar_typing {Γ1 t T} (wt : Typing Γ1 t T) :
  ∀ {c Γ2}, shift_etvar c Γ1 Γ2 → Typing Γ2 (tshiftTm c t) (tshiftTy c T).
Proof.
  induction wt using ind_typing with
  (P0 := fun Γ1 ds Δ wtd =>
           ∀ {c Γ2}, shift_etvar c Γ1 Γ2 →
             DTyping Γ2 (tshiftDecls c ds) (tshiftExt c Δ)); simpl; intros;
    try rewrite tshiftTy_tsubstTy_comm;
    try rewrite tshiftExt_append; econstructor;
    eauto using shift_etvar_wftyp, shift_etvar_lookup_evar, weakenshift_etvar.
Qed.

Lemma shift_evar_typing {Γ1 t T} (wt: Typing Γ1 t T) :
 ∀ {c Γ2}, shift_evar c Γ1 Γ2 → Typing Γ2 (shiftTm c t) T.
Proof.
  induction wt using ind_typing with
  (P0 := fun Γ1 ds Δ wtd =>
           ∀ {c Γ2}, shift_evar c Γ1 Γ2 →
             DTyping Γ2 (shiftDecls c ds) Δ); simpl; intros;
    econstructor; try erewrite DTyping_bind_lengthExt;
    eauto using shift_evar_lookup_evar, shift_evar_wftyp, weakenshift_evar.
Qed.

(******************************************************************************)
(* Substitution lemmas                                                        *)
(******************************************************************************)

Lemma subst_evar_lookup_evar {Γ1 Γ2 : Env} {x t y T} :
  subst_evar x t Γ1 Γ2 → lookup_evar Γ1 y T → Typing Γ2 (substIndex x t y) T.
Proof.
  intros esub gv; revert Γ2 x esub.
  induction gv; intros; dependent destruction esub; simpl;
    eauto using T_Var, shift_evar_typing, shift_etvar_typing.
Qed.

Lemma subst_evar_typing_dtyping_ind :
  ∀ Γ1,
  (∀ {t T} (wt : Typing Γ1 t T),
    ∀ {x t' Γ2}, subst_evar x t' Γ1 Γ2 →
                 Typing Γ2 (substTm x t' t) T) ∧
  (∀ {ds Δ} (wtd : DTyping Γ1 ds Δ),
    ∀ {x t' Γ2}, subst_evar x t' Γ1 Γ2 →
                 DTyping Γ2 (substDecls x t' ds) Δ).
Proof.
  apply ind_typing_dtyping; simpl; intros; eauto using subst_evar_lookup_evar;
    try rewrite (DTyping_bind_lengthExt d); econstructor;
    eauto using subst_evar_wftyp_ind, weaken_subst_evar.
Qed.

Lemma subst_evar_typing_ind {Γ1 t T} (wt : Typing Γ1 t T) :
  ∀ {x t' Γ2}, subst_evar x t' Γ1 Γ2 → Typing Γ2 (substTm x t' t) T.
Proof. apply subst_evar_typing_dtyping_ind; auto. Qed.

Lemma subst_evar_typing {Γ : Env} {s S t T} :
  Typing Γ s S → Typing (evar Γ S) t T → Typing Γ (substTm sub_here s t) T.
Proof.
  intros wt_s wt_t; apply (subst_evar_typing_ind wt_t).
  constructor; auto.
Qed.

Lemma subst_evar_dtyping {Γ : Env} {s S ds Δ} :
  Typing Γ s S → DTyping (evar Γ S) ds Δ →
  DTyping Γ (substDecls sub_here s ds) Δ.
Proof.
  intros wt_s wt_ds.
  apply (proj2 (subst_evar_typing_dtyping_ind _) _ _ wt_ds).
  constructor; auto.
Qed.

Lemma subst_etvar_typing_dtyping_ind :
  ∀ Γ1,
  (∀ {t T} (wt : Typing Γ1 t T),
    ∀ {X T' Γ2}, subst_etvar X T' Γ1 Γ2 →
                 Typing Γ2 (tsubstTm X T' t) (tsubstTy X T' T)) /\
  (∀ {ds Δ} (wtd : DTyping Γ1 ds Δ),
    ∀ {X T' Γ2}, subst_etvar X T' Γ1 Γ2 →
                 DTyping Γ2 (tsubstDecls X T' ds) (tsubstExt X T' Δ)).
Proof.
  apply ind_typing_dtyping; simpl; intros;
    try rewrite tsubstTy_tsubstTy_comm;
    try rewrite tsubstExt_append; econstructor;
    eauto using subst_etvar_lookup_evar, subst_etvar_wftyp_ind.
Qed.

Lemma subst_etvar_typing :
  (∀ {Γ1 t T} (wt : Typing Γ1 t T),
    ∀ {X T' Γ2}, subst_etvar X T' Γ1 Γ2 →
                 Typing Γ2 (tsubstTm X T' t) (tsubstTy X T' T)).
Proof.
  apply subst_etvar_typing_dtyping_ind.
Qed.

(******************************************************************************)
(* Progress                                                                   *)
(******************************************************************************)

Lemma can_form_tarr {t : Tm} {T1 T2} :
  Value t → Typing empty t (tarr T1 T2) →
  exists t' T1', t = abs T1' t'.
Proof.
  intros val_t wt_t.
  depind wt_t; try contradiction.
  exists t, T1; reflexivity.
Qed.

Lemma can_form_tall {t : Tm} {T} :
  Value t → Typing empty t (tall T) →
  exists t', t = tabs t'.
Proof.
  intros val_t wt_t.
  depind wt_t; try contradiction.
  exists t; auto.
Qed.

Lemma progress {Γ t U} (wt: Typing Γ t U) :
  Γ = empty → Value t ∨ ∃ t', red t t'.
Proof with destruct_conjs; subst; simpl; eauto using red, redds'.
  induction wt using ind_typing with
  (P0 := fun Γ ds Δ wds => Γ = empty → ValueH ds ∨ ∃ ds', redds' ds ds');
    intros...
  - inversion l.
  - destruct IHwt1 as [v1|[t1' r1]]...
    destruct IHwt2 as [v2|[t2' r2]]...
    destruct (can_form_tarr v1 wt1)...
  - destruct IHwt as [vt|[t1' r1]]...
    destruct (can_form_tall vt wt)...
  - destruct IHwt as [v1|[t1' r1]]...
    dependent destruction d...
  - destruct IHwt as [vt|[t1' r1]]...
Qed.

(******************************************************************************)
(* Preservation                                                               *)
(******************************************************************************)

Lemma preservation {Γ t U} (wt: Typing Γ t U) :
  ∀ {t'}, red t t' → Typing Γ t' U.
Proof.
  induction wt using ind_typing
  with (P0 := fun Γ ds Δ wds => ∀ ds', redds' ds ds' → DTyping Γ ds' Δ);
    intros t' rd; inversion rd; subst; eauto using Typing, DTyping.
  - dependent destruction wt1; eauto using subst_evar_typing.
  - dependent destruction wt; eauto using subst_etvar_typing.
  - dependent destruction d; auto.
  - dependent destruction d; eauto.
    rewrite (DTyping_bind_lengthExt d).
    rewrite extend_append in *.
    eapply T_Let; eauto using subst_evar_dtyping, subst_evar_typing_ind,
                  subst_evar, weaken_subst_evar.
Qed.
