Require Export DeclarationSyntax.
Require Export InfrastructureOperations.
Set Implicit Arguments.

(******************************************************************************)
(* Lemmas.                                                                    *)
(******************************************************************************)

Lemma tshift_preserves_size T : ∀ c, sizeTy (tshiftTy c T) = sizeTy T.
Proof.
  induction T using ind_Ty with
  (P0 := fun Tr => ∀ c, sizeTRec (tshiftTRec c Tr) = sizeTRec Tr);
    simpl; intuition.
Qed.

Lemma lengthExt_tshiftExt (c : nat) (Δ : Ext) :
  lengthExt (tshiftExt c Δ) = lengthExt Δ.
Proof.
  induction Δ; simpl; f_equal; auto.
Qed.

Lemma lengthExt_tsubstExt (X : nat) (T' : Ty) (Δ : Ext) :
  lengthExt (tsubstExt X T' Δ) = lengthExt Δ.
Proof.
  induction Δ; simpl; f_equal; auto.
Qed.

Ltac specialize_suc :=
  progress
    match goal with
      | [ H : forall _ : nat, _, k : nat |- _ ] => apply (H (S k))
    end.
Ltac var_case :=
  match goal with
    | [ |- forall v i, _ ] =>
      induction v; simpl; intros; destruct i; auto;
      try apply (f_equal (tshiftTy O)) with (y := tvar _); auto
  end.
Ltac ty_case v :=
  simpl; intros; first
     [ try apply f_equal with (f := tvar); apply v
     | apply f_equal2 with (f := tarr); trivial
     | apply f_equal2 with (f := tall); try specialize_suc; trivial
     | apply f_equal with (f := trec); trivial
     | apply f_equal3 with (f := tcons); try specialize_suc; trivial
     | trivial
     ].

Lemma shiftIndex_comm_ind :
  forall v i c,
    shiftIndex (v + S c) (shiftIndex (v + O) i) =
    shiftIndex (v + O  ) (shiftIndex (v + c) i).
Proof.
  var_case.
Qed.

Lemma tshiftTy_comm_ind (T : Ty) :
  forall v c,
    tshiftTy (v + S c) (tshiftTy (v + O) T) =
    tshiftTy (v + O  ) (tshiftTy (v + c) T).
Proof.
  induction T using ind_Ty with
  (P0 := fun Tr => forall v c,
                     tshiftTRec (v + S c) (tshiftTRec (v + O) Tr) =
                     tshiftTRec (v + O  ) (tshiftTRec (v + c) Tr));
    ty_case shiftIndex_comm_ind.
Qed.

Lemma tshiftTy_comm (c : nat) (T : Ty) :
    tshiftTy (S c) (tshiftTy O T)
  = tshiftTy O     (tshiftTy c T).
Proof.
  apply (tshiftTy_comm_ind _ 0).
Qed.

Lemma tsubstIndex_shiftIndex_reflection_ind :
  forall (k X : nat) (T : Ty),
    tsubstIndex (k + 0) T (shiftIndex (k + O) X) = tvar X.
Proof.
  var_case.
Qed.

Lemma tsubstTy_tshiftTy_reflection_ind (T : Ty) :
  forall k T',
    tsubstTy (k + 0) T' (tshiftTy (k + O) T) = T.
Proof.
  induction T using ind_Ty with
  (P0 := fun Tr => forall k T',
                     tsubstTRec (k + 0) T' (tshiftTRec (k + O) Tr) = Tr);
    ty_case tsubstIndex_shiftIndex_reflection_ind.
Qed.

Lemma tsubstTy_tshiftTy_reflection (T' T : Ty) :
  tsubstTy 0 T' (tshiftTy O T) = T.
Proof.
  apply (tsubstTy_tshiftTy_reflection_ind _ 0).
Qed.

Lemma tshiftTy_tsubstIndex_comm_ind :
  forall (k X c : nat) (T : Ty),
    tshiftTy (k + c) (tsubstIndex (k + 0) T X) =
    tsubstIndex (k + 0) (tshiftTy c T) (shiftIndex (k + S c) X).
Proof.
  var_case.
  rewrite tshiftTy_comm; f_equal; auto.
Qed.

Lemma tshiftTy_tsubstTy_comm_ind (T : Ty):
  forall (k c : nat) (T' : Ty),
    tshiftTy (k + c) (tsubstTy (k + 0) T' T) =
    tsubstTy (k + 0) (tshiftTy c T') (tshiftTy (k + S c) T).
Proof.
  induction T using ind_Ty with
  (P0 := fun Tr => forall (k c : nat) (T' : Ty),
    tshiftTRec (k + c) (tsubstTRec (k + 0) T' Tr) =
    tsubstTRec (k + 0) (tshiftTy c T') (tshiftTRec (k + S c) Tr));
    ty_case tshiftTy_tsubstIndex_comm_ind.
Qed.

Lemma tshiftTy_tsubstTy_comm c (T T2 : Ty) :
    tshiftTy c (tsubstTy O T T2)
  = tsubstTy O (tshiftTy c T) (tshiftTy (S c) T2).
Proof.
  apply (tshiftTy_tsubstTy_comm_ind _ 0).
Qed.

Lemma tsubstIndex_shiftIndex_comm_ind :
  forall k Y X T',
    tsubstIndex (k + S X) T' (shiftIndex (k + 0) Y) =
    tshiftTy (k + 0) (tsubstIndex (k + X) T' Y).
Proof.
  var_case.
  rewrite tshiftTy_comm; f_equal; auto.
Qed.

Lemma tsubstTy_tshiftTy_comm_ind (T : Ty) :
  forall k X T',
    tsubstTy (k + S X) T' (tshiftTy (k + 0) T) =
    tshiftTy (k + 0) (tsubstTy (k + X) T' T).
Proof.
  induction T using ind_Ty with
  (P0 := fun Tr => forall k X T',
    tsubstTRec (k + S X) T' (tshiftTRec (k + 0) Tr) =
    tshiftTRec (k + 0) (tsubstTRec (k + X) T' Tr));
    ty_case tsubstIndex_shiftIndex_comm_ind.
Qed.

Lemma tsubstTy_tshiftTy_comm :
  forall X T' T,
    tsubstTy (S X) T' (tshiftTy 0 T) =
    tshiftTy 0 (tsubstTy X T' T).
Proof.
  intros; apply (tsubstTy_tshiftTy_comm_ind _ 0).
Qed.

Lemma tsubstTy_tsubstIndex_comm_ind :
  forall k Y T1 T2 X,
    tsubstTy (k + X) T1 (tsubstIndex (k + 0) T2 Y) =
    tsubstTy (k + 0) (tsubstTy X T1 T2) (tsubstIndex (k + S X) T1 Y).
Proof.
  var_case.
  rewrite tsubstTy_tshiftTy_reflection; auto.
  repeat rewrite tsubstTy_tshiftTy_comm; f_equal; auto.
Qed.

Lemma tsubstTy_tsubstTy_comm_ind (T : Ty) :
  forall k X T1 T2,
    tsubstTy (k + X) T1 (tsubstTy (k + 0) T2 T) =
    tsubstTy (k + 0) (tsubstTy X T1 T2) (tsubstTy (k + (S X)) T1 T).
Proof.
  induction T using ind_Ty with
  (P0 := fun Tr => forall k X T1 T2,
    tsubstTRec (k + X) T1 (tsubstTRec (k + 0) T2 Tr) =
    tsubstTRec (k + 0) (tsubstTy X T1 T2) (tsubstTRec (k + (S X)) T1 Tr));
    ty_case tsubstTy_tsubstIndex_comm_ind.
Qed.

Lemma tsubstTy_tsubstTy_comm :
  forall X T1 T2 T,
    tsubstTy X T1 (tsubstTy 0 T2 T) =
    tsubstTy 0 (tsubstTy X T1 T2) (tsubstTy (S X) T1 T).
Proof.
  intros; apply (tsubstTy_tsubstTy_comm_ind _ 0).
Qed.
