Require Export DeclarationSyntax.
Set Implicit Arguments.

(******************************************************************************)
(* Shifting                                                                   *)
(******************************************************************************)

Fixpoint shiftIndex (c : nat) (i : nat) : nat :=
  match c with
    | O   => S i
    | S c =>
      match i with
        | O   => O
        | S i => S (shiftIndex c i)
      end
  end.

Fixpoint tshiftTRec (c : nat) (SS : TRec) {struct SS} : TRec :=
  match SS with
    | tnil => (tnil)
    | tcons l T TS => (tcons l (tshiftTy c T) (tshiftTRec c TS))
  end
with tshiftTy (c : nat) (S0 : Ty) {struct S0} : Ty :=
  match S0 with
    | tvar X      => tvar (shiftIndex c X)
    | top         => top
    | tarr T1 T2  => tarr (tshiftTy c T1) (tshiftTy c T2)
    | tall T1 T2  => tall (tshiftTy c T1) (tshiftTy (S c) T2)
    | trec TS    => (trec (tshiftTRec c TS))
  end.

Fixpoint tshiftPat (c : nat) (p : Pat) {struct p} : Pat :=
  match p with
    | pvar T  => (pvar (tshiftTy c T))
    | prec ps => (prec (tshiftPRec c ps))
  end
with tshiftPRec (c : nat) (ps : PRec) {struct ps} : PRec :=
  match ps with
    | pnil         => (pnil)
    | pcons l p ps => (pcons l (tshiftPat c p) (tshiftPRec c ps))
  end.

Fixpoint shiftTm (c : nat) (s : Tm) {struct s} : Tm :=
  match s with
    | var x        => (var (shiftIndex c x))
    | abs T t      => (abs T (shiftTm (S c) t))
    | app t1 t2    => (app (shiftTm c t1) (shiftTm c t2))
    | tabs T0 t0   => (tabs T0 (shiftTm c t0))
    | tapp t3 T1   => (tapp (shiftTm c t3) T1)
    | rec ts       => (rec (shiftRec c ts))
    | prj t4 l     => (prj (shiftTm c t4) l)
    | lett p t5 t6 => (lett p (shiftTm c t5) (shiftTm (bindPat p + c) t6))
  end
with shiftRec (c : nat) (ss : Rec) {struct ss} : Rec :=
  match ss with
    | rnil         => (rnil)
    | rcons l t ts => (rcons l (shiftTm c t) (shiftRec c ts))
  end.

Fixpoint tshiftTm (c : nat) (s : Tm) {struct s} : Tm :=
  match s with
    | var x        => (var x)
    | abs T t      => (abs (tshiftTy c T) (tshiftTm c t))
    | app t1 t2    => (app (tshiftTm c t1) (tshiftTm c t2))
    | tabs T0 t0   => (tabs (tshiftTy c T0) (tshiftTm (S c) t0))
    | tapp t3 T1   => (tapp (tshiftTm c t3) (tshiftTy c T1))
    | rec ts       => (rec (tshiftRec c ts))
    | prj t4 l     => (prj (tshiftTm c t4) l)
    | lett p t5 t6 => (lett (tshiftPat c p) (tshiftTm c t5) (tshiftTm c t6))
  end
with tshiftRec (c : nat) (ss : Rec) {struct ss} : Rec :=
  match ss with
    | rnil         => (rnil)
    | rcons l t ts => (rcons l (tshiftTm c t) (tshiftRec c ts))
  end.

Fixpoint weakenTm (t : Tm) (k : nat) : Tm :=
  match k with
    | O   => t
    | S k => shiftTm O (weakenTm t k)
  end.

Fixpoint weakenRec (tr : Rec) (k : nat) : Rec :=
  match k with
    | O   => tr
    | S k => shiftRec O (weakenRec tr k)
  end.

Fixpoint tshiftExt (c : nat) (Δ : Ext) : Ext :=
  match Δ with
    | exempty    => exempty
    | exvar Δ T  => exvar (tshiftExt c Δ) (tshiftTy c T)
  end.

(******************************************************************************)
(* Type substitution.                                                         *)
(******************************************************************************)

Fixpoint tsubstIndex (X : nat) (T' : Ty) (Y : nat) : Ty :=
  match X , Y with
    | O   , O   => T'
    | O   , S Y => tvar Y
    | S X , O   => tvar O
    | S X , S Y => tshiftTy 0 (tsubstIndex X T' Y)
  end.

Fixpoint tsubstTy (d : nat) (S0 : Ty) (S1 : Ty) {struct S1} : Ty :=
  match S1 with
    | tvar X => (tsubstIndex d S0 X)
    | top => (top)
    | tarr T1 T2 => (tarr (tsubstTy d S0 T1) (tsubstTy d S0 T2))
    | tall T0 T3 => (tall (tsubstTy d S0 T0) (tsubstTy (S d) S0 T3))
    | trec TS => (trec (tsubstTRec d S0 TS))
  end
with tsubstTRec (d : nat) (S0 : Ty) (SS : TRec) {struct SS} : TRec :=
  match SS with
    | tnil => (tnil)
    | tcons l T TS => (tcons l (tsubstTy d S0 T) (tsubstTRec d S0 TS))
  end.

Fixpoint tsubstPat (d : nat) (S0 : Ty) (p : Pat) {struct p} : Pat :=
  match p with
    | pvar T => (pvar (tsubstTy d S0 T))
    | prec ps => (prec (tsubstPRec d S0 ps))
  end
with tsubstPRec (d : nat) (S0 : Ty) (ps : PRec) {struct ps} : PRec :=
  match ps with
    | pnil => (pnil)
    | pcons l p ps0 => (pcons l (tsubstPat d S0 p) (tsubstPRec d S0 ps0))
  end.
Fixpoint tsubstTm (d : nat) (S0 : Ty) (s : Tm) {struct s} : Tm :=
  match s with
    | var x => (var x)
    | abs T t => (abs (tsubstTy d S0 T) (tsubstTm d S0 t))
    | app t1 t2 => (app (tsubstTm d S0 t1) (tsubstTm d S0 t2))
    | tabs T0 t0 => (tabs (tsubstTy d S0 T0) (tsubstTm (S d) S0 t0))
    | tapp t3 T1 => (tapp (tsubstTm d S0 t3) (tsubstTy d S0 T1))
    | rec ts => (rec (tsubstRec d S0 ts))
    | prj t4 l => (prj (tsubstTm d S0 t4) l)
    | lett p t5 t6 => (lett (tsubstPat d S0 p) (tsubstTm d S0 t5) (tsubstTm d S0 t6))
  end
with tsubstRec (d : nat) (S0 : Ty) (ss : Rec) {struct ss} : Rec :=
  match ss with
    | rnil => (rnil)
    | rcons l t ts => (rcons l (tsubstTm d S0 t) (tsubstRec d S0 ts))
  end.

Fixpoint tsubstExt (X : nat) (T' : Ty) (Δ : Ext) : Ext :=
  match Δ with
    | exempty    => exempty
    | exvar Δ T  => exvar (tsubstExt X T' Δ) (tsubstTy X T' T)
  end.

(******************************************************************************)
(* Term substitutions.                                                        *)
(******************************************************************************)

Inductive Subst : Set :=
  | sub_here  : Subst
  | sub_var   : Subst → Subst
  | sub_bound : Subst → Subst.

Fixpoint weaken_subst (sub : Subst) (k : nat) : Subst :=
  match k with
   | O   => sub
   | S k => sub_var (weaken_subst sub k)
  end.

Fixpoint substIndex (x : Subst) (t : Tm) (y : nat) : Tm :=
  match x , y with
    | sub_here    , O   => t
    | sub_here    , S y => var y
    | sub_var x   , O   => var O
    | sub_var x   , S y => shiftTm O (substIndex x t y)
    | sub_bound x , y   => tshiftTm O (substIndex x t y)
  end.

Fixpoint substTm (d : Subst) (s : Tm) (s0 : Tm) {struct s0} : Tm :=
  match s0 with
    | var x => (substIndex d s x)
    | abs T1 t2    => abs T1 (substTm (sub_var d) s t2)
    | app t1 t2 => (app (substTm d s t1) (substTm d s t2))
    | tabs T0 t0 => (tabs T0 (substTm (sub_bound d) s t0))
    | tapp t3 T1 => (tapp (substTm d s t3) T1)
    | rec ts => (rec (substRec d s ts))
    | prj t4 l => (prj (substTm d s t4) l)
    | lett p t5 t6 => (lett p (substTm d s t5) (substTm (weaken_subst d (bindPat p)) s t6))
  end
with substRec (d : Subst) (s : Tm) (ss : Rec) {struct ss} : Rec :=
  match ss with
    | rnil => (rnil)
    | rcons l t ts => (rcons l (substTm d s t) (substRec d s ts))
  end.

(******************************************************************************)
(* Context extension.                                                         *)
(******************************************************************************)

Fixpoint extend (Γ : Env) (Δ : Ext) : Env :=
  match Δ with
    | exempty     => Γ
    | exvar Δ T   => evar (extend Γ Δ) T
  end.

Fixpoint append (Δ1 Δ2 : Ext) : Ext :=
  match Δ2 with
    | exempty    => Δ1
    | exvar Δ2 T => exvar (append Δ1 Δ2) T
  end.

Fixpoint lengthExt (Δ : Ext) : nat :=
  match Δ with
    | exempty   => 0
    | exvar Δ _ => 1 + lengthExt Δ
  end.

(******************************************************************************)
(* Context lookups.                                                           *)
(******************************************************************************)

Inductive lookup_etvar : Env → nat → Ty → Prop :=
  | gb_here {Γ T} :
      lookup_etvar (etvar Γ T) O (tshiftTy O T)
  | gb_var {Γ T T' X} :
      lookup_etvar Γ X T →
      lookup_etvar (evar Γ T') X T
  | gb_bound {Γ T T' X} :
      lookup_etvar Γ X T →
      lookup_etvar (etvar Γ T') (S X) (tshiftTy O T).
Hint Constructors lookup_etvar.

Inductive lookup_evar : Env → nat → Ty → Prop :=
  | gv_here {Γ T} :
      lookup_evar (evar Γ T) O T
  | gv_var {Γ T T' X} :
      lookup_evar Γ X T →
      lookup_evar (evar Γ T') (S X) T
  | gv_bound {Γ T T' X} :
      lookup_evar Γ X T →
      lookup_evar (etvar Γ T') X (tshiftTy O T).
Hint Constructors lookup_evar.

(******************************************************************************)
(* Size induction                                                             *)
(******************************************************************************)

Fixpoint sizeTy (S0 : Ty) {struct S0} : nat :=
  match S0 with
    | tvar X     => 1
    | top        => 1
    | tarr T1 T2 => 1 + sizeTy T1 + sizeTy T2
    | tall T0 T3 => 1 + sizeTy T0 + sizeTy T3
    | trec TS    => 1 + sizeTRec TS
  end
with sizeTRec (SS : TRec) {struct SS} : nat :=
  match SS with
    | tnil => 1
    | tcons l T TS => 1 + sizeTy T + sizeTRec TS
  end.

(******************************************************************************)
(* Well-formedness.                                                           *)
(******************************************************************************)

(* These LoC shouldn't be counted here.. *)
(* Inductive wfTy (Γ: Env) : Ty → Prop := *)
(*   | wf_tvar {X T} : *)
(*       lookup_etvar Γ X T → wfTy Γ (tvar X) *)
(*   | wf_top : *)
(*       wfTy Γ top *)
  (* | wf_tarr {T1 T2}: *)
  (*     wfTy Γ T1 → wfTy Γ T2 → wfTy Γ (tarr T1 T2) *)
  (* | wf_tall {T1 T2} : *)
  (*     wfTy Γ T1 → wfTy (etvar Γ T1) T2 → *)
  (*     wfTy Γ (tall T1 T2) *)
(*   | wf_tprod {T1 T2}: *)
(*       wfTy Γ T1 → wfTy Γ T2 → wfTy Γ (tprod T1 T2). *)


Inductive wfTy (Γ : Env) : Ty -> Prop :=
  | wfTy_tvar {X T} :
      lookup_etvar Γ X T → wfTy Γ (tvar X)
  | wfTy_top :
      wfTy Γ top
  | wf_tarr {T1 T2}:
      wfTy Γ T1 → wfTy Γ T2 → wfTy Γ (tarr T1 T2)
  | wf_tall {T1 T2} :
      wfTy Γ T1 → wfTy (etvar Γ T1) T2 →
      wfTy Γ (tall T1 T2)
  | wfTy_trec {Tr} :
      wfTRec Γ Tr → wfTy Γ (trec Tr)
with wfTRec (Γ : Env) : TRec -> Prop :=
  | wfTRec_tnil :
      (wfTRec Γ (tnil))
  | wfTRec_tcons {l T Tr} :
      wfTy Γ T → wfTRec Γ Tr → wfTRec Γ (tcons l T Tr).
Hint Constructors wfTy wfTRec.
Scheme ind_wfTy := Induction for wfTy Sort Prop
with ind_wfTRec := Induction for wfTRec Sort Prop.
Combined Scheme ind_wfty_wftrec from ind_wfTy, ind_wfTRec.
