Require Import Coq.Program.Equality.
Require Import Coq.Program.Tactics.
Require Export Needle.
Section Namespace.
  Inductive Namespace : Type :=
    | TmVar 
    | TyVar .
  Lemma eq_namespace_dec (a : Namespace) (b : Namespace) :
    (sumbool (a =
    b) (not (a =
    b))).
  Proof.
    decide equality .
  Defined.
End Namespace.

Section HVarlist.
  Inductive Hvl : Type :=
    | H0 
    | HS (a : Namespace) (k : Hvl).
  
  Fixpoint appendHvl (k : Hvl) (k0 : Hvl) {struct k0} : Hvl :=
    match k0 with
      | (H0) => k
      | (HS a k0) => (HS a (appendHvl k k0))
    end.
  
  Lemma appendHvl_assoc  :
    (forall (k : Hvl) (k0 : Hvl) (k1 : Hvl) ,
      ((appendHvl (appendHvl k k0) k1) =
      (appendHvl k (appendHvl k0 k1)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
End HVarlist.

Section Index.
  Inductive Index (a : Namespace) : Type :=
    | I0 
    | IS : (Index a) -> (Index a).
  
  Global Arguments I0 [a] .
  Global Arguments IS [a] _ .
  
  Lemma eq_index_dec {a : Namespace} (i : (Index a)) (j : (Index a)) :
    (sumbool (i =
    j) (not (i =
    j))).
  Proof.
    decide equality .
  Qed.
  
  Fixpoint weakenIndexTmVar (x : (Index TmVar)) (k : Hvl) {struct k} : (Index TmVar) :=
    match k with
      | (H0) => x
      | (HS a k) => match a with
        | (TmVar) => (IS (weakenIndexTmVar x k))
        | _ => (weakenIndexTmVar x k)
      end
    end.
  Fixpoint weakenIndexTyVar (X : (Index TyVar)) (k : Hvl) {struct k} : (Index TyVar) :=
    match k with
      | (H0) => X
      | (HS a k) => match a with
        | (TyVar) => (IS (weakenIndexTyVar X k))
        | _ => (weakenIndexTyVar X k)
      end
    end.
  Lemma weakenIndexTmVar_append  :
    (forall (x : (Index TmVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenIndexTmVar (weakenIndexTmVar x k) k0) =
      (weakenIndexTmVar x (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Lemma weakenIndexTyVar_append  :
    (forall (X : (Index TyVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenIndexTyVar (weakenIndexTyVar X k) k0) =
      (weakenIndexTyVar X (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
End Index.

Section Terms.
  Inductive Decls : Set :=
    | DNil 
    | DConsTm (t : Tm) (d : Decls)
    | DConsTy (K : Kind) (d : Decls)
  with Tm : Set :=
    | Var (x : (Index TmVar))
    | Abs (T : Ty) (t : Tm)
    | App (t1 : Tm) (t2 : Tm)
    | Let (d : Decls) (t : Tm)
  with Ty : Set :=
    | TVar (X : (Index TyVar))
    | TPi (T1 : Ty) (T2 : Ty)
    | TApp (T : Ty) (t : Tm)
  with Kind : Set :=
    | Star 
    | KPi (T : Ty) (K : Kind).
  Scheme ind_Decls := Induction for Decls Sort Prop
  with ind_Tm := Induction for Tm Sort Prop
  with ind_Ty := Induction for Ty Sort Prop
  with ind_Kind := Induction for Kind Sort Prop.
  Combined Scheme ind_Decls_Tm_Ty_Kind from ind_Decls, ind_Tm, ind_Ty, ind_Kind.
End Terms.

Section Functions.
  Fixpoint bind (d : Decls) {struct d} : Hvl :=
    match d with
      | (DNil) => H0
      | (DConsTm t d) => (appendHvl (HS TmVar H0) (appendHvl (bind d) H0))
      | (DConsTy K d) => (appendHvl (HS TyVar H0) (appendHvl (bind d) H0))
    end.
  Scheme ind_bind_Decls := Induction for Decls Sort Prop.
End Functions.

Section Shift.
  Inductive Cutoff (a : Namespace) : Type :=
    | C0 
    | CS : (Cutoff a) -> (Cutoff a).
  
  Global Arguments C0 {a} .
  Global Arguments CS {a} _ .
  Fixpoint weakenCutoffTmVar (c : (Cutoff TmVar)) (k : Hvl) {struct k} : (Cutoff TmVar) :=
    match k with
      | (H0) => c
      | (HS a k) => match a with
        | (TmVar) => (CS (weakenCutoffTmVar c k))
        | _ => (weakenCutoffTmVar c k)
      end
    end.
  Fixpoint weakenCutoffTyVar (c : (Cutoff TyVar)) (k : Hvl) {struct k} : (Cutoff TyVar) :=
    match k with
      | (H0) => c
      | (HS a k) => match a with
        | (TyVar) => (CS (weakenCutoffTyVar c k))
        | _ => (weakenCutoffTyVar c k)
      end
    end.
  
  Lemma weakenCutoffTmVar_append  :
    (forall (c : (Cutoff TmVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenCutoffTmVar (weakenCutoffTmVar c k) k0) =
      (weakenCutoffTmVar c (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Lemma weakenCutoffTyVar_append  :
    (forall (c : (Cutoff TyVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenCutoffTyVar (weakenCutoffTyVar c k) k0) =
      (weakenCutoffTyVar c (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Fixpoint shift_TmVar_Index (c : (Cutoff TmVar)) (x : (Index TmVar)) {struct c} : (Index TmVar) :=
    match c with
      | (C0) => (IS x)
      | (CS c) => match x with
        | (I0) => I0
        | (IS x) => (IS (shift_TmVar_Index c x))
      end
    end.
  Fixpoint shift_TyVar_Index (c : (Cutoff TyVar)) (X : (Index TyVar)) {struct c} : (Index TyVar) :=
    match c with
      | (C0) => (IS X)
      | (CS c) => match X with
        | (I0) => I0
        | (IS X) => (IS (shift_TyVar_Index c X))
      end
    end.
  Fixpoint shift_TmVar_Decls (c : (Cutoff TmVar)) (d : Decls) {struct d} : Decls :=
    match d with
      | (DNil) => (DNil)
      | (DConsTm t d0) => (DConsTm (shift_TmVar_Tm (weakenCutoffTmVar c H0) t) (shift_TmVar_Decls (weakenCutoffTmVar c (appendHvl (HS TmVar H0) H0)) d0))
      | (DConsTy K d1) => (DConsTy (shift_TmVar_Kind (weakenCutoffTmVar c H0) K) (shift_TmVar_Decls (weakenCutoffTmVar c (appendHvl (HS TyVar H0) H0)) d1))
    end
  with shift_TmVar_Tm (c : (Cutoff TmVar)) (s : Tm) {struct s} : Tm :=
    match s with
      | (Var x) => (Var (shift_TmVar_Index c x))
      | (Abs T t) => (Abs (shift_TmVar_Ty (weakenCutoffTmVar c H0) T) (shift_TmVar_Tm (weakenCutoffTmVar c (appendHvl (HS TmVar H0) H0)) t))
      | (App t1 t2) => (App (shift_TmVar_Tm (weakenCutoffTmVar c H0) t1) (shift_TmVar_Tm (weakenCutoffTmVar c H0) t2))
      | (Let d t0) => (Let (shift_TmVar_Decls (weakenCutoffTmVar c H0) d) (shift_TmVar_Tm (weakenCutoffTmVar c (appendHvl (bind d) H0)) t0))
    end
  with shift_TmVar_Ty (c : (Cutoff TmVar)) (S0 : Ty) {struct S0} : Ty :=
    match S0 with
      | (TVar X) => (TVar X)
      | (TPi T1 T2) => (TPi (shift_TmVar_Ty (weakenCutoffTmVar c H0) T1) (shift_TmVar_Ty (weakenCutoffTmVar c (appendHvl (HS TmVar H0) H0)) T2))
      | (TApp T t) => (TApp (shift_TmVar_Ty (weakenCutoffTmVar c H0) T) (shift_TmVar_Tm (weakenCutoffTmVar c H0) t))
    end
  with shift_TmVar_Kind (c : (Cutoff TmVar)) (K : Kind) {struct K} : Kind :=
    match K with
      | (Star) => (Star)
      | (KPi T K0) => (KPi (shift_TmVar_Ty (weakenCutoffTmVar c H0) T) (shift_TmVar_Kind (weakenCutoffTmVar c (appendHvl (HS TmVar H0) H0)) K0))
    end.
  Fixpoint shift_TyVar_Decls (c : (Cutoff TyVar)) (d : Decls) {struct d} : Decls :=
    match d with
      | (DNil) => (DNil)
      | (DConsTm t d0) => (DConsTm (shift_TyVar_Tm (weakenCutoffTyVar c H0) t) (shift_TyVar_Decls (weakenCutoffTyVar c (appendHvl (HS TmVar H0) H0)) d0))
      | (DConsTy K d1) => (DConsTy (shift_TyVar_Kind (weakenCutoffTyVar c H0) K) (shift_TyVar_Decls (weakenCutoffTyVar c (appendHvl (HS TyVar H0) H0)) d1))
    end
  with shift_TyVar_Tm (c : (Cutoff TyVar)) (s : Tm) {struct s} : Tm :=
    match s with
      | (Var x) => (Var x)
      | (Abs T t) => (Abs (shift_TyVar_Ty (weakenCutoffTyVar c H0) T) (shift_TyVar_Tm (weakenCutoffTyVar c (appendHvl (HS TmVar H0) H0)) t))
      | (App t1 t2) => (App (shift_TyVar_Tm (weakenCutoffTyVar c H0) t1) (shift_TyVar_Tm (weakenCutoffTyVar c H0) t2))
      | (Let d t0) => (Let (shift_TyVar_Decls (weakenCutoffTyVar c H0) d) (shift_TyVar_Tm (weakenCutoffTyVar c (appendHvl (bind d) H0)) t0))
    end
  with shift_TyVar_Ty (c : (Cutoff TyVar)) (S0 : Ty) {struct S0} : Ty :=
    match S0 with
      | (TVar X) => (TVar (shift_TyVar_Index c X))
      | (TPi T1 T2) => (TPi (shift_TyVar_Ty (weakenCutoffTyVar c H0) T1) (shift_TyVar_Ty (weakenCutoffTyVar c (appendHvl (HS TmVar H0) H0)) T2))
      | (TApp T t) => (TApp (shift_TyVar_Ty (weakenCutoffTyVar c H0) T) (shift_TyVar_Tm (weakenCutoffTyVar c H0) t))
    end
  with shift_TyVar_Kind (c : (Cutoff TyVar)) (K : Kind) {struct K} : Kind :=
    match K with
      | (Star) => (Star)
      | (KPi T K0) => (KPi (shift_TyVar_Ty (weakenCutoffTyVar c H0) T) (shift_TyVar_Kind (weakenCutoffTyVar c (appendHvl (HS TmVar H0) H0)) K0))
    end.
End Shift.

Section Weaken.
  Fixpoint weakenDecls (d : Decls) (k : Hvl) {struct k} : Decls :=
    match k with
      | (H0) => d
      | (HS TmVar k) => (shift_TmVar_Decls C0 (weakenDecls d k))
      | (HS TyVar k) => (shift_TyVar_Decls C0 (weakenDecls d k))
    end.
  Fixpoint weakenTm (s : Tm) (k : Hvl) {struct k} : Tm :=
    match k with
      | (H0) => s
      | (HS TmVar k) => (shift_TmVar_Tm C0 (weakenTm s k))
      | (HS TyVar k) => (shift_TyVar_Tm C0 (weakenTm s k))
    end.
  Fixpoint weakenTy (S0 : Ty) (k : Hvl) {struct k} : Ty :=
    match k with
      | (H0) => S0
      | (HS TmVar k) => (shift_TmVar_Ty C0 (weakenTy S0 k))
      | (HS TyVar k) => (shift_TyVar_Ty C0 (weakenTy S0 k))
    end.
  Fixpoint weakenKind (K : Kind) (k : Hvl) {struct k} : Kind :=
    match k with
      | (H0) => K
      | (HS TmVar k) => (shift_TmVar_Kind C0 (weakenKind K k))
      | (HS TyVar k) => (shift_TyVar_Kind C0 (weakenKind K k))
    end.
End Weaken.

Section Subst.
  Inductive Trace (a : Namespace) : Type :=
    | X0 
    | XS (b : Namespace)
        (T : (Trace a)).
  
  Global Arguments X0 [a] .
  Global Arguments XS [a] b _ .
  Fixpoint weakenTrace {a : Namespace} (x : (Trace a)) (k : Hvl) {struct k} : (Trace a) :=
    match k with
      | (H0) => x
      | (HS b k) => (XS b (weakenTrace x k))
    end.
  Lemma weakenTrace_append (a : Namespace) :
    (forall (x : (Trace a)) (k : Hvl) (k0 : Hvl) ,
      ((weakenTrace (weakenTrace x k) k0) =
      (weakenTrace x (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Fixpoint subst_TmVar_Index (d : (Trace TmVar)) (s : Tm) (x : (Index TmVar)) {struct d} : Tm :=
    match d with
      | (X0) => match x with
        | (I0) => s
        | (IS x) => (Var x)
      end
      | (XS TmVar d) => match x with
        | (I0) => (Var I0)
        | (IS x) => (shift_TmVar_Tm C0 (subst_TmVar_Index d s x))
      end
      | (XS TyVar d) => (shift_TyVar_Tm C0 (subst_TmVar_Index d s x))
    end.
  Fixpoint subst_TyVar_Index (d : (Trace TyVar)) (S0 : Ty) (X : (Index TyVar)) {struct d} : Ty :=
    match d with
      | (X0) => match X with
        | (I0) => S0
        | (IS X) => (TVar X)
      end
      | (XS TmVar d) => (shift_TmVar_Ty C0 (subst_TyVar_Index d S0 X))
      | (XS TyVar d) => match X with
        | (I0) => (TVar I0)
        | (IS X) => (shift_TyVar_Ty C0 (subst_TyVar_Index d S0 X))
      end
    end.
  Fixpoint subst_TmVar_Decls (d : (Trace TmVar)) (s : Tm) (d0 : Decls) {struct d0} : Decls :=
    match d0 with
      | (DNil) => (DNil)
      | (DConsTm t d1) => (DConsTm (subst_TmVar_Tm d s t) (subst_TmVar_Decls (XS TmVar d) s d1))
      | (DConsTy K d2) => (DConsTy (subst_TmVar_Kind d s K) (subst_TmVar_Decls (XS TyVar d) s d2))
    end
  with subst_TmVar_Tm (d : (Trace TmVar)) (s : Tm) (s0 : Tm) {struct s0} : Tm :=
    match s0 with
      | (Var x) => (subst_TmVar_Index d s x)
      | (Abs T t) => (Abs (subst_TmVar_Ty d s T) (subst_TmVar_Tm (XS TmVar d) s t))
      | (App t1 t2) => (App (subst_TmVar_Tm d s t1) (subst_TmVar_Tm d s t2))
      | (Let d0 t0) => (Let (subst_TmVar_Decls d s d0) (subst_TmVar_Tm (weakenTrace d (bind d0)) s t0))
    end
  with subst_TmVar_Ty (d : (Trace TmVar)) (s : Tm) (S0 : Ty) {struct S0} : Ty :=
    match S0 with
      | (TVar X) => (TVar X)
      | (TPi T1 T2) => (TPi (subst_TmVar_Ty d s T1) (subst_TmVar_Ty (XS TmVar d) s T2))
      | (TApp T t) => (TApp (subst_TmVar_Ty d s T) (subst_TmVar_Tm d s t))
    end
  with subst_TmVar_Kind (d : (Trace TmVar)) (s : Tm) (K : Kind) {struct K} : Kind :=
    match K with
      | (Star) => (Star)
      | (KPi T K0) => (KPi (subst_TmVar_Ty d s T) (subst_TmVar_Kind (XS TmVar d) s K0))
    end.
  Fixpoint subst_TyVar_Decls (d : (Trace TyVar)) (S0 : Ty) (d0 : Decls) {struct d0} : Decls :=
    match d0 with
      | (DNil) => (DNil)
      | (DConsTm t d1) => (DConsTm (subst_TyVar_Tm d S0 t) (subst_TyVar_Decls (XS TmVar d) S0 d1))
      | (DConsTy K d2) => (DConsTy (subst_TyVar_Kind d S0 K) (subst_TyVar_Decls (XS TyVar d) S0 d2))
    end
  with subst_TyVar_Tm (d : (Trace TyVar)) (S0 : Ty) (s : Tm) {struct s} : Tm :=
    match s with
      | (Var x) => (Var x)
      | (Abs T t) => (Abs (subst_TyVar_Ty d S0 T) (subst_TyVar_Tm (XS TmVar d) S0 t))
      | (App t1 t2) => (App (subst_TyVar_Tm d S0 t1) (subst_TyVar_Tm d S0 t2))
      | (Let d0 t0) => (Let (subst_TyVar_Decls d S0 d0) (subst_TyVar_Tm (weakenTrace d (bind d0)) S0 t0))
    end
  with subst_TyVar_Ty (d : (Trace TyVar)) (S0 : Ty) (S1 : Ty) {struct S1} : Ty :=
    match S1 with
      | (TVar X) => (subst_TyVar_Index d S0 X)
      | (TPi T1 T2) => (TPi (subst_TyVar_Ty d S0 T1) (subst_TyVar_Ty (XS TmVar d) S0 T2))
      | (TApp T t) => (TApp (subst_TyVar_Ty d S0 T) (subst_TyVar_Tm d S0 t))
    end
  with subst_TyVar_Kind (d : (Trace TyVar)) (S0 : Ty) (K : Kind) {struct K} : Kind :=
    match K with
      | (Star) => (Star)
      | (KPi T K0) => (KPi (subst_TyVar_Ty d S0 T) (subst_TyVar_Kind (XS TmVar d) S0 K0))
    end.
End Subst.

Global Hint Resolve (f_equal (shift_TmVar_Decls C0)) : cong_shift0.
Global Hint Resolve (f_equal (shift_TyVar_Decls C0)) : cong_shift0.
Global Hint Resolve (f_equal (shift_TmVar_Kind C0)) : cong_shift0.
Global Hint Resolve (f_equal (shift_TyVar_Kind C0)) : cong_shift0.
Global Hint Resolve (f_equal (shift_TmVar_Tm C0)) : cong_shift0.
Global Hint Resolve (f_equal (shift_TyVar_Tm C0)) : cong_shift0.
Global Hint Resolve (f_equal (shift_TmVar_Ty C0)) : cong_shift0.
Global Hint Resolve (f_equal (shift_TyVar_Ty C0)) : cong_shift0.
 Hint Rewrite weakenCutoffTmVar_append weakenCutoffTyVar_append : interaction.
 Hint Rewrite weakenCutoffTmVar_append weakenCutoffTyVar_append : weakencutoff_append.
 Hint Rewrite weakenTrace_append : interaction.
 Hint Rewrite weakenTrace_append : weakentrace_append.
Lemma stability_shift_TmVar__bind  :
  (forall (d : Decls) ,
    (forall (c : (Cutoff TmVar)) ,
      ((bind (shift_TmVar_Decls c d)) =
      (bind d)))).
Proof.
  apply_mutual_ind (ind_bind_Decls); simpl; intros; congruence .
Qed.
Lemma stability_shift_TyVar__bind  :
  (forall (d0 : Decls) ,
    (forall (c0 : (Cutoff TyVar)) ,
      ((bind (shift_TyVar_Decls c0 d0)) =
      (bind d0)))).
Proof.
  apply_mutual_ind (ind_bind_Decls); simpl; intros; congruence .
Qed.
 Hint Rewrite stability_shift_TmVar__bind stability_shift_TyVar__bind : interaction.
 Hint Rewrite stability_shift_TmVar__bind stability_shift_TyVar__bind : stability_shift.
Lemma stability_weaken_bind  :
  (forall (k : Hvl) (d1 : Decls) ,
    ((bind (weakenDecls d1 k)) =
    (bind d1))).
Proof.
  needleGenericStabilityWeaken .
Qed.
 Hint Rewrite stability_weaken_bind : interaction.
 Hint Rewrite stability_weaken_bind : stability_weaken.
Lemma stability_subst_TmVar__bind  :
  (forall (d1 : Decls) ,
    (forall (d2 : (Trace TmVar)) (s : Tm) ,
      ((bind (subst_TmVar_Decls d2 s d1)) =
      (bind d1)))).
Proof.
  apply_mutual_ind (ind_bind_Decls); simpl; intros; congruence .
Qed.
Lemma stability_subst_TyVar__bind  :
  (forall (d3 : Decls) ,
    (forall (d4 : (Trace TyVar)) (S0 : Ty) ,
      ((bind (subst_TyVar_Decls d4 S0 d3)) =
      (bind d3)))).
Proof.
  apply_mutual_ind (ind_bind_Decls); simpl; intros; congruence .
Qed.
 Hint Rewrite stability_subst_TmVar__bind stability_subst_TyVar__bind : interaction.
 Hint Rewrite stability_subst_TmVar__bind stability_subst_TyVar__bind : stability_subst.
Section SubstShiftReflection.
  Lemma subst_TmVar_Index0_shift_TmVar_Index0_reflection_ind (s0 : Tm) :
    (forall (k7 : Hvl) (x23 : (Index TmVar)) ,
      ((subst_TmVar_Index (weakenTrace X0 k7) s0 (shift_TmVar_Index (weakenCutoffTmVar C0 k7) x23)) =
      (Var x23))).
  Proof.
    needleGenericSubstIndexShiftIndexReflectionInd .
  Qed.
  Lemma subst_TyVar_Index0_shift_TyVar_Index0_reflection_ind (S1 : Ty) :
    (forall (k7 : Hvl) (X11 : (Index TyVar)) ,
      ((subst_TyVar_Index (weakenTrace X0 k7) S1 (shift_TyVar_Index (weakenCutoffTyVar C0 k7) X11)) =
      (TVar X11))).
  Proof.
    needleGenericSubstIndexShiftIndexReflectionInd .
  Qed.
  Definition subst_TmVar_0_shift_TmVar_0_Decls_Tm_Ty_Kind_reflection_ind := (ind_Decls_Tm_Ty_Kind (fun (d23 : Decls) =>
    (forall (k11 : Hvl) (s2 : Tm) ,
      ((subst_TmVar_Decls (weakenTrace X0 k11) s2 (shift_TmVar_Decls (weakenCutoffTmVar C0 k11) d23)) =
      d23))) (fun (s2 : Tm) =>
    (forall (k11 : Hvl) (s3 : Tm) ,
      ((subst_TmVar_Tm (weakenTrace X0 k11) s3 (shift_TmVar_Tm (weakenCutoffTmVar C0 k11) s2)) =
      s2))) (fun (S3 : Ty) =>
    (forall (k11 : Hvl) (s2 : Tm) ,
      ((subst_TmVar_Ty (weakenTrace X0 k11) s2 (shift_TmVar_Ty (weakenCutoffTmVar C0 k11) S3)) =
      S3))) (fun (K11 : Kind) =>
    (forall (k11 : Hvl) (s2 : Tm) ,
      ((subst_TmVar_Kind (weakenTrace X0 k11) s2 (shift_TmVar_Kind (weakenCutoffTmVar C0 k11) K11)) =
      K11))) (fun (k11 : Hvl) (s2 : Tm) =>
    eq_refl) (fun (t : Tm) IHt35 (d : Decls) IHd (k11 : Hvl) (s2 : Tm) =>
    (f_equal2 DConsTm (IHt35 k11 s2) (eq_trans (f_equal3 subst_TmVar_Decls (weakenTrace_append TmVar X0 k11 (HS TmVar H0)) eq_refl eq_refl) (IHd (HS TmVar k11) s2)))) (fun (K : Kind) IHK11 (d : Decls) IHd (k11 : Hvl) (s2 : Tm) =>
    (f_equal2 DConsTy (IHK11 k11 s2) (eq_trans (f_equal3 subst_TmVar_Decls (weakenTrace_append TmVar X0 k11 (HS TyVar H0)) eq_refl eq_refl) (IHd (HS TyVar k11) s2)))) (fun (x35 : (Index TmVar)) =>
    (fun (k11 : Hvl) (s2 : Tm) =>
      (subst_TmVar_Index0_shift_TmVar_Index0_reflection_ind s2 k11 x35))) (fun (T : Ty) IHT29 (t : Tm) IHt35 (k11 : Hvl) (s2 : Tm) =>
    (f_equal2 Abs (IHT29 k11 s2) (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k11 (HS TmVar H0)) eq_refl eq_refl) (IHt35 (HS TmVar k11) s2)))) (fun (t1 : Tm) IHt35 (t2 : Tm) IHt36 (k11 : Hvl) (s2 : Tm) =>
    (f_equal2 App (IHt35 k11 s2) (IHt36 k11 s2))) (fun (d : Decls) IHd (t : Tm) IHt35 (k11 : Hvl) (s2 : Tm) =>
    (f_equal2 Let (IHd k11 s2) (eq_trans (f_equal3 subst_TmVar_Tm (eq_trans (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (stability_shift_TmVar__bind _ _) (eq_refl H0))) (weakenTrace_append TmVar X0 k11 (bind d))) eq_refl (f_equal2 shift_TmVar_Tm (weakenCutoffTmVar_append C0 k11 (bind d)) eq_refl)) (IHt35 (appendHvl k11 (bind d)) s2)))) (fun (X17 : (Index TyVar)) =>
    (fun (k11 : Hvl) (s2 : Tm) =>
      eq_refl)) (fun (T1 : Ty) IHT29 (T2 : Ty) IHT30 (k11 : Hvl) (s2 : Tm) =>
    (f_equal2 TPi (IHT29 k11 s2) (eq_trans (f_equal3 subst_TmVar_Ty (weakenTrace_append TmVar X0 k11 (HS TmVar H0)) eq_refl eq_refl) (IHT30 (HS TmVar k11) s2)))) (fun (T : Ty) IHT29 (t : Tm) IHt35 (k11 : Hvl) (s2 : Tm) =>
    (f_equal2 TApp (IHT29 k11 s2) (IHt35 k11 s2))) (fun (k11 : Hvl) (s2 : Tm) =>
    eq_refl) (fun (T : Ty) IHT29 (K : Kind) IHK11 (k11 : Hvl) (s2 : Tm) =>
    (f_equal2 KPi (IHT29 k11 s2) (eq_trans (f_equal3 subst_TmVar_Kind (weakenTrace_append TmVar X0 k11 (HS TmVar H0)) eq_refl eq_refl) (IHK11 (HS TmVar k11) s2))))).
  Lemma subst_TmVar_0_shift_TmVar_0_Decls_reflection_ind  :
    (forall (d23 : Decls) ,
      (forall (k11 : Hvl) (s2 : Tm) ,
        ((subst_TmVar_Decls (weakenTrace X0 k11) s2 (shift_TmVar_Decls (weakenCutoffTmVar C0 k11) d23)) =
        d23))).
  Proof.
    pose proof (subst_TmVar_0_shift_TmVar_0_Decls_Tm_Ty_Kind_reflection_ind).
    destruct_conjs .
    easy .
  Qed.
  Lemma subst_TmVar_0_shift_TmVar_0_Tm_reflection_ind  :
    (forall (s2 : Tm) ,
      (forall (k11 : Hvl) (s3 : Tm) ,
        ((subst_TmVar_Tm (weakenTrace X0 k11) s3 (shift_TmVar_Tm (weakenCutoffTmVar C0 k11) s2)) =
        s2))).
  Proof.
    pose proof (subst_TmVar_0_shift_TmVar_0_Decls_Tm_Ty_Kind_reflection_ind).
    destruct_conjs .
    easy .
  Qed.
  Lemma subst_TmVar_0_shift_TmVar_0_Ty_reflection_ind  :
    (forall (S3 : Ty) ,
      (forall (k11 : Hvl) (s3 : Tm) ,
        ((subst_TmVar_Ty (weakenTrace X0 k11) s3 (shift_TmVar_Ty (weakenCutoffTmVar C0 k11) S3)) =
        S3))).
  Proof.
    pose proof (subst_TmVar_0_shift_TmVar_0_Decls_Tm_Ty_Kind_reflection_ind).
    destruct_conjs .
    easy .
  Qed.
  Lemma subst_TmVar_0_shift_TmVar_0_Kind_reflection_ind  :
    (forall (K11 : Kind) ,
      (forall (k11 : Hvl) (s3 : Tm) ,
        ((subst_TmVar_Kind (weakenTrace X0 k11) s3 (shift_TmVar_Kind (weakenCutoffTmVar C0 k11) K11)) =
        K11))).
  Proof.
    pose proof (subst_TmVar_0_shift_TmVar_0_Decls_Tm_Ty_Kind_reflection_ind).
    destruct_conjs .
    easy .
  Qed.
  Definition subst_TyVar_0_shift_TyVar_0_Decls_Tm_Ty_Kind_reflection_ind := (ind_Decls_Tm_Ty_Kind (fun (d23 : Decls) =>
    (forall (k11 : Hvl) (S3 : Ty) ,
      ((subst_TyVar_Decls (weakenTrace X0 k11) S3 (shift_TyVar_Decls (weakenCutoffTyVar C0 k11) d23)) =
      d23))) (fun (s2 : Tm) =>
    (forall (k11 : Hvl) (S3 : Ty) ,
      ((subst_TyVar_Tm (weakenTrace X0 k11) S3 (shift_TyVar_Tm (weakenCutoffTyVar C0 k11) s2)) =
      s2))) (fun (S3 : Ty) =>
    (forall (k11 : Hvl) (S4 : Ty) ,
      ((subst_TyVar_Ty (weakenTrace X0 k11) S4 (shift_TyVar_Ty (weakenCutoffTyVar C0 k11) S3)) =
      S3))) (fun (K11 : Kind) =>
    (forall (k11 : Hvl) (S3 : Ty) ,
      ((subst_TyVar_Kind (weakenTrace X0 k11) S3 (shift_TyVar_Kind (weakenCutoffTyVar C0 k11) K11)) =
      K11))) (fun (k11 : Hvl) (S3 : Ty) =>
    eq_refl) (fun (t : Tm) IHt35 (d : Decls) IHd (k11 : Hvl) (S3 : Ty) =>
    (f_equal2 DConsTm (IHt35 k11 S3) (eq_trans (f_equal3 subst_TyVar_Decls (weakenTrace_append TyVar X0 k11 (HS TmVar H0)) eq_refl eq_refl) (IHd (HS TmVar k11) S3)))) (fun (K : Kind) IHK11 (d : Decls) IHd (k11 : Hvl) (S3 : Ty) =>
    (f_equal2 DConsTy (IHK11 k11 S3) (eq_trans (f_equal3 subst_TyVar_Decls (weakenTrace_append TyVar X0 k11 (HS TyVar H0)) eq_refl eq_refl) (IHd (HS TyVar k11) S3)))) (fun (x35 : (Index TmVar)) =>
    (fun (k11 : Hvl) (S3 : Ty) =>
      eq_refl)) (fun (T : Ty) IHT29 (t : Tm) IHt35 (k11 : Hvl) (S3 : Ty) =>
    (f_equal2 Abs (IHT29 k11 S3) (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k11 (HS TmVar H0)) eq_refl eq_refl) (IHt35 (HS TmVar k11) S3)))) (fun (t1 : Tm) IHt35 (t2 : Tm) IHt36 (k11 : Hvl) (S3 : Ty) =>
    (f_equal2 App (IHt35 k11 S3) (IHt36 k11 S3))) (fun (d : Decls) IHd (t : Tm) IHt35 (k11 : Hvl) (S3 : Ty) =>
    (f_equal2 Let (IHd k11 S3) (eq_trans (f_equal3 subst_TyVar_Tm (eq_trans (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (stability_shift_TyVar__bind _ _) (eq_refl H0))) (weakenTrace_append TyVar X0 k11 (bind d))) eq_refl (f_equal2 shift_TyVar_Tm (weakenCutoffTyVar_append C0 k11 (bind d)) eq_refl)) (IHt35 (appendHvl k11 (bind d)) S3)))) (fun (X17 : (Index TyVar)) =>
    (fun (k11 : Hvl) (S3 : Ty) =>
      (subst_TyVar_Index0_shift_TyVar_Index0_reflection_ind S3 k11 X17))) (fun (T1 : Ty) IHT29 (T2 : Ty) IHT30 (k11 : Hvl) (S3 : Ty) =>
    (f_equal2 TPi (IHT29 k11 S3) (eq_trans (f_equal3 subst_TyVar_Ty (weakenTrace_append TyVar X0 k11 (HS TmVar H0)) eq_refl eq_refl) (IHT30 (HS TmVar k11) S3)))) (fun (T : Ty) IHT29 (t : Tm) IHt35 (k11 : Hvl) (S3 : Ty) =>
    (f_equal2 TApp (IHT29 k11 S3) (IHt35 k11 S3))) (fun (k11 : Hvl) (S3 : Ty) =>
    eq_refl) (fun (T : Ty) IHT29 (K : Kind) IHK11 (k11 : Hvl) (S3 : Ty) =>
    (f_equal2 KPi (IHT29 k11 S3) (eq_trans (f_equal3 subst_TyVar_Kind (weakenTrace_append TyVar X0 k11 (HS TmVar H0)) eq_refl eq_refl) (IHK11 (HS TmVar k11) S3))))).
  Lemma subst_TyVar_0_shift_TyVar_0_Decls_reflection_ind  :
    (forall (d23 : Decls) ,
      (forall (k11 : Hvl) (S3 : Ty) ,
        ((subst_TyVar_Decls (weakenTrace X0 k11) S3 (shift_TyVar_Decls (weakenCutoffTyVar C0 k11) d23)) =
        d23))).
  Proof.
    pose proof (subst_TyVar_0_shift_TyVar_0_Decls_Tm_Ty_Kind_reflection_ind).
    destruct_conjs .
    easy .
  Qed.
  Lemma subst_TyVar_0_shift_TyVar_0_Tm_reflection_ind  :
    (forall (s2 : Tm) ,
      (forall (k11 : Hvl) (S3 : Ty) ,
        ((subst_TyVar_Tm (weakenTrace X0 k11) S3 (shift_TyVar_Tm (weakenCutoffTyVar C0 k11) s2)) =
        s2))).
  Proof.
    pose proof (subst_TyVar_0_shift_TyVar_0_Decls_Tm_Ty_Kind_reflection_ind).
    destruct_conjs .
    easy .
  Qed.
  Lemma subst_TyVar_0_shift_TyVar_0_Ty_reflection_ind  :
    (forall (S3 : Ty) ,
      (forall (k11 : Hvl) (S4 : Ty) ,
        ((subst_TyVar_Ty (weakenTrace X0 k11) S4 (shift_TyVar_Ty (weakenCutoffTyVar C0 k11) S3)) =
        S3))).
  Proof.
    pose proof (subst_TyVar_0_shift_TyVar_0_Decls_Tm_Ty_Kind_reflection_ind).
    destruct_conjs .
    easy .
  Qed.
  Lemma subst_TyVar_0_shift_TyVar_0_Kind_reflection_ind  :
    (forall (K11 : Kind) ,
      (forall (k11 : Hvl) (S4 : Ty) ,
        ((subst_TyVar_Kind (weakenTrace X0 k11) S4 (shift_TyVar_Kind (weakenCutoffTyVar C0 k11) K11)) =
        K11))).
  Proof.
    pose proof (subst_TyVar_0_shift_TyVar_0_Decls_Tm_Ty_Kind_reflection_ind).
    destruct_conjs .
    easy .
  Qed.
  Definition subst_TmVar_Decls0_shift_TmVar_Decls0_reflection (d23 : Decls) : (forall (s2 : Tm) ,
    ((subst_TmVar_Decls X0 s2 (shift_TmVar_Decls C0 d23)) =
    d23)) := (subst_TmVar_0_shift_TmVar_0_Decls_reflection_ind d23 H0).
  Definition subst_TmVar_Tm0_shift_TmVar_Tm0_reflection (s3 : Tm) : (forall (s2 : Tm) ,
    ((subst_TmVar_Tm X0 s2 (shift_TmVar_Tm C0 s3)) =
    s3)) := (subst_TmVar_0_shift_TmVar_0_Tm_reflection_ind s3 H0).
  Definition subst_TmVar_Ty0_shift_TmVar_Ty0_reflection (S3 : Ty) : (forall (s2 : Tm) ,
    ((subst_TmVar_Ty X0 s2 (shift_TmVar_Ty C0 S3)) =
    S3)) := (subst_TmVar_0_shift_TmVar_0_Ty_reflection_ind S3 H0).
  Definition subst_TmVar_Kind0_shift_TmVar_Kind0_reflection (K11 : Kind) : (forall (s2 : Tm) ,
    ((subst_TmVar_Kind X0 s2 (shift_TmVar_Kind C0 K11)) =
    K11)) := (subst_TmVar_0_shift_TmVar_0_Kind_reflection_ind K11 H0).
  Definition subst_TyVar_Decls0_shift_TyVar_Decls0_reflection (d23 : Decls) : (forall (S3 : Ty) ,
    ((subst_TyVar_Decls X0 S3 (shift_TyVar_Decls C0 d23)) =
    d23)) := (subst_TyVar_0_shift_TyVar_0_Decls_reflection_ind d23 H0).
  Definition subst_TyVar_Tm0_shift_TyVar_Tm0_reflection (s2 : Tm) : (forall (S3 : Ty) ,
    ((subst_TyVar_Tm X0 S3 (shift_TyVar_Tm C0 s2)) =
    s2)) := (subst_TyVar_0_shift_TyVar_0_Tm_reflection_ind s2 H0).
  Definition subst_TyVar_Ty0_shift_TyVar_Ty0_reflection (S4 : Ty) : (forall (S3 : Ty) ,
    ((subst_TyVar_Ty X0 S3 (shift_TyVar_Ty C0 S4)) =
    S4)) := (subst_TyVar_0_shift_TyVar_0_Ty_reflection_ind S4 H0).
  Definition subst_TyVar_Kind0_shift_TyVar_Kind0_reflection (K11 : Kind) : (forall (S3 : Ty) ,
    ((subst_TyVar_Kind X0 S3 (shift_TyVar_Kind C0 K11)) =
    K11)) := (subst_TyVar_0_shift_TyVar_0_Kind_reflection_ind K11 H0).
End SubstShiftReflection.
Section ShiftInteraction.
  Section ShiftIndexCommInd.
    Lemma shift_TmVar_Index_shift_TmVar_Index0_comm_ind  :
      (forall (k : Hvl) (c1 : (Cutoff TmVar)) (x : (Index TmVar)) ,
        ((shift_TmVar_Index (weakenCutoffTmVar (CS c1) k) (shift_TmVar_Index (weakenCutoffTmVar C0 k) x)) =
        (shift_TmVar_Index (weakenCutoffTmVar C0 k) (shift_TmVar_Index (weakenCutoffTmVar c1 k) x)))).
    Proof.
      needleGenericShiftIndexCommInd .
    Qed.
    Lemma shift_TyVar_Index_shift_TyVar_Index0_comm_ind  :
      (forall (k : Hvl) (c1 : (Cutoff TyVar)) (X : (Index TyVar)) ,
        ((shift_TyVar_Index (weakenCutoffTyVar (CS c1) k) (shift_TyVar_Index (weakenCutoffTyVar C0 k) X)) =
        (shift_TyVar_Index (weakenCutoffTyVar C0 k) (shift_TyVar_Index (weakenCutoffTyVar c1 k) X)))).
    Proof.
      needleGenericShiftIndexCommInd .
    Qed.
  End ShiftIndexCommInd.
  Section ShiftCommInd.
    Definition shift_TmVar__shift_TmVar_0_Decls_Tm_Ty_Kind_comm_ind := (ind_Decls_Tm_Ty_Kind (fun (d17 : Decls) =>
      (forall (k7 : Hvl) (c9 : (Cutoff TmVar)) ,
        ((shift_TmVar_Decls (weakenCutoffTmVar (CS c9) k7) (shift_TmVar_Decls (weakenCutoffTmVar C0 k7) d17)) =
        (shift_TmVar_Decls (weakenCutoffTmVar C0 k7) (shift_TmVar_Decls (weakenCutoffTmVar c9 k7) d17))))) (fun (s0 : Tm) =>
      (forall (k7 : Hvl) (c9 : (Cutoff TmVar)) ,
        ((shift_TmVar_Tm (weakenCutoffTmVar (CS c9) k7) (shift_TmVar_Tm (weakenCutoffTmVar C0 k7) s0)) =
        (shift_TmVar_Tm (weakenCutoffTmVar C0 k7) (shift_TmVar_Tm (weakenCutoffTmVar c9 k7) s0))))) (fun (S1 : Ty) =>
      (forall (k7 : Hvl) (c9 : (Cutoff TmVar)) ,
        ((shift_TmVar_Ty (weakenCutoffTmVar (CS c9) k7) (shift_TmVar_Ty (weakenCutoffTmVar C0 k7) S1)) =
        (shift_TmVar_Ty (weakenCutoffTmVar C0 k7) (shift_TmVar_Ty (weakenCutoffTmVar c9 k7) S1))))) (fun (K7 : Kind) =>
      (forall (k7 : Hvl) (c9 : (Cutoff TmVar)) ,
        ((shift_TmVar_Kind (weakenCutoffTmVar (CS c9) k7) (shift_TmVar_Kind (weakenCutoffTmVar C0 k7) K7)) =
        (shift_TmVar_Kind (weakenCutoffTmVar C0 k7) (shift_TmVar_Kind (weakenCutoffTmVar c9 k7) K7))))) (fun (k7 : Hvl) (c9 : (Cutoff TmVar)) =>
      eq_refl) (fun (t : Tm) IHt23 (d : Decls) IHd (k7 : Hvl) (c9 : (Cutoff TmVar)) =>
      (f_equal2 DConsTm (IHt23 k7 c9) (IHd (HS TmVar k7) c9))) (fun (K : Kind) IHK7 (d : Decls) IHd (k7 : Hvl) (c9 : (Cutoff TmVar)) =>
      (f_equal2 DConsTy (IHK7 k7 c9) (IHd (HS TyVar k7) c9))) (fun (x23 : (Index TmVar)) =>
      (fun (k7 : Hvl) (c9 : (Cutoff TmVar)) =>
        (f_equal Var (shift_TmVar_Index_shift_TmVar_Index0_comm_ind k7 c9 x23)))) (fun (T : Ty) IHT19 (t : Tm) IHt23 (k7 : Hvl) (c9 : (Cutoff TmVar)) =>
      (f_equal2 Abs (IHT19 k7 c9) (IHt23 (HS TmVar k7) c9))) (fun (t1 : Tm) IHt23 (t2 : Tm) IHt24 (k7 : Hvl) (c9 : (Cutoff TmVar)) =>
      (f_equal2 App (IHt23 k7 c9) (IHt24 k7 c9))) (fun (d : Decls) IHd (t : Tm) IHt23 (k7 : Hvl) (c9 : (Cutoff TmVar)) =>
      (f_equal2 Let (IHd k7 c9) (eq_trans (f_equal2 shift_TmVar_Tm (eq_trans (f_equal2 weakenCutoffTmVar eq_refl (f_equal2 appendHvl (stability_shift_TmVar__bind _ _) (eq_refl H0))) (weakenCutoffTmVar_append (CS c9) k7 (bind d))) (f_equal2 shift_TmVar_Tm (weakenCutoffTmVar_append C0 k7 (bind d)) eq_refl)) (eq_trans (IHt23 (appendHvl k7 (bind d)) c9) (f_equal2 shift_TmVar_Tm (eq_trans (eq_sym (weakenCutoffTmVar_append C0 k7 (bind d))) (f_equal2 weakenCutoffTmVar eq_refl (f_equal2 appendHvl (eq_sym (stability_shift_TmVar__bind _ _)) (eq_refl H0)))) (f_equal2 shift_TmVar_Tm (eq_sym (weakenCutoffTmVar_append c9 k7 (bind d))) eq_refl)))))) (fun (X11 : (Index TyVar)) =>
      (fun (k7 : Hvl) (c9 : (Cutoff TmVar)) =>
        eq_refl)) (fun (T1 : Ty) IHT19 (T2 : Ty) IHT20 (k7 : Hvl) (c9 : (Cutoff TmVar)) =>
      (f_equal2 TPi (IHT19 k7 c9) (IHT20 (HS TmVar k7) c9))) (fun (T : Ty) IHT19 (t : Tm) IHt23 (k7 : Hvl) (c9 : (Cutoff TmVar)) =>
      (f_equal2 TApp (IHT19 k7 c9) (IHt23 k7 c9))) (fun (k7 : Hvl) (c9 : (Cutoff TmVar)) =>
      eq_refl) (fun (T : Ty) IHT19 (K : Kind) IHK7 (k7 : Hvl) (c9 : (Cutoff TmVar)) =>
      (f_equal2 KPi (IHT19 k7 c9) (IHK7 (HS TmVar k7) c9)))).
    Lemma shift_TmVar__shift_TmVar_0_Decls_comm_ind  :
      (forall (d17 : Decls) ,
        (forall (k7 : Hvl) (c9 : (Cutoff TmVar)) ,
          ((shift_TmVar_Decls (weakenCutoffTmVar (CS c9) k7) (shift_TmVar_Decls (weakenCutoffTmVar C0 k7) d17)) =
          (shift_TmVar_Decls (weakenCutoffTmVar C0 k7) (shift_TmVar_Decls (weakenCutoffTmVar c9 k7) d17))))).
    Proof.
      pose proof (shift_TmVar__shift_TmVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TmVar__shift_TmVar_0_Tm_comm_ind  :
      (forall (s0 : Tm) ,
        (forall (k7 : Hvl) (c9 : (Cutoff TmVar)) ,
          ((shift_TmVar_Tm (weakenCutoffTmVar (CS c9) k7) (shift_TmVar_Tm (weakenCutoffTmVar C0 k7) s0)) =
          (shift_TmVar_Tm (weakenCutoffTmVar C0 k7) (shift_TmVar_Tm (weakenCutoffTmVar c9 k7) s0))))).
    Proof.
      pose proof (shift_TmVar__shift_TmVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TmVar__shift_TmVar_0_Ty_comm_ind  :
      (forall (S1 : Ty) ,
        (forall (k7 : Hvl) (c9 : (Cutoff TmVar)) ,
          ((shift_TmVar_Ty (weakenCutoffTmVar (CS c9) k7) (shift_TmVar_Ty (weakenCutoffTmVar C0 k7) S1)) =
          (shift_TmVar_Ty (weakenCutoffTmVar C0 k7) (shift_TmVar_Ty (weakenCutoffTmVar c9 k7) S1))))).
    Proof.
      pose proof (shift_TmVar__shift_TmVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TmVar__shift_TmVar_0_Kind_comm_ind  :
      (forall (K7 : Kind) ,
        (forall (k7 : Hvl) (c9 : (Cutoff TmVar)) ,
          ((shift_TmVar_Kind (weakenCutoffTmVar (CS c9) k7) (shift_TmVar_Kind (weakenCutoffTmVar C0 k7) K7)) =
          (shift_TmVar_Kind (weakenCutoffTmVar C0 k7) (shift_TmVar_Kind (weakenCutoffTmVar c9 k7) K7))))).
    Proof.
      pose proof (shift_TmVar__shift_TmVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_TmVar__shift_TyVar_0_Decls_Tm_Ty_Kind_comm_ind := (ind_Decls_Tm_Ty_Kind (fun (d17 : Decls) =>
      (forall (k7 : Hvl) (c9 : (Cutoff TmVar)) ,
        ((shift_TmVar_Decls (weakenCutoffTmVar c9 k7) (shift_TyVar_Decls (weakenCutoffTyVar C0 k7) d17)) =
        (shift_TyVar_Decls (weakenCutoffTyVar C0 k7) (shift_TmVar_Decls (weakenCutoffTmVar c9 k7) d17))))) (fun (s0 : Tm) =>
      (forall (k7 : Hvl) (c9 : (Cutoff TmVar)) ,
        ((shift_TmVar_Tm (weakenCutoffTmVar c9 k7) (shift_TyVar_Tm (weakenCutoffTyVar C0 k7) s0)) =
        (shift_TyVar_Tm (weakenCutoffTyVar C0 k7) (shift_TmVar_Tm (weakenCutoffTmVar c9 k7) s0))))) (fun (S1 : Ty) =>
      (forall (k7 : Hvl) (c9 : (Cutoff TmVar)) ,
        ((shift_TmVar_Ty (weakenCutoffTmVar c9 k7) (shift_TyVar_Ty (weakenCutoffTyVar C0 k7) S1)) =
        (shift_TyVar_Ty (weakenCutoffTyVar C0 k7) (shift_TmVar_Ty (weakenCutoffTmVar c9 k7) S1))))) (fun (K7 : Kind) =>
      (forall (k7 : Hvl) (c9 : (Cutoff TmVar)) ,
        ((shift_TmVar_Kind (weakenCutoffTmVar c9 k7) (shift_TyVar_Kind (weakenCutoffTyVar C0 k7) K7)) =
        (shift_TyVar_Kind (weakenCutoffTyVar C0 k7) (shift_TmVar_Kind (weakenCutoffTmVar c9 k7) K7))))) (fun (k7 : Hvl) (c9 : (Cutoff TmVar)) =>
      eq_refl) (fun (t : Tm) IHt23 (d : Decls) IHd (k7 : Hvl) (c9 : (Cutoff TmVar)) =>
      (f_equal2 DConsTm (IHt23 k7 c9) (IHd (HS TmVar k7) c9))) (fun (K : Kind) IHK7 (d : Decls) IHd (k7 : Hvl) (c9 : (Cutoff TmVar)) =>
      (f_equal2 DConsTy (IHK7 k7 c9) (IHd (HS TyVar k7) c9))) (fun (x23 : (Index TmVar)) =>
      (fun (k7 : Hvl) (c9 : (Cutoff TmVar)) =>
        eq_refl)) (fun (T : Ty) IHT19 (t : Tm) IHt23 (k7 : Hvl) (c9 : (Cutoff TmVar)) =>
      (f_equal2 Abs (IHT19 k7 c9) (IHt23 (HS TmVar k7) c9))) (fun (t1 : Tm) IHt23 (t2 : Tm) IHt24 (k7 : Hvl) (c9 : (Cutoff TmVar)) =>
      (f_equal2 App (IHt23 k7 c9) (IHt24 k7 c9))) (fun (d : Decls) IHd (t : Tm) IHt23 (k7 : Hvl) (c9 : (Cutoff TmVar)) =>
      (f_equal2 Let (IHd k7 c9) (eq_trans (f_equal2 shift_TmVar_Tm (eq_trans (f_equal2 weakenCutoffTmVar eq_refl (f_equal2 appendHvl (stability_shift_TyVar__bind _ _) (eq_refl H0))) (weakenCutoffTmVar_append c9 k7 (bind d))) (f_equal2 shift_TyVar_Tm (weakenCutoffTyVar_append C0 k7 (bind d)) eq_refl)) (eq_trans (IHt23 (appendHvl k7 (bind d)) c9) (f_equal2 shift_TyVar_Tm (eq_trans (eq_sym (weakenCutoffTyVar_append C0 k7 (bind d))) (f_equal2 weakenCutoffTyVar eq_refl (f_equal2 appendHvl (eq_sym (stability_shift_TmVar__bind _ _)) (eq_refl H0)))) (f_equal2 shift_TmVar_Tm (eq_sym (weakenCutoffTmVar_append c9 k7 (bind d))) eq_refl)))))) (fun (X11 : (Index TyVar)) =>
      (fun (k7 : Hvl) (c9 : (Cutoff TmVar)) =>
        eq_refl)) (fun (T1 : Ty) IHT19 (T2 : Ty) IHT20 (k7 : Hvl) (c9 : (Cutoff TmVar)) =>
      (f_equal2 TPi (IHT19 k7 c9) (IHT20 (HS TmVar k7) c9))) (fun (T : Ty) IHT19 (t : Tm) IHt23 (k7 : Hvl) (c9 : (Cutoff TmVar)) =>
      (f_equal2 TApp (IHT19 k7 c9) (IHt23 k7 c9))) (fun (k7 : Hvl) (c9 : (Cutoff TmVar)) =>
      eq_refl) (fun (T : Ty) IHT19 (K : Kind) IHK7 (k7 : Hvl) (c9 : (Cutoff TmVar)) =>
      (f_equal2 KPi (IHT19 k7 c9) (IHK7 (HS TmVar k7) c9)))).
    Lemma shift_TmVar__shift_TyVar_0_Decls_comm_ind  :
      (forall (d17 : Decls) ,
        (forall (k7 : Hvl) (c9 : (Cutoff TmVar)) ,
          ((shift_TmVar_Decls (weakenCutoffTmVar c9 k7) (shift_TyVar_Decls (weakenCutoffTyVar C0 k7) d17)) =
          (shift_TyVar_Decls (weakenCutoffTyVar C0 k7) (shift_TmVar_Decls (weakenCutoffTmVar c9 k7) d17))))).
    Proof.
      pose proof (shift_TmVar__shift_TyVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TmVar__shift_TyVar_0_Tm_comm_ind  :
      (forall (s0 : Tm) ,
        (forall (k7 : Hvl) (c9 : (Cutoff TmVar)) ,
          ((shift_TmVar_Tm (weakenCutoffTmVar c9 k7) (shift_TyVar_Tm (weakenCutoffTyVar C0 k7) s0)) =
          (shift_TyVar_Tm (weakenCutoffTyVar C0 k7) (shift_TmVar_Tm (weakenCutoffTmVar c9 k7) s0))))).
    Proof.
      pose proof (shift_TmVar__shift_TyVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TmVar__shift_TyVar_0_Ty_comm_ind  :
      (forall (S1 : Ty) ,
        (forall (k7 : Hvl) (c9 : (Cutoff TmVar)) ,
          ((shift_TmVar_Ty (weakenCutoffTmVar c9 k7) (shift_TyVar_Ty (weakenCutoffTyVar C0 k7) S1)) =
          (shift_TyVar_Ty (weakenCutoffTyVar C0 k7) (shift_TmVar_Ty (weakenCutoffTmVar c9 k7) S1))))).
    Proof.
      pose proof (shift_TmVar__shift_TyVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TmVar__shift_TyVar_0_Kind_comm_ind  :
      (forall (K7 : Kind) ,
        (forall (k7 : Hvl) (c9 : (Cutoff TmVar)) ,
          ((shift_TmVar_Kind (weakenCutoffTmVar c9 k7) (shift_TyVar_Kind (weakenCutoffTyVar C0 k7) K7)) =
          (shift_TyVar_Kind (weakenCutoffTyVar C0 k7) (shift_TmVar_Kind (weakenCutoffTmVar c9 k7) K7))))).
    Proof.
      pose proof (shift_TmVar__shift_TyVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_TyVar__shift_TmVar_0_Decls_Tm_Ty_Kind_comm_ind := (ind_Decls_Tm_Ty_Kind (fun (d17 : Decls) =>
      (forall (k7 : Hvl) (c9 : (Cutoff TyVar)) ,
        ((shift_TyVar_Decls (weakenCutoffTyVar c9 k7) (shift_TmVar_Decls (weakenCutoffTmVar C0 k7) d17)) =
        (shift_TmVar_Decls (weakenCutoffTmVar C0 k7) (shift_TyVar_Decls (weakenCutoffTyVar c9 k7) d17))))) (fun (s0 : Tm) =>
      (forall (k7 : Hvl) (c9 : (Cutoff TyVar)) ,
        ((shift_TyVar_Tm (weakenCutoffTyVar c9 k7) (shift_TmVar_Tm (weakenCutoffTmVar C0 k7) s0)) =
        (shift_TmVar_Tm (weakenCutoffTmVar C0 k7) (shift_TyVar_Tm (weakenCutoffTyVar c9 k7) s0))))) (fun (S1 : Ty) =>
      (forall (k7 : Hvl) (c9 : (Cutoff TyVar)) ,
        ((shift_TyVar_Ty (weakenCutoffTyVar c9 k7) (shift_TmVar_Ty (weakenCutoffTmVar C0 k7) S1)) =
        (shift_TmVar_Ty (weakenCutoffTmVar C0 k7) (shift_TyVar_Ty (weakenCutoffTyVar c9 k7) S1))))) (fun (K7 : Kind) =>
      (forall (k7 : Hvl) (c9 : (Cutoff TyVar)) ,
        ((shift_TyVar_Kind (weakenCutoffTyVar c9 k7) (shift_TmVar_Kind (weakenCutoffTmVar C0 k7) K7)) =
        (shift_TmVar_Kind (weakenCutoffTmVar C0 k7) (shift_TyVar_Kind (weakenCutoffTyVar c9 k7) K7))))) (fun (k7 : Hvl) (c9 : (Cutoff TyVar)) =>
      eq_refl) (fun (t : Tm) IHt23 (d : Decls) IHd (k7 : Hvl) (c9 : (Cutoff TyVar)) =>
      (f_equal2 DConsTm (IHt23 k7 c9) (IHd (HS TmVar k7) c9))) (fun (K : Kind) IHK7 (d : Decls) IHd (k7 : Hvl) (c9 : (Cutoff TyVar)) =>
      (f_equal2 DConsTy (IHK7 k7 c9) (IHd (HS TyVar k7) c9))) (fun (x23 : (Index TmVar)) =>
      (fun (k7 : Hvl) (c9 : (Cutoff TyVar)) =>
        eq_refl)) (fun (T : Ty) IHT19 (t : Tm) IHt23 (k7 : Hvl) (c9 : (Cutoff TyVar)) =>
      (f_equal2 Abs (IHT19 k7 c9) (IHt23 (HS TmVar k7) c9))) (fun (t1 : Tm) IHt23 (t2 : Tm) IHt24 (k7 : Hvl) (c9 : (Cutoff TyVar)) =>
      (f_equal2 App (IHt23 k7 c9) (IHt24 k7 c9))) (fun (d : Decls) IHd (t : Tm) IHt23 (k7 : Hvl) (c9 : (Cutoff TyVar)) =>
      (f_equal2 Let (IHd k7 c9) (eq_trans (f_equal2 shift_TyVar_Tm (eq_trans (f_equal2 weakenCutoffTyVar eq_refl (f_equal2 appendHvl (stability_shift_TmVar__bind _ _) (eq_refl H0))) (weakenCutoffTyVar_append c9 k7 (bind d))) (f_equal2 shift_TmVar_Tm (weakenCutoffTmVar_append C0 k7 (bind d)) eq_refl)) (eq_trans (IHt23 (appendHvl k7 (bind d)) c9) (f_equal2 shift_TmVar_Tm (eq_trans (eq_sym (weakenCutoffTmVar_append C0 k7 (bind d))) (f_equal2 weakenCutoffTmVar eq_refl (f_equal2 appendHvl (eq_sym (stability_shift_TyVar__bind _ _)) (eq_refl H0)))) (f_equal2 shift_TyVar_Tm (eq_sym (weakenCutoffTyVar_append c9 k7 (bind d))) eq_refl)))))) (fun (X11 : (Index TyVar)) =>
      (fun (k7 : Hvl) (c9 : (Cutoff TyVar)) =>
        eq_refl)) (fun (T1 : Ty) IHT19 (T2 : Ty) IHT20 (k7 : Hvl) (c9 : (Cutoff TyVar)) =>
      (f_equal2 TPi (IHT19 k7 c9) (IHT20 (HS TmVar k7) c9))) (fun (T : Ty) IHT19 (t : Tm) IHt23 (k7 : Hvl) (c9 : (Cutoff TyVar)) =>
      (f_equal2 TApp (IHT19 k7 c9) (IHt23 k7 c9))) (fun (k7 : Hvl) (c9 : (Cutoff TyVar)) =>
      eq_refl) (fun (T : Ty) IHT19 (K : Kind) IHK7 (k7 : Hvl) (c9 : (Cutoff TyVar)) =>
      (f_equal2 KPi (IHT19 k7 c9) (IHK7 (HS TmVar k7) c9)))).
    Lemma shift_TyVar__shift_TmVar_0_Decls_comm_ind  :
      (forall (d17 : Decls) ,
        (forall (k7 : Hvl) (c9 : (Cutoff TyVar)) ,
          ((shift_TyVar_Decls (weakenCutoffTyVar c9 k7) (shift_TmVar_Decls (weakenCutoffTmVar C0 k7) d17)) =
          (shift_TmVar_Decls (weakenCutoffTmVar C0 k7) (shift_TyVar_Decls (weakenCutoffTyVar c9 k7) d17))))).
    Proof.
      pose proof (shift_TyVar__shift_TmVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TyVar__shift_TmVar_0_Tm_comm_ind  :
      (forall (s0 : Tm) ,
        (forall (k7 : Hvl) (c9 : (Cutoff TyVar)) ,
          ((shift_TyVar_Tm (weakenCutoffTyVar c9 k7) (shift_TmVar_Tm (weakenCutoffTmVar C0 k7) s0)) =
          (shift_TmVar_Tm (weakenCutoffTmVar C0 k7) (shift_TyVar_Tm (weakenCutoffTyVar c9 k7) s0))))).
    Proof.
      pose proof (shift_TyVar__shift_TmVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TyVar__shift_TmVar_0_Ty_comm_ind  :
      (forall (S1 : Ty) ,
        (forall (k7 : Hvl) (c9 : (Cutoff TyVar)) ,
          ((shift_TyVar_Ty (weakenCutoffTyVar c9 k7) (shift_TmVar_Ty (weakenCutoffTmVar C0 k7) S1)) =
          (shift_TmVar_Ty (weakenCutoffTmVar C0 k7) (shift_TyVar_Ty (weakenCutoffTyVar c9 k7) S1))))).
    Proof.
      pose proof (shift_TyVar__shift_TmVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TyVar__shift_TmVar_0_Kind_comm_ind  :
      (forall (K7 : Kind) ,
        (forall (k7 : Hvl) (c9 : (Cutoff TyVar)) ,
          ((shift_TyVar_Kind (weakenCutoffTyVar c9 k7) (shift_TmVar_Kind (weakenCutoffTmVar C0 k7) K7)) =
          (shift_TmVar_Kind (weakenCutoffTmVar C0 k7) (shift_TyVar_Kind (weakenCutoffTyVar c9 k7) K7))))).
    Proof.
      pose proof (shift_TyVar__shift_TmVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_TyVar__shift_TyVar_0_Decls_Tm_Ty_Kind_comm_ind := (ind_Decls_Tm_Ty_Kind (fun (d17 : Decls) =>
      (forall (k7 : Hvl) (c9 : (Cutoff TyVar)) ,
        ((shift_TyVar_Decls (weakenCutoffTyVar (CS c9) k7) (shift_TyVar_Decls (weakenCutoffTyVar C0 k7) d17)) =
        (shift_TyVar_Decls (weakenCutoffTyVar C0 k7) (shift_TyVar_Decls (weakenCutoffTyVar c9 k7) d17))))) (fun (s0 : Tm) =>
      (forall (k7 : Hvl) (c9 : (Cutoff TyVar)) ,
        ((shift_TyVar_Tm (weakenCutoffTyVar (CS c9) k7) (shift_TyVar_Tm (weakenCutoffTyVar C0 k7) s0)) =
        (shift_TyVar_Tm (weakenCutoffTyVar C0 k7) (shift_TyVar_Tm (weakenCutoffTyVar c9 k7) s0))))) (fun (S1 : Ty) =>
      (forall (k7 : Hvl) (c9 : (Cutoff TyVar)) ,
        ((shift_TyVar_Ty (weakenCutoffTyVar (CS c9) k7) (shift_TyVar_Ty (weakenCutoffTyVar C0 k7) S1)) =
        (shift_TyVar_Ty (weakenCutoffTyVar C0 k7) (shift_TyVar_Ty (weakenCutoffTyVar c9 k7) S1))))) (fun (K7 : Kind) =>
      (forall (k7 : Hvl) (c9 : (Cutoff TyVar)) ,
        ((shift_TyVar_Kind (weakenCutoffTyVar (CS c9) k7) (shift_TyVar_Kind (weakenCutoffTyVar C0 k7) K7)) =
        (shift_TyVar_Kind (weakenCutoffTyVar C0 k7) (shift_TyVar_Kind (weakenCutoffTyVar c9 k7) K7))))) (fun (k7 : Hvl) (c9 : (Cutoff TyVar)) =>
      eq_refl) (fun (t : Tm) IHt23 (d : Decls) IHd (k7 : Hvl) (c9 : (Cutoff TyVar)) =>
      (f_equal2 DConsTm (IHt23 k7 c9) (IHd (HS TmVar k7) c9))) (fun (K : Kind) IHK7 (d : Decls) IHd (k7 : Hvl) (c9 : (Cutoff TyVar)) =>
      (f_equal2 DConsTy (IHK7 k7 c9) (IHd (HS TyVar k7) c9))) (fun (x23 : (Index TmVar)) =>
      (fun (k7 : Hvl) (c9 : (Cutoff TyVar)) =>
        eq_refl)) (fun (T : Ty) IHT19 (t : Tm) IHt23 (k7 : Hvl) (c9 : (Cutoff TyVar)) =>
      (f_equal2 Abs (IHT19 k7 c9) (IHt23 (HS TmVar k7) c9))) (fun (t1 : Tm) IHt23 (t2 : Tm) IHt24 (k7 : Hvl) (c9 : (Cutoff TyVar)) =>
      (f_equal2 App (IHt23 k7 c9) (IHt24 k7 c9))) (fun (d : Decls) IHd (t : Tm) IHt23 (k7 : Hvl) (c9 : (Cutoff TyVar)) =>
      (f_equal2 Let (IHd k7 c9) (eq_trans (f_equal2 shift_TyVar_Tm (eq_trans (f_equal2 weakenCutoffTyVar eq_refl (f_equal2 appendHvl (stability_shift_TyVar__bind _ _) (eq_refl H0))) (weakenCutoffTyVar_append (CS c9) k7 (bind d))) (f_equal2 shift_TyVar_Tm (weakenCutoffTyVar_append C0 k7 (bind d)) eq_refl)) (eq_trans (IHt23 (appendHvl k7 (bind d)) c9) (f_equal2 shift_TyVar_Tm (eq_trans (eq_sym (weakenCutoffTyVar_append C0 k7 (bind d))) (f_equal2 weakenCutoffTyVar eq_refl (f_equal2 appendHvl (eq_sym (stability_shift_TyVar__bind _ _)) (eq_refl H0)))) (f_equal2 shift_TyVar_Tm (eq_sym (weakenCutoffTyVar_append c9 k7 (bind d))) eq_refl)))))) (fun (X11 : (Index TyVar)) =>
      (fun (k7 : Hvl) (c9 : (Cutoff TyVar)) =>
        (f_equal TVar (shift_TyVar_Index_shift_TyVar_Index0_comm_ind k7 c9 X11)))) (fun (T1 : Ty) IHT19 (T2 : Ty) IHT20 (k7 : Hvl) (c9 : (Cutoff TyVar)) =>
      (f_equal2 TPi (IHT19 k7 c9) (IHT20 (HS TmVar k7) c9))) (fun (T : Ty) IHT19 (t : Tm) IHt23 (k7 : Hvl) (c9 : (Cutoff TyVar)) =>
      (f_equal2 TApp (IHT19 k7 c9) (IHt23 k7 c9))) (fun (k7 : Hvl) (c9 : (Cutoff TyVar)) =>
      eq_refl) (fun (T : Ty) IHT19 (K : Kind) IHK7 (k7 : Hvl) (c9 : (Cutoff TyVar)) =>
      (f_equal2 KPi (IHT19 k7 c9) (IHK7 (HS TmVar k7) c9)))).
    Lemma shift_TyVar__shift_TyVar_0_Decls_comm_ind  :
      (forall (d17 : Decls) ,
        (forall (k7 : Hvl) (c9 : (Cutoff TyVar)) ,
          ((shift_TyVar_Decls (weakenCutoffTyVar (CS c9) k7) (shift_TyVar_Decls (weakenCutoffTyVar C0 k7) d17)) =
          (shift_TyVar_Decls (weakenCutoffTyVar C0 k7) (shift_TyVar_Decls (weakenCutoffTyVar c9 k7) d17))))).
    Proof.
      pose proof (shift_TyVar__shift_TyVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TyVar__shift_TyVar_0_Tm_comm_ind  :
      (forall (s0 : Tm) ,
        (forall (k7 : Hvl) (c9 : (Cutoff TyVar)) ,
          ((shift_TyVar_Tm (weakenCutoffTyVar (CS c9) k7) (shift_TyVar_Tm (weakenCutoffTyVar C0 k7) s0)) =
          (shift_TyVar_Tm (weakenCutoffTyVar C0 k7) (shift_TyVar_Tm (weakenCutoffTyVar c9 k7) s0))))).
    Proof.
      pose proof (shift_TyVar__shift_TyVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TyVar__shift_TyVar_0_Ty_comm_ind  :
      (forall (S1 : Ty) ,
        (forall (k7 : Hvl) (c9 : (Cutoff TyVar)) ,
          ((shift_TyVar_Ty (weakenCutoffTyVar (CS c9) k7) (shift_TyVar_Ty (weakenCutoffTyVar C0 k7) S1)) =
          (shift_TyVar_Ty (weakenCutoffTyVar C0 k7) (shift_TyVar_Ty (weakenCutoffTyVar c9 k7) S1))))).
    Proof.
      pose proof (shift_TyVar__shift_TyVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TyVar__shift_TyVar_0_Kind_comm_ind  :
      (forall (K7 : Kind) ,
        (forall (k7 : Hvl) (c9 : (Cutoff TyVar)) ,
          ((shift_TyVar_Kind (weakenCutoffTyVar (CS c9) k7) (shift_TyVar_Kind (weakenCutoffTyVar C0 k7) K7)) =
          (shift_TyVar_Kind (weakenCutoffTyVar C0 k7) (shift_TyVar_Kind (weakenCutoffTyVar c9 k7) K7))))).
    Proof.
      pose proof (shift_TyVar__shift_TyVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
  End ShiftCommInd.
  Section ShiftComm.
    Definition shift_TmVar__shift_TmVar_0_Decls_comm (d17 : Decls) : (forall (c9 : (Cutoff TmVar)) ,
      ((shift_TmVar_Decls (CS c9) (shift_TmVar_Decls C0 d17)) =
      (shift_TmVar_Decls C0 (shift_TmVar_Decls c9 d17)))) := (shift_TmVar__shift_TmVar_0_Decls_comm_ind d17 H0).
    Definition shift_TmVar__shift_TmVar_0_Tm_comm (s0 : Tm) : (forall (c9 : (Cutoff TmVar)) ,
      ((shift_TmVar_Tm (CS c9) (shift_TmVar_Tm C0 s0)) =
      (shift_TmVar_Tm C0 (shift_TmVar_Tm c9 s0)))) := (shift_TmVar__shift_TmVar_0_Tm_comm_ind s0 H0).
    Definition shift_TmVar__shift_TmVar_0_Ty_comm (S1 : Ty) : (forall (c9 : (Cutoff TmVar)) ,
      ((shift_TmVar_Ty (CS c9) (shift_TmVar_Ty C0 S1)) =
      (shift_TmVar_Ty C0 (shift_TmVar_Ty c9 S1)))) := (shift_TmVar__shift_TmVar_0_Ty_comm_ind S1 H0).
    Definition shift_TmVar__shift_TmVar_0_Kind_comm (K7 : Kind) : (forall (c9 : (Cutoff TmVar)) ,
      ((shift_TmVar_Kind (CS c9) (shift_TmVar_Kind C0 K7)) =
      (shift_TmVar_Kind C0 (shift_TmVar_Kind c9 K7)))) := (shift_TmVar__shift_TmVar_0_Kind_comm_ind K7 H0).
    Definition shift_TmVar__shift_TyVar_0_Decls_comm (d17 : Decls) : (forall (c9 : (Cutoff TmVar)) ,
      ((shift_TmVar_Decls c9 (shift_TyVar_Decls C0 d17)) =
      (shift_TyVar_Decls C0 (shift_TmVar_Decls c9 d17)))) := (shift_TmVar__shift_TyVar_0_Decls_comm_ind d17 H0).
    Definition shift_TmVar__shift_TyVar_0_Tm_comm (s0 : Tm) : (forall (c9 : (Cutoff TmVar)) ,
      ((shift_TmVar_Tm c9 (shift_TyVar_Tm C0 s0)) =
      (shift_TyVar_Tm C0 (shift_TmVar_Tm c9 s0)))) := (shift_TmVar__shift_TyVar_0_Tm_comm_ind s0 H0).
    Definition shift_TmVar__shift_TyVar_0_Ty_comm (S1 : Ty) : (forall (c9 : (Cutoff TmVar)) ,
      ((shift_TmVar_Ty c9 (shift_TyVar_Ty C0 S1)) =
      (shift_TyVar_Ty C0 (shift_TmVar_Ty c9 S1)))) := (shift_TmVar__shift_TyVar_0_Ty_comm_ind S1 H0).
    Definition shift_TmVar__shift_TyVar_0_Kind_comm (K7 : Kind) : (forall (c9 : (Cutoff TmVar)) ,
      ((shift_TmVar_Kind c9 (shift_TyVar_Kind C0 K7)) =
      (shift_TyVar_Kind C0 (shift_TmVar_Kind c9 K7)))) := (shift_TmVar__shift_TyVar_0_Kind_comm_ind K7 H0).
    Definition shift_TyVar__shift_TmVar_0_Decls_comm (d17 : Decls) : (forall (c9 : (Cutoff TyVar)) ,
      ((shift_TyVar_Decls c9 (shift_TmVar_Decls C0 d17)) =
      (shift_TmVar_Decls C0 (shift_TyVar_Decls c9 d17)))) := (shift_TyVar__shift_TmVar_0_Decls_comm_ind d17 H0).
    Definition shift_TyVar__shift_TmVar_0_Tm_comm (s0 : Tm) : (forall (c9 : (Cutoff TyVar)) ,
      ((shift_TyVar_Tm c9 (shift_TmVar_Tm C0 s0)) =
      (shift_TmVar_Tm C0 (shift_TyVar_Tm c9 s0)))) := (shift_TyVar__shift_TmVar_0_Tm_comm_ind s0 H0).
    Definition shift_TyVar__shift_TmVar_0_Ty_comm (S1 : Ty) : (forall (c9 : (Cutoff TyVar)) ,
      ((shift_TyVar_Ty c9 (shift_TmVar_Ty C0 S1)) =
      (shift_TmVar_Ty C0 (shift_TyVar_Ty c9 S1)))) := (shift_TyVar__shift_TmVar_0_Ty_comm_ind S1 H0).
    Definition shift_TyVar__shift_TmVar_0_Kind_comm (K7 : Kind) : (forall (c9 : (Cutoff TyVar)) ,
      ((shift_TyVar_Kind c9 (shift_TmVar_Kind C0 K7)) =
      (shift_TmVar_Kind C0 (shift_TyVar_Kind c9 K7)))) := (shift_TyVar__shift_TmVar_0_Kind_comm_ind K7 H0).
    Definition shift_TyVar__shift_TyVar_0_Decls_comm (d17 : Decls) : (forall (c9 : (Cutoff TyVar)) ,
      ((shift_TyVar_Decls (CS c9) (shift_TyVar_Decls C0 d17)) =
      (shift_TyVar_Decls C0 (shift_TyVar_Decls c9 d17)))) := (shift_TyVar__shift_TyVar_0_Decls_comm_ind d17 H0).
    Definition shift_TyVar__shift_TyVar_0_Tm_comm (s0 : Tm) : (forall (c9 : (Cutoff TyVar)) ,
      ((shift_TyVar_Tm (CS c9) (shift_TyVar_Tm C0 s0)) =
      (shift_TyVar_Tm C0 (shift_TyVar_Tm c9 s0)))) := (shift_TyVar__shift_TyVar_0_Tm_comm_ind s0 H0).
    Definition shift_TyVar__shift_TyVar_0_Ty_comm (S1 : Ty) : (forall (c9 : (Cutoff TyVar)) ,
      ((shift_TyVar_Ty (CS c9) (shift_TyVar_Ty C0 S1)) =
      (shift_TyVar_Ty C0 (shift_TyVar_Ty c9 S1)))) := (shift_TyVar__shift_TyVar_0_Ty_comm_ind S1 H0).
    Definition shift_TyVar__shift_TyVar_0_Kind_comm (K7 : Kind) : (forall (c9 : (Cutoff TyVar)) ,
      ((shift_TyVar_Kind (CS c9) (shift_TyVar_Kind C0 K7)) =
      (shift_TyVar_Kind C0 (shift_TyVar_Kind c9 K7)))) := (shift_TyVar__shift_TyVar_0_Kind_comm_ind K7 H0).
  End ShiftComm.
End ShiftInteraction.
 Hint Rewrite shift_TmVar__shift_TmVar_0_Decls_comm shift_TmVar__shift_TyVar_0_Decls_comm shift_TyVar__shift_TmVar_0_Decls_comm shift_TyVar__shift_TyVar_0_Decls_comm shift_TmVar__shift_TmVar_0_Kind_comm shift_TmVar__shift_TyVar_0_Kind_comm shift_TyVar__shift_TmVar_0_Kind_comm shift_TyVar__shift_TyVar_0_Kind_comm shift_TmVar__shift_TmVar_0_Tm_comm shift_TmVar__shift_TyVar_0_Tm_comm shift_TyVar__shift_TmVar_0_Tm_comm shift_TyVar__shift_TyVar_0_Tm_comm shift_TmVar__shift_TmVar_0_Ty_comm shift_TmVar__shift_TyVar_0_Ty_comm shift_TyVar__shift_TmVar_0_Ty_comm shift_TyVar__shift_TyVar_0_Ty_comm : interaction.
 Hint Rewrite shift_TmVar__shift_TmVar_0_Decls_comm shift_TmVar__shift_TyVar_0_Decls_comm shift_TyVar__shift_TmVar_0_Decls_comm shift_TyVar__shift_TyVar_0_Decls_comm shift_TmVar__shift_TmVar_0_Kind_comm shift_TmVar__shift_TyVar_0_Kind_comm shift_TyVar__shift_TmVar_0_Kind_comm shift_TyVar__shift_TyVar_0_Kind_comm shift_TmVar__shift_TmVar_0_Tm_comm shift_TmVar__shift_TyVar_0_Tm_comm shift_TyVar__shift_TmVar_0_Tm_comm shift_TyVar__shift_TyVar_0_Tm_comm shift_TmVar__shift_TmVar_0_Ty_comm shift_TmVar__shift_TyVar_0_Ty_comm shift_TyVar__shift_TmVar_0_Ty_comm shift_TyVar__shift_TyVar_0_Ty_comm : shift_shift0.
Section WeakenShift.
  Lemma weakenDecls_shift_TmVar_Decls  :
    (forall (k71 : Hvl) (c17 : (Cutoff TmVar)) (d105 : Decls) ,
      ((weakenDecls (shift_TmVar_Decls c17 d105) k71) =
      (shift_TmVar_Decls (weakenCutoffTmVar c17 k71) (weakenDecls d105 k71)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenTm_shift_TmVar_Tm  :
    (forall (k71 : Hvl) (c17 : (Cutoff TmVar)) (s34 : Tm) ,
      ((weakenTm (shift_TmVar_Tm c17 s34) k71) =
      (shift_TmVar_Tm (weakenCutoffTmVar c17 k71) (weakenTm s34 k71)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenTy_shift_TmVar_Ty  :
    (forall (k71 : Hvl) (c17 : (Cutoff TmVar)) (S35 : Ty) ,
      ((weakenTy (shift_TmVar_Ty c17 S35) k71) =
      (shift_TmVar_Ty (weakenCutoffTmVar c17 k71) (weakenTy S35 k71)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenKind_shift_TmVar_Kind  :
    (forall (k71 : Hvl) (c17 : (Cutoff TmVar)) (K31 : Kind) ,
      ((weakenKind (shift_TmVar_Kind c17 K31) k71) =
      (shift_TmVar_Kind (weakenCutoffTmVar c17 k71) (weakenKind K31 k71)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenDecls_shift_TyVar_Decls  :
    (forall (k71 : Hvl) (c17 : (Cutoff TyVar)) (d105 : Decls) ,
      ((weakenDecls (shift_TyVar_Decls c17 d105) k71) =
      (shift_TyVar_Decls (weakenCutoffTyVar c17 k71) (weakenDecls d105 k71)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenTm_shift_TyVar_Tm  :
    (forall (k71 : Hvl) (c17 : (Cutoff TyVar)) (s34 : Tm) ,
      ((weakenTm (shift_TyVar_Tm c17 s34) k71) =
      (shift_TyVar_Tm (weakenCutoffTyVar c17 k71) (weakenTm s34 k71)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenTy_shift_TyVar_Ty  :
    (forall (k71 : Hvl) (c17 : (Cutoff TyVar)) (S35 : Ty) ,
      ((weakenTy (shift_TyVar_Ty c17 S35) k71) =
      (shift_TyVar_Ty (weakenCutoffTyVar c17 k71) (weakenTy S35 k71)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenKind_shift_TyVar_Kind  :
    (forall (k71 : Hvl) (c17 : (Cutoff TyVar)) (K31 : Kind) ,
      ((weakenKind (shift_TyVar_Kind c17 K31) k71) =
      (shift_TyVar_Kind (weakenCutoffTyVar c17 k71) (weakenKind K31 k71)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
End WeakenShift.
Section ShiftSubstInteraction.
  Section ShiftSubstIndexCommInd.
    Lemma shift_TmVar_Tm_subst_TmVar_Index0_comm_ind (c9 : (Cutoff TmVar)) (s2 : Tm) :
      (forall (k11 : Hvl) (x35 : (Index TmVar)) ,
        ((shift_TmVar_Tm (weakenCutoffTmVar c9 k11) (subst_TmVar_Index (weakenTrace X0 k11) s2 x35)) =
        (subst_TmVar_Index (weakenTrace X0 k11) (shift_TmVar_Tm c9 s2) (shift_TmVar_Index (weakenCutoffTmVar (CS c9) k11) x35)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma shift_TyVar_Tm_subst_TmVar_Index0_comm_ind (c9 : (Cutoff TyVar)) (s2 : Tm) :
      (forall (k11 : Hvl) (x35 : (Index TmVar)) ,
        ((shift_TyVar_Tm (weakenCutoffTyVar c9 k11) (subst_TmVar_Index (weakenTrace X0 k11) s2 x35)) =
        (subst_TmVar_Index (weakenTrace X0 k11) (shift_TyVar_Tm c9 s2) x35))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma shift_TmVar_Ty_subst_TyVar_Index0_comm_ind (c9 : (Cutoff TmVar)) (S3 : Ty) :
      (forall (k11 : Hvl) (X17 : (Index TyVar)) ,
        ((shift_TmVar_Ty (weakenCutoffTmVar c9 k11) (subst_TyVar_Index (weakenTrace X0 k11) S3 X17)) =
        (subst_TyVar_Index (weakenTrace X0 k11) (shift_TmVar_Ty c9 S3) X17))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma shift_TyVar_Ty_subst_TyVar_Index0_comm_ind (c9 : (Cutoff TyVar)) (S3 : Ty) :
      (forall (k11 : Hvl) (X17 : (Index TyVar)) ,
        ((shift_TyVar_Ty (weakenCutoffTyVar c9 k11) (subst_TyVar_Index (weakenTrace X0 k11) S3 X17)) =
        (subst_TyVar_Index (weakenTrace X0 k11) (shift_TyVar_Ty c9 S3) (shift_TyVar_Index (weakenCutoffTyVar (CS c9) k11) X17)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
  End ShiftSubstIndexCommInd.
  Section ShiftSubstCommInd.
    Definition shift_TmVar__subst_TmVar_0_Decls_Tm_Ty_Kind_comm_ind := (ind_Decls_Tm_Ty_Kind (fun (d35 : Decls) =>
      (forall (k19 : Hvl) (c17 : (Cutoff TmVar)) (s6 : Tm) ,
        ((shift_TmVar_Decls (weakenCutoffTmVar c17 k19) (subst_TmVar_Decls (weakenTrace X0 k19) s6 d35)) =
        (subst_TmVar_Decls (weakenTrace X0 k19) (shift_TmVar_Tm c17 s6) (shift_TmVar_Decls (weakenCutoffTmVar (CS c17) k19) d35))))) (fun (s6 : Tm) =>
      (forall (k19 : Hvl) (c17 : (Cutoff TmVar)) (s7 : Tm) ,
        ((shift_TmVar_Tm (weakenCutoffTmVar c17 k19) (subst_TmVar_Tm (weakenTrace X0 k19) s7 s6)) =
        (subst_TmVar_Tm (weakenTrace X0 k19) (shift_TmVar_Tm c17 s7) (shift_TmVar_Tm (weakenCutoffTmVar (CS c17) k19) s6))))) (fun (S7 : Ty) =>
      (forall (k19 : Hvl) (c17 : (Cutoff TmVar)) (s6 : Tm) ,
        ((shift_TmVar_Ty (weakenCutoffTmVar c17 k19) (subst_TmVar_Ty (weakenTrace X0 k19) s6 S7)) =
        (subst_TmVar_Ty (weakenTrace X0 k19) (shift_TmVar_Tm c17 s6) (shift_TmVar_Ty (weakenCutoffTmVar (CS c17) k19) S7))))) (fun (K19 : Kind) =>
      (forall (k19 : Hvl) (c17 : (Cutoff TmVar)) (s6 : Tm) ,
        ((shift_TmVar_Kind (weakenCutoffTmVar c17 k19) (subst_TmVar_Kind (weakenTrace X0 k19) s6 K19)) =
        (subst_TmVar_Kind (weakenTrace X0 k19) (shift_TmVar_Tm c17 s6) (shift_TmVar_Kind (weakenCutoffTmVar (CS c17) k19) K19))))) (fun (k19 : Hvl) (c17 : (Cutoff TmVar)) (s6 : Tm) =>
      eq_refl) (fun (t : Tm) IHt59 (d : Decls) IHd (k19 : Hvl) (c17 : (Cutoff TmVar)) (s6 : Tm) =>
      (f_equal2 DConsTm (IHt59 k19 c17 s6) (eq_trans (f_equal2 shift_TmVar_Decls eq_refl (f_equal3 subst_TmVar_Decls (weakenTrace_append TmVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHd (HS TmVar k19) c17 s6) (f_equal3 subst_TmVar_Decls (eq_sym (weakenTrace_append TmVar X0 k19 (HS TmVar H0))) eq_refl eq_refl))))) (fun (K : Kind) IHK19 (d : Decls) IHd (k19 : Hvl) (c17 : (Cutoff TmVar)) (s6 : Tm) =>
      (f_equal2 DConsTy (IHK19 k19 c17 s6) (eq_trans (f_equal2 shift_TmVar_Decls eq_refl (f_equal3 subst_TmVar_Decls (weakenTrace_append TmVar X0 k19 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHd (HS TyVar k19) c17 s6) (f_equal3 subst_TmVar_Decls (eq_sym (weakenTrace_append TmVar X0 k19 (HS TyVar H0))) eq_refl eq_refl))))) (fun (x59 : (Index TmVar)) =>
      (fun (k19 : Hvl) (c17 : (Cutoff TmVar)) (s6 : Tm) =>
        (shift_TmVar_Tm_subst_TmVar_Index0_comm_ind c17 s6 k19 x59))) (fun (T : Ty) IHT49 (t : Tm) IHt59 (k19 : Hvl) (c17 : (Cutoff TmVar)) (s6 : Tm) =>
      (f_equal2 Abs (IHT49 k19 c17 s6) (eq_trans (f_equal2 shift_TmVar_Tm eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt59 (HS TmVar k19) c17 s6) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k19 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt59 (t2 : Tm) IHt60 (k19 : Hvl) (c17 : (Cutoff TmVar)) (s6 : Tm) =>
      (f_equal2 App (IHt59 k19 c17 s6) (IHt60 k19 c17 s6))) (fun (d : Decls) IHd (t : Tm) IHt59 (k19 : Hvl) (c17 : (Cutoff TmVar)) (s6 : Tm) =>
      (f_equal2 Let (IHd k19 c17 s6) (eq_trans (f_equal2 shift_TmVar_Tm (eq_trans (f_equal2 weakenCutoffTmVar eq_refl (f_equal2 appendHvl (stability_subst_TmVar__bind _ _ _) (eq_refl H0))) (weakenCutoffTmVar_append c17 k19 (bind d))) (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k19 (bind d)) eq_refl eq_refl)) (eq_trans (IHt59 (appendHvl k19 (bind d)) c17 s6) (f_equal3 subst_TmVar_Tm (eq_trans (eq_sym (weakenTrace_append TmVar X0 k19 (bind d))) (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (eq_sym (stability_shift_TmVar__bind _ _)) (eq_refl H0)))) eq_refl (f_equal2 shift_TmVar_Tm (eq_sym (weakenCutoffTmVar_append (CS c17) k19 (bind d))) eq_refl)))))) (fun (X29 : (Index TyVar)) =>
      (fun (k19 : Hvl) (c17 : (Cutoff TmVar)) (s6 : Tm) =>
        eq_refl)) (fun (T1 : Ty) IHT49 (T2 : Ty) IHT50 (k19 : Hvl) (c17 : (Cutoff TmVar)) (s6 : Tm) =>
      (f_equal2 TPi (IHT49 k19 c17 s6) (eq_trans (f_equal2 shift_TmVar_Ty eq_refl (f_equal3 subst_TmVar_Ty (weakenTrace_append TmVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHT50 (HS TmVar k19) c17 s6) (f_equal3 subst_TmVar_Ty (eq_sym (weakenTrace_append TmVar X0 k19 (HS TmVar H0))) eq_refl eq_refl))))) (fun (T : Ty) IHT49 (t : Tm) IHt59 (k19 : Hvl) (c17 : (Cutoff TmVar)) (s6 : Tm) =>
      (f_equal2 TApp (IHT49 k19 c17 s6) (IHt59 k19 c17 s6))) (fun (k19 : Hvl) (c17 : (Cutoff TmVar)) (s6 : Tm) =>
      eq_refl) (fun (T : Ty) IHT49 (K : Kind) IHK19 (k19 : Hvl) (c17 : (Cutoff TmVar)) (s6 : Tm) =>
      (f_equal2 KPi (IHT49 k19 c17 s6) (eq_trans (f_equal2 shift_TmVar_Kind eq_refl (f_equal3 subst_TmVar_Kind (weakenTrace_append TmVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHK19 (HS TmVar k19) c17 s6) (f_equal3 subst_TmVar_Kind (eq_sym (weakenTrace_append TmVar X0 k19 (HS TmVar H0))) eq_refl eq_refl)))))).
    Lemma shift_TmVar__subst_TmVar_0_Decls_comm_ind  :
      (forall (d35 : Decls) ,
        (forall (k19 : Hvl) (c17 : (Cutoff TmVar)) (s6 : Tm) ,
          ((shift_TmVar_Decls (weakenCutoffTmVar c17 k19) (subst_TmVar_Decls (weakenTrace X0 k19) s6 d35)) =
          (subst_TmVar_Decls (weakenTrace X0 k19) (shift_TmVar_Tm c17 s6) (shift_TmVar_Decls (weakenCutoffTmVar (CS c17) k19) d35))))).
    Proof.
      pose proof (shift_TmVar__subst_TmVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TmVar__subst_TmVar_0_Tm_comm_ind  :
      (forall (s6 : Tm) ,
        (forall (k19 : Hvl) (c17 : (Cutoff TmVar)) (s7 : Tm) ,
          ((shift_TmVar_Tm (weakenCutoffTmVar c17 k19) (subst_TmVar_Tm (weakenTrace X0 k19) s7 s6)) =
          (subst_TmVar_Tm (weakenTrace X0 k19) (shift_TmVar_Tm c17 s7) (shift_TmVar_Tm (weakenCutoffTmVar (CS c17) k19) s6))))).
    Proof.
      pose proof (shift_TmVar__subst_TmVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TmVar__subst_TmVar_0_Ty_comm_ind  :
      (forall (S7 : Ty) ,
        (forall (k19 : Hvl) (c17 : (Cutoff TmVar)) (s7 : Tm) ,
          ((shift_TmVar_Ty (weakenCutoffTmVar c17 k19) (subst_TmVar_Ty (weakenTrace X0 k19) s7 S7)) =
          (subst_TmVar_Ty (weakenTrace X0 k19) (shift_TmVar_Tm c17 s7) (shift_TmVar_Ty (weakenCutoffTmVar (CS c17) k19) S7))))).
    Proof.
      pose proof (shift_TmVar__subst_TmVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TmVar__subst_TmVar_0_Kind_comm_ind  :
      (forall (K19 : Kind) ,
        (forall (k19 : Hvl) (c17 : (Cutoff TmVar)) (s7 : Tm) ,
          ((shift_TmVar_Kind (weakenCutoffTmVar c17 k19) (subst_TmVar_Kind (weakenTrace X0 k19) s7 K19)) =
          (subst_TmVar_Kind (weakenTrace X0 k19) (shift_TmVar_Tm c17 s7) (shift_TmVar_Kind (weakenCutoffTmVar (CS c17) k19) K19))))).
    Proof.
      pose proof (shift_TmVar__subst_TmVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_TmVar__subst_TyVar_0_Decls_Tm_Ty_Kind_comm_ind := (ind_Decls_Tm_Ty_Kind (fun (d35 : Decls) =>
      (forall (k19 : Hvl) (c17 : (Cutoff TmVar)) (S7 : Ty) ,
        ((shift_TmVar_Decls (weakenCutoffTmVar c17 k19) (subst_TyVar_Decls (weakenTrace X0 k19) S7 d35)) =
        (subst_TyVar_Decls (weakenTrace X0 k19) (shift_TmVar_Ty c17 S7) (shift_TmVar_Decls (weakenCutoffTmVar c17 k19) d35))))) (fun (s6 : Tm) =>
      (forall (k19 : Hvl) (c17 : (Cutoff TmVar)) (S7 : Ty) ,
        ((shift_TmVar_Tm (weakenCutoffTmVar c17 k19) (subst_TyVar_Tm (weakenTrace X0 k19) S7 s6)) =
        (subst_TyVar_Tm (weakenTrace X0 k19) (shift_TmVar_Ty c17 S7) (shift_TmVar_Tm (weakenCutoffTmVar c17 k19) s6))))) (fun (S7 : Ty) =>
      (forall (k19 : Hvl) (c17 : (Cutoff TmVar)) (S8 : Ty) ,
        ((shift_TmVar_Ty (weakenCutoffTmVar c17 k19) (subst_TyVar_Ty (weakenTrace X0 k19) S8 S7)) =
        (subst_TyVar_Ty (weakenTrace X0 k19) (shift_TmVar_Ty c17 S8) (shift_TmVar_Ty (weakenCutoffTmVar c17 k19) S7))))) (fun (K19 : Kind) =>
      (forall (k19 : Hvl) (c17 : (Cutoff TmVar)) (S7 : Ty) ,
        ((shift_TmVar_Kind (weakenCutoffTmVar c17 k19) (subst_TyVar_Kind (weakenTrace X0 k19) S7 K19)) =
        (subst_TyVar_Kind (weakenTrace X0 k19) (shift_TmVar_Ty c17 S7) (shift_TmVar_Kind (weakenCutoffTmVar c17 k19) K19))))) (fun (k19 : Hvl) (c17 : (Cutoff TmVar)) (S7 : Ty) =>
      eq_refl) (fun (t : Tm) IHt59 (d : Decls) IHd (k19 : Hvl) (c17 : (Cutoff TmVar)) (S7 : Ty) =>
      (f_equal2 DConsTm (IHt59 k19 c17 S7) (eq_trans (f_equal2 shift_TmVar_Decls eq_refl (f_equal3 subst_TyVar_Decls (weakenTrace_append TyVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHd (HS TmVar k19) c17 S7) (f_equal3 subst_TyVar_Decls (eq_sym (weakenTrace_append TyVar X0 k19 (HS TmVar H0))) eq_refl eq_refl))))) (fun (K : Kind) IHK19 (d : Decls) IHd (k19 : Hvl) (c17 : (Cutoff TmVar)) (S7 : Ty) =>
      (f_equal2 DConsTy (IHK19 k19 c17 S7) (eq_trans (f_equal2 shift_TmVar_Decls eq_refl (f_equal3 subst_TyVar_Decls (weakenTrace_append TyVar X0 k19 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHd (HS TyVar k19) c17 S7) (f_equal3 subst_TyVar_Decls (eq_sym (weakenTrace_append TyVar X0 k19 (HS TyVar H0))) eq_refl eq_refl))))) (fun (x59 : (Index TmVar)) =>
      (fun (k19 : Hvl) (c17 : (Cutoff TmVar)) (S7 : Ty) =>
        eq_refl)) (fun (T : Ty) IHT49 (t : Tm) IHt59 (k19 : Hvl) (c17 : (Cutoff TmVar)) (S7 : Ty) =>
      (f_equal2 Abs (IHT49 k19 c17 S7) (eq_trans (f_equal2 shift_TmVar_Tm eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt59 (HS TmVar k19) c17 S7) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k19 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt59 (t2 : Tm) IHt60 (k19 : Hvl) (c17 : (Cutoff TmVar)) (S7 : Ty) =>
      (f_equal2 App (IHt59 k19 c17 S7) (IHt60 k19 c17 S7))) (fun (d : Decls) IHd (t : Tm) IHt59 (k19 : Hvl) (c17 : (Cutoff TmVar)) (S7 : Ty) =>
      (f_equal2 Let (IHd k19 c17 S7) (eq_trans (f_equal2 shift_TmVar_Tm (eq_trans (f_equal2 weakenCutoffTmVar eq_refl (f_equal2 appendHvl (stability_subst_TyVar__bind _ _ _) (eq_refl H0))) (weakenCutoffTmVar_append c17 k19 (bind d))) (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k19 (bind d)) eq_refl eq_refl)) (eq_trans (IHt59 (appendHvl k19 (bind d)) c17 S7) (f_equal3 subst_TyVar_Tm (eq_trans (eq_sym (weakenTrace_append TyVar X0 k19 (bind d))) (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (eq_sym (stability_shift_TmVar__bind _ _)) (eq_refl H0)))) eq_refl (f_equal2 shift_TmVar_Tm (eq_sym (weakenCutoffTmVar_append c17 k19 (bind d))) eq_refl)))))) (fun (X29 : (Index TyVar)) =>
      (fun (k19 : Hvl) (c17 : (Cutoff TmVar)) (S7 : Ty) =>
        (shift_TmVar_Ty_subst_TyVar_Index0_comm_ind c17 S7 k19 X29))) (fun (T1 : Ty) IHT49 (T2 : Ty) IHT50 (k19 : Hvl) (c17 : (Cutoff TmVar)) (S7 : Ty) =>
      (f_equal2 TPi (IHT49 k19 c17 S7) (eq_trans (f_equal2 shift_TmVar_Ty eq_refl (f_equal3 subst_TyVar_Ty (weakenTrace_append TyVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHT50 (HS TmVar k19) c17 S7) (f_equal3 subst_TyVar_Ty (eq_sym (weakenTrace_append TyVar X0 k19 (HS TmVar H0))) eq_refl eq_refl))))) (fun (T : Ty) IHT49 (t : Tm) IHt59 (k19 : Hvl) (c17 : (Cutoff TmVar)) (S7 : Ty) =>
      (f_equal2 TApp (IHT49 k19 c17 S7) (IHt59 k19 c17 S7))) (fun (k19 : Hvl) (c17 : (Cutoff TmVar)) (S7 : Ty) =>
      eq_refl) (fun (T : Ty) IHT49 (K : Kind) IHK19 (k19 : Hvl) (c17 : (Cutoff TmVar)) (S7 : Ty) =>
      (f_equal2 KPi (IHT49 k19 c17 S7) (eq_trans (f_equal2 shift_TmVar_Kind eq_refl (f_equal3 subst_TyVar_Kind (weakenTrace_append TyVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHK19 (HS TmVar k19) c17 S7) (f_equal3 subst_TyVar_Kind (eq_sym (weakenTrace_append TyVar X0 k19 (HS TmVar H0))) eq_refl eq_refl)))))).
    Lemma shift_TmVar__subst_TyVar_0_Decls_comm_ind  :
      (forall (d35 : Decls) ,
        (forall (k19 : Hvl) (c17 : (Cutoff TmVar)) (S7 : Ty) ,
          ((shift_TmVar_Decls (weakenCutoffTmVar c17 k19) (subst_TyVar_Decls (weakenTrace X0 k19) S7 d35)) =
          (subst_TyVar_Decls (weakenTrace X0 k19) (shift_TmVar_Ty c17 S7) (shift_TmVar_Decls (weakenCutoffTmVar c17 k19) d35))))).
    Proof.
      pose proof (shift_TmVar__subst_TyVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TmVar__subst_TyVar_0_Tm_comm_ind  :
      (forall (s6 : Tm) ,
        (forall (k19 : Hvl) (c17 : (Cutoff TmVar)) (S7 : Ty) ,
          ((shift_TmVar_Tm (weakenCutoffTmVar c17 k19) (subst_TyVar_Tm (weakenTrace X0 k19) S7 s6)) =
          (subst_TyVar_Tm (weakenTrace X0 k19) (shift_TmVar_Ty c17 S7) (shift_TmVar_Tm (weakenCutoffTmVar c17 k19) s6))))).
    Proof.
      pose proof (shift_TmVar__subst_TyVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TmVar__subst_TyVar_0_Ty_comm_ind  :
      (forall (S7 : Ty) ,
        (forall (k19 : Hvl) (c17 : (Cutoff TmVar)) (S8 : Ty) ,
          ((shift_TmVar_Ty (weakenCutoffTmVar c17 k19) (subst_TyVar_Ty (weakenTrace X0 k19) S8 S7)) =
          (subst_TyVar_Ty (weakenTrace X0 k19) (shift_TmVar_Ty c17 S8) (shift_TmVar_Ty (weakenCutoffTmVar c17 k19) S7))))).
    Proof.
      pose proof (shift_TmVar__subst_TyVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TmVar__subst_TyVar_0_Kind_comm_ind  :
      (forall (K19 : Kind) ,
        (forall (k19 : Hvl) (c17 : (Cutoff TmVar)) (S8 : Ty) ,
          ((shift_TmVar_Kind (weakenCutoffTmVar c17 k19) (subst_TyVar_Kind (weakenTrace X0 k19) S8 K19)) =
          (subst_TyVar_Kind (weakenTrace X0 k19) (shift_TmVar_Ty c17 S8) (shift_TmVar_Kind (weakenCutoffTmVar c17 k19) K19))))).
    Proof.
      pose proof (shift_TmVar__subst_TyVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_TyVar__subst_TmVar_0_Decls_Tm_Ty_Kind_comm_ind := (ind_Decls_Tm_Ty_Kind (fun (d35 : Decls) =>
      (forall (k19 : Hvl) (c17 : (Cutoff TyVar)) (s6 : Tm) ,
        ((shift_TyVar_Decls (weakenCutoffTyVar c17 k19) (subst_TmVar_Decls (weakenTrace X0 k19) s6 d35)) =
        (subst_TmVar_Decls (weakenTrace X0 k19) (shift_TyVar_Tm c17 s6) (shift_TyVar_Decls (weakenCutoffTyVar c17 k19) d35))))) (fun (s6 : Tm) =>
      (forall (k19 : Hvl) (c17 : (Cutoff TyVar)) (s7 : Tm) ,
        ((shift_TyVar_Tm (weakenCutoffTyVar c17 k19) (subst_TmVar_Tm (weakenTrace X0 k19) s7 s6)) =
        (subst_TmVar_Tm (weakenTrace X0 k19) (shift_TyVar_Tm c17 s7) (shift_TyVar_Tm (weakenCutoffTyVar c17 k19) s6))))) (fun (S7 : Ty) =>
      (forall (k19 : Hvl) (c17 : (Cutoff TyVar)) (s6 : Tm) ,
        ((shift_TyVar_Ty (weakenCutoffTyVar c17 k19) (subst_TmVar_Ty (weakenTrace X0 k19) s6 S7)) =
        (subst_TmVar_Ty (weakenTrace X0 k19) (shift_TyVar_Tm c17 s6) (shift_TyVar_Ty (weakenCutoffTyVar c17 k19) S7))))) (fun (K19 : Kind) =>
      (forall (k19 : Hvl) (c17 : (Cutoff TyVar)) (s6 : Tm) ,
        ((shift_TyVar_Kind (weakenCutoffTyVar c17 k19) (subst_TmVar_Kind (weakenTrace X0 k19) s6 K19)) =
        (subst_TmVar_Kind (weakenTrace X0 k19) (shift_TyVar_Tm c17 s6) (shift_TyVar_Kind (weakenCutoffTyVar c17 k19) K19))))) (fun (k19 : Hvl) (c17 : (Cutoff TyVar)) (s6 : Tm) =>
      eq_refl) (fun (t : Tm) IHt59 (d : Decls) IHd (k19 : Hvl) (c17 : (Cutoff TyVar)) (s6 : Tm) =>
      (f_equal2 DConsTm (IHt59 k19 c17 s6) (eq_trans (f_equal2 shift_TyVar_Decls eq_refl (f_equal3 subst_TmVar_Decls (weakenTrace_append TmVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHd (HS TmVar k19) c17 s6) (f_equal3 subst_TmVar_Decls (eq_sym (weakenTrace_append TmVar X0 k19 (HS TmVar H0))) eq_refl eq_refl))))) (fun (K : Kind) IHK19 (d : Decls) IHd (k19 : Hvl) (c17 : (Cutoff TyVar)) (s6 : Tm) =>
      (f_equal2 DConsTy (IHK19 k19 c17 s6) (eq_trans (f_equal2 shift_TyVar_Decls eq_refl (f_equal3 subst_TmVar_Decls (weakenTrace_append TmVar X0 k19 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHd (HS TyVar k19) c17 s6) (f_equal3 subst_TmVar_Decls (eq_sym (weakenTrace_append TmVar X0 k19 (HS TyVar H0))) eq_refl eq_refl))))) (fun (x59 : (Index TmVar)) =>
      (fun (k19 : Hvl) (c17 : (Cutoff TyVar)) (s6 : Tm) =>
        (shift_TyVar_Tm_subst_TmVar_Index0_comm_ind c17 s6 k19 x59))) (fun (T : Ty) IHT49 (t : Tm) IHt59 (k19 : Hvl) (c17 : (Cutoff TyVar)) (s6 : Tm) =>
      (f_equal2 Abs (IHT49 k19 c17 s6) (eq_trans (f_equal2 shift_TyVar_Tm eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt59 (HS TmVar k19) c17 s6) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k19 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt59 (t2 : Tm) IHt60 (k19 : Hvl) (c17 : (Cutoff TyVar)) (s6 : Tm) =>
      (f_equal2 App (IHt59 k19 c17 s6) (IHt60 k19 c17 s6))) (fun (d : Decls) IHd (t : Tm) IHt59 (k19 : Hvl) (c17 : (Cutoff TyVar)) (s6 : Tm) =>
      (f_equal2 Let (IHd k19 c17 s6) (eq_trans (f_equal2 shift_TyVar_Tm (eq_trans (f_equal2 weakenCutoffTyVar eq_refl (f_equal2 appendHvl (stability_subst_TmVar__bind _ _ _) (eq_refl H0))) (weakenCutoffTyVar_append c17 k19 (bind d))) (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k19 (bind d)) eq_refl eq_refl)) (eq_trans (IHt59 (appendHvl k19 (bind d)) c17 s6) (f_equal3 subst_TmVar_Tm (eq_trans (eq_sym (weakenTrace_append TmVar X0 k19 (bind d))) (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (eq_sym (stability_shift_TyVar__bind _ _)) (eq_refl H0)))) eq_refl (f_equal2 shift_TyVar_Tm (eq_sym (weakenCutoffTyVar_append c17 k19 (bind d))) eq_refl)))))) (fun (X29 : (Index TyVar)) =>
      (fun (k19 : Hvl) (c17 : (Cutoff TyVar)) (s6 : Tm) =>
        eq_refl)) (fun (T1 : Ty) IHT49 (T2 : Ty) IHT50 (k19 : Hvl) (c17 : (Cutoff TyVar)) (s6 : Tm) =>
      (f_equal2 TPi (IHT49 k19 c17 s6) (eq_trans (f_equal2 shift_TyVar_Ty eq_refl (f_equal3 subst_TmVar_Ty (weakenTrace_append TmVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHT50 (HS TmVar k19) c17 s6) (f_equal3 subst_TmVar_Ty (eq_sym (weakenTrace_append TmVar X0 k19 (HS TmVar H0))) eq_refl eq_refl))))) (fun (T : Ty) IHT49 (t : Tm) IHt59 (k19 : Hvl) (c17 : (Cutoff TyVar)) (s6 : Tm) =>
      (f_equal2 TApp (IHT49 k19 c17 s6) (IHt59 k19 c17 s6))) (fun (k19 : Hvl) (c17 : (Cutoff TyVar)) (s6 : Tm) =>
      eq_refl) (fun (T : Ty) IHT49 (K : Kind) IHK19 (k19 : Hvl) (c17 : (Cutoff TyVar)) (s6 : Tm) =>
      (f_equal2 KPi (IHT49 k19 c17 s6) (eq_trans (f_equal2 shift_TyVar_Kind eq_refl (f_equal3 subst_TmVar_Kind (weakenTrace_append TmVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHK19 (HS TmVar k19) c17 s6) (f_equal3 subst_TmVar_Kind (eq_sym (weakenTrace_append TmVar X0 k19 (HS TmVar H0))) eq_refl eq_refl)))))).
    Lemma shift_TyVar__subst_TmVar_0_Decls_comm_ind  :
      (forall (d35 : Decls) ,
        (forall (k19 : Hvl) (c17 : (Cutoff TyVar)) (s6 : Tm) ,
          ((shift_TyVar_Decls (weakenCutoffTyVar c17 k19) (subst_TmVar_Decls (weakenTrace X0 k19) s6 d35)) =
          (subst_TmVar_Decls (weakenTrace X0 k19) (shift_TyVar_Tm c17 s6) (shift_TyVar_Decls (weakenCutoffTyVar c17 k19) d35))))).
    Proof.
      pose proof (shift_TyVar__subst_TmVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TyVar__subst_TmVar_0_Tm_comm_ind  :
      (forall (s6 : Tm) ,
        (forall (k19 : Hvl) (c17 : (Cutoff TyVar)) (s7 : Tm) ,
          ((shift_TyVar_Tm (weakenCutoffTyVar c17 k19) (subst_TmVar_Tm (weakenTrace X0 k19) s7 s6)) =
          (subst_TmVar_Tm (weakenTrace X0 k19) (shift_TyVar_Tm c17 s7) (shift_TyVar_Tm (weakenCutoffTyVar c17 k19) s6))))).
    Proof.
      pose proof (shift_TyVar__subst_TmVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TyVar__subst_TmVar_0_Ty_comm_ind  :
      (forall (S7 : Ty) ,
        (forall (k19 : Hvl) (c17 : (Cutoff TyVar)) (s7 : Tm) ,
          ((shift_TyVar_Ty (weakenCutoffTyVar c17 k19) (subst_TmVar_Ty (weakenTrace X0 k19) s7 S7)) =
          (subst_TmVar_Ty (weakenTrace X0 k19) (shift_TyVar_Tm c17 s7) (shift_TyVar_Ty (weakenCutoffTyVar c17 k19) S7))))).
    Proof.
      pose proof (shift_TyVar__subst_TmVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TyVar__subst_TmVar_0_Kind_comm_ind  :
      (forall (K19 : Kind) ,
        (forall (k19 : Hvl) (c17 : (Cutoff TyVar)) (s7 : Tm) ,
          ((shift_TyVar_Kind (weakenCutoffTyVar c17 k19) (subst_TmVar_Kind (weakenTrace X0 k19) s7 K19)) =
          (subst_TmVar_Kind (weakenTrace X0 k19) (shift_TyVar_Tm c17 s7) (shift_TyVar_Kind (weakenCutoffTyVar c17 k19) K19))))).
    Proof.
      pose proof (shift_TyVar__subst_TmVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_TyVar__subst_TyVar_0_Decls_Tm_Ty_Kind_comm_ind := (ind_Decls_Tm_Ty_Kind (fun (d35 : Decls) =>
      (forall (k19 : Hvl) (c17 : (Cutoff TyVar)) (S7 : Ty) ,
        ((shift_TyVar_Decls (weakenCutoffTyVar c17 k19) (subst_TyVar_Decls (weakenTrace X0 k19) S7 d35)) =
        (subst_TyVar_Decls (weakenTrace X0 k19) (shift_TyVar_Ty c17 S7) (shift_TyVar_Decls (weakenCutoffTyVar (CS c17) k19) d35))))) (fun (s6 : Tm) =>
      (forall (k19 : Hvl) (c17 : (Cutoff TyVar)) (S7 : Ty) ,
        ((shift_TyVar_Tm (weakenCutoffTyVar c17 k19) (subst_TyVar_Tm (weakenTrace X0 k19) S7 s6)) =
        (subst_TyVar_Tm (weakenTrace X0 k19) (shift_TyVar_Ty c17 S7) (shift_TyVar_Tm (weakenCutoffTyVar (CS c17) k19) s6))))) (fun (S7 : Ty) =>
      (forall (k19 : Hvl) (c17 : (Cutoff TyVar)) (S8 : Ty) ,
        ((shift_TyVar_Ty (weakenCutoffTyVar c17 k19) (subst_TyVar_Ty (weakenTrace X0 k19) S8 S7)) =
        (subst_TyVar_Ty (weakenTrace X0 k19) (shift_TyVar_Ty c17 S8) (shift_TyVar_Ty (weakenCutoffTyVar (CS c17) k19) S7))))) (fun (K19 : Kind) =>
      (forall (k19 : Hvl) (c17 : (Cutoff TyVar)) (S7 : Ty) ,
        ((shift_TyVar_Kind (weakenCutoffTyVar c17 k19) (subst_TyVar_Kind (weakenTrace X0 k19) S7 K19)) =
        (subst_TyVar_Kind (weakenTrace X0 k19) (shift_TyVar_Ty c17 S7) (shift_TyVar_Kind (weakenCutoffTyVar (CS c17) k19) K19))))) (fun (k19 : Hvl) (c17 : (Cutoff TyVar)) (S7 : Ty) =>
      eq_refl) (fun (t : Tm) IHt59 (d : Decls) IHd (k19 : Hvl) (c17 : (Cutoff TyVar)) (S7 : Ty) =>
      (f_equal2 DConsTm (IHt59 k19 c17 S7) (eq_trans (f_equal2 shift_TyVar_Decls eq_refl (f_equal3 subst_TyVar_Decls (weakenTrace_append TyVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHd (HS TmVar k19) c17 S7) (f_equal3 subst_TyVar_Decls (eq_sym (weakenTrace_append TyVar X0 k19 (HS TmVar H0))) eq_refl eq_refl))))) (fun (K : Kind) IHK19 (d : Decls) IHd (k19 : Hvl) (c17 : (Cutoff TyVar)) (S7 : Ty) =>
      (f_equal2 DConsTy (IHK19 k19 c17 S7) (eq_trans (f_equal2 shift_TyVar_Decls eq_refl (f_equal3 subst_TyVar_Decls (weakenTrace_append TyVar X0 k19 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHd (HS TyVar k19) c17 S7) (f_equal3 subst_TyVar_Decls (eq_sym (weakenTrace_append TyVar X0 k19 (HS TyVar H0))) eq_refl eq_refl))))) (fun (x59 : (Index TmVar)) =>
      (fun (k19 : Hvl) (c17 : (Cutoff TyVar)) (S7 : Ty) =>
        eq_refl)) (fun (T : Ty) IHT49 (t : Tm) IHt59 (k19 : Hvl) (c17 : (Cutoff TyVar)) (S7 : Ty) =>
      (f_equal2 Abs (IHT49 k19 c17 S7) (eq_trans (f_equal2 shift_TyVar_Tm eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt59 (HS TmVar k19) c17 S7) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k19 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt59 (t2 : Tm) IHt60 (k19 : Hvl) (c17 : (Cutoff TyVar)) (S7 : Ty) =>
      (f_equal2 App (IHt59 k19 c17 S7) (IHt60 k19 c17 S7))) (fun (d : Decls) IHd (t : Tm) IHt59 (k19 : Hvl) (c17 : (Cutoff TyVar)) (S7 : Ty) =>
      (f_equal2 Let (IHd k19 c17 S7) (eq_trans (f_equal2 shift_TyVar_Tm (eq_trans (f_equal2 weakenCutoffTyVar eq_refl (f_equal2 appendHvl (stability_subst_TyVar__bind _ _ _) (eq_refl H0))) (weakenCutoffTyVar_append c17 k19 (bind d))) (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k19 (bind d)) eq_refl eq_refl)) (eq_trans (IHt59 (appendHvl k19 (bind d)) c17 S7) (f_equal3 subst_TyVar_Tm (eq_trans (eq_sym (weakenTrace_append TyVar X0 k19 (bind d))) (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (eq_sym (stability_shift_TyVar__bind _ _)) (eq_refl H0)))) eq_refl (f_equal2 shift_TyVar_Tm (eq_sym (weakenCutoffTyVar_append (CS c17) k19 (bind d))) eq_refl)))))) (fun (X29 : (Index TyVar)) =>
      (fun (k19 : Hvl) (c17 : (Cutoff TyVar)) (S7 : Ty) =>
        (shift_TyVar_Ty_subst_TyVar_Index0_comm_ind c17 S7 k19 X29))) (fun (T1 : Ty) IHT49 (T2 : Ty) IHT50 (k19 : Hvl) (c17 : (Cutoff TyVar)) (S7 : Ty) =>
      (f_equal2 TPi (IHT49 k19 c17 S7) (eq_trans (f_equal2 shift_TyVar_Ty eq_refl (f_equal3 subst_TyVar_Ty (weakenTrace_append TyVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHT50 (HS TmVar k19) c17 S7) (f_equal3 subst_TyVar_Ty (eq_sym (weakenTrace_append TyVar X0 k19 (HS TmVar H0))) eq_refl eq_refl))))) (fun (T : Ty) IHT49 (t : Tm) IHt59 (k19 : Hvl) (c17 : (Cutoff TyVar)) (S7 : Ty) =>
      (f_equal2 TApp (IHT49 k19 c17 S7) (IHt59 k19 c17 S7))) (fun (k19 : Hvl) (c17 : (Cutoff TyVar)) (S7 : Ty) =>
      eq_refl) (fun (T : Ty) IHT49 (K : Kind) IHK19 (k19 : Hvl) (c17 : (Cutoff TyVar)) (S7 : Ty) =>
      (f_equal2 KPi (IHT49 k19 c17 S7) (eq_trans (f_equal2 shift_TyVar_Kind eq_refl (f_equal3 subst_TyVar_Kind (weakenTrace_append TyVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHK19 (HS TmVar k19) c17 S7) (f_equal3 subst_TyVar_Kind (eq_sym (weakenTrace_append TyVar X0 k19 (HS TmVar H0))) eq_refl eq_refl)))))).
    Lemma shift_TyVar__subst_TyVar_0_Decls_comm_ind  :
      (forall (d35 : Decls) ,
        (forall (k19 : Hvl) (c17 : (Cutoff TyVar)) (S7 : Ty) ,
          ((shift_TyVar_Decls (weakenCutoffTyVar c17 k19) (subst_TyVar_Decls (weakenTrace X0 k19) S7 d35)) =
          (subst_TyVar_Decls (weakenTrace X0 k19) (shift_TyVar_Ty c17 S7) (shift_TyVar_Decls (weakenCutoffTyVar (CS c17) k19) d35))))).
    Proof.
      pose proof (shift_TyVar__subst_TyVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TyVar__subst_TyVar_0_Tm_comm_ind  :
      (forall (s6 : Tm) ,
        (forall (k19 : Hvl) (c17 : (Cutoff TyVar)) (S7 : Ty) ,
          ((shift_TyVar_Tm (weakenCutoffTyVar c17 k19) (subst_TyVar_Tm (weakenTrace X0 k19) S7 s6)) =
          (subst_TyVar_Tm (weakenTrace X0 k19) (shift_TyVar_Ty c17 S7) (shift_TyVar_Tm (weakenCutoffTyVar (CS c17) k19) s6))))).
    Proof.
      pose proof (shift_TyVar__subst_TyVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TyVar__subst_TyVar_0_Ty_comm_ind  :
      (forall (S7 : Ty) ,
        (forall (k19 : Hvl) (c17 : (Cutoff TyVar)) (S8 : Ty) ,
          ((shift_TyVar_Ty (weakenCutoffTyVar c17 k19) (subst_TyVar_Ty (weakenTrace X0 k19) S8 S7)) =
          (subst_TyVar_Ty (weakenTrace X0 k19) (shift_TyVar_Ty c17 S8) (shift_TyVar_Ty (weakenCutoffTyVar (CS c17) k19) S7))))).
    Proof.
      pose proof (shift_TyVar__subst_TyVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TyVar__subst_TyVar_0_Kind_comm_ind  :
      (forall (K19 : Kind) ,
        (forall (k19 : Hvl) (c17 : (Cutoff TyVar)) (S8 : Ty) ,
          ((shift_TyVar_Kind (weakenCutoffTyVar c17 k19) (subst_TyVar_Kind (weakenTrace X0 k19) S8 K19)) =
          (subst_TyVar_Kind (weakenTrace X0 k19) (shift_TyVar_Ty c17 S8) (shift_TyVar_Kind (weakenCutoffTyVar (CS c17) k19) K19))))).
    Proof.
      pose proof (shift_TyVar__subst_TyVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
  End ShiftSubstCommInd.
  Section ShiftSubstComm.
    Definition shift_TmVar_Decls_subst_TmVar_Decls0_comm (d35 : Decls) : (forall (c17 : (Cutoff TmVar)) (s6 : Tm) ,
      ((shift_TmVar_Decls c17 (subst_TmVar_Decls X0 s6 d35)) =
      (subst_TmVar_Decls X0 (shift_TmVar_Tm c17 s6) (shift_TmVar_Decls (CS c17) d35)))) := (shift_TmVar__subst_TmVar_0_Decls_comm_ind d35 H0).
    Definition shift_TmVar_Tm_subst_TmVar_Tm0_comm (s7 : Tm) : (forall (c17 : (Cutoff TmVar)) (s6 : Tm) ,
      ((shift_TmVar_Tm c17 (subst_TmVar_Tm X0 s6 s7)) =
      (subst_TmVar_Tm X0 (shift_TmVar_Tm c17 s6) (shift_TmVar_Tm (CS c17) s7)))) := (shift_TmVar__subst_TmVar_0_Tm_comm_ind s7 H0).
    Definition shift_TmVar_Ty_subst_TmVar_Ty0_comm (S7 : Ty) : (forall (c17 : (Cutoff TmVar)) (s6 : Tm) ,
      ((shift_TmVar_Ty c17 (subst_TmVar_Ty X0 s6 S7)) =
      (subst_TmVar_Ty X0 (shift_TmVar_Tm c17 s6) (shift_TmVar_Ty (CS c17) S7)))) := (shift_TmVar__subst_TmVar_0_Ty_comm_ind S7 H0).
    Definition shift_TmVar_Kind_subst_TmVar_Kind0_comm (K19 : Kind) : (forall (c17 : (Cutoff TmVar)) (s6 : Tm) ,
      ((shift_TmVar_Kind c17 (subst_TmVar_Kind X0 s6 K19)) =
      (subst_TmVar_Kind X0 (shift_TmVar_Tm c17 s6) (shift_TmVar_Kind (CS c17) K19)))) := (shift_TmVar__subst_TmVar_0_Kind_comm_ind K19 H0).
    Definition shift_TmVar_Decls_subst_TyVar_Decls0_comm (d35 : Decls) : (forall (c17 : (Cutoff TmVar)) (S7 : Ty) ,
      ((shift_TmVar_Decls c17 (subst_TyVar_Decls X0 S7 d35)) =
      (subst_TyVar_Decls X0 (shift_TmVar_Ty c17 S7) (shift_TmVar_Decls c17 d35)))) := (shift_TmVar__subst_TyVar_0_Decls_comm_ind d35 H0).
    Definition shift_TmVar_Tm_subst_TyVar_Tm0_comm (s6 : Tm) : (forall (c17 : (Cutoff TmVar)) (S7 : Ty) ,
      ((shift_TmVar_Tm c17 (subst_TyVar_Tm X0 S7 s6)) =
      (subst_TyVar_Tm X0 (shift_TmVar_Ty c17 S7) (shift_TmVar_Tm c17 s6)))) := (shift_TmVar__subst_TyVar_0_Tm_comm_ind s6 H0).
    Definition shift_TmVar_Ty_subst_TyVar_Ty0_comm (S8 : Ty) : (forall (c17 : (Cutoff TmVar)) (S7 : Ty) ,
      ((shift_TmVar_Ty c17 (subst_TyVar_Ty X0 S7 S8)) =
      (subst_TyVar_Ty X0 (shift_TmVar_Ty c17 S7) (shift_TmVar_Ty c17 S8)))) := (shift_TmVar__subst_TyVar_0_Ty_comm_ind S8 H0).
    Definition shift_TmVar_Kind_subst_TyVar_Kind0_comm (K19 : Kind) : (forall (c17 : (Cutoff TmVar)) (S7 : Ty) ,
      ((shift_TmVar_Kind c17 (subst_TyVar_Kind X0 S7 K19)) =
      (subst_TyVar_Kind X0 (shift_TmVar_Ty c17 S7) (shift_TmVar_Kind c17 K19)))) := (shift_TmVar__subst_TyVar_0_Kind_comm_ind K19 H0).
    Definition shift_TyVar_Decls_subst_TmVar_Decls0_comm (d35 : Decls) : (forall (c17 : (Cutoff TyVar)) (s6 : Tm) ,
      ((shift_TyVar_Decls c17 (subst_TmVar_Decls X0 s6 d35)) =
      (subst_TmVar_Decls X0 (shift_TyVar_Tm c17 s6) (shift_TyVar_Decls c17 d35)))) := (shift_TyVar__subst_TmVar_0_Decls_comm_ind d35 H0).
    Definition shift_TyVar_Tm_subst_TmVar_Tm0_comm (s7 : Tm) : (forall (c17 : (Cutoff TyVar)) (s6 : Tm) ,
      ((shift_TyVar_Tm c17 (subst_TmVar_Tm X0 s6 s7)) =
      (subst_TmVar_Tm X0 (shift_TyVar_Tm c17 s6) (shift_TyVar_Tm c17 s7)))) := (shift_TyVar__subst_TmVar_0_Tm_comm_ind s7 H0).
    Definition shift_TyVar_Ty_subst_TmVar_Ty0_comm (S7 : Ty) : (forall (c17 : (Cutoff TyVar)) (s6 : Tm) ,
      ((shift_TyVar_Ty c17 (subst_TmVar_Ty X0 s6 S7)) =
      (subst_TmVar_Ty X0 (shift_TyVar_Tm c17 s6) (shift_TyVar_Ty c17 S7)))) := (shift_TyVar__subst_TmVar_0_Ty_comm_ind S7 H0).
    Definition shift_TyVar_Kind_subst_TmVar_Kind0_comm (K19 : Kind) : (forall (c17 : (Cutoff TyVar)) (s6 : Tm) ,
      ((shift_TyVar_Kind c17 (subst_TmVar_Kind X0 s6 K19)) =
      (subst_TmVar_Kind X0 (shift_TyVar_Tm c17 s6) (shift_TyVar_Kind c17 K19)))) := (shift_TyVar__subst_TmVar_0_Kind_comm_ind K19 H0).
    Definition shift_TyVar_Decls_subst_TyVar_Decls0_comm (d35 : Decls) : (forall (c17 : (Cutoff TyVar)) (S7 : Ty) ,
      ((shift_TyVar_Decls c17 (subst_TyVar_Decls X0 S7 d35)) =
      (subst_TyVar_Decls X0 (shift_TyVar_Ty c17 S7) (shift_TyVar_Decls (CS c17) d35)))) := (shift_TyVar__subst_TyVar_0_Decls_comm_ind d35 H0).
    Definition shift_TyVar_Tm_subst_TyVar_Tm0_comm (s6 : Tm) : (forall (c17 : (Cutoff TyVar)) (S7 : Ty) ,
      ((shift_TyVar_Tm c17 (subst_TyVar_Tm X0 S7 s6)) =
      (subst_TyVar_Tm X0 (shift_TyVar_Ty c17 S7) (shift_TyVar_Tm (CS c17) s6)))) := (shift_TyVar__subst_TyVar_0_Tm_comm_ind s6 H0).
    Definition shift_TyVar_Ty_subst_TyVar_Ty0_comm (S8 : Ty) : (forall (c17 : (Cutoff TyVar)) (S7 : Ty) ,
      ((shift_TyVar_Ty c17 (subst_TyVar_Ty X0 S7 S8)) =
      (subst_TyVar_Ty X0 (shift_TyVar_Ty c17 S7) (shift_TyVar_Ty (CS c17) S8)))) := (shift_TyVar__subst_TyVar_0_Ty_comm_ind S8 H0).
    Definition shift_TyVar_Kind_subst_TyVar_Kind0_comm (K19 : Kind) : (forall (c17 : (Cutoff TyVar)) (S7 : Ty) ,
      ((shift_TyVar_Kind c17 (subst_TyVar_Kind X0 S7 K19)) =
      (subst_TyVar_Kind X0 (shift_TyVar_Ty c17 S7) (shift_TyVar_Kind (CS c17) K19)))) := (shift_TyVar__subst_TyVar_0_Kind_comm_ind K19 H0).
  End ShiftSubstComm.
  Section SubstShiftIndexCommInd.
    Lemma subst_TmVar_Index_shift_TmVar_Tm0_comm_ind (d35 : (Trace TmVar)) (s6 : Tm) :
      (forall (k19 : Hvl) (x59 : (Index TmVar)) ,
        ((subst_TmVar_Index (weakenTrace (XS TmVar d35) k19) s6 (shift_TmVar_Index (weakenCutoffTmVar C0 k19) x59)) =
        (shift_TmVar_Tm (weakenCutoffTmVar C0 k19) (subst_TmVar_Index (weakenTrace d35 k19) s6 x59)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma subst_TmVar_Index_shift_TyVar_Tm0_comm_ind (d35 : (Trace TmVar)) (s6 : Tm) :
      (forall (k19 : Hvl) (x59 : (Index TmVar)) ,
        ((subst_TmVar_Index (weakenTrace (XS TyVar d35) k19) s6 x59) =
        (shift_TyVar_Tm (weakenCutoffTyVar C0 k19) (subst_TmVar_Index (weakenTrace d35 k19) s6 x59)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma subst_TyVar_Index_shift_TmVar_Ty0_comm_ind (d35 : (Trace TyVar)) (S7 : Ty) :
      (forall (k19 : Hvl) (X29 : (Index TyVar)) ,
        ((subst_TyVar_Index (weakenTrace (XS TmVar d35) k19) S7 X29) =
        (shift_TmVar_Ty (weakenCutoffTmVar C0 k19) (subst_TyVar_Index (weakenTrace d35 k19) S7 X29)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma subst_TyVar_Index_shift_TyVar_Ty0_comm_ind (d35 : (Trace TyVar)) (S7 : Ty) :
      (forall (k19 : Hvl) (X29 : (Index TyVar)) ,
        ((subst_TyVar_Index (weakenTrace (XS TyVar d35) k19) S7 (shift_TyVar_Index (weakenCutoffTyVar C0 k19) X29)) =
        (shift_TyVar_Ty (weakenCutoffTyVar C0 k19) (subst_TyVar_Index (weakenTrace d35 k19) S7 X29)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
  End SubstShiftIndexCommInd.
  Section SubstShiftCommInd.
    Definition subst_TmVar__shift_TmVar_0_Decls_Tm_Ty_Kind_comm_ind := (ind_Decls_Tm_Ty_Kind (fun (d95 : Decls) =>
      (forall (k67 : Hvl) (d96 : (Trace TmVar)) (s30 : Tm) ,
        ((subst_TmVar_Decls (weakenTrace (XS TmVar d96) k67) s30 (shift_TmVar_Decls (weakenCutoffTmVar C0 k67) d95)) =
        (shift_TmVar_Decls (weakenCutoffTmVar C0 k67) (subst_TmVar_Decls (weakenTrace d96 k67) s30 d95))))) (fun (s30 : Tm) =>
      (forall (k67 : Hvl) (d95 : (Trace TmVar)) (s31 : Tm) ,
        ((subst_TmVar_Tm (weakenTrace (XS TmVar d95) k67) s31 (shift_TmVar_Tm (weakenCutoffTmVar C0 k67) s30)) =
        (shift_TmVar_Tm (weakenCutoffTmVar C0 k67) (subst_TmVar_Tm (weakenTrace d95 k67) s31 s30))))) (fun (S31 : Ty) =>
      (forall (k67 : Hvl) (d95 : (Trace TmVar)) (s30 : Tm) ,
        ((subst_TmVar_Ty (weakenTrace (XS TmVar d95) k67) s30 (shift_TmVar_Ty (weakenCutoffTmVar C0 k67) S31)) =
        (shift_TmVar_Ty (weakenCutoffTmVar C0 k67) (subst_TmVar_Ty (weakenTrace d95 k67) s30 S31))))) (fun (K27 : Kind) =>
      (forall (k67 : Hvl) (d95 : (Trace TmVar)) (s30 : Tm) ,
        ((subst_TmVar_Kind (weakenTrace (XS TmVar d95) k67) s30 (shift_TmVar_Kind (weakenCutoffTmVar C0 k67) K27)) =
        (shift_TmVar_Kind (weakenCutoffTmVar C0 k67) (subst_TmVar_Kind (weakenTrace d95 k67) s30 K27))))) (fun (k67 : Hvl) (d95 : (Trace TmVar)) (s30 : Tm) =>
      eq_refl) (fun (t : Tm) IHt83 (d : Decls) IHd (k67 : Hvl) (d95 : (Trace TmVar)) (s30 : Tm) =>
      (f_equal2 DConsTm (IHt83 k67 d95 s30) (eq_trans (f_equal3 subst_TmVar_Decls (weakenTrace_append TmVar (XS TmVar d95) k67 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHd (HS TmVar k67) d95 s30) (f_equal2 shift_TmVar_Decls eq_refl (f_equal3 subst_TmVar_Decls (eq_sym (weakenTrace_append TmVar d95 k67 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (K : Kind) IHK27 (d : Decls) IHd (k67 : Hvl) (d95 : (Trace TmVar)) (s30 : Tm) =>
      (f_equal2 DConsTy (IHK27 k67 d95 s30) (eq_trans (f_equal3 subst_TmVar_Decls (weakenTrace_append TmVar (XS TmVar d95) k67 (HS TyVar H0)) eq_refl eq_refl) (eq_trans (IHd (HS TyVar k67) d95 s30) (f_equal2 shift_TmVar_Decls eq_refl (f_equal3 subst_TmVar_Decls (eq_sym (weakenTrace_append TmVar d95 k67 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (x83 : (Index TmVar)) =>
      (fun (k67 : Hvl) (d95 : (Trace TmVar)) (s30 : Tm) =>
        (subst_TmVar_Index_shift_TmVar_Tm0_comm_ind d95 s30 k67 x83))) (fun (T : Ty) IHT69 (t : Tm) IHt83 (k67 : Hvl) (d95 : (Trace TmVar)) (s30 : Tm) =>
      (f_equal2 Abs (IHT69 k67 d95 s30) (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar (XS TmVar d95) k67 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt83 (HS TmVar k67) d95 s30) (f_equal2 shift_TmVar_Tm eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar d95 k67 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt83 (t2 : Tm) IHt84 (k67 : Hvl) (d95 : (Trace TmVar)) (s30 : Tm) =>
      (f_equal2 App (IHt83 k67 d95 s30) (IHt84 k67 d95 s30))) (fun (d : Decls) IHd (t : Tm) IHt83 (k67 : Hvl) (d95 : (Trace TmVar)) (s30 : Tm) =>
      (f_equal2 Let (IHd k67 d95 s30) (eq_trans (f_equal3 subst_TmVar_Tm (eq_trans (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (stability_shift_TmVar__bind _ _) (eq_refl H0))) (weakenTrace_append TmVar (XS TmVar d95) k67 (bind d))) eq_refl (f_equal2 shift_TmVar_Tm (weakenCutoffTmVar_append C0 k67 (bind d)) eq_refl)) (eq_trans (IHt83 (appendHvl k67 (bind d)) d95 s30) (f_equal2 shift_TmVar_Tm (eq_trans (eq_sym (weakenCutoffTmVar_append C0 k67 (bind d))) (f_equal2 weakenCutoffTmVar eq_refl (f_equal2 appendHvl (eq_sym (stability_subst_TmVar__bind _ _ _)) (eq_refl H0)))) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar d95 k67 (bind d))) eq_refl eq_refl)))))) (fun (X41 : (Index TyVar)) =>
      (fun (k67 : Hvl) (d95 : (Trace TmVar)) (s30 : Tm) =>
        eq_refl)) (fun (T1 : Ty) IHT69 (T2 : Ty) IHT70 (k67 : Hvl) (d95 : (Trace TmVar)) (s30 : Tm) =>
      (f_equal2 TPi (IHT69 k67 d95 s30) (eq_trans (f_equal3 subst_TmVar_Ty (weakenTrace_append TmVar (XS TmVar d95) k67 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHT70 (HS TmVar k67) d95 s30) (f_equal2 shift_TmVar_Ty eq_refl (f_equal3 subst_TmVar_Ty (eq_sym (weakenTrace_append TmVar d95 k67 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (T : Ty) IHT69 (t : Tm) IHt83 (k67 : Hvl) (d95 : (Trace TmVar)) (s30 : Tm) =>
      (f_equal2 TApp (IHT69 k67 d95 s30) (IHt83 k67 d95 s30))) (fun (k67 : Hvl) (d95 : (Trace TmVar)) (s30 : Tm) =>
      eq_refl) (fun (T : Ty) IHT69 (K : Kind) IHK27 (k67 : Hvl) (d95 : (Trace TmVar)) (s30 : Tm) =>
      (f_equal2 KPi (IHT69 k67 d95 s30) (eq_trans (f_equal3 subst_TmVar_Kind (weakenTrace_append TmVar (XS TmVar d95) k67 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHK27 (HS TmVar k67) d95 s30) (f_equal2 shift_TmVar_Kind eq_refl (f_equal3 subst_TmVar_Kind (eq_sym (weakenTrace_append TmVar d95 k67 (HS TmVar H0))) eq_refl eq_refl))))))).
    Lemma subst_TmVar__shift_TmVar_0_Decls_comm_ind  :
      (forall (d95 : Decls) ,
        (forall (k67 : Hvl) (d96 : (Trace TmVar)) (s30 : Tm) ,
          ((subst_TmVar_Decls (weakenTrace (XS TmVar d96) k67) s30 (shift_TmVar_Decls (weakenCutoffTmVar C0 k67) d95)) =
          (shift_TmVar_Decls (weakenCutoffTmVar C0 k67) (subst_TmVar_Decls (weakenTrace d96 k67) s30 d95))))).
    Proof.
      pose proof (subst_TmVar__shift_TmVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TmVar__shift_TmVar_0_Tm_comm_ind  :
      (forall (s30 : Tm) ,
        (forall (k67 : Hvl) (d96 : (Trace TmVar)) (s31 : Tm) ,
          ((subst_TmVar_Tm (weakenTrace (XS TmVar d96) k67) s31 (shift_TmVar_Tm (weakenCutoffTmVar C0 k67) s30)) =
          (shift_TmVar_Tm (weakenCutoffTmVar C0 k67) (subst_TmVar_Tm (weakenTrace d96 k67) s31 s30))))).
    Proof.
      pose proof (subst_TmVar__shift_TmVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TmVar__shift_TmVar_0_Ty_comm_ind  :
      (forall (S31 : Ty) ,
        (forall (k67 : Hvl) (d96 : (Trace TmVar)) (s31 : Tm) ,
          ((subst_TmVar_Ty (weakenTrace (XS TmVar d96) k67) s31 (shift_TmVar_Ty (weakenCutoffTmVar C0 k67) S31)) =
          (shift_TmVar_Ty (weakenCutoffTmVar C0 k67) (subst_TmVar_Ty (weakenTrace d96 k67) s31 S31))))).
    Proof.
      pose proof (subst_TmVar__shift_TmVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TmVar__shift_TmVar_0_Kind_comm_ind  :
      (forall (K27 : Kind) ,
        (forall (k67 : Hvl) (d96 : (Trace TmVar)) (s31 : Tm) ,
          ((subst_TmVar_Kind (weakenTrace (XS TmVar d96) k67) s31 (shift_TmVar_Kind (weakenCutoffTmVar C0 k67) K27)) =
          (shift_TmVar_Kind (weakenCutoffTmVar C0 k67) (subst_TmVar_Kind (weakenTrace d96 k67) s31 K27))))).
    Proof.
      pose proof (subst_TmVar__shift_TmVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_TmVar__shift_TyVar_0_Decls_Tm_Ty_Kind_comm_ind := (ind_Decls_Tm_Ty_Kind (fun (d95 : Decls) =>
      (forall (k67 : Hvl) (d96 : (Trace TmVar)) (s30 : Tm) ,
        ((subst_TmVar_Decls (weakenTrace (XS TyVar d96) k67) s30 (shift_TyVar_Decls (weakenCutoffTyVar C0 k67) d95)) =
        (shift_TyVar_Decls (weakenCutoffTyVar C0 k67) (subst_TmVar_Decls (weakenTrace d96 k67) s30 d95))))) (fun (s30 : Tm) =>
      (forall (k67 : Hvl) (d95 : (Trace TmVar)) (s31 : Tm) ,
        ((subst_TmVar_Tm (weakenTrace (XS TyVar d95) k67) s31 (shift_TyVar_Tm (weakenCutoffTyVar C0 k67) s30)) =
        (shift_TyVar_Tm (weakenCutoffTyVar C0 k67) (subst_TmVar_Tm (weakenTrace d95 k67) s31 s30))))) (fun (S31 : Ty) =>
      (forall (k67 : Hvl) (d95 : (Trace TmVar)) (s30 : Tm) ,
        ((subst_TmVar_Ty (weakenTrace (XS TyVar d95) k67) s30 (shift_TyVar_Ty (weakenCutoffTyVar C0 k67) S31)) =
        (shift_TyVar_Ty (weakenCutoffTyVar C0 k67) (subst_TmVar_Ty (weakenTrace d95 k67) s30 S31))))) (fun (K27 : Kind) =>
      (forall (k67 : Hvl) (d95 : (Trace TmVar)) (s30 : Tm) ,
        ((subst_TmVar_Kind (weakenTrace (XS TyVar d95) k67) s30 (shift_TyVar_Kind (weakenCutoffTyVar C0 k67) K27)) =
        (shift_TyVar_Kind (weakenCutoffTyVar C0 k67) (subst_TmVar_Kind (weakenTrace d95 k67) s30 K27))))) (fun (k67 : Hvl) (d95 : (Trace TmVar)) (s30 : Tm) =>
      eq_refl) (fun (t : Tm) IHt83 (d : Decls) IHd (k67 : Hvl) (d95 : (Trace TmVar)) (s30 : Tm) =>
      (f_equal2 DConsTm (IHt83 k67 d95 s30) (eq_trans (f_equal3 subst_TmVar_Decls (weakenTrace_append TmVar (XS TyVar d95) k67 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHd (HS TmVar k67) d95 s30) (f_equal2 shift_TyVar_Decls eq_refl (f_equal3 subst_TmVar_Decls (eq_sym (weakenTrace_append TmVar d95 k67 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (K : Kind) IHK27 (d : Decls) IHd (k67 : Hvl) (d95 : (Trace TmVar)) (s30 : Tm) =>
      (f_equal2 DConsTy (IHK27 k67 d95 s30) (eq_trans (f_equal3 subst_TmVar_Decls (weakenTrace_append TmVar (XS TyVar d95) k67 (HS TyVar H0)) eq_refl eq_refl) (eq_trans (IHd (HS TyVar k67) d95 s30) (f_equal2 shift_TyVar_Decls eq_refl (f_equal3 subst_TmVar_Decls (eq_sym (weakenTrace_append TmVar d95 k67 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (x83 : (Index TmVar)) =>
      (fun (k67 : Hvl) (d95 : (Trace TmVar)) (s30 : Tm) =>
        (subst_TmVar_Index_shift_TyVar_Tm0_comm_ind d95 s30 k67 x83))) (fun (T : Ty) IHT69 (t : Tm) IHt83 (k67 : Hvl) (d95 : (Trace TmVar)) (s30 : Tm) =>
      (f_equal2 Abs (IHT69 k67 d95 s30) (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar (XS TyVar d95) k67 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt83 (HS TmVar k67) d95 s30) (f_equal2 shift_TyVar_Tm eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar d95 k67 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt83 (t2 : Tm) IHt84 (k67 : Hvl) (d95 : (Trace TmVar)) (s30 : Tm) =>
      (f_equal2 App (IHt83 k67 d95 s30) (IHt84 k67 d95 s30))) (fun (d : Decls) IHd (t : Tm) IHt83 (k67 : Hvl) (d95 : (Trace TmVar)) (s30 : Tm) =>
      (f_equal2 Let (IHd k67 d95 s30) (eq_trans (f_equal3 subst_TmVar_Tm (eq_trans (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (stability_shift_TyVar__bind _ _) (eq_refl H0))) (weakenTrace_append TmVar (XS TyVar d95) k67 (bind d))) eq_refl (f_equal2 shift_TyVar_Tm (weakenCutoffTyVar_append C0 k67 (bind d)) eq_refl)) (eq_trans (IHt83 (appendHvl k67 (bind d)) d95 s30) (f_equal2 shift_TyVar_Tm (eq_trans (eq_sym (weakenCutoffTyVar_append C0 k67 (bind d))) (f_equal2 weakenCutoffTyVar eq_refl (f_equal2 appendHvl (eq_sym (stability_subst_TmVar__bind _ _ _)) (eq_refl H0)))) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar d95 k67 (bind d))) eq_refl eq_refl)))))) (fun (X41 : (Index TyVar)) =>
      (fun (k67 : Hvl) (d95 : (Trace TmVar)) (s30 : Tm) =>
        eq_refl)) (fun (T1 : Ty) IHT69 (T2 : Ty) IHT70 (k67 : Hvl) (d95 : (Trace TmVar)) (s30 : Tm) =>
      (f_equal2 TPi (IHT69 k67 d95 s30) (eq_trans (f_equal3 subst_TmVar_Ty (weakenTrace_append TmVar (XS TyVar d95) k67 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHT70 (HS TmVar k67) d95 s30) (f_equal2 shift_TyVar_Ty eq_refl (f_equal3 subst_TmVar_Ty (eq_sym (weakenTrace_append TmVar d95 k67 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (T : Ty) IHT69 (t : Tm) IHt83 (k67 : Hvl) (d95 : (Trace TmVar)) (s30 : Tm) =>
      (f_equal2 TApp (IHT69 k67 d95 s30) (IHt83 k67 d95 s30))) (fun (k67 : Hvl) (d95 : (Trace TmVar)) (s30 : Tm) =>
      eq_refl) (fun (T : Ty) IHT69 (K : Kind) IHK27 (k67 : Hvl) (d95 : (Trace TmVar)) (s30 : Tm) =>
      (f_equal2 KPi (IHT69 k67 d95 s30) (eq_trans (f_equal3 subst_TmVar_Kind (weakenTrace_append TmVar (XS TyVar d95) k67 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHK27 (HS TmVar k67) d95 s30) (f_equal2 shift_TyVar_Kind eq_refl (f_equal3 subst_TmVar_Kind (eq_sym (weakenTrace_append TmVar d95 k67 (HS TmVar H0))) eq_refl eq_refl))))))).
    Lemma subst_TmVar__shift_TyVar_0_Decls_comm_ind  :
      (forall (d95 : Decls) ,
        (forall (k67 : Hvl) (d96 : (Trace TmVar)) (s30 : Tm) ,
          ((subst_TmVar_Decls (weakenTrace (XS TyVar d96) k67) s30 (shift_TyVar_Decls (weakenCutoffTyVar C0 k67) d95)) =
          (shift_TyVar_Decls (weakenCutoffTyVar C0 k67) (subst_TmVar_Decls (weakenTrace d96 k67) s30 d95))))).
    Proof.
      pose proof (subst_TmVar__shift_TyVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TmVar__shift_TyVar_0_Tm_comm_ind  :
      (forall (s30 : Tm) ,
        (forall (k67 : Hvl) (d96 : (Trace TmVar)) (s31 : Tm) ,
          ((subst_TmVar_Tm (weakenTrace (XS TyVar d96) k67) s31 (shift_TyVar_Tm (weakenCutoffTyVar C0 k67) s30)) =
          (shift_TyVar_Tm (weakenCutoffTyVar C0 k67) (subst_TmVar_Tm (weakenTrace d96 k67) s31 s30))))).
    Proof.
      pose proof (subst_TmVar__shift_TyVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TmVar__shift_TyVar_0_Ty_comm_ind  :
      (forall (S31 : Ty) ,
        (forall (k67 : Hvl) (d96 : (Trace TmVar)) (s31 : Tm) ,
          ((subst_TmVar_Ty (weakenTrace (XS TyVar d96) k67) s31 (shift_TyVar_Ty (weakenCutoffTyVar C0 k67) S31)) =
          (shift_TyVar_Ty (weakenCutoffTyVar C0 k67) (subst_TmVar_Ty (weakenTrace d96 k67) s31 S31))))).
    Proof.
      pose proof (subst_TmVar__shift_TyVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TmVar__shift_TyVar_0_Kind_comm_ind  :
      (forall (K27 : Kind) ,
        (forall (k67 : Hvl) (d96 : (Trace TmVar)) (s31 : Tm) ,
          ((subst_TmVar_Kind (weakenTrace (XS TyVar d96) k67) s31 (shift_TyVar_Kind (weakenCutoffTyVar C0 k67) K27)) =
          (shift_TyVar_Kind (weakenCutoffTyVar C0 k67) (subst_TmVar_Kind (weakenTrace d96 k67) s31 K27))))).
    Proof.
      pose proof (subst_TmVar__shift_TyVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_TyVar__shift_TmVar_0_Decls_Tm_Ty_Kind_comm_ind := (ind_Decls_Tm_Ty_Kind (fun (d95 : Decls) =>
      (forall (k67 : Hvl) (d96 : (Trace TyVar)) (S31 : Ty) ,
        ((subst_TyVar_Decls (weakenTrace (XS TmVar d96) k67) S31 (shift_TmVar_Decls (weakenCutoffTmVar C0 k67) d95)) =
        (shift_TmVar_Decls (weakenCutoffTmVar C0 k67) (subst_TyVar_Decls (weakenTrace d96 k67) S31 d95))))) (fun (s30 : Tm) =>
      (forall (k67 : Hvl) (d95 : (Trace TyVar)) (S31 : Ty) ,
        ((subst_TyVar_Tm (weakenTrace (XS TmVar d95) k67) S31 (shift_TmVar_Tm (weakenCutoffTmVar C0 k67) s30)) =
        (shift_TmVar_Tm (weakenCutoffTmVar C0 k67) (subst_TyVar_Tm (weakenTrace d95 k67) S31 s30))))) (fun (S31 : Ty) =>
      (forall (k67 : Hvl) (d95 : (Trace TyVar)) (S32 : Ty) ,
        ((subst_TyVar_Ty (weakenTrace (XS TmVar d95) k67) S32 (shift_TmVar_Ty (weakenCutoffTmVar C0 k67) S31)) =
        (shift_TmVar_Ty (weakenCutoffTmVar C0 k67) (subst_TyVar_Ty (weakenTrace d95 k67) S32 S31))))) (fun (K27 : Kind) =>
      (forall (k67 : Hvl) (d95 : (Trace TyVar)) (S31 : Ty) ,
        ((subst_TyVar_Kind (weakenTrace (XS TmVar d95) k67) S31 (shift_TmVar_Kind (weakenCutoffTmVar C0 k67) K27)) =
        (shift_TmVar_Kind (weakenCutoffTmVar C0 k67) (subst_TyVar_Kind (weakenTrace d95 k67) S31 K27))))) (fun (k67 : Hvl) (d95 : (Trace TyVar)) (S31 : Ty) =>
      eq_refl) (fun (t : Tm) IHt83 (d : Decls) IHd (k67 : Hvl) (d95 : (Trace TyVar)) (S31 : Ty) =>
      (f_equal2 DConsTm (IHt83 k67 d95 S31) (eq_trans (f_equal3 subst_TyVar_Decls (weakenTrace_append TyVar (XS TmVar d95) k67 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHd (HS TmVar k67) d95 S31) (f_equal2 shift_TmVar_Decls eq_refl (f_equal3 subst_TyVar_Decls (eq_sym (weakenTrace_append TyVar d95 k67 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (K : Kind) IHK27 (d : Decls) IHd (k67 : Hvl) (d95 : (Trace TyVar)) (S31 : Ty) =>
      (f_equal2 DConsTy (IHK27 k67 d95 S31) (eq_trans (f_equal3 subst_TyVar_Decls (weakenTrace_append TyVar (XS TmVar d95) k67 (HS TyVar H0)) eq_refl eq_refl) (eq_trans (IHd (HS TyVar k67) d95 S31) (f_equal2 shift_TmVar_Decls eq_refl (f_equal3 subst_TyVar_Decls (eq_sym (weakenTrace_append TyVar d95 k67 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (x83 : (Index TmVar)) =>
      (fun (k67 : Hvl) (d95 : (Trace TyVar)) (S31 : Ty) =>
        eq_refl)) (fun (T : Ty) IHT69 (t : Tm) IHt83 (k67 : Hvl) (d95 : (Trace TyVar)) (S31 : Ty) =>
      (f_equal2 Abs (IHT69 k67 d95 S31) (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar (XS TmVar d95) k67 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt83 (HS TmVar k67) d95 S31) (f_equal2 shift_TmVar_Tm eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d95 k67 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt83 (t2 : Tm) IHt84 (k67 : Hvl) (d95 : (Trace TyVar)) (S31 : Ty) =>
      (f_equal2 App (IHt83 k67 d95 S31) (IHt84 k67 d95 S31))) (fun (d : Decls) IHd (t : Tm) IHt83 (k67 : Hvl) (d95 : (Trace TyVar)) (S31 : Ty) =>
      (f_equal2 Let (IHd k67 d95 S31) (eq_trans (f_equal3 subst_TyVar_Tm (eq_trans (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (stability_shift_TmVar__bind _ _) (eq_refl H0))) (weakenTrace_append TyVar (XS TmVar d95) k67 (bind d))) eq_refl (f_equal2 shift_TmVar_Tm (weakenCutoffTmVar_append C0 k67 (bind d)) eq_refl)) (eq_trans (IHt83 (appendHvl k67 (bind d)) d95 S31) (f_equal2 shift_TmVar_Tm (eq_trans (eq_sym (weakenCutoffTmVar_append C0 k67 (bind d))) (f_equal2 weakenCutoffTmVar eq_refl (f_equal2 appendHvl (eq_sym (stability_subst_TyVar__bind _ _ _)) (eq_refl H0)))) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d95 k67 (bind d))) eq_refl eq_refl)))))) (fun (X41 : (Index TyVar)) =>
      (fun (k67 : Hvl) (d95 : (Trace TyVar)) (S31 : Ty) =>
        (subst_TyVar_Index_shift_TmVar_Ty0_comm_ind d95 S31 k67 X41))) (fun (T1 : Ty) IHT69 (T2 : Ty) IHT70 (k67 : Hvl) (d95 : (Trace TyVar)) (S31 : Ty) =>
      (f_equal2 TPi (IHT69 k67 d95 S31) (eq_trans (f_equal3 subst_TyVar_Ty (weakenTrace_append TyVar (XS TmVar d95) k67 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHT70 (HS TmVar k67) d95 S31) (f_equal2 shift_TmVar_Ty eq_refl (f_equal3 subst_TyVar_Ty (eq_sym (weakenTrace_append TyVar d95 k67 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (T : Ty) IHT69 (t : Tm) IHt83 (k67 : Hvl) (d95 : (Trace TyVar)) (S31 : Ty) =>
      (f_equal2 TApp (IHT69 k67 d95 S31) (IHt83 k67 d95 S31))) (fun (k67 : Hvl) (d95 : (Trace TyVar)) (S31 : Ty) =>
      eq_refl) (fun (T : Ty) IHT69 (K : Kind) IHK27 (k67 : Hvl) (d95 : (Trace TyVar)) (S31 : Ty) =>
      (f_equal2 KPi (IHT69 k67 d95 S31) (eq_trans (f_equal3 subst_TyVar_Kind (weakenTrace_append TyVar (XS TmVar d95) k67 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHK27 (HS TmVar k67) d95 S31) (f_equal2 shift_TmVar_Kind eq_refl (f_equal3 subst_TyVar_Kind (eq_sym (weakenTrace_append TyVar d95 k67 (HS TmVar H0))) eq_refl eq_refl))))))).
    Lemma subst_TyVar__shift_TmVar_0_Decls_comm_ind  :
      (forall (d95 : Decls) ,
        (forall (k67 : Hvl) (d96 : (Trace TyVar)) (S31 : Ty) ,
          ((subst_TyVar_Decls (weakenTrace (XS TmVar d96) k67) S31 (shift_TmVar_Decls (weakenCutoffTmVar C0 k67) d95)) =
          (shift_TmVar_Decls (weakenCutoffTmVar C0 k67) (subst_TyVar_Decls (weakenTrace d96 k67) S31 d95))))).
    Proof.
      pose proof (subst_TyVar__shift_TmVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TyVar__shift_TmVar_0_Tm_comm_ind  :
      (forall (s30 : Tm) ,
        (forall (k67 : Hvl) (d96 : (Trace TyVar)) (S31 : Ty) ,
          ((subst_TyVar_Tm (weakenTrace (XS TmVar d96) k67) S31 (shift_TmVar_Tm (weakenCutoffTmVar C0 k67) s30)) =
          (shift_TmVar_Tm (weakenCutoffTmVar C0 k67) (subst_TyVar_Tm (weakenTrace d96 k67) S31 s30))))).
    Proof.
      pose proof (subst_TyVar__shift_TmVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TyVar__shift_TmVar_0_Ty_comm_ind  :
      (forall (S31 : Ty) ,
        (forall (k67 : Hvl) (d96 : (Trace TyVar)) (S32 : Ty) ,
          ((subst_TyVar_Ty (weakenTrace (XS TmVar d96) k67) S32 (shift_TmVar_Ty (weakenCutoffTmVar C0 k67) S31)) =
          (shift_TmVar_Ty (weakenCutoffTmVar C0 k67) (subst_TyVar_Ty (weakenTrace d96 k67) S32 S31))))).
    Proof.
      pose proof (subst_TyVar__shift_TmVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TyVar__shift_TmVar_0_Kind_comm_ind  :
      (forall (K27 : Kind) ,
        (forall (k67 : Hvl) (d96 : (Trace TyVar)) (S32 : Ty) ,
          ((subst_TyVar_Kind (weakenTrace (XS TmVar d96) k67) S32 (shift_TmVar_Kind (weakenCutoffTmVar C0 k67) K27)) =
          (shift_TmVar_Kind (weakenCutoffTmVar C0 k67) (subst_TyVar_Kind (weakenTrace d96 k67) S32 K27))))).
    Proof.
      pose proof (subst_TyVar__shift_TmVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_TyVar__shift_TyVar_0_Decls_Tm_Ty_Kind_comm_ind := (ind_Decls_Tm_Ty_Kind (fun (d95 : Decls) =>
      (forall (k67 : Hvl) (d96 : (Trace TyVar)) (S31 : Ty) ,
        ((subst_TyVar_Decls (weakenTrace (XS TyVar d96) k67) S31 (shift_TyVar_Decls (weakenCutoffTyVar C0 k67) d95)) =
        (shift_TyVar_Decls (weakenCutoffTyVar C0 k67) (subst_TyVar_Decls (weakenTrace d96 k67) S31 d95))))) (fun (s30 : Tm) =>
      (forall (k67 : Hvl) (d95 : (Trace TyVar)) (S31 : Ty) ,
        ((subst_TyVar_Tm (weakenTrace (XS TyVar d95) k67) S31 (shift_TyVar_Tm (weakenCutoffTyVar C0 k67) s30)) =
        (shift_TyVar_Tm (weakenCutoffTyVar C0 k67) (subst_TyVar_Tm (weakenTrace d95 k67) S31 s30))))) (fun (S31 : Ty) =>
      (forall (k67 : Hvl) (d95 : (Trace TyVar)) (S32 : Ty) ,
        ((subst_TyVar_Ty (weakenTrace (XS TyVar d95) k67) S32 (shift_TyVar_Ty (weakenCutoffTyVar C0 k67) S31)) =
        (shift_TyVar_Ty (weakenCutoffTyVar C0 k67) (subst_TyVar_Ty (weakenTrace d95 k67) S32 S31))))) (fun (K27 : Kind) =>
      (forall (k67 : Hvl) (d95 : (Trace TyVar)) (S31 : Ty) ,
        ((subst_TyVar_Kind (weakenTrace (XS TyVar d95) k67) S31 (shift_TyVar_Kind (weakenCutoffTyVar C0 k67) K27)) =
        (shift_TyVar_Kind (weakenCutoffTyVar C0 k67) (subst_TyVar_Kind (weakenTrace d95 k67) S31 K27))))) (fun (k67 : Hvl) (d95 : (Trace TyVar)) (S31 : Ty) =>
      eq_refl) (fun (t : Tm) IHt83 (d : Decls) IHd (k67 : Hvl) (d95 : (Trace TyVar)) (S31 : Ty) =>
      (f_equal2 DConsTm (IHt83 k67 d95 S31) (eq_trans (f_equal3 subst_TyVar_Decls (weakenTrace_append TyVar (XS TyVar d95) k67 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHd (HS TmVar k67) d95 S31) (f_equal2 shift_TyVar_Decls eq_refl (f_equal3 subst_TyVar_Decls (eq_sym (weakenTrace_append TyVar d95 k67 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (K : Kind) IHK27 (d : Decls) IHd (k67 : Hvl) (d95 : (Trace TyVar)) (S31 : Ty) =>
      (f_equal2 DConsTy (IHK27 k67 d95 S31) (eq_trans (f_equal3 subst_TyVar_Decls (weakenTrace_append TyVar (XS TyVar d95) k67 (HS TyVar H0)) eq_refl eq_refl) (eq_trans (IHd (HS TyVar k67) d95 S31) (f_equal2 shift_TyVar_Decls eq_refl (f_equal3 subst_TyVar_Decls (eq_sym (weakenTrace_append TyVar d95 k67 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (x83 : (Index TmVar)) =>
      (fun (k67 : Hvl) (d95 : (Trace TyVar)) (S31 : Ty) =>
        eq_refl)) (fun (T : Ty) IHT69 (t : Tm) IHt83 (k67 : Hvl) (d95 : (Trace TyVar)) (S31 : Ty) =>
      (f_equal2 Abs (IHT69 k67 d95 S31) (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar (XS TyVar d95) k67 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt83 (HS TmVar k67) d95 S31) (f_equal2 shift_TyVar_Tm eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d95 k67 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt83 (t2 : Tm) IHt84 (k67 : Hvl) (d95 : (Trace TyVar)) (S31 : Ty) =>
      (f_equal2 App (IHt83 k67 d95 S31) (IHt84 k67 d95 S31))) (fun (d : Decls) IHd (t : Tm) IHt83 (k67 : Hvl) (d95 : (Trace TyVar)) (S31 : Ty) =>
      (f_equal2 Let (IHd k67 d95 S31) (eq_trans (f_equal3 subst_TyVar_Tm (eq_trans (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (stability_shift_TyVar__bind _ _) (eq_refl H0))) (weakenTrace_append TyVar (XS TyVar d95) k67 (bind d))) eq_refl (f_equal2 shift_TyVar_Tm (weakenCutoffTyVar_append C0 k67 (bind d)) eq_refl)) (eq_trans (IHt83 (appendHvl k67 (bind d)) d95 S31) (f_equal2 shift_TyVar_Tm (eq_trans (eq_sym (weakenCutoffTyVar_append C0 k67 (bind d))) (f_equal2 weakenCutoffTyVar eq_refl (f_equal2 appendHvl (eq_sym (stability_subst_TyVar__bind _ _ _)) (eq_refl H0)))) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d95 k67 (bind d))) eq_refl eq_refl)))))) (fun (X41 : (Index TyVar)) =>
      (fun (k67 : Hvl) (d95 : (Trace TyVar)) (S31 : Ty) =>
        (subst_TyVar_Index_shift_TyVar_Ty0_comm_ind d95 S31 k67 X41))) (fun (T1 : Ty) IHT69 (T2 : Ty) IHT70 (k67 : Hvl) (d95 : (Trace TyVar)) (S31 : Ty) =>
      (f_equal2 TPi (IHT69 k67 d95 S31) (eq_trans (f_equal3 subst_TyVar_Ty (weakenTrace_append TyVar (XS TyVar d95) k67 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHT70 (HS TmVar k67) d95 S31) (f_equal2 shift_TyVar_Ty eq_refl (f_equal3 subst_TyVar_Ty (eq_sym (weakenTrace_append TyVar d95 k67 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (T : Ty) IHT69 (t : Tm) IHt83 (k67 : Hvl) (d95 : (Trace TyVar)) (S31 : Ty) =>
      (f_equal2 TApp (IHT69 k67 d95 S31) (IHt83 k67 d95 S31))) (fun (k67 : Hvl) (d95 : (Trace TyVar)) (S31 : Ty) =>
      eq_refl) (fun (T : Ty) IHT69 (K : Kind) IHK27 (k67 : Hvl) (d95 : (Trace TyVar)) (S31 : Ty) =>
      (f_equal2 KPi (IHT69 k67 d95 S31) (eq_trans (f_equal3 subst_TyVar_Kind (weakenTrace_append TyVar (XS TyVar d95) k67 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHK27 (HS TmVar k67) d95 S31) (f_equal2 shift_TyVar_Kind eq_refl (f_equal3 subst_TyVar_Kind (eq_sym (weakenTrace_append TyVar d95 k67 (HS TmVar H0))) eq_refl eq_refl))))))).
    Lemma subst_TyVar__shift_TyVar_0_Decls_comm_ind  :
      (forall (d95 : Decls) ,
        (forall (k67 : Hvl) (d96 : (Trace TyVar)) (S31 : Ty) ,
          ((subst_TyVar_Decls (weakenTrace (XS TyVar d96) k67) S31 (shift_TyVar_Decls (weakenCutoffTyVar C0 k67) d95)) =
          (shift_TyVar_Decls (weakenCutoffTyVar C0 k67) (subst_TyVar_Decls (weakenTrace d96 k67) S31 d95))))).
    Proof.
      pose proof (subst_TyVar__shift_TyVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TyVar__shift_TyVar_0_Tm_comm_ind  :
      (forall (s30 : Tm) ,
        (forall (k67 : Hvl) (d96 : (Trace TyVar)) (S31 : Ty) ,
          ((subst_TyVar_Tm (weakenTrace (XS TyVar d96) k67) S31 (shift_TyVar_Tm (weakenCutoffTyVar C0 k67) s30)) =
          (shift_TyVar_Tm (weakenCutoffTyVar C0 k67) (subst_TyVar_Tm (weakenTrace d96 k67) S31 s30))))).
    Proof.
      pose proof (subst_TyVar__shift_TyVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TyVar__shift_TyVar_0_Ty_comm_ind  :
      (forall (S31 : Ty) ,
        (forall (k67 : Hvl) (d96 : (Trace TyVar)) (S32 : Ty) ,
          ((subst_TyVar_Ty (weakenTrace (XS TyVar d96) k67) S32 (shift_TyVar_Ty (weakenCutoffTyVar C0 k67) S31)) =
          (shift_TyVar_Ty (weakenCutoffTyVar C0 k67) (subst_TyVar_Ty (weakenTrace d96 k67) S32 S31))))).
    Proof.
      pose proof (subst_TyVar__shift_TyVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TyVar__shift_TyVar_0_Kind_comm_ind  :
      (forall (K27 : Kind) ,
        (forall (k67 : Hvl) (d96 : (Trace TyVar)) (S32 : Ty) ,
          ((subst_TyVar_Kind (weakenTrace (XS TyVar d96) k67) S32 (shift_TyVar_Kind (weakenCutoffTyVar C0 k67) K27)) =
          (shift_TyVar_Kind (weakenCutoffTyVar C0 k67) (subst_TyVar_Kind (weakenTrace d96 k67) S32 K27))))).
    Proof.
      pose proof (subst_TyVar__shift_TyVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
  End SubstShiftCommInd.
  Section SubstShiftComm.
    Definition subst_TmVar_Decls_shift_TmVar_Decls0_comm (d96 : Decls) : (forall (d95 : (Trace TmVar)) (s30 : Tm) ,
      ((subst_TmVar_Decls (XS TmVar d95) s30 (shift_TmVar_Decls C0 d96)) =
      (shift_TmVar_Decls C0 (subst_TmVar_Decls d95 s30 d96)))) := (subst_TmVar__shift_TmVar_0_Decls_comm_ind d96 H0).
    Definition subst_TmVar_Tm_shift_TmVar_Tm0_comm (s31 : Tm) : (forall (d95 : (Trace TmVar)) (s30 : Tm) ,
      ((subst_TmVar_Tm (XS TmVar d95) s30 (shift_TmVar_Tm C0 s31)) =
      (shift_TmVar_Tm C0 (subst_TmVar_Tm d95 s30 s31)))) := (subst_TmVar__shift_TmVar_0_Tm_comm_ind s31 H0).
    Definition subst_TmVar_Ty_shift_TmVar_Ty0_comm (S31 : Ty) : (forall (d95 : (Trace TmVar)) (s30 : Tm) ,
      ((subst_TmVar_Ty (XS TmVar d95) s30 (shift_TmVar_Ty C0 S31)) =
      (shift_TmVar_Ty C0 (subst_TmVar_Ty d95 s30 S31)))) := (subst_TmVar__shift_TmVar_0_Ty_comm_ind S31 H0).
    Definition subst_TmVar_Kind_shift_TmVar_Kind0_comm (K27 : Kind) : (forall (d95 : (Trace TmVar)) (s30 : Tm) ,
      ((subst_TmVar_Kind (XS TmVar d95) s30 (shift_TmVar_Kind C0 K27)) =
      (shift_TmVar_Kind C0 (subst_TmVar_Kind d95 s30 K27)))) := (subst_TmVar__shift_TmVar_0_Kind_comm_ind K27 H0).
    Definition subst_TmVar_Decls_shift_TyVar_Decls0_comm (d96 : Decls) : (forall (d95 : (Trace TmVar)) (s30 : Tm) ,
      ((subst_TmVar_Decls (XS TyVar d95) s30 (shift_TyVar_Decls C0 d96)) =
      (shift_TyVar_Decls C0 (subst_TmVar_Decls d95 s30 d96)))) := (subst_TmVar__shift_TyVar_0_Decls_comm_ind d96 H0).
    Definition subst_TmVar_Tm_shift_TyVar_Tm0_comm (s31 : Tm) : (forall (d95 : (Trace TmVar)) (s30 : Tm) ,
      ((subst_TmVar_Tm (XS TyVar d95) s30 (shift_TyVar_Tm C0 s31)) =
      (shift_TyVar_Tm C0 (subst_TmVar_Tm d95 s30 s31)))) := (subst_TmVar__shift_TyVar_0_Tm_comm_ind s31 H0).
    Definition subst_TmVar_Ty_shift_TyVar_Ty0_comm (S31 : Ty) : (forall (d95 : (Trace TmVar)) (s30 : Tm) ,
      ((subst_TmVar_Ty (XS TyVar d95) s30 (shift_TyVar_Ty C0 S31)) =
      (shift_TyVar_Ty C0 (subst_TmVar_Ty d95 s30 S31)))) := (subst_TmVar__shift_TyVar_0_Ty_comm_ind S31 H0).
    Definition subst_TmVar_Kind_shift_TyVar_Kind0_comm (K27 : Kind) : (forall (d95 : (Trace TmVar)) (s30 : Tm) ,
      ((subst_TmVar_Kind (XS TyVar d95) s30 (shift_TyVar_Kind C0 K27)) =
      (shift_TyVar_Kind C0 (subst_TmVar_Kind d95 s30 K27)))) := (subst_TmVar__shift_TyVar_0_Kind_comm_ind K27 H0).
    Definition subst_TyVar_Decls_shift_TmVar_Decls0_comm (d96 : Decls) : (forall (d95 : (Trace TyVar)) (S31 : Ty) ,
      ((subst_TyVar_Decls (XS TmVar d95) S31 (shift_TmVar_Decls C0 d96)) =
      (shift_TmVar_Decls C0 (subst_TyVar_Decls d95 S31 d96)))) := (subst_TyVar__shift_TmVar_0_Decls_comm_ind d96 H0).
    Definition subst_TyVar_Tm_shift_TmVar_Tm0_comm (s30 : Tm) : (forall (d95 : (Trace TyVar)) (S31 : Ty) ,
      ((subst_TyVar_Tm (XS TmVar d95) S31 (shift_TmVar_Tm C0 s30)) =
      (shift_TmVar_Tm C0 (subst_TyVar_Tm d95 S31 s30)))) := (subst_TyVar__shift_TmVar_0_Tm_comm_ind s30 H0).
    Definition subst_TyVar_Ty_shift_TmVar_Ty0_comm (S32 : Ty) : (forall (d95 : (Trace TyVar)) (S31 : Ty) ,
      ((subst_TyVar_Ty (XS TmVar d95) S31 (shift_TmVar_Ty C0 S32)) =
      (shift_TmVar_Ty C0 (subst_TyVar_Ty d95 S31 S32)))) := (subst_TyVar__shift_TmVar_0_Ty_comm_ind S32 H0).
    Definition subst_TyVar_Kind_shift_TmVar_Kind0_comm (K27 : Kind) : (forall (d95 : (Trace TyVar)) (S31 : Ty) ,
      ((subst_TyVar_Kind (XS TmVar d95) S31 (shift_TmVar_Kind C0 K27)) =
      (shift_TmVar_Kind C0 (subst_TyVar_Kind d95 S31 K27)))) := (subst_TyVar__shift_TmVar_0_Kind_comm_ind K27 H0).
    Definition subst_TyVar_Decls_shift_TyVar_Decls0_comm (d96 : Decls) : (forall (d95 : (Trace TyVar)) (S31 : Ty) ,
      ((subst_TyVar_Decls (XS TyVar d95) S31 (shift_TyVar_Decls C0 d96)) =
      (shift_TyVar_Decls C0 (subst_TyVar_Decls d95 S31 d96)))) := (subst_TyVar__shift_TyVar_0_Decls_comm_ind d96 H0).
    Definition subst_TyVar_Tm_shift_TyVar_Tm0_comm (s30 : Tm) : (forall (d95 : (Trace TyVar)) (S31 : Ty) ,
      ((subst_TyVar_Tm (XS TyVar d95) S31 (shift_TyVar_Tm C0 s30)) =
      (shift_TyVar_Tm C0 (subst_TyVar_Tm d95 S31 s30)))) := (subst_TyVar__shift_TyVar_0_Tm_comm_ind s30 H0).
    Definition subst_TyVar_Ty_shift_TyVar_Ty0_comm (S32 : Ty) : (forall (d95 : (Trace TyVar)) (S31 : Ty) ,
      ((subst_TyVar_Ty (XS TyVar d95) S31 (shift_TyVar_Ty C0 S32)) =
      (shift_TyVar_Ty C0 (subst_TyVar_Ty d95 S31 S32)))) := (subst_TyVar__shift_TyVar_0_Ty_comm_ind S32 H0).
    Definition subst_TyVar_Kind_shift_TyVar_Kind0_comm (K27 : Kind) : (forall (d95 : (Trace TyVar)) (S31 : Ty) ,
      ((subst_TyVar_Kind (XS TyVar d95) S31 (shift_TyVar_Kind C0 K27)) =
      (shift_TyVar_Kind C0 (subst_TyVar_Kind d95 S31 K27)))) := (subst_TyVar__shift_TyVar_0_Kind_comm_ind K27 H0).
  End SubstShiftComm.
  Section SubstSubordInd.
    
  End SubstSubordInd.
  Section SubstSubord.
    
  End SubstSubord.
End ShiftSubstInteraction.
 Hint Rewrite subst_TmVar_Decls0_shift_TmVar_Decls0_reflection subst_TyVar_Decls0_shift_TyVar_Decls0_reflection subst_TmVar_Kind0_shift_TmVar_Kind0_reflection subst_TyVar_Kind0_shift_TyVar_Kind0_reflection subst_TmVar_Tm0_shift_TmVar_Tm0_reflection subst_TyVar_Tm0_shift_TyVar_Tm0_reflection subst_TmVar_Ty0_shift_TmVar_Ty0_reflection subst_TyVar_Ty0_shift_TyVar_Ty0_reflection : interaction.
 Hint Rewrite subst_TmVar_Decls0_shift_TmVar_Decls0_reflection subst_TyVar_Decls0_shift_TyVar_Decls0_reflection subst_TmVar_Kind0_shift_TmVar_Kind0_reflection subst_TyVar_Kind0_shift_TyVar_Kind0_reflection subst_TmVar_Tm0_shift_TmVar_Tm0_reflection subst_TyVar_Tm0_shift_TyVar_Tm0_reflection subst_TmVar_Ty0_shift_TmVar_Ty0_reflection subst_TyVar_Ty0_shift_TyVar_Ty0_reflection : reflection.
 Hint Rewrite subst_TmVar_Decls_shift_TmVar_Decls0_comm subst_TmVar_Decls_shift_TyVar_Decls0_comm subst_TyVar_Decls_shift_TmVar_Decls0_comm subst_TyVar_Decls_shift_TyVar_Decls0_comm subst_TmVar_Kind_shift_TmVar_Kind0_comm subst_TmVar_Kind_shift_TyVar_Kind0_comm subst_TyVar_Kind_shift_TmVar_Kind0_comm subst_TyVar_Kind_shift_TyVar_Kind0_comm subst_TmVar_Tm_shift_TmVar_Tm0_comm subst_TmVar_Tm_shift_TyVar_Tm0_comm subst_TyVar_Tm_shift_TmVar_Tm0_comm subst_TyVar_Tm_shift_TyVar_Tm0_comm subst_TmVar_Ty_shift_TmVar_Ty0_comm subst_TmVar_Ty_shift_TyVar_Ty0_comm subst_TyVar_Ty_shift_TmVar_Ty0_comm subst_TyVar_Ty_shift_TyVar_Ty0_comm : interaction.
 Hint Rewrite subst_TmVar_Decls_shift_TmVar_Decls0_comm subst_TmVar_Decls_shift_TyVar_Decls0_comm subst_TyVar_Decls_shift_TmVar_Decls0_comm subst_TyVar_Decls_shift_TyVar_Decls0_comm subst_TmVar_Kind_shift_TmVar_Kind0_comm subst_TmVar_Kind_shift_TyVar_Kind0_comm subst_TyVar_Kind_shift_TmVar_Kind0_comm subst_TyVar_Kind_shift_TyVar_Kind0_comm subst_TmVar_Tm_shift_TmVar_Tm0_comm subst_TmVar_Tm_shift_TyVar_Tm0_comm subst_TyVar_Tm_shift_TmVar_Tm0_comm subst_TyVar_Tm_shift_TyVar_Tm0_comm subst_TmVar_Ty_shift_TmVar_Ty0_comm subst_TmVar_Ty_shift_TyVar_Ty0_comm subst_TyVar_Ty_shift_TmVar_Ty0_comm subst_TyVar_Ty_shift_TyVar_Ty0_comm : subst_shift0.
 Hint Rewrite shift_TmVar_Decls_subst_TmVar_Decls0_comm shift_TmVar_Decls_subst_TyVar_Decls0_comm shift_TyVar_Decls_subst_TmVar_Decls0_comm shift_TyVar_Decls_subst_TyVar_Decls0_comm shift_TmVar_Kind_subst_TmVar_Kind0_comm shift_TmVar_Kind_subst_TyVar_Kind0_comm shift_TyVar_Kind_subst_TmVar_Kind0_comm shift_TyVar_Kind_subst_TyVar_Kind0_comm shift_TmVar_Tm_subst_TmVar_Tm0_comm shift_TmVar_Tm_subst_TyVar_Tm0_comm shift_TyVar_Tm_subst_TmVar_Tm0_comm shift_TyVar_Tm_subst_TyVar_Tm0_comm shift_TmVar_Ty_subst_TmVar_Ty0_comm shift_TmVar_Ty_subst_TyVar_Ty0_comm shift_TyVar_Ty_subst_TmVar_Ty0_comm shift_TyVar_Ty_subst_TyVar_Ty0_comm : interaction.
 Hint Rewrite shift_TmVar_Decls_subst_TmVar_Decls0_comm shift_TmVar_Decls_subst_TyVar_Decls0_comm shift_TyVar_Decls_subst_TmVar_Decls0_comm shift_TyVar_Decls_subst_TyVar_Decls0_comm shift_TmVar_Kind_subst_TmVar_Kind0_comm shift_TmVar_Kind_subst_TyVar_Kind0_comm shift_TyVar_Kind_subst_TmVar_Kind0_comm shift_TyVar_Kind_subst_TyVar_Kind0_comm shift_TmVar_Tm_subst_TmVar_Tm0_comm shift_TmVar_Tm_subst_TyVar_Tm0_comm shift_TyVar_Tm_subst_TmVar_Tm0_comm shift_TyVar_Tm_subst_TyVar_Tm0_comm shift_TmVar_Ty_subst_TmVar_Ty0_comm shift_TmVar_Ty_subst_TyVar_Ty0_comm shift_TyVar_Ty_subst_TmVar_Ty0_comm shift_TyVar_Ty_subst_TyVar_Ty0_comm : shift_subst0.
Section SubstSubstInteraction.
  Section SubstSubstIndexCommInd.
    Lemma subst_TmVar_Tm_subst_TmVar_Index0_commright_ind (d95 : (Trace TmVar)) (s30 : Tm) (s31 : Tm) :
      (forall (k67 : Hvl) (x83 : (Index TmVar)) ,
        ((subst_TmVar_Tm (weakenTrace d95 k67) s30 (subst_TmVar_Index (weakenTrace X0 k67) s31 x83)) =
        (subst_TmVar_Tm (weakenTrace X0 k67) (subst_TmVar_Tm d95 s30 s31) (subst_TmVar_Index (weakenTrace (XS TmVar d95) k67) s30 x83)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_TyVar_Tm_subst_TmVar_Index0_commright_ind (d95 : (Trace TyVar)) (S31 : Ty) (s30 : Tm) :
      (forall (k67 : Hvl) (x83 : (Index TmVar)) ,
        ((subst_TyVar_Tm (weakenTrace d95 k67) S31 (subst_TmVar_Index (weakenTrace X0 k67) s30 x83)) =
        (subst_TmVar_Index (weakenTrace X0 k67) (subst_TyVar_Tm d95 S31 s30) x83))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_TmVar_Ty_subst_TyVar_Index0_commright_ind (d95 : (Trace TmVar)) (s30 : Tm) (S31 : Ty) :
      (forall (k67 : Hvl) (X41 : (Index TyVar)) ,
        ((subst_TmVar_Ty (weakenTrace d95 k67) s30 (subst_TyVar_Index (weakenTrace X0 k67) S31 X41)) =
        (subst_TyVar_Index (weakenTrace X0 k67) (subst_TmVar_Ty d95 s30 S31) X41))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_TyVar_Ty_subst_TyVar_Index0_commright_ind (d95 : (Trace TyVar)) (S31 : Ty) (S32 : Ty) :
      (forall (k67 : Hvl) (X41 : (Index TyVar)) ,
        ((subst_TyVar_Ty (weakenTrace d95 k67) S31 (subst_TyVar_Index (weakenTrace X0 k67) S32 X41)) =
        (subst_TyVar_Ty (weakenTrace X0 k67) (subst_TyVar_Ty d95 S31 S32) (subst_TyVar_Index (weakenTrace (XS TyVar d95) k67) S31 X41)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_TmVar_Ty_subst_TyVar_Index0_commleft_ind (d95 : (Trace TmVar)) (s30 : Tm) (S31 : Ty) :
      (forall (k67 : Hvl) (x83 : (Index TmVar)) ,
        ((subst_TmVar_Index (weakenTrace d95 k67) s30 x83) =
        (subst_TyVar_Tm (weakenTrace X0 k67) (subst_TmVar_Ty d95 s30 S31) (subst_TmVar_Index (weakenTrace (XS TyVar d95) k67) s30 x83)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_TyVar_Tm_subst_TmVar_Index0_commleft_ind (d95 : (Trace TyVar)) (S31 : Ty) (s30 : Tm) :
      (forall (k67 : Hvl) (X41 : (Index TyVar)) ,
        ((subst_TyVar_Index (weakenTrace d95 k67) S31 X41) =
        (subst_TmVar_Ty (weakenTrace X0 k67) (subst_TyVar_Tm d95 S31 s30) (subst_TyVar_Index (weakenTrace (XS TmVar d95) k67) S31 X41)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
  End SubstSubstIndexCommInd.
  Section SubstSubstCommInd.
    Definition subst_TmVar__subst_TmVar_0_Decls_Tm_Ty_Kind_comm_ind := (ind_Decls_Tm_Ty_Kind (fun (d105 : Decls) =>
      (forall (k71 : Hvl) (d106 : (Trace TmVar)) (s34 : Tm) (s35 : Tm) ,
        ((subst_TmVar_Decls (weakenTrace d106 k71) s34 (subst_TmVar_Decls (weakenTrace X0 k71) s35 d105)) =
        (subst_TmVar_Decls (weakenTrace X0 k71) (subst_TmVar_Tm d106 s34 s35) (subst_TmVar_Decls (weakenTrace (XS TmVar d106) k71) s34 d105))))) (fun (s34 : Tm) =>
      (forall (k71 : Hvl) (d105 : (Trace TmVar)) (s35 : Tm) (s36 : Tm) ,
        ((subst_TmVar_Tm (weakenTrace d105 k71) s35 (subst_TmVar_Tm (weakenTrace X0 k71) s36 s34)) =
        (subst_TmVar_Tm (weakenTrace X0 k71) (subst_TmVar_Tm d105 s35 s36) (subst_TmVar_Tm (weakenTrace (XS TmVar d105) k71) s35 s34))))) (fun (S35 : Ty) =>
      (forall (k71 : Hvl) (d105 : (Trace TmVar)) (s34 : Tm) (s35 : Tm) ,
        ((subst_TmVar_Ty (weakenTrace d105 k71) s34 (subst_TmVar_Ty (weakenTrace X0 k71) s35 S35)) =
        (subst_TmVar_Ty (weakenTrace X0 k71) (subst_TmVar_Tm d105 s34 s35) (subst_TmVar_Ty (weakenTrace (XS TmVar d105) k71) s34 S35))))) (fun (K31 : Kind) =>
      (forall (k71 : Hvl) (d105 : (Trace TmVar)) (s34 : Tm) (s35 : Tm) ,
        ((subst_TmVar_Kind (weakenTrace d105 k71) s34 (subst_TmVar_Kind (weakenTrace X0 k71) s35 K31)) =
        (subst_TmVar_Kind (weakenTrace X0 k71) (subst_TmVar_Tm d105 s34 s35) (subst_TmVar_Kind (weakenTrace (XS TmVar d105) k71) s34 K31))))) (fun (k71 : Hvl) (d105 : (Trace TmVar)) (s34 : Tm) (s35 : Tm) =>
      eq_refl) (fun (t : Tm) IHt95 (d : Decls) IHd (k71 : Hvl) (d105 : (Trace TmVar)) (s34 : Tm) (s35 : Tm) =>
      (f_equal2 DConsTm (IHt95 k71 d105 s34 s35) (eq_trans (f_equal3 subst_TmVar_Decls (weakenTrace_append TmVar d105 k71 (HS TmVar H0)) eq_refl (f_equal3 subst_TmVar_Decls (weakenTrace_append TmVar X0 k71 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHd (HS TmVar k71) d105 s34 s35) (f_equal3 subst_TmVar_Decls (eq_sym (weakenTrace_append TmVar X0 k71 (HS TmVar H0))) eq_refl (f_equal3 subst_TmVar_Decls (eq_sym (weakenTrace_append TmVar (XS TmVar d105) k71 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (K : Kind) IHK31 (d : Decls) IHd (k71 : Hvl) (d105 : (Trace TmVar)) (s34 : Tm) (s35 : Tm) =>
      (f_equal2 DConsTy (IHK31 k71 d105 s34 s35) (eq_trans (f_equal3 subst_TmVar_Decls (weakenTrace_append TmVar d105 k71 (HS TyVar H0)) eq_refl (f_equal3 subst_TmVar_Decls (weakenTrace_append TmVar X0 k71 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHd (HS TyVar k71) d105 s34 s35) (f_equal3 subst_TmVar_Decls (eq_sym (weakenTrace_append TmVar X0 k71 (HS TyVar H0))) eq_refl (f_equal3 subst_TmVar_Decls (eq_sym (weakenTrace_append TmVar (XS TmVar d105) k71 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (x95 : (Index TmVar)) =>
      (fun (k71 : Hvl) (d105 : (Trace TmVar)) (s34 : Tm) (s35 : Tm) =>
        (subst_TmVar_Tm_subst_TmVar_Index0_commright_ind d105 s34 s35 k71 x95))) (fun (T : Ty) IHT79 (t : Tm) IHt95 (k71 : Hvl) (d105 : (Trace TmVar)) (s34 : Tm) (s35 : Tm) =>
      (f_equal2 Abs (IHT79 k71 d105 s34 s35) (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar d105 k71 (HS TmVar H0)) eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k71 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt95 (HS TmVar k71) d105 s34 s35) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k71 (HS TmVar H0))) eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar (XS TmVar d105) k71 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt95 (t2 : Tm) IHt96 (k71 : Hvl) (d105 : (Trace TmVar)) (s34 : Tm) (s35 : Tm) =>
      (f_equal2 App (IHt95 k71 d105 s34 s35) (IHt96 k71 d105 s34 s35))) (fun (d : Decls) IHd (t : Tm) IHt95 (k71 : Hvl) (d105 : (Trace TmVar)) (s34 : Tm) (s35 : Tm) =>
      (f_equal2 Let (IHd k71 d105 s34 s35) (eq_trans (f_equal3 subst_TmVar_Tm (eq_trans (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (stability_subst_TmVar__bind _ _ _) (eq_refl H0))) (weakenTrace_append TmVar d105 k71 (bind d))) eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k71 (bind d)) eq_refl eq_refl)) (eq_trans (IHt95 (appendHvl k71 (bind d)) d105 s34 s35) (f_equal3 subst_TmVar_Tm (eq_trans (eq_sym (weakenTrace_append TmVar X0 k71 (bind d))) (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (eq_sym (stability_subst_TmVar__bind _ _ _)) (eq_refl H0)))) eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar (XS TmVar d105) k71 (bind d))) eq_refl eq_refl)))))) (fun (X47 : (Index TyVar)) =>
      (fun (k71 : Hvl) (d105 : (Trace TmVar)) (s34 : Tm) (s35 : Tm) =>
        eq_refl)) (fun (T1 : Ty) IHT79 (T2 : Ty) IHT80 (k71 : Hvl) (d105 : (Trace TmVar)) (s34 : Tm) (s35 : Tm) =>
      (f_equal2 TPi (IHT79 k71 d105 s34 s35) (eq_trans (f_equal3 subst_TmVar_Ty (weakenTrace_append TmVar d105 k71 (HS TmVar H0)) eq_refl (f_equal3 subst_TmVar_Ty (weakenTrace_append TmVar X0 k71 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHT80 (HS TmVar k71) d105 s34 s35) (f_equal3 subst_TmVar_Ty (eq_sym (weakenTrace_append TmVar X0 k71 (HS TmVar H0))) eq_refl (f_equal3 subst_TmVar_Ty (eq_sym (weakenTrace_append TmVar (XS TmVar d105) k71 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (T : Ty) IHT79 (t : Tm) IHt95 (k71 : Hvl) (d105 : (Trace TmVar)) (s34 : Tm) (s35 : Tm) =>
      (f_equal2 TApp (IHT79 k71 d105 s34 s35) (IHt95 k71 d105 s34 s35))) (fun (k71 : Hvl) (d105 : (Trace TmVar)) (s34 : Tm) (s35 : Tm) =>
      eq_refl) (fun (T : Ty) IHT79 (K : Kind) IHK31 (k71 : Hvl) (d105 : (Trace TmVar)) (s34 : Tm) (s35 : Tm) =>
      (f_equal2 KPi (IHT79 k71 d105 s34 s35) (eq_trans (f_equal3 subst_TmVar_Kind (weakenTrace_append TmVar d105 k71 (HS TmVar H0)) eq_refl (f_equal3 subst_TmVar_Kind (weakenTrace_append TmVar X0 k71 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHK31 (HS TmVar k71) d105 s34 s35) (f_equal3 subst_TmVar_Kind (eq_sym (weakenTrace_append TmVar X0 k71 (HS TmVar H0))) eq_refl (f_equal3 subst_TmVar_Kind (eq_sym (weakenTrace_append TmVar (XS TmVar d105) k71 (HS TmVar H0))) eq_refl eq_refl))))))).
    Lemma subst_TmVar__subst_TmVar_0_Decls_comm_ind  :
      (forall (d105 : Decls) ,
        (forall (k71 : Hvl) (d106 : (Trace TmVar)) (s34 : Tm) (s35 : Tm) ,
          ((subst_TmVar_Decls (weakenTrace d106 k71) s34 (subst_TmVar_Decls (weakenTrace X0 k71) s35 d105)) =
          (subst_TmVar_Decls (weakenTrace X0 k71) (subst_TmVar_Tm d106 s34 s35) (subst_TmVar_Decls (weakenTrace (XS TmVar d106) k71) s34 d105))))).
    Proof.
      pose proof (subst_TmVar__subst_TmVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TmVar__subst_TmVar_0_Tm_comm_ind  :
      (forall (s34 : Tm) ,
        (forall (k71 : Hvl) (d106 : (Trace TmVar)) (s35 : Tm) (s36 : Tm) ,
          ((subst_TmVar_Tm (weakenTrace d106 k71) s35 (subst_TmVar_Tm (weakenTrace X0 k71) s36 s34)) =
          (subst_TmVar_Tm (weakenTrace X0 k71) (subst_TmVar_Tm d106 s35 s36) (subst_TmVar_Tm (weakenTrace (XS TmVar d106) k71) s35 s34))))).
    Proof.
      pose proof (subst_TmVar__subst_TmVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TmVar__subst_TmVar_0_Ty_comm_ind  :
      (forall (S35 : Ty) ,
        (forall (k71 : Hvl) (d106 : (Trace TmVar)) (s35 : Tm) (s36 : Tm) ,
          ((subst_TmVar_Ty (weakenTrace d106 k71) s35 (subst_TmVar_Ty (weakenTrace X0 k71) s36 S35)) =
          (subst_TmVar_Ty (weakenTrace X0 k71) (subst_TmVar_Tm d106 s35 s36) (subst_TmVar_Ty (weakenTrace (XS TmVar d106) k71) s35 S35))))).
    Proof.
      pose proof (subst_TmVar__subst_TmVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TmVar__subst_TmVar_0_Kind_comm_ind  :
      (forall (K31 : Kind) ,
        (forall (k71 : Hvl) (d106 : (Trace TmVar)) (s35 : Tm) (s36 : Tm) ,
          ((subst_TmVar_Kind (weakenTrace d106 k71) s35 (subst_TmVar_Kind (weakenTrace X0 k71) s36 K31)) =
          (subst_TmVar_Kind (weakenTrace X0 k71) (subst_TmVar_Tm d106 s35 s36) (subst_TmVar_Kind (weakenTrace (XS TmVar d106) k71) s35 K31))))).
    Proof.
      pose proof (subst_TmVar__subst_TmVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_TmVar__subst_TyVar_0_Decls_Tm_Ty_Kind_comm_ind := (ind_Decls_Tm_Ty_Kind (fun (d105 : Decls) =>
      (forall (k71 : Hvl) (d106 : (Trace TmVar)) (s34 : Tm) (S35 : Ty) ,
        ((subst_TmVar_Decls (weakenTrace d106 k71) s34 (subst_TyVar_Decls (weakenTrace X0 k71) S35 d105)) =
        (subst_TyVar_Decls (weakenTrace X0 k71) (subst_TmVar_Ty d106 s34 S35) (subst_TmVar_Decls (weakenTrace (XS TyVar d106) k71) s34 d105))))) (fun (s34 : Tm) =>
      (forall (k71 : Hvl) (d105 : (Trace TmVar)) (s35 : Tm) (S35 : Ty) ,
        ((subst_TmVar_Tm (weakenTrace d105 k71) s35 (subst_TyVar_Tm (weakenTrace X0 k71) S35 s34)) =
        (subst_TyVar_Tm (weakenTrace X0 k71) (subst_TmVar_Ty d105 s35 S35) (subst_TmVar_Tm (weakenTrace (XS TyVar d105) k71) s35 s34))))) (fun (S35 : Ty) =>
      (forall (k71 : Hvl) (d105 : (Trace TmVar)) (s34 : Tm) (S36 : Ty) ,
        ((subst_TmVar_Ty (weakenTrace d105 k71) s34 (subst_TyVar_Ty (weakenTrace X0 k71) S36 S35)) =
        (subst_TyVar_Ty (weakenTrace X0 k71) (subst_TmVar_Ty d105 s34 S36) (subst_TmVar_Ty (weakenTrace (XS TyVar d105) k71) s34 S35))))) (fun (K31 : Kind) =>
      (forall (k71 : Hvl) (d105 : (Trace TmVar)) (s34 : Tm) (S35 : Ty) ,
        ((subst_TmVar_Kind (weakenTrace d105 k71) s34 (subst_TyVar_Kind (weakenTrace X0 k71) S35 K31)) =
        (subst_TyVar_Kind (weakenTrace X0 k71) (subst_TmVar_Ty d105 s34 S35) (subst_TmVar_Kind (weakenTrace (XS TyVar d105) k71) s34 K31))))) (fun (k71 : Hvl) (d105 : (Trace TmVar)) (s34 : Tm) (S35 : Ty) =>
      eq_refl) (fun (t : Tm) IHt95 (d : Decls) IHd (k71 : Hvl) (d105 : (Trace TmVar)) (s34 : Tm) (S35 : Ty) =>
      (f_equal2 DConsTm (IHt95 k71 d105 s34 S35) (eq_trans (f_equal3 subst_TmVar_Decls (weakenTrace_append TmVar d105 k71 (HS TmVar H0)) eq_refl (f_equal3 subst_TyVar_Decls (weakenTrace_append TyVar X0 k71 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHd (HS TmVar k71) d105 s34 S35) (f_equal3 subst_TyVar_Decls (eq_sym (weakenTrace_append TyVar X0 k71 (HS TmVar H0))) eq_refl (f_equal3 subst_TmVar_Decls (eq_sym (weakenTrace_append TmVar (XS TyVar d105) k71 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (K : Kind) IHK31 (d : Decls) IHd (k71 : Hvl) (d105 : (Trace TmVar)) (s34 : Tm) (S35 : Ty) =>
      (f_equal2 DConsTy (IHK31 k71 d105 s34 S35) (eq_trans (f_equal3 subst_TmVar_Decls (weakenTrace_append TmVar d105 k71 (HS TyVar H0)) eq_refl (f_equal3 subst_TyVar_Decls (weakenTrace_append TyVar X0 k71 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHd (HS TyVar k71) d105 s34 S35) (f_equal3 subst_TyVar_Decls (eq_sym (weakenTrace_append TyVar X0 k71 (HS TyVar H0))) eq_refl (f_equal3 subst_TmVar_Decls (eq_sym (weakenTrace_append TmVar (XS TyVar d105) k71 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (x95 : (Index TmVar)) =>
      (fun (k71 : Hvl) (d105 : (Trace TmVar)) (s34 : Tm) (S35 : Ty) =>
        (subst_TmVar_Ty_subst_TyVar_Index0_commleft_ind d105 s34 S35 k71 x95))) (fun (T : Ty) IHT79 (t : Tm) IHt95 (k71 : Hvl) (d105 : (Trace TmVar)) (s34 : Tm) (S35 : Ty) =>
      (f_equal2 Abs (IHT79 k71 d105 s34 S35) (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar d105 k71 (HS TmVar H0)) eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k71 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt95 (HS TmVar k71) d105 s34 S35) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k71 (HS TmVar H0))) eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar (XS TyVar d105) k71 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt95 (t2 : Tm) IHt96 (k71 : Hvl) (d105 : (Trace TmVar)) (s34 : Tm) (S35 : Ty) =>
      (f_equal2 App (IHt95 k71 d105 s34 S35) (IHt96 k71 d105 s34 S35))) (fun (d : Decls) IHd (t : Tm) IHt95 (k71 : Hvl) (d105 : (Trace TmVar)) (s34 : Tm) (S35 : Ty) =>
      (f_equal2 Let (IHd k71 d105 s34 S35) (eq_trans (f_equal3 subst_TmVar_Tm (eq_trans (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (stability_subst_TyVar__bind _ _ _) (eq_refl H0))) (weakenTrace_append TmVar d105 k71 (bind d))) eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k71 (bind d)) eq_refl eq_refl)) (eq_trans (IHt95 (appendHvl k71 (bind d)) d105 s34 S35) (f_equal3 subst_TyVar_Tm (eq_trans (eq_sym (weakenTrace_append TyVar X0 k71 (bind d))) (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (eq_sym (stability_subst_TmVar__bind _ _ _)) (eq_refl H0)))) eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar (XS TyVar d105) k71 (bind d))) eq_refl eq_refl)))))) (fun (X47 : (Index TyVar)) =>
      (fun (k71 : Hvl) (d105 : (Trace TmVar)) (s34 : Tm) (S35 : Ty) =>
        (subst_TmVar_Ty_subst_TyVar_Index0_commright_ind d105 s34 S35 k71 X47))) (fun (T1 : Ty) IHT79 (T2 : Ty) IHT80 (k71 : Hvl) (d105 : (Trace TmVar)) (s34 : Tm) (S35 : Ty) =>
      (f_equal2 TPi (IHT79 k71 d105 s34 S35) (eq_trans (f_equal3 subst_TmVar_Ty (weakenTrace_append TmVar d105 k71 (HS TmVar H0)) eq_refl (f_equal3 subst_TyVar_Ty (weakenTrace_append TyVar X0 k71 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHT80 (HS TmVar k71) d105 s34 S35) (f_equal3 subst_TyVar_Ty (eq_sym (weakenTrace_append TyVar X0 k71 (HS TmVar H0))) eq_refl (f_equal3 subst_TmVar_Ty (eq_sym (weakenTrace_append TmVar (XS TyVar d105) k71 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (T : Ty) IHT79 (t : Tm) IHt95 (k71 : Hvl) (d105 : (Trace TmVar)) (s34 : Tm) (S35 : Ty) =>
      (f_equal2 TApp (IHT79 k71 d105 s34 S35) (IHt95 k71 d105 s34 S35))) (fun (k71 : Hvl) (d105 : (Trace TmVar)) (s34 : Tm) (S35 : Ty) =>
      eq_refl) (fun (T : Ty) IHT79 (K : Kind) IHK31 (k71 : Hvl) (d105 : (Trace TmVar)) (s34 : Tm) (S35 : Ty) =>
      (f_equal2 KPi (IHT79 k71 d105 s34 S35) (eq_trans (f_equal3 subst_TmVar_Kind (weakenTrace_append TmVar d105 k71 (HS TmVar H0)) eq_refl (f_equal3 subst_TyVar_Kind (weakenTrace_append TyVar X0 k71 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHK31 (HS TmVar k71) d105 s34 S35) (f_equal3 subst_TyVar_Kind (eq_sym (weakenTrace_append TyVar X0 k71 (HS TmVar H0))) eq_refl (f_equal3 subst_TmVar_Kind (eq_sym (weakenTrace_append TmVar (XS TyVar d105) k71 (HS TmVar H0))) eq_refl eq_refl))))))).
    Lemma subst_TmVar__subst_TyVar_0_Decls_comm_ind  :
      (forall (d105 : Decls) ,
        (forall (k71 : Hvl) (d106 : (Trace TmVar)) (s34 : Tm) (S35 : Ty) ,
          ((subst_TmVar_Decls (weakenTrace d106 k71) s34 (subst_TyVar_Decls (weakenTrace X0 k71) S35 d105)) =
          (subst_TyVar_Decls (weakenTrace X0 k71) (subst_TmVar_Ty d106 s34 S35) (subst_TmVar_Decls (weakenTrace (XS TyVar d106) k71) s34 d105))))).
    Proof.
      pose proof (subst_TmVar__subst_TyVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TmVar__subst_TyVar_0_Tm_comm_ind  :
      (forall (s34 : Tm) ,
        (forall (k71 : Hvl) (d106 : (Trace TmVar)) (s35 : Tm) (S35 : Ty) ,
          ((subst_TmVar_Tm (weakenTrace d106 k71) s35 (subst_TyVar_Tm (weakenTrace X0 k71) S35 s34)) =
          (subst_TyVar_Tm (weakenTrace X0 k71) (subst_TmVar_Ty d106 s35 S35) (subst_TmVar_Tm (weakenTrace (XS TyVar d106) k71) s35 s34))))).
    Proof.
      pose proof (subst_TmVar__subst_TyVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TmVar__subst_TyVar_0_Ty_comm_ind  :
      (forall (S35 : Ty) ,
        (forall (k71 : Hvl) (d106 : (Trace TmVar)) (s35 : Tm) (S36 : Ty) ,
          ((subst_TmVar_Ty (weakenTrace d106 k71) s35 (subst_TyVar_Ty (weakenTrace X0 k71) S36 S35)) =
          (subst_TyVar_Ty (weakenTrace X0 k71) (subst_TmVar_Ty d106 s35 S36) (subst_TmVar_Ty (weakenTrace (XS TyVar d106) k71) s35 S35))))).
    Proof.
      pose proof (subst_TmVar__subst_TyVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TmVar__subst_TyVar_0_Kind_comm_ind  :
      (forall (K31 : Kind) ,
        (forall (k71 : Hvl) (d106 : (Trace TmVar)) (s35 : Tm) (S36 : Ty) ,
          ((subst_TmVar_Kind (weakenTrace d106 k71) s35 (subst_TyVar_Kind (weakenTrace X0 k71) S36 K31)) =
          (subst_TyVar_Kind (weakenTrace X0 k71) (subst_TmVar_Ty d106 s35 S36) (subst_TmVar_Kind (weakenTrace (XS TyVar d106) k71) s35 K31))))).
    Proof.
      pose proof (subst_TmVar__subst_TyVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_TyVar__subst_TmVar_0_Decls_Tm_Ty_Kind_comm_ind := (ind_Decls_Tm_Ty_Kind (fun (d105 : Decls) =>
      (forall (k71 : Hvl) (d106 : (Trace TyVar)) (S35 : Ty) (s34 : Tm) ,
        ((subst_TyVar_Decls (weakenTrace d106 k71) S35 (subst_TmVar_Decls (weakenTrace X0 k71) s34 d105)) =
        (subst_TmVar_Decls (weakenTrace X0 k71) (subst_TyVar_Tm d106 S35 s34) (subst_TyVar_Decls (weakenTrace (XS TmVar d106) k71) S35 d105))))) (fun (s34 : Tm) =>
      (forall (k71 : Hvl) (d105 : (Trace TyVar)) (S35 : Ty) (s35 : Tm) ,
        ((subst_TyVar_Tm (weakenTrace d105 k71) S35 (subst_TmVar_Tm (weakenTrace X0 k71) s35 s34)) =
        (subst_TmVar_Tm (weakenTrace X0 k71) (subst_TyVar_Tm d105 S35 s35) (subst_TyVar_Tm (weakenTrace (XS TmVar d105) k71) S35 s34))))) (fun (S35 : Ty) =>
      (forall (k71 : Hvl) (d105 : (Trace TyVar)) (S36 : Ty) (s34 : Tm) ,
        ((subst_TyVar_Ty (weakenTrace d105 k71) S36 (subst_TmVar_Ty (weakenTrace X0 k71) s34 S35)) =
        (subst_TmVar_Ty (weakenTrace X0 k71) (subst_TyVar_Tm d105 S36 s34) (subst_TyVar_Ty (weakenTrace (XS TmVar d105) k71) S36 S35))))) (fun (K31 : Kind) =>
      (forall (k71 : Hvl) (d105 : (Trace TyVar)) (S35 : Ty) (s34 : Tm) ,
        ((subst_TyVar_Kind (weakenTrace d105 k71) S35 (subst_TmVar_Kind (weakenTrace X0 k71) s34 K31)) =
        (subst_TmVar_Kind (weakenTrace X0 k71) (subst_TyVar_Tm d105 S35 s34) (subst_TyVar_Kind (weakenTrace (XS TmVar d105) k71) S35 K31))))) (fun (k71 : Hvl) (d105 : (Trace TyVar)) (S35 : Ty) (s34 : Tm) =>
      eq_refl) (fun (t : Tm) IHt95 (d : Decls) IHd (k71 : Hvl) (d105 : (Trace TyVar)) (S35 : Ty) (s34 : Tm) =>
      (f_equal2 DConsTm (IHt95 k71 d105 S35 s34) (eq_trans (f_equal3 subst_TyVar_Decls (weakenTrace_append TyVar d105 k71 (HS TmVar H0)) eq_refl (f_equal3 subst_TmVar_Decls (weakenTrace_append TmVar X0 k71 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHd (HS TmVar k71) d105 S35 s34) (f_equal3 subst_TmVar_Decls (eq_sym (weakenTrace_append TmVar X0 k71 (HS TmVar H0))) eq_refl (f_equal3 subst_TyVar_Decls (eq_sym (weakenTrace_append TyVar (XS TmVar d105) k71 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (K : Kind) IHK31 (d : Decls) IHd (k71 : Hvl) (d105 : (Trace TyVar)) (S35 : Ty) (s34 : Tm) =>
      (f_equal2 DConsTy (IHK31 k71 d105 S35 s34) (eq_trans (f_equal3 subst_TyVar_Decls (weakenTrace_append TyVar d105 k71 (HS TyVar H0)) eq_refl (f_equal3 subst_TmVar_Decls (weakenTrace_append TmVar X0 k71 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHd (HS TyVar k71) d105 S35 s34) (f_equal3 subst_TmVar_Decls (eq_sym (weakenTrace_append TmVar X0 k71 (HS TyVar H0))) eq_refl (f_equal3 subst_TyVar_Decls (eq_sym (weakenTrace_append TyVar (XS TmVar d105) k71 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (x95 : (Index TmVar)) =>
      (fun (k71 : Hvl) (d105 : (Trace TyVar)) (S35 : Ty) (s34 : Tm) =>
        (subst_TyVar_Tm_subst_TmVar_Index0_commright_ind d105 S35 s34 k71 x95))) (fun (T : Ty) IHT79 (t : Tm) IHt95 (k71 : Hvl) (d105 : (Trace TyVar)) (S35 : Ty) (s34 : Tm) =>
      (f_equal2 Abs (IHT79 k71 d105 S35 s34) (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar d105 k71 (HS TmVar H0)) eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k71 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt95 (HS TmVar k71) d105 S35 s34) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k71 (HS TmVar H0))) eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar (XS TmVar d105) k71 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt95 (t2 : Tm) IHt96 (k71 : Hvl) (d105 : (Trace TyVar)) (S35 : Ty) (s34 : Tm) =>
      (f_equal2 App (IHt95 k71 d105 S35 s34) (IHt96 k71 d105 S35 s34))) (fun (d : Decls) IHd (t : Tm) IHt95 (k71 : Hvl) (d105 : (Trace TyVar)) (S35 : Ty) (s34 : Tm) =>
      (f_equal2 Let (IHd k71 d105 S35 s34) (eq_trans (f_equal3 subst_TyVar_Tm (eq_trans (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (stability_subst_TmVar__bind _ _ _) (eq_refl H0))) (weakenTrace_append TyVar d105 k71 (bind d))) eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k71 (bind d)) eq_refl eq_refl)) (eq_trans (IHt95 (appendHvl k71 (bind d)) d105 S35 s34) (f_equal3 subst_TmVar_Tm (eq_trans (eq_sym (weakenTrace_append TmVar X0 k71 (bind d))) (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (eq_sym (stability_subst_TyVar__bind _ _ _)) (eq_refl H0)))) eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar (XS TmVar d105) k71 (bind d))) eq_refl eq_refl)))))) (fun (X47 : (Index TyVar)) =>
      (fun (k71 : Hvl) (d105 : (Trace TyVar)) (S35 : Ty) (s34 : Tm) =>
        (subst_TyVar_Tm_subst_TmVar_Index0_commleft_ind d105 S35 s34 k71 X47))) (fun (T1 : Ty) IHT79 (T2 : Ty) IHT80 (k71 : Hvl) (d105 : (Trace TyVar)) (S35 : Ty) (s34 : Tm) =>
      (f_equal2 TPi (IHT79 k71 d105 S35 s34) (eq_trans (f_equal3 subst_TyVar_Ty (weakenTrace_append TyVar d105 k71 (HS TmVar H0)) eq_refl (f_equal3 subst_TmVar_Ty (weakenTrace_append TmVar X0 k71 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHT80 (HS TmVar k71) d105 S35 s34) (f_equal3 subst_TmVar_Ty (eq_sym (weakenTrace_append TmVar X0 k71 (HS TmVar H0))) eq_refl (f_equal3 subst_TyVar_Ty (eq_sym (weakenTrace_append TyVar (XS TmVar d105) k71 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (T : Ty) IHT79 (t : Tm) IHt95 (k71 : Hvl) (d105 : (Trace TyVar)) (S35 : Ty) (s34 : Tm) =>
      (f_equal2 TApp (IHT79 k71 d105 S35 s34) (IHt95 k71 d105 S35 s34))) (fun (k71 : Hvl) (d105 : (Trace TyVar)) (S35 : Ty) (s34 : Tm) =>
      eq_refl) (fun (T : Ty) IHT79 (K : Kind) IHK31 (k71 : Hvl) (d105 : (Trace TyVar)) (S35 : Ty) (s34 : Tm) =>
      (f_equal2 KPi (IHT79 k71 d105 S35 s34) (eq_trans (f_equal3 subst_TyVar_Kind (weakenTrace_append TyVar d105 k71 (HS TmVar H0)) eq_refl (f_equal3 subst_TmVar_Kind (weakenTrace_append TmVar X0 k71 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHK31 (HS TmVar k71) d105 S35 s34) (f_equal3 subst_TmVar_Kind (eq_sym (weakenTrace_append TmVar X0 k71 (HS TmVar H0))) eq_refl (f_equal3 subst_TyVar_Kind (eq_sym (weakenTrace_append TyVar (XS TmVar d105) k71 (HS TmVar H0))) eq_refl eq_refl))))))).
    Lemma subst_TyVar__subst_TmVar_0_Decls_comm_ind  :
      (forall (d105 : Decls) ,
        (forall (k71 : Hvl) (d106 : (Trace TyVar)) (S35 : Ty) (s34 : Tm) ,
          ((subst_TyVar_Decls (weakenTrace d106 k71) S35 (subst_TmVar_Decls (weakenTrace X0 k71) s34 d105)) =
          (subst_TmVar_Decls (weakenTrace X0 k71) (subst_TyVar_Tm d106 S35 s34) (subst_TyVar_Decls (weakenTrace (XS TmVar d106) k71) S35 d105))))).
    Proof.
      pose proof (subst_TyVar__subst_TmVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TyVar__subst_TmVar_0_Tm_comm_ind  :
      (forall (s34 : Tm) ,
        (forall (k71 : Hvl) (d106 : (Trace TyVar)) (S35 : Ty) (s35 : Tm) ,
          ((subst_TyVar_Tm (weakenTrace d106 k71) S35 (subst_TmVar_Tm (weakenTrace X0 k71) s35 s34)) =
          (subst_TmVar_Tm (weakenTrace X0 k71) (subst_TyVar_Tm d106 S35 s35) (subst_TyVar_Tm (weakenTrace (XS TmVar d106) k71) S35 s34))))).
    Proof.
      pose proof (subst_TyVar__subst_TmVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TyVar__subst_TmVar_0_Ty_comm_ind  :
      (forall (S35 : Ty) ,
        (forall (k71 : Hvl) (d106 : (Trace TyVar)) (S36 : Ty) (s35 : Tm) ,
          ((subst_TyVar_Ty (weakenTrace d106 k71) S36 (subst_TmVar_Ty (weakenTrace X0 k71) s35 S35)) =
          (subst_TmVar_Ty (weakenTrace X0 k71) (subst_TyVar_Tm d106 S36 s35) (subst_TyVar_Ty (weakenTrace (XS TmVar d106) k71) S36 S35))))).
    Proof.
      pose proof (subst_TyVar__subst_TmVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TyVar__subst_TmVar_0_Kind_comm_ind  :
      (forall (K31 : Kind) ,
        (forall (k71 : Hvl) (d106 : (Trace TyVar)) (S36 : Ty) (s35 : Tm) ,
          ((subst_TyVar_Kind (weakenTrace d106 k71) S36 (subst_TmVar_Kind (weakenTrace X0 k71) s35 K31)) =
          (subst_TmVar_Kind (weakenTrace X0 k71) (subst_TyVar_Tm d106 S36 s35) (subst_TyVar_Kind (weakenTrace (XS TmVar d106) k71) S36 K31))))).
    Proof.
      pose proof (subst_TyVar__subst_TmVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_TyVar__subst_TyVar_0_Decls_Tm_Ty_Kind_comm_ind := (ind_Decls_Tm_Ty_Kind (fun (d105 : Decls) =>
      (forall (k71 : Hvl) (d106 : (Trace TyVar)) (S35 : Ty) (S36 : Ty) ,
        ((subst_TyVar_Decls (weakenTrace d106 k71) S35 (subst_TyVar_Decls (weakenTrace X0 k71) S36 d105)) =
        (subst_TyVar_Decls (weakenTrace X0 k71) (subst_TyVar_Ty d106 S35 S36) (subst_TyVar_Decls (weakenTrace (XS TyVar d106) k71) S35 d105))))) (fun (s34 : Tm) =>
      (forall (k71 : Hvl) (d105 : (Trace TyVar)) (S35 : Ty) (S36 : Ty) ,
        ((subst_TyVar_Tm (weakenTrace d105 k71) S35 (subst_TyVar_Tm (weakenTrace X0 k71) S36 s34)) =
        (subst_TyVar_Tm (weakenTrace X0 k71) (subst_TyVar_Ty d105 S35 S36) (subst_TyVar_Tm (weakenTrace (XS TyVar d105) k71) S35 s34))))) (fun (S35 : Ty) =>
      (forall (k71 : Hvl) (d105 : (Trace TyVar)) (S36 : Ty) (S37 : Ty) ,
        ((subst_TyVar_Ty (weakenTrace d105 k71) S36 (subst_TyVar_Ty (weakenTrace X0 k71) S37 S35)) =
        (subst_TyVar_Ty (weakenTrace X0 k71) (subst_TyVar_Ty d105 S36 S37) (subst_TyVar_Ty (weakenTrace (XS TyVar d105) k71) S36 S35))))) (fun (K31 : Kind) =>
      (forall (k71 : Hvl) (d105 : (Trace TyVar)) (S35 : Ty) (S36 : Ty) ,
        ((subst_TyVar_Kind (weakenTrace d105 k71) S35 (subst_TyVar_Kind (weakenTrace X0 k71) S36 K31)) =
        (subst_TyVar_Kind (weakenTrace X0 k71) (subst_TyVar_Ty d105 S35 S36) (subst_TyVar_Kind (weakenTrace (XS TyVar d105) k71) S35 K31))))) (fun (k71 : Hvl) (d105 : (Trace TyVar)) (S35 : Ty) (S36 : Ty) =>
      eq_refl) (fun (t : Tm) IHt95 (d : Decls) IHd (k71 : Hvl) (d105 : (Trace TyVar)) (S35 : Ty) (S36 : Ty) =>
      (f_equal2 DConsTm (IHt95 k71 d105 S35 S36) (eq_trans (f_equal3 subst_TyVar_Decls (weakenTrace_append TyVar d105 k71 (HS TmVar H0)) eq_refl (f_equal3 subst_TyVar_Decls (weakenTrace_append TyVar X0 k71 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHd (HS TmVar k71) d105 S35 S36) (f_equal3 subst_TyVar_Decls (eq_sym (weakenTrace_append TyVar X0 k71 (HS TmVar H0))) eq_refl (f_equal3 subst_TyVar_Decls (eq_sym (weakenTrace_append TyVar (XS TyVar d105) k71 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (K : Kind) IHK31 (d : Decls) IHd (k71 : Hvl) (d105 : (Trace TyVar)) (S35 : Ty) (S36 : Ty) =>
      (f_equal2 DConsTy (IHK31 k71 d105 S35 S36) (eq_trans (f_equal3 subst_TyVar_Decls (weakenTrace_append TyVar d105 k71 (HS TyVar H0)) eq_refl (f_equal3 subst_TyVar_Decls (weakenTrace_append TyVar X0 k71 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHd (HS TyVar k71) d105 S35 S36) (f_equal3 subst_TyVar_Decls (eq_sym (weakenTrace_append TyVar X0 k71 (HS TyVar H0))) eq_refl (f_equal3 subst_TyVar_Decls (eq_sym (weakenTrace_append TyVar (XS TyVar d105) k71 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (x95 : (Index TmVar)) =>
      (fun (k71 : Hvl) (d105 : (Trace TyVar)) (S35 : Ty) (S36 : Ty) =>
        eq_refl)) (fun (T : Ty) IHT79 (t : Tm) IHt95 (k71 : Hvl) (d105 : (Trace TyVar)) (S35 : Ty) (S36 : Ty) =>
      (f_equal2 Abs (IHT79 k71 d105 S35 S36) (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar d105 k71 (HS TmVar H0)) eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k71 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt95 (HS TmVar k71) d105 S35 S36) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k71 (HS TmVar H0))) eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar (XS TyVar d105) k71 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt95 (t2 : Tm) IHt96 (k71 : Hvl) (d105 : (Trace TyVar)) (S35 : Ty) (S36 : Ty) =>
      (f_equal2 App (IHt95 k71 d105 S35 S36) (IHt96 k71 d105 S35 S36))) (fun (d : Decls) IHd (t : Tm) IHt95 (k71 : Hvl) (d105 : (Trace TyVar)) (S35 : Ty) (S36 : Ty) =>
      (f_equal2 Let (IHd k71 d105 S35 S36) (eq_trans (f_equal3 subst_TyVar_Tm (eq_trans (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (stability_subst_TyVar__bind _ _ _) (eq_refl H0))) (weakenTrace_append TyVar d105 k71 (bind d))) eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k71 (bind d)) eq_refl eq_refl)) (eq_trans (IHt95 (appendHvl k71 (bind d)) d105 S35 S36) (f_equal3 subst_TyVar_Tm (eq_trans (eq_sym (weakenTrace_append TyVar X0 k71 (bind d))) (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (eq_sym (stability_subst_TyVar__bind _ _ _)) (eq_refl H0)))) eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar (XS TyVar d105) k71 (bind d))) eq_refl eq_refl)))))) (fun (X47 : (Index TyVar)) =>
      (fun (k71 : Hvl) (d105 : (Trace TyVar)) (S35 : Ty) (S36 : Ty) =>
        (subst_TyVar_Ty_subst_TyVar_Index0_commright_ind d105 S35 S36 k71 X47))) (fun (T1 : Ty) IHT79 (T2 : Ty) IHT80 (k71 : Hvl) (d105 : (Trace TyVar)) (S35 : Ty) (S36 : Ty) =>
      (f_equal2 TPi (IHT79 k71 d105 S35 S36) (eq_trans (f_equal3 subst_TyVar_Ty (weakenTrace_append TyVar d105 k71 (HS TmVar H0)) eq_refl (f_equal3 subst_TyVar_Ty (weakenTrace_append TyVar X0 k71 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHT80 (HS TmVar k71) d105 S35 S36) (f_equal3 subst_TyVar_Ty (eq_sym (weakenTrace_append TyVar X0 k71 (HS TmVar H0))) eq_refl (f_equal3 subst_TyVar_Ty (eq_sym (weakenTrace_append TyVar (XS TyVar d105) k71 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (T : Ty) IHT79 (t : Tm) IHt95 (k71 : Hvl) (d105 : (Trace TyVar)) (S35 : Ty) (S36 : Ty) =>
      (f_equal2 TApp (IHT79 k71 d105 S35 S36) (IHt95 k71 d105 S35 S36))) (fun (k71 : Hvl) (d105 : (Trace TyVar)) (S35 : Ty) (S36 : Ty) =>
      eq_refl) (fun (T : Ty) IHT79 (K : Kind) IHK31 (k71 : Hvl) (d105 : (Trace TyVar)) (S35 : Ty) (S36 : Ty) =>
      (f_equal2 KPi (IHT79 k71 d105 S35 S36) (eq_trans (f_equal3 subst_TyVar_Kind (weakenTrace_append TyVar d105 k71 (HS TmVar H0)) eq_refl (f_equal3 subst_TyVar_Kind (weakenTrace_append TyVar X0 k71 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHK31 (HS TmVar k71) d105 S35 S36) (f_equal3 subst_TyVar_Kind (eq_sym (weakenTrace_append TyVar X0 k71 (HS TmVar H0))) eq_refl (f_equal3 subst_TyVar_Kind (eq_sym (weakenTrace_append TyVar (XS TyVar d105) k71 (HS TmVar H0))) eq_refl eq_refl))))))).
    Lemma subst_TyVar__subst_TyVar_0_Decls_comm_ind  :
      (forall (d105 : Decls) ,
        (forall (k71 : Hvl) (d106 : (Trace TyVar)) (S35 : Ty) (S36 : Ty) ,
          ((subst_TyVar_Decls (weakenTrace d106 k71) S35 (subst_TyVar_Decls (weakenTrace X0 k71) S36 d105)) =
          (subst_TyVar_Decls (weakenTrace X0 k71) (subst_TyVar_Ty d106 S35 S36) (subst_TyVar_Decls (weakenTrace (XS TyVar d106) k71) S35 d105))))).
    Proof.
      pose proof (subst_TyVar__subst_TyVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TyVar__subst_TyVar_0_Tm_comm_ind  :
      (forall (s34 : Tm) ,
        (forall (k71 : Hvl) (d106 : (Trace TyVar)) (S35 : Ty) (S36 : Ty) ,
          ((subst_TyVar_Tm (weakenTrace d106 k71) S35 (subst_TyVar_Tm (weakenTrace X0 k71) S36 s34)) =
          (subst_TyVar_Tm (weakenTrace X0 k71) (subst_TyVar_Ty d106 S35 S36) (subst_TyVar_Tm (weakenTrace (XS TyVar d106) k71) S35 s34))))).
    Proof.
      pose proof (subst_TyVar__subst_TyVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TyVar__subst_TyVar_0_Ty_comm_ind  :
      (forall (S35 : Ty) ,
        (forall (k71 : Hvl) (d106 : (Trace TyVar)) (S36 : Ty) (S37 : Ty) ,
          ((subst_TyVar_Ty (weakenTrace d106 k71) S36 (subst_TyVar_Ty (weakenTrace X0 k71) S37 S35)) =
          (subst_TyVar_Ty (weakenTrace X0 k71) (subst_TyVar_Ty d106 S36 S37) (subst_TyVar_Ty (weakenTrace (XS TyVar d106) k71) S36 S35))))).
    Proof.
      pose proof (subst_TyVar__subst_TyVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TyVar__subst_TyVar_0_Kind_comm_ind  :
      (forall (K31 : Kind) ,
        (forall (k71 : Hvl) (d106 : (Trace TyVar)) (S36 : Ty) (S37 : Ty) ,
          ((subst_TyVar_Kind (weakenTrace d106 k71) S36 (subst_TyVar_Kind (weakenTrace X0 k71) S37 K31)) =
          (subst_TyVar_Kind (weakenTrace X0 k71) (subst_TyVar_Ty d106 S36 S37) (subst_TyVar_Kind (weakenTrace (XS TyVar d106) k71) S36 K31))))).
    Proof.
      pose proof (subst_TyVar__subst_TyVar_0_Decls_Tm_Ty_Kind_comm_ind).
      destruct_conjs .
      easy .
    Qed.
  End SubstSubstCommInd.
  Section SubstSubstComm.
    Definition subst_TmVar_Decls_subst_TmVar_Decls0_comm (d106 : Decls) : (forall (d105 : (Trace TmVar)) (s34 : Tm) (s35 : Tm) ,
      ((subst_TmVar_Decls d105 s34 (subst_TmVar_Decls X0 s35 d106)) =
      (subst_TmVar_Decls X0 (subst_TmVar_Tm d105 s34 s35) (subst_TmVar_Decls (XS TmVar d105) s34 d106)))) := (subst_TmVar__subst_TmVar_0_Decls_comm_ind d106 H0).
    Definition subst_TmVar_Tm_subst_TmVar_Tm0_comm (s36 : Tm) : (forall (d105 : (Trace TmVar)) (s34 : Tm) (s35 : Tm) ,
      ((subst_TmVar_Tm d105 s34 (subst_TmVar_Tm X0 s35 s36)) =
      (subst_TmVar_Tm X0 (subst_TmVar_Tm d105 s34 s35) (subst_TmVar_Tm (XS TmVar d105) s34 s36)))) := (subst_TmVar__subst_TmVar_0_Tm_comm_ind s36 H0).
    Definition subst_TmVar_Ty_subst_TmVar_Ty0_comm (S35 : Ty) : (forall (d105 : (Trace TmVar)) (s34 : Tm) (s35 : Tm) ,
      ((subst_TmVar_Ty d105 s34 (subst_TmVar_Ty X0 s35 S35)) =
      (subst_TmVar_Ty X0 (subst_TmVar_Tm d105 s34 s35) (subst_TmVar_Ty (XS TmVar d105) s34 S35)))) := (subst_TmVar__subst_TmVar_0_Ty_comm_ind S35 H0).
    Definition subst_TmVar_Kind_subst_TmVar_Kind0_comm (K31 : Kind) : (forall (d105 : (Trace TmVar)) (s34 : Tm) (s35 : Tm) ,
      ((subst_TmVar_Kind d105 s34 (subst_TmVar_Kind X0 s35 K31)) =
      (subst_TmVar_Kind X0 (subst_TmVar_Tm d105 s34 s35) (subst_TmVar_Kind (XS TmVar d105) s34 K31)))) := (subst_TmVar__subst_TmVar_0_Kind_comm_ind K31 H0).
    Definition subst_TmVar_Decls_subst_TyVar_Decls0_comm (d106 : Decls) : (forall (d105 : (Trace TmVar)) (s34 : Tm) (S35 : Ty) ,
      ((subst_TmVar_Decls d105 s34 (subst_TyVar_Decls X0 S35 d106)) =
      (subst_TyVar_Decls X0 (subst_TmVar_Ty d105 s34 S35) (subst_TmVar_Decls (XS TyVar d105) s34 d106)))) := (subst_TmVar__subst_TyVar_0_Decls_comm_ind d106 H0).
    Definition subst_TmVar_Tm_subst_TyVar_Tm0_comm (s35 : Tm) : (forall (d105 : (Trace TmVar)) (s34 : Tm) (S35 : Ty) ,
      ((subst_TmVar_Tm d105 s34 (subst_TyVar_Tm X0 S35 s35)) =
      (subst_TyVar_Tm X0 (subst_TmVar_Ty d105 s34 S35) (subst_TmVar_Tm (XS TyVar d105) s34 s35)))) := (subst_TmVar__subst_TyVar_0_Tm_comm_ind s35 H0).
    Definition subst_TmVar_Ty_subst_TyVar_Ty0_comm (S36 : Ty) : (forall (d105 : (Trace TmVar)) (s34 : Tm) (S35 : Ty) ,
      ((subst_TmVar_Ty d105 s34 (subst_TyVar_Ty X0 S35 S36)) =
      (subst_TyVar_Ty X0 (subst_TmVar_Ty d105 s34 S35) (subst_TmVar_Ty (XS TyVar d105) s34 S36)))) := (subst_TmVar__subst_TyVar_0_Ty_comm_ind S36 H0).
    Definition subst_TmVar_Kind_subst_TyVar_Kind0_comm (K31 : Kind) : (forall (d105 : (Trace TmVar)) (s34 : Tm) (S35 : Ty) ,
      ((subst_TmVar_Kind d105 s34 (subst_TyVar_Kind X0 S35 K31)) =
      (subst_TyVar_Kind X0 (subst_TmVar_Ty d105 s34 S35) (subst_TmVar_Kind (XS TyVar d105) s34 K31)))) := (subst_TmVar__subst_TyVar_0_Kind_comm_ind K31 H0).
    Definition subst_TyVar_Decls_subst_TmVar_Decls0_comm (d106 : Decls) : (forall (d105 : (Trace TyVar)) (S35 : Ty) (s34 : Tm) ,
      ((subst_TyVar_Decls d105 S35 (subst_TmVar_Decls X0 s34 d106)) =
      (subst_TmVar_Decls X0 (subst_TyVar_Tm d105 S35 s34) (subst_TyVar_Decls (XS TmVar d105) S35 d106)))) := (subst_TyVar__subst_TmVar_0_Decls_comm_ind d106 H0).
    Definition subst_TyVar_Tm_subst_TmVar_Tm0_comm (s35 : Tm) : (forall (d105 : (Trace TyVar)) (S35 : Ty) (s34 : Tm) ,
      ((subst_TyVar_Tm d105 S35 (subst_TmVar_Tm X0 s34 s35)) =
      (subst_TmVar_Tm X0 (subst_TyVar_Tm d105 S35 s34) (subst_TyVar_Tm (XS TmVar d105) S35 s35)))) := (subst_TyVar__subst_TmVar_0_Tm_comm_ind s35 H0).
    Definition subst_TyVar_Ty_subst_TmVar_Ty0_comm (S36 : Ty) : (forall (d105 : (Trace TyVar)) (S35 : Ty) (s34 : Tm) ,
      ((subst_TyVar_Ty d105 S35 (subst_TmVar_Ty X0 s34 S36)) =
      (subst_TmVar_Ty X0 (subst_TyVar_Tm d105 S35 s34) (subst_TyVar_Ty (XS TmVar d105) S35 S36)))) := (subst_TyVar__subst_TmVar_0_Ty_comm_ind S36 H0).
    Definition subst_TyVar_Kind_subst_TmVar_Kind0_comm (K31 : Kind) : (forall (d105 : (Trace TyVar)) (S35 : Ty) (s34 : Tm) ,
      ((subst_TyVar_Kind d105 S35 (subst_TmVar_Kind X0 s34 K31)) =
      (subst_TmVar_Kind X0 (subst_TyVar_Tm d105 S35 s34) (subst_TyVar_Kind (XS TmVar d105) S35 K31)))) := (subst_TyVar__subst_TmVar_0_Kind_comm_ind K31 H0).
    Definition subst_TyVar_Decls_subst_TyVar_Decls0_comm (d106 : Decls) : (forall (d105 : (Trace TyVar)) (S35 : Ty) (S36 : Ty) ,
      ((subst_TyVar_Decls d105 S35 (subst_TyVar_Decls X0 S36 d106)) =
      (subst_TyVar_Decls X0 (subst_TyVar_Ty d105 S35 S36) (subst_TyVar_Decls (XS TyVar d105) S35 d106)))) := (subst_TyVar__subst_TyVar_0_Decls_comm_ind d106 H0).
    Definition subst_TyVar_Tm_subst_TyVar_Tm0_comm (s34 : Tm) : (forall (d105 : (Trace TyVar)) (S35 : Ty) (S36 : Ty) ,
      ((subst_TyVar_Tm d105 S35 (subst_TyVar_Tm X0 S36 s34)) =
      (subst_TyVar_Tm X0 (subst_TyVar_Ty d105 S35 S36) (subst_TyVar_Tm (XS TyVar d105) S35 s34)))) := (subst_TyVar__subst_TyVar_0_Tm_comm_ind s34 H0).
    Definition subst_TyVar_Ty_subst_TyVar_Ty0_comm (S37 : Ty) : (forall (d105 : (Trace TyVar)) (S35 : Ty) (S36 : Ty) ,
      ((subst_TyVar_Ty d105 S35 (subst_TyVar_Ty X0 S36 S37)) =
      (subst_TyVar_Ty X0 (subst_TyVar_Ty d105 S35 S36) (subst_TyVar_Ty (XS TyVar d105) S35 S37)))) := (subst_TyVar__subst_TyVar_0_Ty_comm_ind S37 H0).
    Definition subst_TyVar_Kind_subst_TyVar_Kind0_comm (K31 : Kind) : (forall (d105 : (Trace TyVar)) (S35 : Ty) (S36 : Ty) ,
      ((subst_TyVar_Kind d105 S35 (subst_TyVar_Kind X0 S36 K31)) =
      (subst_TyVar_Kind X0 (subst_TyVar_Ty d105 S35 S36) (subst_TyVar_Kind (XS TyVar d105) S35 K31)))) := (subst_TyVar__subst_TyVar_0_Kind_comm_ind K31 H0).
  End SubstSubstComm.
  Section WeakenSubst.
    Lemma weakenDecls_subst_TmVar_Decls  :
      (forall (k71 : Hvl) (d105 : (Trace TmVar)) (s34 : Tm) (d106 : Decls) ,
        ((weakenDecls (subst_TmVar_Decls d105 s34 d106) k71) =
        (subst_TmVar_Decls (weakenTrace d105 k71) s34 (weakenDecls d106 k71)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenTm_subst_TmVar_Tm  :
      (forall (k71 : Hvl) (d105 : (Trace TmVar)) (s34 : Tm) (s35 : Tm) ,
        ((weakenTm (subst_TmVar_Tm d105 s34 s35) k71) =
        (subst_TmVar_Tm (weakenTrace d105 k71) s34 (weakenTm s35 k71)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenTy_subst_TmVar_Ty  :
      (forall (k71 : Hvl) (d105 : (Trace TmVar)) (s34 : Tm) (S35 : Ty) ,
        ((weakenTy (subst_TmVar_Ty d105 s34 S35) k71) =
        (subst_TmVar_Ty (weakenTrace d105 k71) s34 (weakenTy S35 k71)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenKind_subst_TmVar_Kind  :
      (forall (k71 : Hvl) (d105 : (Trace TmVar)) (s34 : Tm) (K31 : Kind) ,
        ((weakenKind (subst_TmVar_Kind d105 s34 K31) k71) =
        (subst_TmVar_Kind (weakenTrace d105 k71) s34 (weakenKind K31 k71)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenDecls_subst_TyVar_Decls  :
      (forall (k71 : Hvl) (d105 : (Trace TyVar)) (S35 : Ty) (d106 : Decls) ,
        ((weakenDecls (subst_TyVar_Decls d105 S35 d106) k71) =
        (subst_TyVar_Decls (weakenTrace d105 k71) S35 (weakenDecls d106 k71)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenTm_subst_TyVar_Tm  :
      (forall (k71 : Hvl) (d105 : (Trace TyVar)) (S35 : Ty) (s34 : Tm) ,
        ((weakenTm (subst_TyVar_Tm d105 S35 s34) k71) =
        (subst_TyVar_Tm (weakenTrace d105 k71) S35 (weakenTm s34 k71)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenTy_subst_TyVar_Ty  :
      (forall (k71 : Hvl) (d105 : (Trace TyVar)) (S35 : Ty) (S36 : Ty) ,
        ((weakenTy (subst_TyVar_Ty d105 S35 S36) k71) =
        (subst_TyVar_Ty (weakenTrace d105 k71) S35 (weakenTy S36 k71)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenKind_subst_TyVar_Kind  :
      (forall (k71 : Hvl) (d105 : (Trace TyVar)) (S35 : Ty) (K31 : Kind) ,
        ((weakenKind (subst_TyVar_Kind d105 S35 K31) k71) =
        (subst_TyVar_Kind (weakenTrace d105 k71) S35 (weakenKind K31 k71)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
  End WeakenSubst.
  Section WeakenAppend.
    Lemma weakenDecls_append  :
      (forall (d105 : Decls) (k71 : Hvl) (k72 : Hvl) ,
        ((weakenDecls (weakenDecls d105 k71) k72) =
        (weakenDecls d105 (appendHvl k71 k72)))).
    Proof.
      needleGenericWeakenAppend .
    Qed.
    Lemma weakenTm_append  :
      (forall (s34 : Tm) (k71 : Hvl) (k72 : Hvl) ,
        ((weakenTm (weakenTm s34 k71) k72) =
        (weakenTm s34 (appendHvl k71 k72)))).
    Proof.
      needleGenericWeakenAppend .
    Qed.
    Lemma weakenTy_append  :
      (forall (S35 : Ty) (k71 : Hvl) (k72 : Hvl) ,
        ((weakenTy (weakenTy S35 k71) k72) =
        (weakenTy S35 (appendHvl k71 k72)))).
    Proof.
      needleGenericWeakenAppend .
    Qed.
    Lemma weakenKind_append  :
      (forall (K31 : Kind) (k71 : Hvl) (k72 : Hvl) ,
        ((weakenKind (weakenKind K31 k71) k72) =
        (weakenKind K31 (appendHvl k71 k72)))).
    Proof.
      needleGenericWeakenAppend .
    Qed.
  End WeakenAppend.
End SubstSubstInteraction.
 Hint Rewrite subst_TmVar_Decls_subst_TmVar_Decls0_comm subst_TyVar_Decls_subst_TyVar_Decls0_comm subst_TmVar_Kind_subst_TmVar_Kind0_comm subst_TyVar_Kind_subst_TyVar_Kind0_comm subst_TmVar_Tm_subst_TmVar_Tm0_comm subst_TyVar_Tm_subst_TyVar_Tm0_comm subst_TmVar_Ty_subst_TmVar_Ty0_comm subst_TyVar_Ty_subst_TyVar_Ty0_comm : interaction.
 Hint Rewrite subst_TmVar_Decls_subst_TmVar_Decls0_comm subst_TyVar_Decls_subst_TyVar_Decls0_comm subst_TmVar_Kind_subst_TmVar_Kind0_comm subst_TyVar_Kind_subst_TyVar_Kind0_comm subst_TmVar_Tm_subst_TmVar_Tm0_comm subst_TyVar_Tm_subst_TyVar_Tm0_comm subst_TmVar_Ty_subst_TmVar_Ty0_comm subst_TyVar_Ty_subst_TyVar_Ty0_comm : subst_subst0.
 Hint Rewrite weakenDecls_shift_TmVar_Decls weakenDecls_shift_TyVar_Decls weakenKind_shift_TmVar_Kind weakenKind_shift_TyVar_Kind weakenTm_shift_TmVar_Tm weakenTm_shift_TyVar_Tm weakenTy_shift_TmVar_Ty weakenTy_shift_TyVar_Ty : weaken_shift.
 Hint Rewrite weakenDecls_subst_TmVar_Decls weakenDecls_subst_TyVar_Decls weakenKind_subst_TmVar_Kind weakenKind_subst_TyVar_Kind weakenTm_subst_TmVar_Tm weakenTm_subst_TyVar_Tm weakenTy_subst_TmVar_Ty weakenTy_subst_TyVar_Ty : weaken_subst.
Section WellFormed.
  Fixpoint wfindex {a : Namespace} (k71 : Hvl) (i : (Index a)) {struct k71} : Prop :=
    match k71 with
      | (H0) => False
      | (HS b k71) => match (eq_namespace_dec a b) with
        | (left e) => match i with
          | (I0) => True
          | (IS i) => (wfindex k71 i)
        end
        | (right n) => (wfindex k71 i)
      end
    end.
  Inductive wfDecls (k71 : Hvl) : Decls -> Prop :=
    | wfDecls_DNil :
        (wfDecls k71 (DNil))
    | wfDecls_DConsTm {t107 : Tm}
        (wf : (wfTm k71 t107))
        {d111 : Decls}
        (wf0 : (wfDecls (HS TmVar k71) d111))
        :
        (wfDecls k71 (DConsTm t107 d111))
    | wfDecls_DConsTy {K35 : Kind}
        (wf : (wfKind k71 K35))
        {d112 : Decls}
        (wf0 : (wfDecls (HS TyVar k71) d112))
        :
        (wfDecls k71 (DConsTy K35 d112))
  with wfTm (k71 : Hvl) : Tm -> Prop :=
    | wfTm_Var
        {x107 : (Index TmVar)}
        (wfi : (wfindex k71 x107)) :
        (wfTm k71 (Var x107))
    | wfTm_Abs {T89 : Ty}
        (wf : (wfTy k71 T89))
        {t107 : Tm}
        (wf0 : (wfTm (HS TmVar k71) t107))
        : (wfTm k71 (Abs T89 t107))
    | wfTm_App {t108 : Tm}
        (wf : (wfTm k71 t108))
        {t109 : Tm}
        (wf0 : (wfTm k71 t109)) :
        (wfTm k71 (App t108 t109))
    | wfTm_Let {d111 : Decls}
        (wf : (wfDecls k71 d111))
        {t110 : Tm}
        (wf0 : (wfTm (appendHvl k71 (bind d111)) t110))
        : (wfTm k71 (Let d111 t110))
  with wfTy (k71 : Hvl) : Ty -> Prop :=
    | wfTy_TVar
        {X53 : (Index TyVar)}
        (wfi : (wfindex k71 X53)) :
        (wfTy k71 (TVar X53))
    | wfTy_TPi {T89 : Ty}
        (wf : (wfTy k71 T89)) {T90 : Ty}
        (wf0 : (wfTy (HS TmVar k71) T90))
        : (wfTy k71 (TPi T89 T90))
    | wfTy_TApp {T91 : Ty}
        (wf : (wfTy k71 T91))
        {t107 : Tm}
        (wf0 : (wfTm k71 t107)) :
        (wfTy k71 (TApp T91 t107))
  with wfKind (k71 : Hvl) : Kind -> Prop :=
    | wfKind_Star :
        (wfKind k71 (Star))
    | wfKind_KPi {T89 : Ty}
        (wf : (wfTy k71 T89))
        {K35 : Kind}
        (wf0 : (wfKind (HS TmVar k71) K35))
        : (wfKind k71 (KPi T89 K35)).
  Scheme ind_wfDecls := Induction for wfDecls Sort Prop
  with ind_wfTm := Induction for wfTm Sort Prop
  with ind_wfTy := Induction for wfTy Sort Prop
  with ind_wfKind := Induction for wfKind Sort Prop.
  Combined Scheme ind_wfDecls_Tm_Ty_Kind from ind_wfDecls, ind_wfTm, ind_wfTy, ind_wfKind.
End WellFormed.
Section ShiftWellFormed.
  Inductive shifthvl_TmVar : (forall (c21 : (Cutoff TmVar)) (k71 : Hvl) (k72 : Hvl) ,
    Prop) :=
    | shifthvl_TmVar_here
        (k71 : Hvl) :
        (shifthvl_TmVar C0 k71 (HS TmVar k71))
    | shifthvl_TmVar_there_TmVar
        (c21 : (Cutoff TmVar))
        (k71 : Hvl) (k72 : Hvl) :
        (shifthvl_TmVar c21 k71 k72) -> (shifthvl_TmVar (CS c21) (HS TmVar k71) (HS TmVar k72))
    | shifthvl_TmVar_there_TyVar
        (c21 : (Cutoff TmVar))
        (k71 : Hvl) (k72 : Hvl) :
        (shifthvl_TmVar c21 k71 k72) -> (shifthvl_TmVar c21 (HS TyVar k71) (HS TyVar k72)).
  Inductive shifthvl_TyVar : (forall (c21 : (Cutoff TyVar)) (k71 : Hvl) (k72 : Hvl) ,
    Prop) :=
    | shifthvl_TyVar_here
        (k71 : Hvl) :
        (shifthvl_TyVar C0 k71 (HS TyVar k71))
    | shifthvl_TyVar_there_TmVar
        (c21 : (Cutoff TyVar))
        (k71 : Hvl) (k72 : Hvl) :
        (shifthvl_TyVar c21 k71 k72) -> (shifthvl_TyVar c21 (HS TmVar k71) (HS TmVar k72))
    | shifthvl_TyVar_there_TyVar
        (c21 : (Cutoff TyVar))
        (k71 : Hvl) (k72 : Hvl) :
        (shifthvl_TyVar c21 k71 k72) -> (shifthvl_TyVar (CS c21) (HS TyVar k71) (HS TyVar k72)).
  Lemma weaken_shifthvl_TmVar  :
    (forall (k71 : Hvl) {c21 : (Cutoff TmVar)} {k72 : Hvl} {k73 : Hvl} ,
      (shifthvl_TmVar c21 k72 k73) -> (shifthvl_TmVar (weakenCutoffTmVar c21 k71) (appendHvl k72 k71) (appendHvl k73 k71))).
  Proof.
    needleGenericWeakenHVarlistInsert .
  Qed.
  Lemma weaken_shifthvl_TyVar  :
    (forall (k71 : Hvl) {c21 : (Cutoff TyVar)} {k72 : Hvl} {k73 : Hvl} ,
      (shifthvl_TyVar c21 k72 k73) -> (shifthvl_TyVar (weakenCutoffTyVar c21 k71) (appendHvl k72 k71) (appendHvl k73 k71))).
  Proof.
    needleGenericWeakenHVarlistInsert .
  Qed.
  Lemma shift_TmVar__wfindex_TmVar  :
    (forall (c21 : (Cutoff TmVar)) (k71 : Hvl) (k72 : Hvl) (ins : (shifthvl_TmVar c21 k71 k72)) (x107 : (Index TmVar)) ,
      (wfindex k71 x107) -> (wfindex k72 (shift_TmVar_Index c21 x107))).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma shift_TmVar__wfindex_TyVar  :
    (forall (c21 : (Cutoff TmVar)) (k71 : Hvl) (k72 : Hvl) (ins : (shifthvl_TmVar c21 k71 k72)) (X53 : (Index TyVar)) ,
      (wfindex k71 X53) -> (wfindex k72 X53)).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma shift_TyVar__wfindex_TmVar  :
    (forall (c21 : (Cutoff TyVar)) (k71 : Hvl) (k72 : Hvl) (ins : (shifthvl_TyVar c21 k71 k72)) (x107 : (Index TmVar)) ,
      (wfindex k71 x107) -> (wfindex k72 x107)).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma shift_TyVar__wfindex_TyVar  :
    (forall (c21 : (Cutoff TyVar)) (k71 : Hvl) (k72 : Hvl) (ins : (shifthvl_TyVar c21 k71 k72)) (X53 : (Index TyVar)) ,
      (wfindex k71 X53) -> (wfindex k72 (shift_TyVar_Index c21 X53))).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Definition shift_TmVar__wfDecls_Tm_Ty_Kind : (forall (k71 : Hvl) ,
    (forall (d111 : Decls) (wf : (wfDecls k71 d111)) ,
      (forall (c21 : (Cutoff TmVar)) (k72 : Hvl) ,
        (shifthvl_TmVar c21 k71 k72) -> (wfDecls k72 (shift_TmVar_Decls c21 d111)))) /\
    (forall (s34 : Tm) (wf : (wfTm k71 s34)) ,
      (forall (c21 : (Cutoff TmVar)) (k72 : Hvl) ,
        (shifthvl_TmVar c21 k71 k72) -> (wfTm k72 (shift_TmVar_Tm c21 s34)))) /\
    (forall (S35 : Ty) (wf : (wfTy k71 S35)) ,
      (forall (c21 : (Cutoff TmVar)) (k72 : Hvl) ,
        (shifthvl_TmVar c21 k71 k72) -> (wfTy k72 (shift_TmVar_Ty c21 S35)))) /\
    (forall (K35 : Kind) (wf : (wfKind k71 K35)) ,
      (forall (c21 : (Cutoff TmVar)) (k72 : Hvl) ,
        (shifthvl_TmVar c21 k71 k72) -> (wfKind k72 (shift_TmVar_Kind c21 K35))))) := (ind_wfDecls_Tm_Ty_Kind (fun (k71 : Hvl) (d111 : Decls) (wf : (wfDecls k71 d111)) =>
    (forall (c21 : (Cutoff TmVar)) (k72 : Hvl) ,
      (shifthvl_TmVar c21 k71 k72) -> (wfDecls k72 (shift_TmVar_Decls c21 d111)))) (fun (k71 : Hvl) (s34 : Tm) (wf : (wfTm k71 s34)) =>
    (forall (c21 : (Cutoff TmVar)) (k72 : Hvl) ,
      (shifthvl_TmVar c21 k71 k72) -> (wfTm k72 (shift_TmVar_Tm c21 s34)))) (fun (k71 : Hvl) (S35 : Ty) (wf : (wfTy k71 S35)) =>
    (forall (c21 : (Cutoff TmVar)) (k72 : Hvl) ,
      (shifthvl_TmVar c21 k71 k72) -> (wfTy k72 (shift_TmVar_Ty c21 S35)))) (fun (k71 : Hvl) (K35 : Kind) (wf : (wfKind k71 K35)) =>
    (forall (c21 : (Cutoff TmVar)) (k72 : Hvl) ,
      (shifthvl_TmVar c21 k71 k72) -> (wfKind k72 (shift_TmVar_Kind c21 K35)))) (fun (k71 : Hvl) (c21 : (Cutoff TmVar)) (k72 : Hvl) (ins : (shifthvl_TmVar c21 k71 k72)) =>
    (wfDecls_DNil k72)) (fun (k71 : Hvl) (t : Tm) (wf : (wfTm k71 t)) IHt107 (d : Decls) (wf0 : (wfDecls (HS TmVar k71) d)) IHd (c21 : (Cutoff TmVar)) (k72 : Hvl) (ins : (shifthvl_TmVar c21 k71 k72)) =>
    (wfDecls_DConsTm k72 (IHt107 c21 k72 (weaken_shifthvl_TmVar H0 ins)) (IHd (CS c21) (HS TmVar k72) (weaken_shifthvl_TmVar (HS TmVar H0) ins)))) (fun (k71 : Hvl) (K : Kind) (wf : (wfKind k71 K)) IHK35 (d : Decls) (wf0 : (wfDecls (HS TyVar k71) d)) IHd (c21 : (Cutoff TmVar)) (k72 : Hvl) (ins : (shifthvl_TmVar c21 k71 k72)) =>
    (wfDecls_DConsTy k72 (IHK35 c21 k72 (weaken_shifthvl_TmVar H0 ins)) (IHd c21 (HS TyVar k72) (weaken_shifthvl_TmVar (HS TyVar H0) ins)))) (fun (k71 : Hvl) (x107 : (Index TmVar)) (wfi : (wfindex k71 x107)) (c21 : (Cutoff TmVar)) (k72 : Hvl) (ins : (shifthvl_TmVar c21 k71 k72)) =>
    (wfTm_Var k72 (shift_TmVar__wfindex_TmVar c21 k71 k72 ins x107 wfi))) (fun (k71 : Hvl) (T : Ty) (wf : (wfTy k71 T)) IHT89 (t : Tm) (wf0 : (wfTm (HS TmVar k71) t)) IHt107 (c21 : (Cutoff TmVar)) (k72 : Hvl) (ins : (shifthvl_TmVar c21 k71 k72)) =>
    (wfTm_Abs k72 (IHT89 c21 k72 (weaken_shifthvl_TmVar H0 ins)) (IHt107 (CS c21) (HS TmVar k72) (weaken_shifthvl_TmVar (HS TmVar H0) ins)))) (fun (k71 : Hvl) (t1 : Tm) (wf : (wfTm k71 t1)) IHt107 (t2 : Tm) (wf0 : (wfTm k71 t2)) IHt108 (c21 : (Cutoff TmVar)) (k72 : Hvl) (ins : (shifthvl_TmVar c21 k71 k72)) =>
    (wfTm_App k72 (IHt107 c21 k72 (weaken_shifthvl_TmVar H0 ins)) (IHt108 c21 k72 (weaken_shifthvl_TmVar H0 ins)))) (fun (k71 : Hvl) (d : Decls) (wf : (wfDecls k71 d)) IHd (t : Tm) (wf0 : (wfTm (appendHvl k71 (bind d)) t)) IHt107 (c21 : (Cutoff TmVar)) (k72 : Hvl) (ins : (shifthvl_TmVar c21 k71 k72)) =>
    (wfTm_Let k72 (IHd c21 k72 (weaken_shifthvl_TmVar H0 ins)) (eq_ind2 wfTm (f_equal2 appendHvl (eq_refl k72) (f_equal2 appendHvl (eq_sym (stability_shift_TmVar__bind _ _)) (eq_refl H0))) eq_refl (IHt107 (weakenCutoffTmVar c21 (bind d)) (appendHvl k72 (bind d)) (weaken_shifthvl_TmVar (bind d) ins))))) (fun (k71 : Hvl) (X53 : (Index TyVar)) (wfi : (wfindex k71 X53)) (c21 : (Cutoff TmVar)) (k72 : Hvl) (ins : (shifthvl_TmVar c21 k71 k72)) =>
    (wfTy_TVar k72 (shift_TmVar__wfindex_TyVar c21 k71 k72 ins X53 wfi))) (fun (k71 : Hvl) (T1 : Ty) (wf : (wfTy k71 T1)) IHT89 (T2 : Ty) (wf0 : (wfTy (HS TmVar k71) T2)) IHT90 (c21 : (Cutoff TmVar)) (k72 : Hvl) (ins : (shifthvl_TmVar c21 k71 k72)) =>
    (wfTy_TPi k72 (IHT89 c21 k72 (weaken_shifthvl_TmVar H0 ins)) (IHT90 (CS c21) (HS TmVar k72) (weaken_shifthvl_TmVar (HS TmVar H0) ins)))) (fun (k71 : Hvl) (T : Ty) (wf : (wfTy k71 T)) IHT89 (t : Tm) (wf0 : (wfTm k71 t)) IHt107 (c21 : (Cutoff TmVar)) (k72 : Hvl) (ins : (shifthvl_TmVar c21 k71 k72)) =>
    (wfTy_TApp k72 (IHT89 c21 k72 (weaken_shifthvl_TmVar H0 ins)) (IHt107 c21 k72 (weaken_shifthvl_TmVar H0 ins)))) (fun (k71 : Hvl) (c21 : (Cutoff TmVar)) (k72 : Hvl) (ins : (shifthvl_TmVar c21 k71 k72)) =>
    (wfKind_Star k72)) (fun (k71 : Hvl) (T : Ty) (wf : (wfTy k71 T)) IHT89 (K : Kind) (wf0 : (wfKind (HS TmVar k71) K)) IHK35 (c21 : (Cutoff TmVar)) (k72 : Hvl) (ins : (shifthvl_TmVar c21 k71 k72)) =>
    (wfKind_KPi k72 (IHT89 c21 k72 (weaken_shifthvl_TmVar H0 ins)) (IHK35 (CS c21) (HS TmVar k72) (weaken_shifthvl_TmVar (HS TmVar H0) ins))))).
  Lemma shift_TmVar__wfDecls (k71 : Hvl) :
    (forall (d111 : Decls) (wf : (wfDecls k71 d111)) ,
      (forall (c21 : (Cutoff TmVar)) (k72 : Hvl) ,
        (shifthvl_TmVar c21 k71 k72) -> (wfDecls k72 (shift_TmVar_Decls c21 d111)))).
  Proof.
    pose proof ((shift_TmVar__wfDecls_Tm_Ty_Kind k71)).
    destruct_conjs .
    easy .
  Qed.
  Lemma shift_TmVar__wfTm (k71 : Hvl) :
    (forall (s34 : Tm) (wf0 : (wfTm k71 s34)) ,
      (forall (c22 : (Cutoff TmVar)) (k73 : Hvl) ,
        (shifthvl_TmVar c22 k71 k73) -> (wfTm k73 (shift_TmVar_Tm c22 s34)))).
  Proof.
    pose proof ((shift_TmVar__wfDecls_Tm_Ty_Kind k71)).
    destruct_conjs .
    easy .
  Qed.
  Lemma shift_TmVar__wfTy (k71 : Hvl) :
    (forall (S35 : Ty) (wf1 : (wfTy k71 S35)) ,
      (forall (c23 : (Cutoff TmVar)) (k74 : Hvl) ,
        (shifthvl_TmVar c23 k71 k74) -> (wfTy k74 (shift_TmVar_Ty c23 S35)))).
  Proof.
    pose proof ((shift_TmVar__wfDecls_Tm_Ty_Kind k71)).
    destruct_conjs .
    easy .
  Qed.
  Lemma shift_TmVar__wfKind (k71 : Hvl) :
    (forall (K35 : Kind) (wf2 : (wfKind k71 K35)) ,
      (forall (c24 : (Cutoff TmVar)) (k75 : Hvl) ,
        (shifthvl_TmVar c24 k71 k75) -> (wfKind k75 (shift_TmVar_Kind c24 K35)))).
  Proof.
    pose proof ((shift_TmVar__wfDecls_Tm_Ty_Kind k71)).
    destruct_conjs .
    easy .
  Qed.
  Definition shift_TyVar__wfDecls_Tm_Ty_Kind : (forall (k71 : Hvl) ,
    (forall (d111 : Decls) (wf : (wfDecls k71 d111)) ,
      (forall (c21 : (Cutoff TyVar)) (k72 : Hvl) ,
        (shifthvl_TyVar c21 k71 k72) -> (wfDecls k72 (shift_TyVar_Decls c21 d111)))) /\
    (forall (s34 : Tm) (wf : (wfTm k71 s34)) ,
      (forall (c21 : (Cutoff TyVar)) (k72 : Hvl) ,
        (shifthvl_TyVar c21 k71 k72) -> (wfTm k72 (shift_TyVar_Tm c21 s34)))) /\
    (forall (S35 : Ty) (wf : (wfTy k71 S35)) ,
      (forall (c21 : (Cutoff TyVar)) (k72 : Hvl) ,
        (shifthvl_TyVar c21 k71 k72) -> (wfTy k72 (shift_TyVar_Ty c21 S35)))) /\
    (forall (K35 : Kind) (wf : (wfKind k71 K35)) ,
      (forall (c21 : (Cutoff TyVar)) (k72 : Hvl) ,
        (shifthvl_TyVar c21 k71 k72) -> (wfKind k72 (shift_TyVar_Kind c21 K35))))) := (ind_wfDecls_Tm_Ty_Kind (fun (k71 : Hvl) (d111 : Decls) (wf : (wfDecls k71 d111)) =>
    (forall (c21 : (Cutoff TyVar)) (k72 : Hvl) ,
      (shifthvl_TyVar c21 k71 k72) -> (wfDecls k72 (shift_TyVar_Decls c21 d111)))) (fun (k71 : Hvl) (s34 : Tm) (wf : (wfTm k71 s34)) =>
    (forall (c21 : (Cutoff TyVar)) (k72 : Hvl) ,
      (shifthvl_TyVar c21 k71 k72) -> (wfTm k72 (shift_TyVar_Tm c21 s34)))) (fun (k71 : Hvl) (S35 : Ty) (wf : (wfTy k71 S35)) =>
    (forall (c21 : (Cutoff TyVar)) (k72 : Hvl) ,
      (shifthvl_TyVar c21 k71 k72) -> (wfTy k72 (shift_TyVar_Ty c21 S35)))) (fun (k71 : Hvl) (K35 : Kind) (wf : (wfKind k71 K35)) =>
    (forall (c21 : (Cutoff TyVar)) (k72 : Hvl) ,
      (shifthvl_TyVar c21 k71 k72) -> (wfKind k72 (shift_TyVar_Kind c21 K35)))) (fun (k71 : Hvl) (c21 : (Cutoff TyVar)) (k72 : Hvl) (ins : (shifthvl_TyVar c21 k71 k72)) =>
    (wfDecls_DNil k72)) (fun (k71 : Hvl) (t : Tm) (wf : (wfTm k71 t)) IHt107 (d : Decls) (wf0 : (wfDecls (HS TmVar k71) d)) IHd (c21 : (Cutoff TyVar)) (k72 : Hvl) (ins : (shifthvl_TyVar c21 k71 k72)) =>
    (wfDecls_DConsTm k72 (IHt107 c21 k72 (weaken_shifthvl_TyVar H0 ins)) (IHd c21 (HS TmVar k72) (weaken_shifthvl_TyVar (HS TmVar H0) ins)))) (fun (k71 : Hvl) (K : Kind) (wf : (wfKind k71 K)) IHK35 (d : Decls) (wf0 : (wfDecls (HS TyVar k71) d)) IHd (c21 : (Cutoff TyVar)) (k72 : Hvl) (ins : (shifthvl_TyVar c21 k71 k72)) =>
    (wfDecls_DConsTy k72 (IHK35 c21 k72 (weaken_shifthvl_TyVar H0 ins)) (IHd (CS c21) (HS TyVar k72) (weaken_shifthvl_TyVar (HS TyVar H0) ins)))) (fun (k71 : Hvl) (x107 : (Index TmVar)) (wfi : (wfindex k71 x107)) (c21 : (Cutoff TyVar)) (k72 : Hvl) (ins : (shifthvl_TyVar c21 k71 k72)) =>
    (wfTm_Var k72 (shift_TyVar__wfindex_TmVar c21 k71 k72 ins x107 wfi))) (fun (k71 : Hvl) (T : Ty) (wf : (wfTy k71 T)) IHT89 (t : Tm) (wf0 : (wfTm (HS TmVar k71) t)) IHt107 (c21 : (Cutoff TyVar)) (k72 : Hvl) (ins : (shifthvl_TyVar c21 k71 k72)) =>
    (wfTm_Abs k72 (IHT89 c21 k72 (weaken_shifthvl_TyVar H0 ins)) (IHt107 c21 (HS TmVar k72) (weaken_shifthvl_TyVar (HS TmVar H0) ins)))) (fun (k71 : Hvl) (t1 : Tm) (wf : (wfTm k71 t1)) IHt107 (t2 : Tm) (wf0 : (wfTm k71 t2)) IHt108 (c21 : (Cutoff TyVar)) (k72 : Hvl) (ins : (shifthvl_TyVar c21 k71 k72)) =>
    (wfTm_App k72 (IHt107 c21 k72 (weaken_shifthvl_TyVar H0 ins)) (IHt108 c21 k72 (weaken_shifthvl_TyVar H0 ins)))) (fun (k71 : Hvl) (d : Decls) (wf : (wfDecls k71 d)) IHd (t : Tm) (wf0 : (wfTm (appendHvl k71 (bind d)) t)) IHt107 (c21 : (Cutoff TyVar)) (k72 : Hvl) (ins : (shifthvl_TyVar c21 k71 k72)) =>
    (wfTm_Let k72 (IHd c21 k72 (weaken_shifthvl_TyVar H0 ins)) (eq_ind2 wfTm (f_equal2 appendHvl (eq_refl k72) (f_equal2 appendHvl (eq_sym (stability_shift_TyVar__bind _ _)) (eq_refl H0))) eq_refl (IHt107 (weakenCutoffTyVar c21 (bind d)) (appendHvl k72 (bind d)) (weaken_shifthvl_TyVar (bind d) ins))))) (fun (k71 : Hvl) (X53 : (Index TyVar)) (wfi : (wfindex k71 X53)) (c21 : (Cutoff TyVar)) (k72 : Hvl) (ins : (shifthvl_TyVar c21 k71 k72)) =>
    (wfTy_TVar k72 (shift_TyVar__wfindex_TyVar c21 k71 k72 ins X53 wfi))) (fun (k71 : Hvl) (T1 : Ty) (wf : (wfTy k71 T1)) IHT89 (T2 : Ty) (wf0 : (wfTy (HS TmVar k71) T2)) IHT90 (c21 : (Cutoff TyVar)) (k72 : Hvl) (ins : (shifthvl_TyVar c21 k71 k72)) =>
    (wfTy_TPi k72 (IHT89 c21 k72 (weaken_shifthvl_TyVar H0 ins)) (IHT90 c21 (HS TmVar k72) (weaken_shifthvl_TyVar (HS TmVar H0) ins)))) (fun (k71 : Hvl) (T : Ty) (wf : (wfTy k71 T)) IHT89 (t : Tm) (wf0 : (wfTm k71 t)) IHt107 (c21 : (Cutoff TyVar)) (k72 : Hvl) (ins : (shifthvl_TyVar c21 k71 k72)) =>
    (wfTy_TApp k72 (IHT89 c21 k72 (weaken_shifthvl_TyVar H0 ins)) (IHt107 c21 k72 (weaken_shifthvl_TyVar H0 ins)))) (fun (k71 : Hvl) (c21 : (Cutoff TyVar)) (k72 : Hvl) (ins : (shifthvl_TyVar c21 k71 k72)) =>
    (wfKind_Star k72)) (fun (k71 : Hvl) (T : Ty) (wf : (wfTy k71 T)) IHT89 (K : Kind) (wf0 : (wfKind (HS TmVar k71) K)) IHK35 (c21 : (Cutoff TyVar)) (k72 : Hvl) (ins : (shifthvl_TyVar c21 k71 k72)) =>
    (wfKind_KPi k72 (IHT89 c21 k72 (weaken_shifthvl_TyVar H0 ins)) (IHK35 c21 (HS TmVar k72) (weaken_shifthvl_TyVar (HS TmVar H0) ins))))).
  Lemma shift_TyVar__wfDecls (k71 : Hvl) :
    (forall (d111 : Decls) (wf : (wfDecls k71 d111)) ,
      (forall (c21 : (Cutoff TyVar)) (k72 : Hvl) ,
        (shifthvl_TyVar c21 k71 k72) -> (wfDecls k72 (shift_TyVar_Decls c21 d111)))).
  Proof.
    pose proof ((shift_TyVar__wfDecls_Tm_Ty_Kind k71)).
    destruct_conjs .
    easy .
  Qed.
  Lemma shift_TyVar__wfTm (k71 : Hvl) :
    (forall (s34 : Tm) (wf0 : (wfTm k71 s34)) ,
      (forall (c22 : (Cutoff TyVar)) (k73 : Hvl) ,
        (shifthvl_TyVar c22 k71 k73) -> (wfTm k73 (shift_TyVar_Tm c22 s34)))).
  Proof.
    pose proof ((shift_TyVar__wfDecls_Tm_Ty_Kind k71)).
    destruct_conjs .
    easy .
  Qed.
  Lemma shift_TyVar__wfTy (k71 : Hvl) :
    (forall (S35 : Ty) (wf1 : (wfTy k71 S35)) ,
      (forall (c23 : (Cutoff TyVar)) (k74 : Hvl) ,
        (shifthvl_TyVar c23 k71 k74) -> (wfTy k74 (shift_TyVar_Ty c23 S35)))).
  Proof.
    pose proof ((shift_TyVar__wfDecls_Tm_Ty_Kind k71)).
    destruct_conjs .
    easy .
  Qed.
  Lemma shift_TyVar__wfKind (k71 : Hvl) :
    (forall (K35 : Kind) (wf2 : (wfKind k71 K35)) ,
      (forall (c24 : (Cutoff TyVar)) (k75 : Hvl) ,
        (shifthvl_TyVar c24 k71 k75) -> (wfKind k75 (shift_TyVar_Kind c24 K35)))).
  Proof.
    pose proof ((shift_TyVar__wfDecls_Tm_Ty_Kind k71)).
    destruct_conjs .
    easy .
  Qed.
End ShiftWellFormed.
 Hint Resolve shift_TmVar__wfindex_TmVar shift_TmVar__wfindex_TyVar shift_TyVar__wfindex_TmVar shift_TyVar__wfindex_TyVar : infra.
 Hint Resolve shift_TmVar__wfindex_TmVar shift_TmVar__wfindex_TyVar shift_TyVar__wfindex_TmVar shift_TyVar__wfindex_TyVar : shift.
 Hint Resolve shift_TmVar__wfindex_TmVar shift_TmVar__wfindex_TyVar shift_TyVar__wfindex_TmVar shift_TyVar__wfindex_TyVar : shift_wf.
 Hint Resolve shift_TmVar__wfindex_TmVar shift_TmVar__wfindex_TyVar shift_TyVar__wfindex_TmVar shift_TyVar__wfindex_TyVar : wf.
 Hint Resolve shift_TmVar__wfDecls shift_TmVar__wfKind shift_TmVar__wfTm shift_TmVar__wfTy shift_TyVar__wfDecls shift_TyVar__wfKind shift_TyVar__wfTm shift_TyVar__wfTy : infra.
 Hint Resolve shift_TmVar__wfDecls shift_TmVar__wfKind shift_TmVar__wfTm shift_TmVar__wfTy shift_TyVar__wfDecls shift_TyVar__wfKind shift_TyVar__wfTm shift_TyVar__wfTy : shift.
 Hint Resolve shift_TmVar__wfDecls shift_TmVar__wfKind shift_TmVar__wfTm shift_TmVar__wfTy shift_TyVar__wfDecls shift_TyVar__wfKind shift_TyVar__wfTm shift_TyVar__wfTy : shift_wf.
 Hint Resolve shift_TmVar__wfDecls shift_TmVar__wfKind shift_TmVar__wfTm shift_TmVar__wfTy shift_TyVar__wfDecls shift_TyVar__wfKind shift_TyVar__wfTm shift_TyVar__wfTy : wf.
 Hint Constructors shifthvl_TmVar shifthvl_TyVar : infra.
 Hint Constructors shifthvl_TmVar shifthvl_TyVar : shift.
 Hint Constructors shifthvl_TmVar shifthvl_TyVar : shift_wf.
 Hint Constructors shifthvl_TmVar shifthvl_TyVar : wf.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : infra.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : shift.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : shift_wf.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : weaken.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : wf.
Section SubstWellFormed.
  Inductive substhvl_TmVar (k71 : Hvl) : (forall (d111 : (Trace TmVar)) (k72 : Hvl) (k73 : Hvl) ,
    Prop) :=
    | substhvl_TmVar_here :
        (substhvl_TmVar k71 X0 (HS TmVar k71) k71)
    | substhvl_TmVar_there_TmVar
        {d111 : (Trace TmVar)}
        {k72 : Hvl} {k73 : Hvl} :
        (substhvl_TmVar k71 d111 k72 k73) -> (substhvl_TmVar k71 (XS TmVar d111) (HS TmVar k72) (HS TmVar k73))
    | substhvl_TmVar_there_TyVar
        {d111 : (Trace TmVar)}
        {k72 : Hvl} {k73 : Hvl} :
        (substhvl_TmVar k71 d111 k72 k73) -> (substhvl_TmVar k71 (XS TyVar d111) (HS TyVar k72) (HS TyVar k73)).
  Inductive substhvl_TyVar (k71 : Hvl) : (forall (d111 : (Trace TyVar)) (k72 : Hvl) (k73 : Hvl) ,
    Prop) :=
    | substhvl_TyVar_here :
        (substhvl_TyVar k71 X0 (HS TyVar k71) k71)
    | substhvl_TyVar_there_TmVar
        {d111 : (Trace TyVar)}
        {k72 : Hvl} {k73 : Hvl} :
        (substhvl_TyVar k71 d111 k72 k73) -> (substhvl_TyVar k71 (XS TmVar d111) (HS TmVar k72) (HS TmVar k73))
    | substhvl_TyVar_there_TyVar
        {d111 : (Trace TyVar)}
        {k72 : Hvl} {k73 : Hvl} :
        (substhvl_TyVar k71 d111 k72 k73) -> (substhvl_TyVar k71 (XS TyVar d111) (HS TyVar k72) (HS TyVar k73)).
  Lemma weaken_substhvl_TmVar  :
    (forall {k72 : Hvl} (k71 : Hvl) {d111 : (Trace TmVar)} {k73 : Hvl} {k74 : Hvl} ,
      (substhvl_TmVar k72 d111 k73 k74) -> (substhvl_TmVar k72 (weakenTrace d111 k71) (appendHvl k73 k71) (appendHvl k74 k71))).
  Proof.
    needleGenericWeakenSubstHvl .
  Qed.
  Lemma weaken_substhvl_TyVar  :
    (forall {k72 : Hvl} (k71 : Hvl) {d111 : (Trace TyVar)} {k73 : Hvl} {k74 : Hvl} ,
      (substhvl_TyVar k72 d111 k73 k74) -> (substhvl_TyVar k72 (weakenTrace d111 k71) (appendHvl k73 k71) (appendHvl k74 k71))).
  Proof.
    needleGenericWeakenSubstHvl .
  Qed.
  Lemma substhvl_TmVar_wfindex_TmVar {k71 : Hvl} {s34 : Tm} (wft : (wfTm k71 s34)) :
    (forall {d111 : (Trace TmVar)} {k72 : Hvl} {k73 : Hvl} ,
      (substhvl_TmVar k71 d111 k72 k73) -> (forall {x107 : (Index TmVar)} ,
        (wfindex k72 x107) -> (wfTm k73 (subst_TmVar_Index d111 s34 x107)))).
  Proof.
    needleGenericSubstHvlWfIndexHom .
  Qed.
  Lemma substhvl_TyVar_wfindex_TyVar {k71 : Hvl} {S35 : Ty} (wft : (wfTy k71 S35)) :
    (forall {d111 : (Trace TyVar)} {k72 : Hvl} {k73 : Hvl} ,
      (substhvl_TyVar k71 d111 k72 k73) -> (forall {X53 : (Index TyVar)} ,
        (wfindex k72 X53) -> (wfTy k73 (subst_TyVar_Index d111 S35 X53)))).
  Proof.
    needleGenericSubstHvlWfIndexHom .
  Qed.
  Lemma substhvl_TmVar_wfindex_TyVar {k71 : Hvl} :
    (forall {d111 : (Trace TmVar)} {k72 : Hvl} {k73 : Hvl} ,
      (substhvl_TmVar k71 d111 k72 k73) -> (forall {X53 : (Index TyVar)} ,
        (wfindex k72 X53) -> (wfindex k73 X53))).
  Proof.
    needleGenericSubstHvlWfIndexHet .
  Qed.
  Lemma substhvl_TyVar_wfindex_TmVar {k71 : Hvl} :
    (forall {d111 : (Trace TyVar)} {k72 : Hvl} {k73 : Hvl} ,
      (substhvl_TyVar k71 d111 k72 k73) -> (forall {x107 : (Index TmVar)} ,
        (wfindex k72 x107) -> (wfindex k73 x107))).
  Proof.
    needleGenericSubstHvlWfIndexHet .
  Qed.
  Definition substhvl_TmVar_wfDecls_Tm_Ty_Kind {k71 : Hvl} {s34 : Tm} (wf : (wfTm k71 s34)) : (forall (k72 : Hvl) ,
    (forall (d111 : Decls) (wf0 : (wfDecls k72 d111)) ,
      (forall {d112 : (Trace TmVar)} {k73 : Hvl} ,
        (substhvl_TmVar k71 d112 k72 k73) -> (wfDecls k73 (subst_TmVar_Decls d112 s34 d111)))) /\
    (forall (s35 : Tm) (wf0 : (wfTm k72 s35)) ,
      (forall {d111 : (Trace TmVar)} {k73 : Hvl} ,
        (substhvl_TmVar k71 d111 k72 k73) -> (wfTm k73 (subst_TmVar_Tm d111 s34 s35)))) /\
    (forall (S35 : Ty) (wf0 : (wfTy k72 S35)) ,
      (forall {d111 : (Trace TmVar)} {k73 : Hvl} ,
        (substhvl_TmVar k71 d111 k72 k73) -> (wfTy k73 (subst_TmVar_Ty d111 s34 S35)))) /\
    (forall (K35 : Kind) (wf0 : (wfKind k72 K35)) ,
      (forall {d111 : (Trace TmVar)} {k73 : Hvl} ,
        (substhvl_TmVar k71 d111 k72 k73) -> (wfKind k73 (subst_TmVar_Kind d111 s34 K35))))) := (ind_wfDecls_Tm_Ty_Kind (fun (k72 : Hvl) (d111 : Decls) (wf0 : (wfDecls k72 d111)) =>
    (forall {d112 : (Trace TmVar)} {k73 : Hvl} ,
      (substhvl_TmVar k71 d112 k72 k73) -> (wfDecls k73 (subst_TmVar_Decls d112 s34 d111)))) (fun (k72 : Hvl) (s35 : Tm) (wf0 : (wfTm k72 s35)) =>
    (forall {d111 : (Trace TmVar)} {k73 : Hvl} ,
      (substhvl_TmVar k71 d111 k72 k73) -> (wfTm k73 (subst_TmVar_Tm d111 s34 s35)))) (fun (k72 : Hvl) (S35 : Ty) (wf0 : (wfTy k72 S35)) =>
    (forall {d111 : (Trace TmVar)} {k73 : Hvl} ,
      (substhvl_TmVar k71 d111 k72 k73) -> (wfTy k73 (subst_TmVar_Ty d111 s34 S35)))) (fun (k72 : Hvl) (K35 : Kind) (wf0 : (wfKind k72 K35)) =>
    (forall {d111 : (Trace TmVar)} {k73 : Hvl} ,
      (substhvl_TmVar k71 d111 k72 k73) -> (wfKind k73 (subst_TmVar_Kind d111 s34 K35)))) (fun (k72 : Hvl) {d111 : (Trace TmVar)} {k73 : Hvl} (del : (substhvl_TmVar k71 d111 k72 k73)) =>
    (wfDecls_DNil k73)) (fun (k72 : Hvl) (t : Tm) (wf0 : (wfTm k72 t)) IHt107 (d : Decls) (wf1 : (wfDecls (HS TmVar k72) d)) IHd {d111 : (Trace TmVar)} {k73 : Hvl} (del : (substhvl_TmVar k71 d111 k72 k73)) =>
    (wfDecls_DConsTm k73 (IHt107 (weakenTrace d111 H0) k73 (weaken_substhvl_TmVar H0 del)) (IHd (weakenTrace d111 (HS TmVar H0)) (HS TmVar k73) (weaken_substhvl_TmVar (HS TmVar H0) del)))) (fun (k72 : Hvl) (K : Kind) (wf0 : (wfKind k72 K)) IHK35 (d : Decls) (wf1 : (wfDecls (HS TyVar k72) d)) IHd {d111 : (Trace TmVar)} {k73 : Hvl} (del : (substhvl_TmVar k71 d111 k72 k73)) =>
    (wfDecls_DConsTy k73 (IHK35 (weakenTrace d111 H0) k73 (weaken_substhvl_TmVar H0 del)) (IHd (weakenTrace d111 (HS TyVar H0)) (HS TyVar k73) (weaken_substhvl_TmVar (HS TyVar H0) del)))) (fun (k72 : Hvl) {x107 : (Index TmVar)} (wfi : (wfindex k72 x107)) {d111 : (Trace TmVar)} {k73 : Hvl} (del : (substhvl_TmVar k71 d111 k72 k73)) =>
    (substhvl_TmVar_wfindex_TmVar wf del wfi)) (fun (k72 : Hvl) (T : Ty) (wf0 : (wfTy k72 T)) IHT89 (t : Tm) (wf1 : (wfTm (HS TmVar k72) t)) IHt107 {d111 : (Trace TmVar)} {k73 : Hvl} (del : (substhvl_TmVar k71 d111 k72 k73)) =>
    (wfTm_Abs k73 (IHT89 (weakenTrace d111 H0) k73 (weaken_substhvl_TmVar H0 del)) (IHt107 (weakenTrace d111 (HS TmVar H0)) (HS TmVar k73) (weaken_substhvl_TmVar (HS TmVar H0) del)))) (fun (k72 : Hvl) (t1 : Tm) (wf0 : (wfTm k72 t1)) IHt107 (t2 : Tm) (wf1 : (wfTm k72 t2)) IHt108 {d111 : (Trace TmVar)} {k73 : Hvl} (del : (substhvl_TmVar k71 d111 k72 k73)) =>
    (wfTm_App k73 (IHt107 (weakenTrace d111 H0) k73 (weaken_substhvl_TmVar H0 del)) (IHt108 (weakenTrace d111 H0) k73 (weaken_substhvl_TmVar H0 del)))) (fun (k72 : Hvl) (d : Decls) (wf0 : (wfDecls k72 d)) IHd (t : Tm) (wf1 : (wfTm (appendHvl k72 (bind d)) t)) IHt107 {d111 : (Trace TmVar)} {k73 : Hvl} (del : (substhvl_TmVar k71 d111 k72 k73)) =>
    (wfTm_Let k73 (IHd (weakenTrace d111 H0) k73 (weaken_substhvl_TmVar H0 del)) (eq_ind2 wfTm (f_equal2 appendHvl (eq_refl k73) (f_equal2 appendHvl (eq_sym (stability_subst_TmVar__bind _ _ _)) (eq_refl H0))) eq_refl (IHt107 (weakenTrace d111 (bind d)) (appendHvl k73 (bind d)) (weaken_substhvl_TmVar (bind d) del))))) (fun (k72 : Hvl) {X53 : (Index TyVar)} (wfi : (wfindex k72 X53)) {d111 : (Trace TmVar)} {k73 : Hvl} (del : (substhvl_TmVar k71 d111 k72 k73)) =>
    (wfTy_TVar k73 (substhvl_TmVar_wfindex_TyVar del wfi))) (fun (k72 : Hvl) (T1 : Ty) (wf0 : (wfTy k72 T1)) IHT89 (T2 : Ty) (wf1 : (wfTy (HS TmVar k72) T2)) IHT90 {d111 : (Trace TmVar)} {k73 : Hvl} (del : (substhvl_TmVar k71 d111 k72 k73)) =>
    (wfTy_TPi k73 (IHT89 (weakenTrace d111 H0) k73 (weaken_substhvl_TmVar H0 del)) (IHT90 (weakenTrace d111 (HS TmVar H0)) (HS TmVar k73) (weaken_substhvl_TmVar (HS TmVar H0) del)))) (fun (k72 : Hvl) (T : Ty) (wf0 : (wfTy k72 T)) IHT89 (t : Tm) (wf1 : (wfTm k72 t)) IHt107 {d111 : (Trace TmVar)} {k73 : Hvl} (del : (substhvl_TmVar k71 d111 k72 k73)) =>
    (wfTy_TApp k73 (IHT89 (weakenTrace d111 H0) k73 (weaken_substhvl_TmVar H0 del)) (IHt107 (weakenTrace d111 H0) k73 (weaken_substhvl_TmVar H0 del)))) (fun (k72 : Hvl) {d111 : (Trace TmVar)} {k73 : Hvl} (del : (substhvl_TmVar k71 d111 k72 k73)) =>
    (wfKind_Star k73)) (fun (k72 : Hvl) (T : Ty) (wf0 : (wfTy k72 T)) IHT89 (K : Kind) (wf1 : (wfKind (HS TmVar k72) K)) IHK35 {d111 : (Trace TmVar)} {k73 : Hvl} (del : (substhvl_TmVar k71 d111 k72 k73)) =>
    (wfKind_KPi k73 (IHT89 (weakenTrace d111 H0) k73 (weaken_substhvl_TmVar H0 del)) (IHK35 (weakenTrace d111 (HS TmVar H0)) (HS TmVar k73) (weaken_substhvl_TmVar (HS TmVar H0) del))))).
  Lemma substhvl_TmVar_wfDecls {k71 : Hvl} {s34 : Tm} (wf : (wfTm k71 s34)) (k72 : Hvl) (d111 : Decls) (wf0 : (wfDecls k72 d111)) :
    (forall {d112 : (Trace TmVar)} {k73 : Hvl} ,
      (substhvl_TmVar k71 d112 k72 k73) -> (wfDecls k73 (subst_TmVar_Decls d112 s34 d111))).
  Proof.
    apply ((substhvl_TmVar_wfDecls_Tm_Ty_Kind wf k72)).
    auto .
  Qed.
  Lemma substhvl_TmVar_wfTm {k71 : Hvl} {s34 : Tm} (wf : (wfTm k71 s34)) (k72 : Hvl) (s35 : Tm) (wf1 : (wfTm k72 s35)) :
    (forall {d113 : (Trace TmVar)} {k74 : Hvl} ,
      (substhvl_TmVar k71 d113 k72 k74) -> (wfTm k74 (subst_TmVar_Tm d113 s34 s35))).
  Proof.
    apply ((substhvl_TmVar_wfDecls_Tm_Ty_Kind wf k72)).
    auto .
  Qed.
  Lemma substhvl_TmVar_wfTy {k71 : Hvl} {s34 : Tm} (wf : (wfTm k71 s34)) (k72 : Hvl) (S35 : Ty) (wf2 : (wfTy k72 S35)) :
    (forall {d114 : (Trace TmVar)} {k75 : Hvl} ,
      (substhvl_TmVar k71 d114 k72 k75) -> (wfTy k75 (subst_TmVar_Ty d114 s34 S35))).
  Proof.
    apply ((substhvl_TmVar_wfDecls_Tm_Ty_Kind wf k72)).
    auto .
  Qed.
  Lemma substhvl_TmVar_wfKind {k71 : Hvl} {s34 : Tm} (wf : (wfTm k71 s34)) (k72 : Hvl) (K35 : Kind) (wf3 : (wfKind k72 K35)) :
    (forall {d115 : (Trace TmVar)} {k76 : Hvl} ,
      (substhvl_TmVar k71 d115 k72 k76) -> (wfKind k76 (subst_TmVar_Kind d115 s34 K35))).
  Proof.
    apply ((substhvl_TmVar_wfDecls_Tm_Ty_Kind wf k72)).
    auto .
  Qed.
  Definition substhvl_TyVar_wfDecls_Tm_Ty_Kind {k71 : Hvl} {S35 : Ty} (wf : (wfTy k71 S35)) : (forall (k72 : Hvl) ,
    (forall (d111 : Decls) (wf0 : (wfDecls k72 d111)) ,
      (forall {d112 : (Trace TyVar)} {k73 : Hvl} ,
        (substhvl_TyVar k71 d112 k72 k73) -> (wfDecls k73 (subst_TyVar_Decls d112 S35 d111)))) /\
    (forall (s34 : Tm) (wf0 : (wfTm k72 s34)) ,
      (forall {d111 : (Trace TyVar)} {k73 : Hvl} ,
        (substhvl_TyVar k71 d111 k72 k73) -> (wfTm k73 (subst_TyVar_Tm d111 S35 s34)))) /\
    (forall (S36 : Ty) (wf0 : (wfTy k72 S36)) ,
      (forall {d111 : (Trace TyVar)} {k73 : Hvl} ,
        (substhvl_TyVar k71 d111 k72 k73) -> (wfTy k73 (subst_TyVar_Ty d111 S35 S36)))) /\
    (forall (K35 : Kind) (wf0 : (wfKind k72 K35)) ,
      (forall {d111 : (Trace TyVar)} {k73 : Hvl} ,
        (substhvl_TyVar k71 d111 k72 k73) -> (wfKind k73 (subst_TyVar_Kind d111 S35 K35))))) := (ind_wfDecls_Tm_Ty_Kind (fun (k72 : Hvl) (d111 : Decls) (wf0 : (wfDecls k72 d111)) =>
    (forall {d112 : (Trace TyVar)} {k73 : Hvl} ,
      (substhvl_TyVar k71 d112 k72 k73) -> (wfDecls k73 (subst_TyVar_Decls d112 S35 d111)))) (fun (k72 : Hvl) (s34 : Tm) (wf0 : (wfTm k72 s34)) =>
    (forall {d111 : (Trace TyVar)} {k73 : Hvl} ,
      (substhvl_TyVar k71 d111 k72 k73) -> (wfTm k73 (subst_TyVar_Tm d111 S35 s34)))) (fun (k72 : Hvl) (S36 : Ty) (wf0 : (wfTy k72 S36)) =>
    (forall {d111 : (Trace TyVar)} {k73 : Hvl} ,
      (substhvl_TyVar k71 d111 k72 k73) -> (wfTy k73 (subst_TyVar_Ty d111 S35 S36)))) (fun (k72 : Hvl) (K35 : Kind) (wf0 : (wfKind k72 K35)) =>
    (forall {d111 : (Trace TyVar)} {k73 : Hvl} ,
      (substhvl_TyVar k71 d111 k72 k73) -> (wfKind k73 (subst_TyVar_Kind d111 S35 K35)))) (fun (k72 : Hvl) {d111 : (Trace TyVar)} {k73 : Hvl} (del : (substhvl_TyVar k71 d111 k72 k73)) =>
    (wfDecls_DNil k73)) (fun (k72 : Hvl) (t : Tm) (wf0 : (wfTm k72 t)) IHt107 (d : Decls) (wf1 : (wfDecls (HS TmVar k72) d)) IHd {d111 : (Trace TyVar)} {k73 : Hvl} (del : (substhvl_TyVar k71 d111 k72 k73)) =>
    (wfDecls_DConsTm k73 (IHt107 (weakenTrace d111 H0) k73 (weaken_substhvl_TyVar H0 del)) (IHd (weakenTrace d111 (HS TmVar H0)) (HS TmVar k73) (weaken_substhvl_TyVar (HS TmVar H0) del)))) (fun (k72 : Hvl) (K : Kind) (wf0 : (wfKind k72 K)) IHK35 (d : Decls) (wf1 : (wfDecls (HS TyVar k72) d)) IHd {d111 : (Trace TyVar)} {k73 : Hvl} (del : (substhvl_TyVar k71 d111 k72 k73)) =>
    (wfDecls_DConsTy k73 (IHK35 (weakenTrace d111 H0) k73 (weaken_substhvl_TyVar H0 del)) (IHd (weakenTrace d111 (HS TyVar H0)) (HS TyVar k73) (weaken_substhvl_TyVar (HS TyVar H0) del)))) (fun (k72 : Hvl) {x107 : (Index TmVar)} (wfi : (wfindex k72 x107)) {d111 : (Trace TyVar)} {k73 : Hvl} (del : (substhvl_TyVar k71 d111 k72 k73)) =>
    (wfTm_Var k73 (substhvl_TyVar_wfindex_TmVar del wfi))) (fun (k72 : Hvl) (T : Ty) (wf0 : (wfTy k72 T)) IHT89 (t : Tm) (wf1 : (wfTm (HS TmVar k72) t)) IHt107 {d111 : (Trace TyVar)} {k73 : Hvl} (del : (substhvl_TyVar k71 d111 k72 k73)) =>
    (wfTm_Abs k73 (IHT89 (weakenTrace d111 H0) k73 (weaken_substhvl_TyVar H0 del)) (IHt107 (weakenTrace d111 (HS TmVar H0)) (HS TmVar k73) (weaken_substhvl_TyVar (HS TmVar H0) del)))) (fun (k72 : Hvl) (t1 : Tm) (wf0 : (wfTm k72 t1)) IHt107 (t2 : Tm) (wf1 : (wfTm k72 t2)) IHt108 {d111 : (Trace TyVar)} {k73 : Hvl} (del : (substhvl_TyVar k71 d111 k72 k73)) =>
    (wfTm_App k73 (IHt107 (weakenTrace d111 H0) k73 (weaken_substhvl_TyVar H0 del)) (IHt108 (weakenTrace d111 H0) k73 (weaken_substhvl_TyVar H0 del)))) (fun (k72 : Hvl) (d : Decls) (wf0 : (wfDecls k72 d)) IHd (t : Tm) (wf1 : (wfTm (appendHvl k72 (bind d)) t)) IHt107 {d111 : (Trace TyVar)} {k73 : Hvl} (del : (substhvl_TyVar k71 d111 k72 k73)) =>
    (wfTm_Let k73 (IHd (weakenTrace d111 H0) k73 (weaken_substhvl_TyVar H0 del)) (eq_ind2 wfTm (f_equal2 appendHvl (eq_refl k73) (f_equal2 appendHvl (eq_sym (stability_subst_TyVar__bind _ _ _)) (eq_refl H0))) eq_refl (IHt107 (weakenTrace d111 (bind d)) (appendHvl k73 (bind d)) (weaken_substhvl_TyVar (bind d) del))))) (fun (k72 : Hvl) {X53 : (Index TyVar)} (wfi : (wfindex k72 X53)) {d111 : (Trace TyVar)} {k73 : Hvl} (del : (substhvl_TyVar k71 d111 k72 k73)) =>
    (substhvl_TyVar_wfindex_TyVar wf del wfi)) (fun (k72 : Hvl) (T1 : Ty) (wf0 : (wfTy k72 T1)) IHT89 (T2 : Ty) (wf1 : (wfTy (HS TmVar k72) T2)) IHT90 {d111 : (Trace TyVar)} {k73 : Hvl} (del : (substhvl_TyVar k71 d111 k72 k73)) =>
    (wfTy_TPi k73 (IHT89 (weakenTrace d111 H0) k73 (weaken_substhvl_TyVar H0 del)) (IHT90 (weakenTrace d111 (HS TmVar H0)) (HS TmVar k73) (weaken_substhvl_TyVar (HS TmVar H0) del)))) (fun (k72 : Hvl) (T : Ty) (wf0 : (wfTy k72 T)) IHT89 (t : Tm) (wf1 : (wfTm k72 t)) IHt107 {d111 : (Trace TyVar)} {k73 : Hvl} (del : (substhvl_TyVar k71 d111 k72 k73)) =>
    (wfTy_TApp k73 (IHT89 (weakenTrace d111 H0) k73 (weaken_substhvl_TyVar H0 del)) (IHt107 (weakenTrace d111 H0) k73 (weaken_substhvl_TyVar H0 del)))) (fun (k72 : Hvl) {d111 : (Trace TyVar)} {k73 : Hvl} (del : (substhvl_TyVar k71 d111 k72 k73)) =>
    (wfKind_Star k73)) (fun (k72 : Hvl) (T : Ty) (wf0 : (wfTy k72 T)) IHT89 (K : Kind) (wf1 : (wfKind (HS TmVar k72) K)) IHK35 {d111 : (Trace TyVar)} {k73 : Hvl} (del : (substhvl_TyVar k71 d111 k72 k73)) =>
    (wfKind_KPi k73 (IHT89 (weakenTrace d111 H0) k73 (weaken_substhvl_TyVar H0 del)) (IHK35 (weakenTrace d111 (HS TmVar H0)) (HS TmVar k73) (weaken_substhvl_TyVar (HS TmVar H0) del))))).
  Lemma substhvl_TyVar_wfDecls {k71 : Hvl} {S35 : Ty} (wf : (wfTy k71 S35)) (k72 : Hvl) (d111 : Decls) (wf0 : (wfDecls k72 d111)) :
    (forall {d112 : (Trace TyVar)} {k73 : Hvl} ,
      (substhvl_TyVar k71 d112 k72 k73) -> (wfDecls k73 (subst_TyVar_Decls d112 S35 d111))).
  Proof.
    apply ((substhvl_TyVar_wfDecls_Tm_Ty_Kind wf k72)).
    auto .
  Qed.
  Lemma substhvl_TyVar_wfTm {k71 : Hvl} {S35 : Ty} (wf : (wfTy k71 S35)) (k72 : Hvl) (s34 : Tm) (wf1 : (wfTm k72 s34)) :
    (forall {d113 : (Trace TyVar)} {k74 : Hvl} ,
      (substhvl_TyVar k71 d113 k72 k74) -> (wfTm k74 (subst_TyVar_Tm d113 S35 s34))).
  Proof.
    apply ((substhvl_TyVar_wfDecls_Tm_Ty_Kind wf k72)).
    auto .
  Qed.
  Lemma substhvl_TyVar_wfTy {k71 : Hvl} {S35 : Ty} (wf : (wfTy k71 S35)) (k72 : Hvl) (S36 : Ty) (wf2 : (wfTy k72 S36)) :
    (forall {d114 : (Trace TyVar)} {k75 : Hvl} ,
      (substhvl_TyVar k71 d114 k72 k75) -> (wfTy k75 (subst_TyVar_Ty d114 S35 S36))).
  Proof.
    apply ((substhvl_TyVar_wfDecls_Tm_Ty_Kind wf k72)).
    auto .
  Qed.
  Lemma substhvl_TyVar_wfKind {k71 : Hvl} {S35 : Ty} (wf : (wfTy k71 S35)) (k72 : Hvl) (K35 : Kind) (wf3 : (wfKind k72 K35)) :
    (forall {d115 : (Trace TyVar)} {k76 : Hvl} ,
      (substhvl_TyVar k71 d115 k72 k76) -> (wfKind k76 (subst_TyVar_Kind d115 S35 K35))).
  Proof.
    apply ((substhvl_TyVar_wfDecls_Tm_Ty_Kind wf k72)).
    auto .
  Qed.
End SubstWellFormed.
 Hint Resolve substhvl_TmVar_wfindex_TmVar substhvl_TmVar_wfindex_TyVar substhvl_TyVar_wfindex_TmVar substhvl_TyVar_wfindex_TyVar : infra.
 Hint Resolve substhvl_TmVar_wfindex_TmVar substhvl_TmVar_wfindex_TyVar substhvl_TyVar_wfindex_TmVar substhvl_TyVar_wfindex_TyVar : subst.
 Hint Resolve substhvl_TmVar_wfindex_TmVar substhvl_TmVar_wfindex_TyVar substhvl_TyVar_wfindex_TmVar substhvl_TyVar_wfindex_TyVar : subst_wf.
 Hint Resolve substhvl_TmVar_wfindex_TmVar substhvl_TmVar_wfindex_TyVar substhvl_TyVar_wfindex_TmVar substhvl_TyVar_wfindex_TyVar : wf.
 Hint Resolve substhvl_TmVar_wfDecls substhvl_TmVar_wfKind substhvl_TmVar_wfTm substhvl_TmVar_wfTy substhvl_TyVar_wfDecls substhvl_TyVar_wfKind substhvl_TyVar_wfTm substhvl_TyVar_wfTy : infra.
 Hint Resolve substhvl_TmVar_wfDecls substhvl_TmVar_wfKind substhvl_TmVar_wfTm substhvl_TmVar_wfTy substhvl_TyVar_wfDecls substhvl_TyVar_wfKind substhvl_TyVar_wfTm substhvl_TyVar_wfTy : subst.
 Hint Resolve substhvl_TmVar_wfDecls substhvl_TmVar_wfKind substhvl_TmVar_wfTm substhvl_TmVar_wfTy substhvl_TyVar_wfDecls substhvl_TyVar_wfKind substhvl_TyVar_wfTm substhvl_TyVar_wfTy : subst_wf.
 Hint Resolve substhvl_TmVar_wfDecls substhvl_TmVar_wfKind substhvl_TmVar_wfTm substhvl_TmVar_wfTy substhvl_TyVar_wfDecls substhvl_TyVar_wfKind substhvl_TyVar_wfTm substhvl_TyVar_wfTy : wf.
 Hint Constructors substhvl_TmVar substhvl_TyVar : infra.
 Hint Constructors substhvl_TmVar substhvl_TyVar : subst.
 Hint Constructors substhvl_TmVar substhvl_TyVar : subst_wf.
 Hint Constructors substhvl_TmVar substhvl_TyVar : wf.
Fixpoint subhvl_TmVar_TyVar (k71 : Hvl) {struct k71} : Prop :=
  match k71 with
    | (H0) => True
    | (HS a k71) => match a with
      | (TmVar) => (subhvl_TmVar_TyVar k71)
      | (TyVar) => (subhvl_TmVar_TyVar k71)
    end
  end.
Lemma subhvl_TmVar_TyVar_append  :
  (forall (k71 : Hvl) (k72 : Hvl) ,
    (subhvl_TmVar_TyVar k71) -> (subhvl_TmVar_TyVar k72) -> (subhvl_TmVar_TyVar (appendHvl k71 k72))).
Proof.
  needleGenericSubHvlAppend .
Qed.
 Hint Resolve subhvl_TmVar_TyVar_append : infra.
 Hint Resolve subhvl_TmVar_TyVar_append : wf.
Section Context.
  Inductive Ctx : Type :=
    | empty 
    | evar (G : Ctx) (T : Ty)
    | etvar (G : Ctx) (K : Kind).
  Fixpoint appendCtx (G : Ctx) (G0 : Ctx) : Ctx :=
    match G0 with
      | (empty) => G
      | (evar G1 T) => (evar (appendCtx G G1) T)
      | (etvar G1 K) => (etvar (appendCtx G G1) K)
    end.
  Fixpoint domainCtx (G : Ctx) : Hvl :=
    match G with
      | (empty) => H0
      | (evar G0 T) => (HS TmVar (domainCtx G0))
      | (etvar G0 K) => (HS TyVar (domainCtx G0))
    end.
  Lemma appendCtx_assoc  :
    (forall (G : Ctx) (G0 : Ctx) (G1 : Ctx) ,
      ((appendCtx G (appendCtx G0 G1)) =
      (appendCtx (appendCtx G G0) G1))).
  Proof.
    needleGenericAppendEnvAssoc .
  Qed.
  Lemma domainCtx_appendCtx  :
    (forall (G : Ctx) (G0 : Ctx) ,
      ((domainCtx (appendCtx G G0)) =
      (appendHvl (domainCtx G) (domainCtx G0)))).
  Proof.
    needleGenericDomainEnvAppendEnv .
  Qed.
  Fixpoint shift_TmVar_Ctx (c21 : (Cutoff TmVar)) (G : Ctx) : Ctx :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (shift_TmVar_Ctx c21 G0) (shift_TmVar_Ty (weakenCutoffTmVar c21 (domainCtx G0)) T))
      | (etvar G0 K) => (etvar (shift_TmVar_Ctx c21 G0) (shift_TmVar_Kind (weakenCutoffTmVar c21 (domainCtx G0)) K))
    end.
  Fixpoint shift_TyVar_Ctx (c21 : (Cutoff TyVar)) (G : Ctx) : Ctx :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (shift_TyVar_Ctx c21 G0) (shift_TyVar_Ty (weakenCutoffTyVar c21 (domainCtx G0)) T))
      | (etvar G0 K) => (etvar (shift_TyVar_Ctx c21 G0) (shift_TyVar_Kind (weakenCutoffTyVar c21 (domainCtx G0)) K))
    end.
  Fixpoint subst_TmVar_Ctx (d111 : (Trace TmVar)) (s34 : Tm) (G : Ctx) : Ctx :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (subst_TmVar_Ctx d111 s34 G0) (subst_TmVar_Ty (weakenTrace d111 (domainCtx G0)) s34 T))
      | (etvar G0 K) => (etvar (subst_TmVar_Ctx d111 s34 G0) (subst_TmVar_Kind (weakenTrace d111 (domainCtx G0)) s34 K))
    end.
  Fixpoint subst_TyVar_Ctx (d111 : (Trace TyVar)) (S35 : Ty) (G : Ctx) : Ctx :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (subst_TyVar_Ctx d111 S35 G0) (subst_TyVar_Ty (weakenTrace d111 (domainCtx G0)) S35 T))
      | (etvar G0 K) => (etvar (subst_TyVar_Ctx d111 S35 G0) (subst_TyVar_Kind (weakenTrace d111 (domainCtx G0)) S35 K))
    end.
  Lemma domainCtx_shift_TmVar_Ctx  :
    (forall (c21 : (Cutoff TmVar)) (G : Ctx) ,
      ((domainCtx (shift_TmVar_Ctx c21 G)) =
      (domainCtx G))).
  Proof.
    needleGenericDomainEnvShiftEnv .
  Qed.
  Lemma domainCtx_shift_TyVar_Ctx  :
    (forall (c21 : (Cutoff TyVar)) (G : Ctx) ,
      ((domainCtx (shift_TyVar_Ctx c21 G)) =
      (domainCtx G))).
  Proof.
    needleGenericDomainEnvShiftEnv .
  Qed.
  Lemma domainCtx_subst_TmVar_Ctx  :
    (forall (d111 : (Trace TmVar)) (s34 : Tm) (G : Ctx) ,
      ((domainCtx (subst_TmVar_Ctx d111 s34 G)) =
      (domainCtx G))).
  Proof.
    needleGenericDomainEnvSubstEnv .
  Qed.
  Lemma domainCtx_subst_TyVar_Ctx  :
    (forall (d111 : (Trace TyVar)) (S35 : Ty) (G : Ctx) ,
      ((domainCtx (subst_TyVar_Ctx d111 S35 G)) =
      (domainCtx G))).
  Proof.
    needleGenericDomainEnvSubstEnv .
  Qed.
End Context.
 Hint Rewrite domainCtx_appendCtx : interaction.
 Hint Rewrite domainCtx_appendCtx : env_domain_append.
Section ContextStuff.
  Section ShiftEnvAppendEnv.
    Lemma shift_TmVar_Ctx_appendCtx  :
      (forall (c21 : (Cutoff TmVar)) (G : Ctx) (G0 : Ctx) ,
        ((shift_TmVar_Ctx c21 (appendCtx G G0)) =
        (appendCtx (shift_TmVar_Ctx c21 G) (shift_TmVar_Ctx (weakenCutoffTmVar c21 (domainCtx G)) G0)))).
    Proof.
      needleGenericShiftEnvAppendEnv .
    Qed.
    Lemma shift_TyVar_Ctx_appendCtx  :
      (forall (c21 : (Cutoff TyVar)) (G : Ctx) (G0 : Ctx) ,
        ((shift_TyVar_Ctx c21 (appendCtx G G0)) =
        (appendCtx (shift_TyVar_Ctx c21 G) (shift_TyVar_Ctx (weakenCutoffTyVar c21 (domainCtx G)) G0)))).
    Proof.
      needleGenericShiftEnvAppendEnv .
    Qed.
  End ShiftEnvAppendEnv.
  Section SubstEnvAppendEnv.
    Lemma subst_TmVar_Ctx_appendCtx  :
      (forall (d111 : (Trace TmVar)) (s34 : Tm) (G : Ctx) (G0 : Ctx) ,
        ((subst_TmVar_Ctx d111 s34 (appendCtx G G0)) =
        (appendCtx (subst_TmVar_Ctx d111 s34 G) (subst_TmVar_Ctx (weakenTrace d111 (domainCtx G)) s34 G0)))).
    Proof.
      needleGenericSubstEnvAppendEnv .
    Qed.
    Lemma subst_TyVar_Ctx_appendCtx  :
      (forall (d111 : (Trace TyVar)) (S35 : Ty) (G : Ctx) (G0 : Ctx) ,
        ((subst_TyVar_Ctx d111 S35 (appendCtx G G0)) =
        (appendCtx (subst_TyVar_Ctx d111 S35 G) (subst_TyVar_Ctx (weakenTrace d111 (domainCtx G)) S35 G0)))).
    Proof.
      needleGenericSubstEnvAppendEnv .
    Qed.
  End SubstEnvAppendEnv.
  Section Lookups.
    Inductive lookup_evar : Ctx -> (Index TmVar) -> Ty -> Prop :=
      | lookup_evar_here {G : Ctx}
          {T89 : Ty} :
          (wfTy (domainCtx G) T89) -> (lookup_evar (evar G T89) I0 (shift_TmVar_Ty C0 T89))
      | lookup_evar_there_evar
          {G : Ctx} {x107 : (Index TmVar)}
          {T89 : Ty} {T90 : Ty} :
          (lookup_evar G x107 T89) -> (lookup_evar (evar G T90) (IS x107) (shift_TmVar_Ty C0 T89))
      | lookup_evar_there_etvar
          {G : Ctx} {x107 : (Index TmVar)}
          {T89 : Ty} {K35 : Kind} :
          (lookup_evar G x107 T89) -> (lookup_evar (etvar G K35) x107 (shift_TyVar_Ty C0 T89)).
    Inductive lookup_etvar : Ctx -> (Index TyVar) -> Kind -> Prop :=
      | lookup_etvar_here {G : Ctx}
          {K35 : Kind} :
          (wfKind (domainCtx G) K35) -> (lookup_etvar (etvar G K35) I0 (shift_TyVar_Kind C0 K35))
      | lookup_etvar_there_evar
          {G : Ctx} {X53 : (Index TyVar)}
          {K35 : Kind} {T89 : Ty} :
          (lookup_etvar G X53 K35) -> (lookup_etvar (evar G T89) X53 (shift_TmVar_Kind C0 K35))
      | lookup_etvar_there_etvar
          {G : Ctx} {X53 : (Index TyVar)}
          {K35 : Kind} {K36 : Kind} :
          (lookup_etvar G X53 K35) -> (lookup_etvar (etvar G K36) (IS X53) (shift_TyVar_Kind C0 K35)).
    Lemma lookup_evar_inversion_here  :
      (forall (G : Ctx) (S35 : Ty) (S36 : Ty) ,
        (lookup_evar (evar G S35) I0 S36) -> ((shift_TmVar_Ty C0 S35) =
        S36)).
    Proof.
      needleGenericLookupInversion .
    Qed.
    Lemma lookup_etvar_inversion_here  :
      (forall (G : Ctx) (K35 : Kind) (K36 : Kind) ,
        (lookup_etvar (etvar G K35) I0 K36) -> ((shift_TyVar_Kind C0 K35) =
        K36)).
    Proof.
      needleGenericLookupInversion .
    Qed.
    Lemma lookup_evar_functional  :
      (forall {G : Ctx} {x107 : (Index TmVar)} ,
        (forall {S35 : Ty} ,
          (lookup_evar G x107 S35) -> (forall {S36 : Ty} ,
            (lookup_evar G x107 S36) -> (S35 =
            S36)))).
    Proof.
      needleGenericLookupFunctional .
    Qed.
    Lemma lookup_etvar_functional  :
      (forall {G : Ctx} {X53 : (Index TyVar)} ,
        (forall {K35 : Kind} ,
          (lookup_etvar G X53 K35) -> (forall {K36 : Kind} ,
            (lookup_etvar G X53 K36) -> (K35 =
            K36)))).
    Proof.
      needleGenericLookupFunctional .
    Qed.
    Lemma lookup_evar_wf  :
      (forall {G : Ctx} {x107 : (Index TmVar)} {S35 : Ty} ,
        (lookup_evar G x107 S35) -> (wfTy (domainCtx G) S35)).
    Proof.
      needleGenericLookupWellformedData .
    Qed.
    Lemma lookup_etvar_wf  :
      (forall {G : Ctx} {X53 : (Index TyVar)} {K35 : Kind} ,
        (lookup_etvar G X53 K35) -> (wfKind (domainCtx G) K35)).
    Proof.
      needleGenericLookupWellformedData .
    Qed.
    Lemma weaken_lookup_evar  :
      (forall {G : Ctx} (G0 : Ctx) {x107 : (Index TmVar)} {T89 : Ty} ,
        (lookup_evar G x107 T89) -> (lookup_evar (appendCtx G G0) (weakenIndexTmVar x107 (domainCtx G0)) (weakenTy T89 (domainCtx G0)))).
    Proof.
      needleGenericWeakenLookup .
    Qed.
    Lemma weaken_lookup_etvar  :
      (forall {G : Ctx} (G0 : Ctx) {X53 : (Index TyVar)} {K35 : Kind} ,
        (lookup_etvar G X53 K35) -> (lookup_etvar (appendCtx G G0) (weakenIndexTyVar X53 (domainCtx G0)) (weakenKind K35 (domainCtx G0)))).
    Proof.
      needleGenericWeakenLookup .
    Qed.
    Lemma lookup_evar_wfindex  :
      (forall {G : Ctx} {x107 : (Index TmVar)} {S35 : Ty} ,
        (lookup_evar G x107 S35) -> (wfindex (domainCtx G) x107)).
    Proof.
      needleGenericLookupWellformedIndex .
    Qed.
    Lemma lookup_etvar_wfindex  :
      (forall {G : Ctx} {X53 : (Index TyVar)} {K36 : Kind} ,
        (lookup_etvar G X53 K36) -> (wfindex (domainCtx G) X53)).
    Proof.
      needleGenericLookupWellformedIndex .
    Qed.
  End Lookups.
  Inductive shift_evar : (Cutoff TmVar) -> Ctx -> Ctx -> Prop :=
    | shift_evar_here {G : Ctx}
        {T89 : Ty} :
        (shift_evar C0 G (evar G T89))
    | shiftevar_there_evar
        {c21 : (Cutoff TmVar)} {G : Ctx}
        {G0 : Ctx} {T : Ty} :
        (shift_evar c21 G G0) -> (shift_evar (CS c21) (evar G T) (evar G0 (shift_TmVar_Ty c21 T)))
    | shiftevar_there_etvar
        {c21 : (Cutoff TmVar)} {G : Ctx}
        {G0 : Ctx} {K : Kind} :
        (shift_evar c21 G G0) -> (shift_evar c21 (etvar G K) (etvar G0 (shift_TmVar_Kind c21 K))).
  Lemma weaken_shift_evar  :
    (forall (G : Ctx) {c21 : (Cutoff TmVar)} {G0 : Ctx} {G1 : Ctx} ,
      (shift_evar c21 G0 G1) -> (shift_evar (weakenCutoffTmVar c21 (domainCtx G)) (appendCtx G0 G) (appendCtx G1 (shift_TmVar_Ctx c21 G)))).
  Proof.
    needleGenericWeakenInsertEnv .
  Qed.
  Inductive shift_etvar : (Cutoff TyVar) -> Ctx -> Ctx -> Prop :=
    | shift_etvar_here {G : Ctx}
        {K35 : Kind} :
        (shift_etvar C0 G (etvar G K35))
    | shiftetvar_there_evar
        {c21 : (Cutoff TyVar)} {G : Ctx}
        {G0 : Ctx} {T : Ty} :
        (shift_etvar c21 G G0) -> (shift_etvar c21 (evar G T) (evar G0 (shift_TyVar_Ty c21 T)))
    | shiftetvar_there_etvar
        {c21 : (Cutoff TyVar)} {G : Ctx}
        {G0 : Ctx} {K : Kind} :
        (shift_etvar c21 G G0) -> (shift_etvar (CS c21) (etvar G K) (etvar G0 (shift_TyVar_Kind c21 K))).
  Lemma weaken_shift_etvar  :
    (forall (G : Ctx) {c21 : (Cutoff TyVar)} {G0 : Ctx} {G1 : Ctx} ,
      (shift_etvar c21 G0 G1) -> (shift_etvar (weakenCutoffTyVar c21 (domainCtx G)) (appendCtx G0 G) (appendCtx G1 (shift_TyVar_Ctx c21 G)))).
  Proof.
    needleGenericWeakenInsertEnv .
  Qed.
  Lemma shift_evar_shifthvl_TmVar  :
    (forall {c21 : (Cutoff TmVar)} {G : Ctx} {G0 : Ctx} ,
      (shift_evar c21 G G0) -> (shifthvl_TmVar c21 (domainCtx G) (domainCtx G0))).
  Proof.
    needleGenericInsertEnvInsertHvl .
  Qed.
  Lemma shift_etvar_shifthvl_TyVar  :
    (forall {c21 : (Cutoff TyVar)} {G : Ctx} {G0 : Ctx} ,
      (shift_etvar c21 G G0) -> (shifthvl_TyVar c21 (domainCtx G) (domainCtx G0))).
  Proof.
    needleGenericInsertEnvInsertHvl .
  Qed.
  Inductive subst_evar (G : Ctx) (T89 : Ty) (s34 : Tm) : (Trace TmVar) -> Ctx -> Ctx -> Prop :=
    | subst_evar_here :
        (subst_evar G T89 s34 X0 (evar G T89) G)
    | subst_evar_there_evar
        {d111 : (Trace TmVar)}
        {G0 : Ctx} {G1 : Ctx} {T : Ty} :
        (subst_evar G T89 s34 d111 G0 G1) -> (subst_evar G T89 s34 (XS TmVar d111) (evar G0 T) (evar G1 (subst_TmVar_Ty d111 s34 T)))
    | subst_evar_there_etvar
        {d111 : (Trace TmVar)}
        {G0 : Ctx} {G1 : Ctx} {K : Kind}
        :
        (subst_evar G T89 s34 d111 G0 G1) -> (subst_evar G T89 s34 (XS TyVar d111) (etvar G0 K) (etvar G1 (subst_TmVar_Kind d111 s34 K))).
  Lemma weaken_subst_evar {G : Ctx} {T89 : Ty} {s34 : Tm} :
    (forall (G0 : Ctx) {d111 : (Trace TmVar)} {G1 : Ctx} {G2 : Ctx} ,
      (subst_evar G T89 s34 d111 G1 G2) -> (subst_evar G T89 s34 (weakenTrace d111 (domainCtx G0)) (appendCtx G1 G0) (appendCtx G2 (subst_TmVar_Ctx d111 s34 G0)))).
  Proof.
    needleGenericWeakenSubstEnv .
  Qed.
  Inductive subst_etvar (G : Ctx) (K35 : Kind) (S35 : Ty) : (Trace TyVar) -> Ctx -> Ctx -> Prop :=
    | subst_etvar_here :
        (subst_etvar G K35 S35 X0 (etvar G K35) G)
    | subst_etvar_there_evar
        {d111 : (Trace TyVar)}
        {G0 : Ctx} {G1 : Ctx} {T : Ty} :
        (subst_etvar G K35 S35 d111 G0 G1) -> (subst_etvar G K35 S35 (XS TmVar d111) (evar G0 T) (evar G1 (subst_TyVar_Ty d111 S35 T)))
    | subst_etvar_there_etvar
        {d111 : (Trace TyVar)}
        {G0 : Ctx} {G1 : Ctx} {K : Kind}
        :
        (subst_etvar G K35 S35 d111 G0 G1) -> (subst_etvar G K35 S35 (XS TyVar d111) (etvar G0 K) (etvar G1 (subst_TyVar_Kind d111 S35 K))).
  Lemma weaken_subst_etvar {G : Ctx} {K35 : Kind} {S35 : Ty} :
    (forall (G0 : Ctx) {d111 : (Trace TyVar)} {G1 : Ctx} {G2 : Ctx} ,
      (subst_etvar G K35 S35 d111 G1 G2) -> (subst_etvar G K35 S35 (weakenTrace d111 (domainCtx G0)) (appendCtx G1 G0) (appendCtx G2 (subst_TyVar_Ctx d111 S35 G0)))).
  Proof.
    needleGenericWeakenSubstEnv .
  Qed.
  Lemma subst_evar_substhvl_TmVar {G : Ctx} {s34 : Tm} {T89 : Ty} :
    (forall {d111 : (Trace TmVar)} {G0 : Ctx} {G1 : Ctx} ,
      (subst_evar G T89 s34 d111 G0 G1) -> (substhvl_TmVar (domainCtx G) d111 (domainCtx G0) (domainCtx G1))).
  Proof.
    needleGenericSubstEnvSubstHvl .
  Qed.
  Lemma subst_etvar_substhvl_TyVar {G : Ctx} {S35 : Ty} {K35 : Kind} :
    (forall {d111 : (Trace TyVar)} {G0 : Ctx} {G1 : Ctx} ,
      (subst_etvar G K35 S35 d111 G0 G1) -> (substhvl_TyVar (domainCtx G) d111 (domainCtx G0) (domainCtx G1))).
  Proof.
    needleGenericSubstEnvSubstHvl .
  Qed.
End ContextStuff.
 Hint Rewrite domainCtx_shift_TmVar_Ctx domainCtx_shift_TyVar_Ctx : interaction.
 Hint Rewrite domainCtx_shift_TmVar_Ctx domainCtx_shift_TyVar_Ctx : env_domain_shift.
 Hint Rewrite domainCtx_subst_TmVar_Ctx domainCtx_subst_TyVar_Ctx : interaction.
 Hint Rewrite domainCtx_subst_TmVar_Ctx domainCtx_subst_TyVar_Ctx : env_domain_subst.
 Hint Rewrite shift_TmVar_Ctx_appendCtx shift_TyVar_Ctx_appendCtx : interaction.
 Hint Rewrite shift_TmVar_Ctx_appendCtx shift_TyVar_Ctx_appendCtx : env_shift_append.
 Hint Rewrite subst_TmVar_Ctx_appendCtx subst_TyVar_Ctx_appendCtx : interaction.
 Hint Rewrite subst_TmVar_Ctx_appendCtx subst_TyVar_Ctx_appendCtx : env_shift_append.
 Hint Constructors lookup_evar lookup_etvar : infra.
 Hint Constructors lookup_evar lookup_etvar : lookup.
 Hint Constructors lookup_evar lookup_etvar : shift.
 Hint Constructors lookup_evar lookup_etvar : subst.
 Hint Resolve weaken_lookup_evar weaken_lookup_etvar : infra.
 Hint Resolve weaken_lookup_evar weaken_lookup_etvar : lookup.
 Hint Resolve weaken_lookup_evar weaken_lookup_etvar : shift.
 Hint Resolve weaken_lookup_evar weaken_lookup_etvar : weaken.
Lemma weaken_lookup_evar_here  :
  (forall {G : Ctx} (G0 : Ctx) {T89 : Ty} (wf : (wfTy (domainCtx G) T89)) ,
    (lookup_evar (appendCtx (evar G T89) G0) (weakenIndexTmVar I0 (domainCtx G0)) (weakenTy (shift_TmVar_Ty C0 T89) (domainCtx G0)))).
Proof.
  eauto with lookup .
Qed.
Lemma weaken_lookup_etvar_here  :
  (forall {G : Ctx} (G0 : Ctx) {K35 : Kind} (wf : (wfKind (domainCtx G) K35)) ,
    (lookup_etvar (appendCtx (etvar G K35) G0) (weakenIndexTyVar I0 (domainCtx G0)) (weakenKind (shift_TyVar_Kind C0 K35) (domainCtx G0)))).
Proof.
  eauto with lookup .
Qed.
 Hint Constructors wfDecls wfKind wfTm wfTy : infra.
 Hint Constructors wfDecls wfKind wfTm wfTy : wf.
 Hint Extern 10 ((wfDecls _ _)) => autorewrite with env_domain_append in *  : infra wf.
 Hint Extern 10 ((wfKind _ _)) => autorewrite with env_domain_append in *  : infra wf.
 Hint Extern 10 ((wfTm _ _)) => autorewrite with env_domain_append in *  : infra wf.
 Hint Extern 10 ((wfTy _ _)) => autorewrite with env_domain_append in *  : infra wf.
 Hint Extern 2 ((wfDecls _ _)) => match goal with
  | H : (wfDecls _ (DNil)) |- _ => inversion H; subst; clear H
  | H : (wfDecls _ (DConsTm _ _)) |- _ => inversion H; subst; clear H
  | H : (wfDecls _ (DConsTy _ _)) |- _ => inversion H; subst; clear H
end : infra wf.
 Hint Extern 2 ((wfTm _ _)) => match goal with
  | H : (wfTm _ (Var _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (Abs _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (App _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (Let _ _)) |- _ => inversion H; subst; clear H
end : infra wf.
 Hint Extern 2 ((wfTy _ _)) => match goal with
  | H : (wfTy _ (TVar _)) |- _ => inversion H; subst; clear H
  | H : (wfTy _ (TPi _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTy _ (TApp _ _)) |- _ => inversion H; subst; clear H
end : infra wf.
 Hint Extern 2 ((wfKind _ _)) => match goal with
  | H : (wfKind _ (Star)) |- _ => inversion H; subst; clear H
  | H : (wfKind _ (KPi _ _)) |- _ => inversion H; subst; clear H
end : infra wf.
 Hint Resolve lookup_evar_wf lookup_etvar_wf : infra.
 Hint Resolve lookup_evar_wf lookup_etvar_wf : wf.
 Hint Resolve lookup_evar_wfindex lookup_etvar_wfindex : infra.
 Hint Resolve lookup_evar_wfindex lookup_etvar_wfindex : lookup.
 Hint Resolve lookup_evar_wfindex lookup_etvar_wfindex : wf.
 Hint Constructors shift_evar shift_etvar : infra.
 Hint Constructors shift_evar shift_etvar : shift.
 Hint Constructors shift_evar shift_etvar : subst.
 Hint Resolve weaken_shift_evar weaken_shift_etvar : infra.
 Hint Resolve weaken_shift_evar weaken_shift_etvar : shift.
 Hint Resolve shift_evar_shifthvl_TmVar shift_etvar_shifthvl_TyVar : infra.
 Hint Resolve shift_evar_shifthvl_TmVar shift_etvar_shifthvl_TyVar : shift.
 Hint Resolve shift_evar_shifthvl_TmVar shift_etvar_shifthvl_TyVar : shift_wf.
 Hint Resolve shift_evar_shifthvl_TmVar shift_etvar_shifthvl_TyVar : wf.
 Hint Constructors subst_evar subst_etvar : infra.
 Hint Constructors subst_evar subst_etvar : subst.
 Hint Resolve weaken_subst_evar weaken_subst_etvar : infra.
 Hint Resolve weaken_subst_evar weaken_subst_etvar : subst.
 Hint Resolve subst_evar_substhvl_TmVar subst_etvar_substhvl_TyVar : infra.
 Hint Resolve subst_evar_substhvl_TmVar subst_etvar_substhvl_TyVar : subst.
 Hint Resolve subst_evar_substhvl_TmVar subst_etvar_substhvl_TyVar : subst_wf.
 Hint Resolve subst_evar_substhvl_TmVar subst_etvar_substhvl_TyVar : wf.
 Hint Resolve subst_evar_substhvl_TmVar subst_etvar_substhvl_TyVar : substenv_substhvl.
Lemma shift_evar_lookup_evar  :
  (forall {c21 : (Cutoff TmVar)} {G : Ctx} {G0 : Ctx} (ins : (shift_evar c21 G G0)) {x107 : (Index TmVar)} {T89 : Ty} ,
    (lookup_evar G x107 T89) -> (lookup_evar G0 (shift_TmVar_Index c21 x107) (shift_TmVar_Ty c21 T89))).
Proof.
  needleGenericInsertLookup .
Qed.
Lemma shift_evar_lookup_etvar  :
  (forall {c21 : (Cutoff TmVar)} {G : Ctx} {G0 : Ctx} (ins : (shift_evar c21 G G0)) {X53 : (Index TyVar)} {K35 : Kind} ,
    (lookup_etvar G X53 K35) -> (lookup_etvar G0 X53 (shift_TmVar_Kind c21 K35))).
Proof.
  needleGenericInsertLookup .
Qed.
Lemma shift_etvar_lookup_evar  :
  (forall {c21 : (Cutoff TyVar)} {G : Ctx} {G0 : Ctx} (ins : (shift_etvar c21 G G0)) {x107 : (Index TmVar)} {T89 : Ty} ,
    (lookup_evar G x107 T89) -> (lookup_evar G0 x107 (shift_TyVar_Ty c21 T89))).
Proof.
  needleGenericInsertLookup .
Qed.
Lemma shift_etvar_lookup_etvar  :
  (forall {c21 : (Cutoff TyVar)} {G : Ctx} {G0 : Ctx} (ins : (shift_etvar c21 G G0)) {X53 : (Index TyVar)} {K35 : Kind} ,
    (lookup_etvar G X53 K35) -> (lookup_etvar G0 (shift_TyVar_Index c21 X53) (shift_TyVar_Kind c21 K35))).
Proof.
  needleGenericInsertLookup .
Qed.
 Hint Resolve shift_evar_lookup_evar shift_evar_lookup_etvar shift_etvar_lookup_evar shift_etvar_lookup_etvar : infra.
 Hint Resolve shift_evar_lookup_evar shift_evar_lookup_etvar shift_etvar_lookup_evar shift_etvar_lookup_etvar : lookup.
 Hint Resolve shift_evar_lookup_evar shift_evar_lookup_etvar shift_etvar_lookup_evar shift_etvar_lookup_etvar : shift.
Lemma subst_evar_lookup_etvar {G : Ctx} {T90 : Ty} {s34 : Tm} (wf : (wfTm (domainCtx G) s34)) :
  (forall {d111 : (Trace TmVar)} {G0 : Ctx} {G1 : Ctx} (sub : (subst_evar G T90 s34 d111 G0 G1)) {X53 : (Index TyVar)} {K36 : Kind} ,
    (lookup_etvar G0 X53 K36) -> (lookup_etvar G1 X53 (subst_TmVar_Kind d111 s34 K36))).
Proof.
  needleGenericSubstEnvLookup .
Qed.
Lemma subst_etvar_lookup_evar {G : Ctx} {K36 : Kind} {S35 : Ty} (wf : (wfTy (domainCtx G) S35)) :
  (forall {d111 : (Trace TyVar)} {G0 : Ctx} {G1 : Ctx} (sub : (subst_etvar G K36 S35 d111 G0 G1)) {x107 : (Index TmVar)} {T90 : Ty} ,
    (lookup_evar G0 x107 T90) -> (lookup_evar G1 x107 (subst_TyVar_Ty d111 S35 T90))).
Proof.
  needleGenericSubstEnvLookup .
Qed.
 Hint Resolve subst_evar_lookup_etvar subst_etvar_lookup_evar : infra.
 Hint Resolve subst_evar_lookup_etvar subst_etvar_lookup_evar : lookup.
 Hint Resolve subst_evar_lookup_etvar subst_etvar_lookup_evar : subst.
Fixpoint size_Decls (d : Decls) {struct d} : nat :=
  match d with
    | (DNil) => 1
    | (DConsTm t d0) => (plus 1 (plus (size_Tm t) (size_Decls d0)))
    | (DConsTy K d1) => (plus 1 (plus (size_Kind K) (size_Decls d1)))
  end
with size_Tm (s : Tm) {struct s} : nat :=
  match s with
    | (Var x) => 1
    | (Abs T t) => (plus 1 (plus (size_Ty T) (size_Tm t)))
    | (App t1 t2) => (plus 1 (plus (size_Tm t1) (size_Tm t2)))
    | (Let d t0) => (plus 1 (plus (size_Decls d) (size_Tm t0)))
  end
with size_Ty (S0 : Ty) {struct S0} : nat :=
  match S0 with
    | (TVar X) => 1
    | (TPi T1 T2) => (plus 1 (plus (size_Ty T1) (size_Ty T2)))
    | (TApp T t) => (plus 1 (plus (size_Ty T) (size_Tm t)))
  end
with size_Kind (K : Kind) {struct K} : nat :=
  match K with
    | (Star) => 1
    | (KPi T K0) => (plus 1 (plus (size_Ty T) (size_Kind K0)))
  end.
Lemma shift_TmVar__size_Decls_Tm_Ty_Kind  :
  (forall (d105 : Decls) (c17 : (Cutoff TmVar)) ,
    ((size_Decls (shift_TmVar_Decls c17 d105)) =
    (size_Decls d105))) /\
  (forall (s34 : Tm) (c17 : (Cutoff TmVar)) ,
    ((size_Tm (shift_TmVar_Tm c17 s34)) =
    (size_Tm s34))) /\
  (forall (S35 : Ty) (c17 : (Cutoff TmVar)) ,
    ((size_Ty (shift_TmVar_Ty c17 S35)) =
    (size_Ty S35))) /\
  (forall (K31 : Kind) (c17 : (Cutoff TmVar)) ,
    ((size_Kind (shift_TmVar_Kind c17 K31)) =
    (size_Kind K31))).
Proof.
  apply_mutual_ind ind_Decls_Tm_Ty_Kind.
  - intros c17; simpl; apply ((f_equal S)); reflexivity.
  - intros t95 IHt95 d105 IHd105.
  intros c17; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt95 c17)).
  + exact ((IHd105 (CS c17))).
  - intros K31 IHK31 d106 IHd106.
  intros c17; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHK31 c17)).
  + exact ((IHd106 c17)).
  - intros x97 c17.
  reflexivity.
  - intros T79 IHT79 t96 IHt96.
  intros c18; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT79 c18)).
  + exact ((IHt96 (CS c18))).
  - intros t97 IHt97 t98 IHt98.
  intros c18; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt97 c18)).
  + exact ((IHt98 c18)).
  - intros d107 IHd107 t99 IHt99.
  intros c18; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHd107 c18)).
  + exact ((IHt99 (weakenCutoffTmVar c18 (bind d107)))).
  - intros x99 c18.
  reflexivity.
  - intros T80 IHT80 T81 IHT81.
  intros c19; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT80 c19)).
  + exact ((IHT81 (CS c19))).
  - intros T82 IHT82 t100 IHt100.
  intros c19; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT82 c19)).
  + exact ((IHt100 c19)).
  - intros c19; simpl; apply ((f_equal S)); reflexivity.
  - intros T83 IHT83 K32 IHK32.
  intros c19; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT83 c19)).
  + exact ((IHK32 (CS c19))).
Qed.

Lemma shift_TmVar__size_Decls  :
  (forall (d108 : Decls) (c19 : (Cutoff TmVar)) ,
    ((size_Decls (shift_TmVar_Decls c19 d108)) =
    (size_Decls d108))).
Proof.
  pose proof (shift_TmVar__size_Decls_Tm_Ty_Kind).
  destruct_conjs .
  easy .
Qed.
Lemma shift_TmVar__size_Tm  :
  (forall (s34 : Tm) (c19 : (Cutoff TmVar)) ,
    ((size_Tm (shift_TmVar_Tm c19 s34)) =
    (size_Tm s34))).
Proof.
  pose proof (shift_TmVar__size_Decls_Tm_Ty_Kind).
  destruct_conjs .
  easy .
Qed.
Lemma shift_TmVar__size_Ty  :
  (forall (S35 : Ty) (c19 : (Cutoff TmVar)) ,
    ((size_Ty (shift_TmVar_Ty c19 S35)) =
    (size_Ty S35))).
Proof.
  pose proof (shift_TmVar__size_Decls_Tm_Ty_Kind).
  destruct_conjs .
  easy .
Qed.
Lemma shift_TmVar__size_Kind  :
  (forall (K33 : Kind) (c19 : (Cutoff TmVar)) ,
    ((size_Kind (shift_TmVar_Kind c19 K33)) =
    (size_Kind K33))).
Proof.
  pose proof (shift_TmVar__size_Decls_Tm_Ty_Kind).
  destruct_conjs .
  easy .
Qed.
Lemma shift_TyVar__size_Decls_Tm_Ty_Kind  :
  (forall (d108 : Decls) (c19 : (Cutoff TyVar)) ,
    ((size_Decls (shift_TyVar_Decls c19 d108)) =
    (size_Decls d108))) /\
  (forall (s34 : Tm) (c19 : (Cutoff TyVar)) ,
    ((size_Tm (shift_TyVar_Tm c19 s34)) =
    (size_Tm s34))) /\
  (forall (S35 : Ty) (c19 : (Cutoff TyVar)) ,
    ((size_Ty (shift_TyVar_Ty c19 S35)) =
    (size_Ty S35))) /\
  (forall (K33 : Kind) (c19 : (Cutoff TyVar)) ,
    ((size_Kind (shift_TyVar_Kind c19 K33)) =
    (size_Kind K33))).
Proof.
  apply_mutual_ind ind_Decls_Tm_Ty_Kind.
  - intros c19; simpl; apply ((f_equal S)); reflexivity.
  - intros t101 IHt101 d108 IHd108.
  intros c19; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt101 c19)).
  + exact ((IHd108 c19)).
  - intros K33 IHK33 d109 IHd109.
  intros c19; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHK33 c19)).
  + exact ((IHd109 (CS c19))).
  - intros X50 c19.
  reflexivity.
  - intros T84 IHT84 t102 IHt102.
  intros c20; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT84 c20)).
  + exact ((IHt102 c20)).
  - intros t103 IHt103 t104 IHt104.
  intros c20; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt103 c20)).
  + exact ((IHt104 c20)).
  - intros d110 IHd110 t105 IHt105.
  intros c20; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHd110 c20)).
  + exact ((IHt105 (weakenCutoffTyVar c20 (bind d110)))).
  - intros X52 c20.
  reflexivity.
  - intros T85 IHT85 T86 IHT86.
  intros c21; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT85 c21)).
  + exact ((IHT86 c21)).
  - intros T87 IHT87 t106 IHt106.
  intros c21; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT87 c21)).
  + exact ((IHt106 c21)).
  - intros c21; simpl; apply ((f_equal S)); reflexivity.
  - intros T88 IHT88 K34 IHK34.
  intros c21; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT88 c21)).
  + exact ((IHK34 c21)).
Qed.

Lemma shift_TyVar__size_Decls  :
  (forall (d111 : Decls) (c21 : (Cutoff TyVar)) ,
    ((size_Decls (shift_TyVar_Decls c21 d111)) =
    (size_Decls d111))).
Proof.
  pose proof (shift_TyVar__size_Decls_Tm_Ty_Kind).
  destruct_conjs .
  easy .
Qed.
Lemma shift_TyVar__size_Tm  :
  (forall (s34 : Tm) (c21 : (Cutoff TyVar)) ,
    ((size_Tm (shift_TyVar_Tm c21 s34)) =
    (size_Tm s34))).
Proof.
  pose proof (shift_TyVar__size_Decls_Tm_Ty_Kind).
  destruct_conjs .
  easy .
Qed.
Lemma shift_TyVar__size_Ty  :
  (forall (S35 : Ty) (c21 : (Cutoff TyVar)) ,
    ((size_Ty (shift_TyVar_Ty c21 S35)) =
    (size_Ty S35))).
Proof.
  pose proof (shift_TyVar__size_Decls_Tm_Ty_Kind).
  destruct_conjs .
  easy .
Qed.
Lemma shift_TyVar__size_Kind  :
  (forall (K35 : Kind) (c21 : (Cutoff TyVar)) ,
    ((size_Kind (shift_TyVar_Kind c21 K35)) =
    (size_Kind K35))).
Proof.
  pose proof (shift_TyVar__size_Decls_Tm_Ty_Kind).
  destruct_conjs .
  easy .
Qed.
 Hint Rewrite shift_TmVar__size_Decls shift_TyVar__size_Decls shift_TmVar__size_Kind shift_TyVar__size_Kind shift_TmVar__size_Tm shift_TyVar__size_Tm shift_TmVar__size_Ty shift_TyVar__size_Ty : interaction.
 Hint Rewrite shift_TmVar__size_Decls shift_TyVar__size_Decls shift_TmVar__size_Kind shift_TyVar__size_Kind shift_TmVar__size_Tm shift_TyVar__size_Tm shift_TmVar__size_Ty shift_TyVar__size_Ty : shift_size.
Lemma weaken_size_Decls  :
  (forall (k71 : Hvl) (d111 : Decls) ,
    ((size_Decls (weakenDecls d111 k71)) =
    (size_Decls d111))).
Proof.
  needleGenericWeakenSize .
Qed.
Lemma weaken_size_Kind  :
  (forall (k71 : Hvl) (K35 : Kind) ,
    ((size_Kind (weakenKind K35 k71)) =
    (size_Kind K35))).
Proof.
  needleGenericWeakenSize .
Qed.
Lemma weaken_size_Tm  :
  (forall (k71 : Hvl) (s34 : Tm) ,
    ((size_Tm (weakenTm s34 k71)) =
    (size_Tm s34))).
Proof.
  needleGenericWeakenSize .
Qed.
Lemma weaken_size_Ty  :
  (forall (k71 : Hvl) (S35 : Ty) ,
    ((size_Ty (weakenTy S35 k71)) =
    (size_Ty S35))).
Proof.
  needleGenericWeakenSize .
Qed.
 Hint Rewrite weaken_size_Decls weaken_size_Kind weaken_size_Tm weaken_size_Ty : interaction.
 Hint Rewrite weaken_size_Decls weaken_size_Kind weaken_size_Tm weaken_size_Ty : weaken_size.