Require Import Coq.Program.Equality.
Require Import Coq.Program.Tactics.
Require Export Needle.
Section Namespace.
  Inductive Namespace : Type :=
    | A 
    | B 
    | Z .
  Lemma eq_namespace_dec (a : Namespace) (b : Namespace) :
    (sumbool (a =
    b) (not (a =
    b))).
  Proof.
    decide equality .
  Defined.
End Namespace.

Section HVarlist.
  Inductive Hvl : Type :=
    | H0 
    | HS (a : Namespace) (k : Hvl).
  
  Fixpoint appendHvl (k : Hvl) (k0 : Hvl) {struct k0} : Hvl :=
    match k0 with
      | (H0) => k
      | (HS a k0) => (HS a (appendHvl k k0))
    end.
  
  Lemma appendHvl_assoc  :
    (forall (k : Hvl) (k0 : Hvl) (k1 : Hvl) ,
      ((appendHvl (appendHvl k k0) k1) =
      (appendHvl k (appendHvl k0 k1)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
End HVarlist.

Section Index.
  Inductive Index (a : Namespace) : Type :=
    | I0 
    | IS : (Index a) -> (Index a).
  
  Global Arguments I0 [a] .
  Global Arguments IS [a] _ .
  
  Lemma eq_index_dec {a : Namespace} (i : (Index a)) (j : (Index a)) :
    (sumbool (i =
    j) (not (i =
    j))).
  Proof.
    decide equality .
  Qed.
  
  Fixpoint weakenIndexA (a0 : (Index A)) (k : Hvl) {struct k} : (Index A) :=
    match k with
      | (H0) => a0
      | (HS a k) => match a with
        | (A) => (IS (weakenIndexA a0 k))
        | _ => (weakenIndexA a0 k)
      end
    end.
  Fixpoint weakenIndexB (b : (Index B)) (k : Hvl) {struct k} : (Index B) :=
    match k with
      | (H0) => b
      | (HS a k) => match a with
        | (B) => (IS (weakenIndexB b k))
        | _ => (weakenIndexB b k)
      end
    end.
  Fixpoint weakenIndexZ (z : (Index Z)) (k : Hvl) {struct k} : (Index Z) :=
    match k with
      | (H0) => z
      | (HS a k) => match a with
        | (Z) => (IS (weakenIndexZ z k))
        | _ => (weakenIndexZ z k)
      end
    end.
  Lemma weakenIndexA_append  :
    (forall (a : (Index A)) (k : Hvl) (k0 : Hvl) ,
      ((weakenIndexA (weakenIndexA a k) k0) =
      (weakenIndexA a (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Lemma weakenIndexB_append  :
    (forall (b : (Index B)) (k : Hvl) (k0 : Hvl) ,
      ((weakenIndexB (weakenIndexB b k) k0) =
      (weakenIndexB b (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Lemma weakenIndexZ_append  :
    (forall (z : (Index Z)) (k : Hvl) (k0 : Hvl) ,
      ((weakenIndexZ (weakenIndexZ z k) k0) =
      (weakenIndexZ z (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
End Index.

Section Terms.
  Inductive AA : Set :=
    | avar (a : (Index A))
    | arec (zz : ZZ)
  with ZZ : Set :=
    | zvar (z : (Index Z))
    | aabs (aa : AA)
    | babs (bb : BB)
  with BB : Set :=
    | bvar (b : (Index B))
    | brec (zz : ZZ).
  Scheme ind_AA := Induction for AA Sort Prop
  with ind_ZZ := Induction for ZZ Sort Prop
  with ind_BB := Induction for BB Sort Prop.
  Combined Scheme ind_AA_ZZ_BB from ind_AA, ind_ZZ, ind_BB.
End Terms.

Section Functions.
  
End Functions.

Section Shift.
  Inductive Cutoff (a : Namespace) : Type :=
    | C0 
    | CS : (Cutoff a) -> (Cutoff a).
  
  Global Arguments C0 {a} .
  Global Arguments CS {a} _ .
  Fixpoint weakenCutoffA (c : (Cutoff A)) (k : Hvl) {struct k} : (Cutoff A) :=
    match k with
      | (H0) => c
      | (HS a k) => match a with
        | (A) => (CS (weakenCutoffA c k))
        | _ => (weakenCutoffA c k)
      end
    end.
  Fixpoint weakenCutoffB (c : (Cutoff B)) (k : Hvl) {struct k} : (Cutoff B) :=
    match k with
      | (H0) => c
      | (HS a k) => match a with
        | (B) => (CS (weakenCutoffB c k))
        | _ => (weakenCutoffB c k)
      end
    end.
  Fixpoint weakenCutoffZ (c : (Cutoff Z)) (k : Hvl) {struct k} : (Cutoff Z) :=
    match k with
      | (H0) => c
      | (HS a k) => match a with
        | (Z) => (CS (weakenCutoffZ c k))
        | _ => (weakenCutoffZ c k)
      end
    end.
  
  Lemma weakenCutoffA_append  :
    (forall (c : (Cutoff A)) (k : Hvl) (k0 : Hvl) ,
      ((weakenCutoffA (weakenCutoffA c k) k0) =
      (weakenCutoffA c (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Lemma weakenCutoffB_append  :
    (forall (c : (Cutoff B)) (k : Hvl) (k0 : Hvl) ,
      ((weakenCutoffB (weakenCutoffB c k) k0) =
      (weakenCutoffB c (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Lemma weakenCutoffZ_append  :
    (forall (c : (Cutoff Z)) (k : Hvl) (k0 : Hvl) ,
      ((weakenCutoffZ (weakenCutoffZ c k) k0) =
      (weakenCutoffZ c (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Fixpoint shift_A_Index (c : (Cutoff A)) (a : (Index A)) {struct c} : (Index A) :=
    match c with
      | (C0) => (IS a)
      | (CS c) => match a with
        | (I0) => I0
        | (IS a) => (IS (shift_A_Index c a))
      end
    end.
  Fixpoint shift_B_Index (c : (Cutoff B)) (b : (Index B)) {struct c} : (Index B) :=
    match c with
      | (C0) => (IS b)
      | (CS c) => match b with
        | (I0) => I0
        | (IS b) => (IS (shift_B_Index c b))
      end
    end.
  Fixpoint shift_Z_Index (c : (Cutoff Z)) (z : (Index Z)) {struct c} : (Index Z) :=
    match c with
      | (C0) => (IS z)
      | (CS c) => match z with
        | (I0) => I0
        | (IS z) => (IS (shift_Z_Index c z))
      end
    end.
  Fixpoint shift_A_AA (c : (Cutoff A)) (aa : AA) {struct aa} : AA :=
    match aa with
      | (avar a) => (avar (shift_A_Index c a))
      | (arec zz) => (arec (shift_A_ZZ (weakenCutoffA c H0) zz))
    end
  with shift_A_ZZ (c : (Cutoff A)) (zz : ZZ) {struct zz} : ZZ :=
    match zz with
      | (zvar z) => (zvar z)
      | (aabs aa) => (aabs (shift_A_AA (weakenCutoffA c (appendHvl (HS A H0) H0)) aa))
      | (babs bb) => (babs (shift_A_BB (weakenCutoffA c (appendHvl (HS B H0) H0)) bb))
    end
  with shift_A_BB (c : (Cutoff A)) (bb : BB) {struct bb} : BB :=
    match bb with
      | (bvar b) => (bvar b)
      | (brec zz) => (brec (shift_A_ZZ (weakenCutoffA c H0) zz))
    end.
  Fixpoint shift_B_AA (c : (Cutoff B)) (aa : AA) {struct aa} : AA :=
    match aa with
      | (avar a) => (avar a)
      | (arec zz) => (arec (shift_B_ZZ (weakenCutoffB c H0) zz))
    end
  with shift_B_ZZ (c : (Cutoff B)) (zz : ZZ) {struct zz} : ZZ :=
    match zz with
      | (zvar z) => (zvar z)
      | (aabs aa) => (aabs (shift_B_AA (weakenCutoffB c (appendHvl (HS A H0) H0)) aa))
      | (babs bb) => (babs (shift_B_BB (weakenCutoffB c (appendHvl (HS B H0) H0)) bb))
    end
  with shift_B_BB (c : (Cutoff B)) (bb : BB) {struct bb} : BB :=
    match bb with
      | (bvar b) => (bvar (shift_B_Index c b))
      | (brec zz) => (brec (shift_B_ZZ (weakenCutoffB c H0) zz))
    end.
  Fixpoint shift_Z_AA (c : (Cutoff Z)) (aa : AA) {struct aa} : AA :=
    match aa with
      | (avar a) => (avar a)
      | (arec zz) => (arec (shift_Z_ZZ (weakenCutoffZ c H0) zz))
    end
  with shift_Z_ZZ (c : (Cutoff Z)) (zz : ZZ) {struct zz} : ZZ :=
    match zz with
      | (zvar z) => (zvar (shift_Z_Index c z))
      | (aabs aa) => (aabs (shift_Z_AA (weakenCutoffZ c (appendHvl (HS A H0) H0)) aa))
      | (babs bb) => (babs (shift_Z_BB (weakenCutoffZ c (appendHvl (HS B H0) H0)) bb))
    end
  with shift_Z_BB (c : (Cutoff Z)) (bb : BB) {struct bb} : BB :=
    match bb with
      | (bvar b) => (bvar b)
      | (brec zz) => (brec (shift_Z_ZZ (weakenCutoffZ c H0) zz))
    end.
End Shift.

Section Weaken.
  Fixpoint weakenAA (aa : AA) (k : Hvl) {struct k} : AA :=
    match k with
      | (H0) => aa
      | (HS A k) => (shift_A_AA C0 (weakenAA aa k))
      | (HS B k) => (shift_B_AA C0 (weakenAA aa k))
      | (HS Z k) => (shift_Z_AA C0 (weakenAA aa k))
    end.
  Fixpoint weakenZZ (zz : ZZ) (k : Hvl) {struct k} : ZZ :=
    match k with
      | (H0) => zz
      | (HS A k) => (shift_A_ZZ C0 (weakenZZ zz k))
      | (HS B k) => (shift_B_ZZ C0 (weakenZZ zz k))
      | (HS Z k) => (shift_Z_ZZ C0 (weakenZZ zz k))
    end.
  Fixpoint weakenBB (bb : BB) (k : Hvl) {struct k} : BB :=
    match k with
      | (H0) => bb
      | (HS A k) => (shift_A_BB C0 (weakenBB bb k))
      | (HS B k) => (shift_B_BB C0 (weakenBB bb k))
      | (HS Z k) => (shift_Z_BB C0 (weakenBB bb k))
    end.
End Weaken.

Section Subst.
  Inductive Trace (a : Namespace) : Type :=
    | X0 
    | XS (b : Namespace)
        (T : (Trace a)).
  
  Global Arguments X0 [a] .
  Global Arguments XS [a] b _ .
  Fixpoint weakenTrace {a : Namespace} (x : (Trace a)) (k : Hvl) {struct k} : (Trace a) :=
    match k with
      | (H0) => x
      | (HS b k) => (XS b (weakenTrace x k))
    end.
  Lemma weakenTrace_append (a : Namespace) :
    (forall (x : (Trace a)) (k : Hvl) (k0 : Hvl) ,
      ((weakenTrace (weakenTrace x k) k0) =
      (weakenTrace x (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Fixpoint subst_A_Index (d : (Trace A)) (aa : AA) (a : (Index A)) {struct d} : AA :=
    match d with
      | (X0) => match a with
        | (I0) => aa
        | (IS a) => (avar a)
      end
      | (XS A d) => match a with
        | (I0) => (avar I0)
        | (IS a) => (shift_A_AA C0 (subst_A_Index d aa a))
      end
      | (XS B d) => (shift_B_AA C0 (subst_A_Index d aa a))
      | (XS Z d) => (shift_Z_AA C0 (subst_A_Index d aa a))
    end.
  Fixpoint subst_B_Index (d : (Trace B)) (bb : BB) (b : (Index B)) {struct d} : BB :=
    match d with
      | (X0) => match b with
        | (I0) => bb
        | (IS b) => (bvar b)
      end
      | (XS A d) => (shift_A_BB C0 (subst_B_Index d bb b))
      | (XS B d) => match b with
        | (I0) => (bvar I0)
        | (IS b) => (shift_B_BB C0 (subst_B_Index d bb b))
      end
      | (XS Z d) => (shift_Z_BB C0 (subst_B_Index d bb b))
    end.
  Fixpoint subst_Z_Index (d : (Trace Z)) (zz : ZZ) (z : (Index Z)) {struct d} : ZZ :=
    match d with
      | (X0) => match z with
        | (I0) => zz
        | (IS z) => (zvar z)
      end
      | (XS A d) => (shift_A_ZZ C0 (subst_Z_Index d zz z))
      | (XS B d) => (shift_B_ZZ C0 (subst_Z_Index d zz z))
      | (XS Z d) => match z with
        | (I0) => (zvar I0)
        | (IS z) => (shift_Z_ZZ C0 (subst_Z_Index d zz z))
      end
    end.
  Fixpoint subst_A_AA (d : (Trace A)) (aa : AA) (aa0 : AA) {struct aa0} : AA :=
    match aa0 with
      | (avar a) => (subst_A_Index d aa a)
      | (arec zz) => (arec (subst_A_ZZ d aa zz))
    end
  with subst_A_ZZ (d : (Trace A)) (aa : AA) (zz : ZZ) {struct zz} : ZZ :=
    match zz with
      | (zvar z) => (zvar z)
      | (aabs aa0) => (aabs (subst_A_AA (XS A d) aa aa0))
      | (babs bb) => (babs (subst_A_BB (XS B d) aa bb))
    end
  with subst_A_BB (d : (Trace A)) (aa : AA) (bb : BB) {struct bb} : BB :=
    match bb with
      | (bvar b) => (bvar b)
      | (brec zz) => (brec (subst_A_ZZ d aa zz))
    end.
  Fixpoint subst_B_AA (d : (Trace B)) (bb : BB) (aa : AA) {struct aa} : AA :=
    match aa with
      | (avar a) => (avar a)
      | (arec zz) => (arec (subst_B_ZZ d bb zz))
    end
  with subst_B_ZZ (d : (Trace B)) (bb : BB) (zz : ZZ) {struct zz} : ZZ :=
    match zz with
      | (zvar z) => (zvar z)
      | (aabs aa) => (aabs (subst_B_AA (XS A d) bb aa))
      | (babs bb0) => (babs (subst_B_BB (XS B d) bb bb0))
    end
  with subst_B_BB (d : (Trace B)) (bb : BB) (bb0 : BB) {struct bb0} : BB :=
    match bb0 with
      | (bvar b) => (subst_B_Index d bb b)
      | (brec zz) => (brec (subst_B_ZZ d bb zz))
    end.
  Fixpoint subst_Z_AA (d : (Trace Z)) (zz : ZZ) (aa : AA) {struct aa} : AA :=
    match aa with
      | (avar a) => (avar a)
      | (arec zz0) => (arec (subst_Z_ZZ d zz zz0))
    end
  with subst_Z_ZZ (d : (Trace Z)) (zz : ZZ) (zz0 : ZZ) {struct zz0} : ZZ :=
    match zz0 with
      | (zvar z) => (subst_Z_Index d zz z)
      | (aabs aa) => (aabs (subst_Z_AA (XS A d) zz aa))
      | (babs bb) => (babs (subst_Z_BB (XS B d) zz bb))
    end
  with subst_Z_BB (d : (Trace Z)) (zz : ZZ) (bb : BB) {struct bb} : BB :=
    match bb with
      | (bvar b) => (bvar b)
      | (brec zz0) => (brec (subst_Z_ZZ d zz zz0))
    end.
End Subst.

Global Hint Resolve (f_equal (shift_A_AA C0)) : cong_shift0.
Global Hint Resolve (f_equal (shift_B_AA C0)) : cong_shift0.
Global Hint Resolve (f_equal (shift_Z_AA C0)) : cong_shift0.
Global Hint Resolve (f_equal (shift_A_BB C0)) : cong_shift0.
Global Hint Resolve (f_equal (shift_B_BB C0)) : cong_shift0.
Global Hint Resolve (f_equal (shift_Z_BB C0)) : cong_shift0.
Global Hint Resolve (f_equal (shift_A_ZZ C0)) : cong_shift0.
Global Hint Resolve (f_equal (shift_B_ZZ C0)) : cong_shift0.
Global Hint Resolve (f_equal (shift_Z_ZZ C0)) : cong_shift0.
 Hint Rewrite weakenCutoffA_append weakenCutoffB_append weakenCutoffZ_append : interaction.
 Hint Rewrite weakenCutoffA_append weakenCutoffB_append weakenCutoffZ_append : weakencutoff_append.
 Hint Rewrite weakenTrace_append : interaction.
 Hint Rewrite weakenTrace_append : weakentrace_append.
Section SubstShiftReflection.
  Lemma subst_A_Index0_shift_A_Index0_reflection_ind (aa8 : AA) :
    (forall (k26 : Hvl) (a26 : (Index A)) ,
      ((subst_A_Index (weakenTrace X0 k26) aa8 (shift_A_Index (weakenCutoffA C0 k26) a26)) =
      (avar a26))).
  Proof.
    needleGenericSubstIndexShiftIndexReflectionInd .
  Qed.
  Lemma subst_B_Index0_shift_B_Index0_reflection_ind (bb8 : BB) :
    (forall (k26 : Hvl) (b26 : (Index B)) ,
      ((subst_B_Index (weakenTrace X0 k26) bb8 (shift_B_Index (weakenCutoffB C0 k26) b26)) =
      (bvar b26))).
  Proof.
    needleGenericSubstIndexShiftIndexReflectionInd .
  Qed.
  Lemma subst_Z_Index0_shift_Z_Index0_reflection_ind (zz17 : ZZ) :
    (forall (k26 : Hvl) (z17 : (Index Z)) ,
      ((subst_Z_Index (weakenTrace X0 k26) zz17 (shift_Z_Index (weakenCutoffZ C0 k26) z17)) =
      (zvar z17))).
  Proof.
    needleGenericSubstIndexShiftIndexReflectionInd .
  Qed.
  Definition subst_A_0_shift_A_0_AA_ZZ_BB_reflection_ind := (ind_AA_ZZ_BB (fun (aa14 : AA) =>
    (forall (k35 : Hvl) (aa15 : AA) ,
      ((subst_A_AA (weakenTrace X0 k35) aa15 (shift_A_AA (weakenCutoffA C0 k35) aa14)) =
      aa14))) (fun (zz26 : ZZ) =>
    (forall (k35 : Hvl) (aa14 : AA) ,
      ((subst_A_ZZ (weakenTrace X0 k35) aa14 (shift_A_ZZ (weakenCutoffA C0 k35) zz26)) =
      zz26))) (fun (bb14 : BB) =>
    (forall (k35 : Hvl) (aa14 : AA) ,
      ((subst_A_BB (weakenTrace X0 k35) aa14 (shift_A_BB (weakenCutoffA C0 k35) bb14)) =
      bb14))) (fun (a35 : (Index A)) =>
    (fun (k35 : Hvl) (aa14 : AA) =>
      (subst_A_Index0_shift_A_Index0_reflection_ind aa14 k35 a35))) (fun (zz : ZZ) IHzz21 (k35 : Hvl) (aa14 : AA) =>
    (f_equal arec (IHzz21 k35 aa14))) (fun (z23 : (Index Z)) =>
    (fun (k35 : Hvl) (aa14 : AA) =>
      eq_refl)) (fun (aa : AA) IHaa8 (k35 : Hvl) (aa14 : AA) =>
    (f_equal aabs (eq_trans (f_equal3 subst_A_AA (weakenTrace_append A X0 k35 (HS A H0)) eq_refl eq_refl) (IHaa8 (HS A k35) aa14)))) (fun (bb : BB) IHbb9 (k35 : Hvl) (aa14 : AA) =>
    (f_equal babs (eq_trans (f_equal3 subst_A_BB (weakenTrace_append A X0 k35 (HS B H0)) eq_refl eq_refl) (IHbb9 (HS B k35) aa14)))) (fun (b35 : (Index B)) =>
    (fun (k35 : Hvl) (aa14 : AA) =>
      eq_refl)) (fun (zz : ZZ) IHzz21 (k35 : Hvl) (aa14 : AA) =>
    (f_equal brec (IHzz21 k35 aa14)))).
  Lemma subst_A_0_shift_A_0_AA_reflection_ind  :
    (forall (aa14 : AA) ,
      (forall (k35 : Hvl) (aa15 : AA) ,
        ((subst_A_AA (weakenTrace X0 k35) aa15 (shift_A_AA (weakenCutoffA C0 k35) aa14)) =
        aa14))).
  Proof.
    pose proof (subst_A_0_shift_A_0_AA_ZZ_BB_reflection_ind).
    destruct_conjs .
    easy .
  Qed.
  Lemma subst_A_0_shift_A_0_ZZ_reflection_ind  :
    (forall (zz26 : ZZ) ,
      (forall (k35 : Hvl) (aa15 : AA) ,
        ((subst_A_ZZ (weakenTrace X0 k35) aa15 (shift_A_ZZ (weakenCutoffA C0 k35) zz26)) =
        zz26))).
  Proof.
    pose proof (subst_A_0_shift_A_0_AA_ZZ_BB_reflection_ind).
    destruct_conjs .
    easy .
  Qed.
  Lemma subst_A_0_shift_A_0_BB_reflection_ind  :
    (forall (bb14 : BB) ,
      (forall (k35 : Hvl) (aa15 : AA) ,
        ((subst_A_BB (weakenTrace X0 k35) aa15 (shift_A_BB (weakenCutoffA C0 k35) bb14)) =
        bb14))).
  Proof.
    pose proof (subst_A_0_shift_A_0_AA_ZZ_BB_reflection_ind).
    destruct_conjs .
    easy .
  Qed.
  Definition subst_B_0_shift_B_0_AA_ZZ_BB_reflection_ind := (ind_AA_ZZ_BB (fun (aa14 : AA) =>
    (forall (k35 : Hvl) (bb14 : BB) ,
      ((subst_B_AA (weakenTrace X0 k35) bb14 (shift_B_AA (weakenCutoffB C0 k35) aa14)) =
      aa14))) (fun (zz26 : ZZ) =>
    (forall (k35 : Hvl) (bb14 : BB) ,
      ((subst_B_ZZ (weakenTrace X0 k35) bb14 (shift_B_ZZ (weakenCutoffB C0 k35) zz26)) =
      zz26))) (fun (bb14 : BB) =>
    (forall (k35 : Hvl) (bb15 : BB) ,
      ((subst_B_BB (weakenTrace X0 k35) bb15 (shift_B_BB (weakenCutoffB C0 k35) bb14)) =
      bb14))) (fun (a35 : (Index A)) =>
    (fun (k35 : Hvl) (bb14 : BB) =>
      eq_refl)) (fun (zz : ZZ) IHzz21 (k35 : Hvl) (bb14 : BB) =>
    (f_equal arec (IHzz21 k35 bb14))) (fun (z23 : (Index Z)) =>
    (fun (k35 : Hvl) (bb14 : BB) =>
      eq_refl)) (fun (aa : AA) IHaa8 (k35 : Hvl) (bb14 : BB) =>
    (f_equal aabs (eq_trans (f_equal3 subst_B_AA (weakenTrace_append B X0 k35 (HS A H0)) eq_refl eq_refl) (IHaa8 (HS A k35) bb14)))) (fun (bb : BB) IHbb9 (k35 : Hvl) (bb14 : BB) =>
    (f_equal babs (eq_trans (f_equal3 subst_B_BB (weakenTrace_append B X0 k35 (HS B H0)) eq_refl eq_refl) (IHbb9 (HS B k35) bb14)))) (fun (b35 : (Index B)) =>
    (fun (k35 : Hvl) (bb14 : BB) =>
      (subst_B_Index0_shift_B_Index0_reflection_ind bb14 k35 b35))) (fun (zz : ZZ) IHzz21 (k35 : Hvl) (bb14 : BB) =>
    (f_equal brec (IHzz21 k35 bb14)))).
  Lemma subst_B_0_shift_B_0_AA_reflection_ind  :
    (forall (aa14 : AA) ,
      (forall (k35 : Hvl) (bb14 : BB) ,
        ((subst_B_AA (weakenTrace X0 k35) bb14 (shift_B_AA (weakenCutoffB C0 k35) aa14)) =
        aa14))).
  Proof.
    pose proof (subst_B_0_shift_B_0_AA_ZZ_BB_reflection_ind).
    destruct_conjs .
    easy .
  Qed.
  Lemma subst_B_0_shift_B_0_ZZ_reflection_ind  :
    (forall (zz26 : ZZ) ,
      (forall (k35 : Hvl) (bb14 : BB) ,
        ((subst_B_ZZ (weakenTrace X0 k35) bb14 (shift_B_ZZ (weakenCutoffB C0 k35) zz26)) =
        zz26))).
  Proof.
    pose proof (subst_B_0_shift_B_0_AA_ZZ_BB_reflection_ind).
    destruct_conjs .
    easy .
  Qed.
  Lemma subst_B_0_shift_B_0_BB_reflection_ind  :
    (forall (bb14 : BB) ,
      (forall (k35 : Hvl) (bb15 : BB) ,
        ((subst_B_BB (weakenTrace X0 k35) bb15 (shift_B_BB (weakenCutoffB C0 k35) bb14)) =
        bb14))).
  Proof.
    pose proof (subst_B_0_shift_B_0_AA_ZZ_BB_reflection_ind).
    destruct_conjs .
    easy .
  Qed.
  Definition subst_Z_0_shift_Z_0_AA_ZZ_BB_reflection_ind := (ind_AA_ZZ_BB (fun (aa14 : AA) =>
    (forall (k35 : Hvl) (zz26 : ZZ) ,
      ((subst_Z_AA (weakenTrace X0 k35) zz26 (shift_Z_AA (weakenCutoffZ C0 k35) aa14)) =
      aa14))) (fun (zz26 : ZZ) =>
    (forall (k35 : Hvl) (zz27 : ZZ) ,
      ((subst_Z_ZZ (weakenTrace X0 k35) zz27 (shift_Z_ZZ (weakenCutoffZ C0 k35) zz26)) =
      zz26))) (fun (bb14 : BB) =>
    (forall (k35 : Hvl) (zz26 : ZZ) ,
      ((subst_Z_BB (weakenTrace X0 k35) zz26 (shift_Z_BB (weakenCutoffZ C0 k35) bb14)) =
      bb14))) (fun (a35 : (Index A)) =>
    (fun (k35 : Hvl) (zz26 : ZZ) =>
      eq_refl)) (fun (zz : ZZ) IHzz21 (k35 : Hvl) (zz26 : ZZ) =>
    (f_equal arec (IHzz21 k35 zz26))) (fun (z23 : (Index Z)) =>
    (fun (k35 : Hvl) (zz26 : ZZ) =>
      (subst_Z_Index0_shift_Z_Index0_reflection_ind zz26 k35 z23))) (fun (aa : AA) IHaa8 (k35 : Hvl) (zz26 : ZZ) =>
    (f_equal aabs (eq_trans (f_equal3 subst_Z_AA (weakenTrace_append Z X0 k35 (HS A H0)) eq_refl eq_refl) (IHaa8 (HS A k35) zz26)))) (fun (bb : BB) IHbb9 (k35 : Hvl) (zz26 : ZZ) =>
    (f_equal babs (eq_trans (f_equal3 subst_Z_BB (weakenTrace_append Z X0 k35 (HS B H0)) eq_refl eq_refl) (IHbb9 (HS B k35) zz26)))) (fun (b35 : (Index B)) =>
    (fun (k35 : Hvl) (zz26 : ZZ) =>
      eq_refl)) (fun (zz : ZZ) IHzz21 (k35 : Hvl) (zz26 : ZZ) =>
    (f_equal brec (IHzz21 k35 zz26)))).
  Lemma subst_Z_0_shift_Z_0_AA_reflection_ind  :
    (forall (aa14 : AA) ,
      (forall (k35 : Hvl) (zz26 : ZZ) ,
        ((subst_Z_AA (weakenTrace X0 k35) zz26 (shift_Z_AA (weakenCutoffZ C0 k35) aa14)) =
        aa14))).
  Proof.
    pose proof (subst_Z_0_shift_Z_0_AA_ZZ_BB_reflection_ind).
    destruct_conjs .
    easy .
  Qed.
  Lemma subst_Z_0_shift_Z_0_ZZ_reflection_ind  :
    (forall (zz26 : ZZ) ,
      (forall (k35 : Hvl) (zz27 : ZZ) ,
        ((subst_Z_ZZ (weakenTrace X0 k35) zz27 (shift_Z_ZZ (weakenCutoffZ C0 k35) zz26)) =
        zz26))).
  Proof.
    pose proof (subst_Z_0_shift_Z_0_AA_ZZ_BB_reflection_ind).
    destruct_conjs .
    easy .
  Qed.
  Lemma subst_Z_0_shift_Z_0_BB_reflection_ind  :
    (forall (bb14 : BB) ,
      (forall (k35 : Hvl) (zz27 : ZZ) ,
        ((subst_Z_BB (weakenTrace X0 k35) zz27 (shift_Z_BB (weakenCutoffZ C0 k35) bb14)) =
        bb14))).
  Proof.
    pose proof (subst_Z_0_shift_Z_0_AA_ZZ_BB_reflection_ind).
    destruct_conjs .
    easy .
  Qed.
  Definition subst_A_AA0_shift_A_AA0_reflection (aa15 : AA) : (forall (aa14 : AA) ,
    ((subst_A_AA X0 aa14 (shift_A_AA C0 aa15)) =
    aa15)) := (subst_A_0_shift_A_0_AA_reflection_ind aa15 H0).
  Definition subst_A_ZZ0_shift_A_ZZ0_reflection (zz26 : ZZ) : (forall (aa14 : AA) ,
    ((subst_A_ZZ X0 aa14 (shift_A_ZZ C0 zz26)) =
    zz26)) := (subst_A_0_shift_A_0_ZZ_reflection_ind zz26 H0).
  Definition subst_A_BB0_shift_A_BB0_reflection (bb14 : BB) : (forall (aa14 : AA) ,
    ((subst_A_BB X0 aa14 (shift_A_BB C0 bb14)) =
    bb14)) := (subst_A_0_shift_A_0_BB_reflection_ind bb14 H0).
  Definition subst_B_AA0_shift_B_AA0_reflection (aa14 : AA) : (forall (bb14 : BB) ,
    ((subst_B_AA X0 bb14 (shift_B_AA C0 aa14)) =
    aa14)) := (subst_B_0_shift_B_0_AA_reflection_ind aa14 H0).
  Definition subst_B_ZZ0_shift_B_ZZ0_reflection (zz26 : ZZ) : (forall (bb14 : BB) ,
    ((subst_B_ZZ X0 bb14 (shift_B_ZZ C0 zz26)) =
    zz26)) := (subst_B_0_shift_B_0_ZZ_reflection_ind zz26 H0).
  Definition subst_B_BB0_shift_B_BB0_reflection (bb15 : BB) : (forall (bb14 : BB) ,
    ((subst_B_BB X0 bb14 (shift_B_BB C0 bb15)) =
    bb15)) := (subst_B_0_shift_B_0_BB_reflection_ind bb15 H0).
  Definition subst_Z_AA0_shift_Z_AA0_reflection (aa14 : AA) : (forall (zz26 : ZZ) ,
    ((subst_Z_AA X0 zz26 (shift_Z_AA C0 aa14)) =
    aa14)) := (subst_Z_0_shift_Z_0_AA_reflection_ind aa14 H0).
  Definition subst_Z_ZZ0_shift_Z_ZZ0_reflection (zz27 : ZZ) : (forall (zz26 : ZZ) ,
    ((subst_Z_ZZ X0 zz26 (shift_Z_ZZ C0 zz27)) =
    zz27)) := (subst_Z_0_shift_Z_0_ZZ_reflection_ind zz27 H0).
  Definition subst_Z_BB0_shift_Z_BB0_reflection (bb14 : BB) : (forall (zz26 : ZZ) ,
    ((subst_Z_BB X0 zz26 (shift_Z_BB C0 bb14)) =
    bb14)) := (subst_Z_0_shift_Z_0_BB_reflection_ind bb14 H0).
End SubstShiftReflection.
Section ShiftInteraction.
  Section ShiftIndexCommInd.
    Lemma shift_A_Index_shift_A_Index0_comm_ind  :
      (forall (k : Hvl) (c : (Cutoff A)) (a : (Index A)) ,
        ((shift_A_Index (weakenCutoffA (CS c) k) (shift_A_Index (weakenCutoffA C0 k) a)) =
        (shift_A_Index (weakenCutoffA C0 k) (shift_A_Index (weakenCutoffA c k) a)))).
    Proof.
      needleGenericShiftIndexCommInd .
    Qed.
    Lemma shift_B_Index_shift_B_Index0_comm_ind  :
      (forall (k : Hvl) (c : (Cutoff B)) (b : (Index B)) ,
        ((shift_B_Index (weakenCutoffB (CS c) k) (shift_B_Index (weakenCutoffB C0 k) b)) =
        (shift_B_Index (weakenCutoffB C0 k) (shift_B_Index (weakenCutoffB c k) b)))).
    Proof.
      needleGenericShiftIndexCommInd .
    Qed.
    Lemma shift_Z_Index_shift_Z_Index0_comm_ind  :
      (forall (k : Hvl) (c : (Cutoff Z)) (z : (Index Z)) ,
        ((shift_Z_Index (weakenCutoffZ (CS c) k) (shift_Z_Index (weakenCutoffZ C0 k) z)) =
        (shift_Z_Index (weakenCutoffZ C0 k) (shift_Z_Index (weakenCutoffZ c k) z)))).
    Proof.
      needleGenericShiftIndexCommInd .
    Qed.
  End ShiftIndexCommInd.
  Section ShiftCommInd.
    Definition shift_A__shift_A_0_AA_ZZ_BB_comm_ind := (ind_AA_ZZ_BB (fun (aa8 : AA) =>
      (forall (k26 : Hvl) (c26 : (Cutoff A)) ,
        ((shift_A_AA (weakenCutoffA (CS c26) k26) (shift_A_AA (weakenCutoffA C0 k26) aa8)) =
        (shift_A_AA (weakenCutoffA C0 k26) (shift_A_AA (weakenCutoffA c26 k26) aa8))))) (fun (zz17 : ZZ) =>
      (forall (k26 : Hvl) (c26 : (Cutoff A)) ,
        ((shift_A_ZZ (weakenCutoffA (CS c26) k26) (shift_A_ZZ (weakenCutoffA C0 k26) zz17)) =
        (shift_A_ZZ (weakenCutoffA C0 k26) (shift_A_ZZ (weakenCutoffA c26 k26) zz17))))) (fun (bb8 : BB) =>
      (forall (k26 : Hvl) (c26 : (Cutoff A)) ,
        ((shift_A_BB (weakenCutoffA (CS c26) k26) (shift_A_BB (weakenCutoffA C0 k26) bb8)) =
        (shift_A_BB (weakenCutoffA C0 k26) (shift_A_BB (weakenCutoffA c26 k26) bb8))))) (fun (a26 : (Index A)) =>
      (fun (k26 : Hvl) (c26 : (Cutoff A)) =>
        (f_equal avar (shift_A_Index_shift_A_Index0_comm_ind k26 c26 a26)))) (fun (zz : ZZ) IHzz17 (k26 : Hvl) (c26 : (Cutoff A)) =>
      (f_equal arec (IHzz17 k26 c26))) (fun (z17 : (Index Z)) =>
      (fun (k26 : Hvl) (c26 : (Cutoff A)) =>
        eq_refl)) (fun (aa : AA) IHaa8 (k26 : Hvl) (c26 : (Cutoff A)) =>
      (f_equal aabs (IHaa8 (HS A k26) c26))) (fun (bb : BB) IHbb8 (k26 : Hvl) (c26 : (Cutoff A)) =>
      (f_equal babs (IHbb8 (HS B k26) c26))) (fun (b26 : (Index B)) =>
      (fun (k26 : Hvl) (c26 : (Cutoff A)) =>
        eq_refl)) (fun (zz : ZZ) IHzz17 (k26 : Hvl) (c26 : (Cutoff A)) =>
      (f_equal brec (IHzz17 k26 c26)))).
    Lemma shift_A__shift_A_0_AA_comm_ind  :
      (forall (aa8 : AA) ,
        (forall (k26 : Hvl) (c26 : (Cutoff A)) ,
          ((shift_A_AA (weakenCutoffA (CS c26) k26) (shift_A_AA (weakenCutoffA C0 k26) aa8)) =
          (shift_A_AA (weakenCutoffA C0 k26) (shift_A_AA (weakenCutoffA c26 k26) aa8))))).
    Proof.
      pose proof (shift_A__shift_A_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_A__shift_A_0_ZZ_comm_ind  :
      (forall (zz17 : ZZ) ,
        (forall (k26 : Hvl) (c26 : (Cutoff A)) ,
          ((shift_A_ZZ (weakenCutoffA (CS c26) k26) (shift_A_ZZ (weakenCutoffA C0 k26) zz17)) =
          (shift_A_ZZ (weakenCutoffA C0 k26) (shift_A_ZZ (weakenCutoffA c26 k26) zz17))))).
    Proof.
      pose proof (shift_A__shift_A_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_A__shift_A_0_BB_comm_ind  :
      (forall (bb8 : BB) ,
        (forall (k26 : Hvl) (c26 : (Cutoff A)) ,
          ((shift_A_BB (weakenCutoffA (CS c26) k26) (shift_A_BB (weakenCutoffA C0 k26) bb8)) =
          (shift_A_BB (weakenCutoffA C0 k26) (shift_A_BB (weakenCutoffA c26 k26) bb8))))).
    Proof.
      pose proof (shift_A__shift_A_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_A__shift_B_0_AA_ZZ_BB_comm_ind := (ind_AA_ZZ_BB (fun (aa8 : AA) =>
      (forall (k26 : Hvl) (c26 : (Cutoff A)) ,
        ((shift_A_AA (weakenCutoffA c26 k26) (shift_B_AA (weakenCutoffB C0 k26) aa8)) =
        (shift_B_AA (weakenCutoffB C0 k26) (shift_A_AA (weakenCutoffA c26 k26) aa8))))) (fun (zz17 : ZZ) =>
      (forall (k26 : Hvl) (c26 : (Cutoff A)) ,
        ((shift_A_ZZ (weakenCutoffA c26 k26) (shift_B_ZZ (weakenCutoffB C0 k26) zz17)) =
        (shift_B_ZZ (weakenCutoffB C0 k26) (shift_A_ZZ (weakenCutoffA c26 k26) zz17))))) (fun (bb8 : BB) =>
      (forall (k26 : Hvl) (c26 : (Cutoff A)) ,
        ((shift_A_BB (weakenCutoffA c26 k26) (shift_B_BB (weakenCutoffB C0 k26) bb8)) =
        (shift_B_BB (weakenCutoffB C0 k26) (shift_A_BB (weakenCutoffA c26 k26) bb8))))) (fun (a26 : (Index A)) =>
      (fun (k26 : Hvl) (c26 : (Cutoff A)) =>
        eq_refl)) (fun (zz : ZZ) IHzz17 (k26 : Hvl) (c26 : (Cutoff A)) =>
      (f_equal arec (IHzz17 k26 c26))) (fun (z17 : (Index Z)) =>
      (fun (k26 : Hvl) (c26 : (Cutoff A)) =>
        eq_refl)) (fun (aa : AA) IHaa8 (k26 : Hvl) (c26 : (Cutoff A)) =>
      (f_equal aabs (IHaa8 (HS A k26) c26))) (fun (bb : BB) IHbb8 (k26 : Hvl) (c26 : (Cutoff A)) =>
      (f_equal babs (IHbb8 (HS B k26) c26))) (fun (b26 : (Index B)) =>
      (fun (k26 : Hvl) (c26 : (Cutoff A)) =>
        eq_refl)) (fun (zz : ZZ) IHzz17 (k26 : Hvl) (c26 : (Cutoff A)) =>
      (f_equal brec (IHzz17 k26 c26)))).
    Lemma shift_A__shift_B_0_AA_comm_ind  :
      (forall (aa8 : AA) ,
        (forall (k26 : Hvl) (c26 : (Cutoff A)) ,
          ((shift_A_AA (weakenCutoffA c26 k26) (shift_B_AA (weakenCutoffB C0 k26) aa8)) =
          (shift_B_AA (weakenCutoffB C0 k26) (shift_A_AA (weakenCutoffA c26 k26) aa8))))).
    Proof.
      pose proof (shift_A__shift_B_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_A__shift_B_0_ZZ_comm_ind  :
      (forall (zz17 : ZZ) ,
        (forall (k26 : Hvl) (c26 : (Cutoff A)) ,
          ((shift_A_ZZ (weakenCutoffA c26 k26) (shift_B_ZZ (weakenCutoffB C0 k26) zz17)) =
          (shift_B_ZZ (weakenCutoffB C0 k26) (shift_A_ZZ (weakenCutoffA c26 k26) zz17))))).
    Proof.
      pose proof (shift_A__shift_B_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_A__shift_B_0_BB_comm_ind  :
      (forall (bb8 : BB) ,
        (forall (k26 : Hvl) (c26 : (Cutoff A)) ,
          ((shift_A_BB (weakenCutoffA c26 k26) (shift_B_BB (weakenCutoffB C0 k26) bb8)) =
          (shift_B_BB (weakenCutoffB C0 k26) (shift_A_BB (weakenCutoffA c26 k26) bb8))))).
    Proof.
      pose proof (shift_A__shift_B_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_A__shift_Z_0_AA_ZZ_BB_comm_ind := (ind_AA_ZZ_BB (fun (aa8 : AA) =>
      (forall (k26 : Hvl) (c26 : (Cutoff A)) ,
        ((shift_A_AA (weakenCutoffA c26 k26) (shift_Z_AA (weakenCutoffZ C0 k26) aa8)) =
        (shift_Z_AA (weakenCutoffZ C0 k26) (shift_A_AA (weakenCutoffA c26 k26) aa8))))) (fun (zz17 : ZZ) =>
      (forall (k26 : Hvl) (c26 : (Cutoff A)) ,
        ((shift_A_ZZ (weakenCutoffA c26 k26) (shift_Z_ZZ (weakenCutoffZ C0 k26) zz17)) =
        (shift_Z_ZZ (weakenCutoffZ C0 k26) (shift_A_ZZ (weakenCutoffA c26 k26) zz17))))) (fun (bb8 : BB) =>
      (forall (k26 : Hvl) (c26 : (Cutoff A)) ,
        ((shift_A_BB (weakenCutoffA c26 k26) (shift_Z_BB (weakenCutoffZ C0 k26) bb8)) =
        (shift_Z_BB (weakenCutoffZ C0 k26) (shift_A_BB (weakenCutoffA c26 k26) bb8))))) (fun (a26 : (Index A)) =>
      (fun (k26 : Hvl) (c26 : (Cutoff A)) =>
        eq_refl)) (fun (zz : ZZ) IHzz17 (k26 : Hvl) (c26 : (Cutoff A)) =>
      (f_equal arec (IHzz17 k26 c26))) (fun (z17 : (Index Z)) =>
      (fun (k26 : Hvl) (c26 : (Cutoff A)) =>
        eq_refl)) (fun (aa : AA) IHaa8 (k26 : Hvl) (c26 : (Cutoff A)) =>
      (f_equal aabs (IHaa8 (HS A k26) c26))) (fun (bb : BB) IHbb8 (k26 : Hvl) (c26 : (Cutoff A)) =>
      (f_equal babs (IHbb8 (HS B k26) c26))) (fun (b26 : (Index B)) =>
      (fun (k26 : Hvl) (c26 : (Cutoff A)) =>
        eq_refl)) (fun (zz : ZZ) IHzz17 (k26 : Hvl) (c26 : (Cutoff A)) =>
      (f_equal brec (IHzz17 k26 c26)))).
    Lemma shift_A__shift_Z_0_AA_comm_ind  :
      (forall (aa8 : AA) ,
        (forall (k26 : Hvl) (c26 : (Cutoff A)) ,
          ((shift_A_AA (weakenCutoffA c26 k26) (shift_Z_AA (weakenCutoffZ C0 k26) aa8)) =
          (shift_Z_AA (weakenCutoffZ C0 k26) (shift_A_AA (weakenCutoffA c26 k26) aa8))))).
    Proof.
      pose proof (shift_A__shift_Z_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_A__shift_Z_0_ZZ_comm_ind  :
      (forall (zz17 : ZZ) ,
        (forall (k26 : Hvl) (c26 : (Cutoff A)) ,
          ((shift_A_ZZ (weakenCutoffA c26 k26) (shift_Z_ZZ (weakenCutoffZ C0 k26) zz17)) =
          (shift_Z_ZZ (weakenCutoffZ C0 k26) (shift_A_ZZ (weakenCutoffA c26 k26) zz17))))).
    Proof.
      pose proof (shift_A__shift_Z_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_A__shift_Z_0_BB_comm_ind  :
      (forall (bb8 : BB) ,
        (forall (k26 : Hvl) (c26 : (Cutoff A)) ,
          ((shift_A_BB (weakenCutoffA c26 k26) (shift_Z_BB (weakenCutoffZ C0 k26) bb8)) =
          (shift_Z_BB (weakenCutoffZ C0 k26) (shift_A_BB (weakenCutoffA c26 k26) bb8))))).
    Proof.
      pose proof (shift_A__shift_Z_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_B__shift_A_0_AA_ZZ_BB_comm_ind := (ind_AA_ZZ_BB (fun (aa8 : AA) =>
      (forall (k26 : Hvl) (c26 : (Cutoff B)) ,
        ((shift_B_AA (weakenCutoffB c26 k26) (shift_A_AA (weakenCutoffA C0 k26) aa8)) =
        (shift_A_AA (weakenCutoffA C0 k26) (shift_B_AA (weakenCutoffB c26 k26) aa8))))) (fun (zz17 : ZZ) =>
      (forall (k26 : Hvl) (c26 : (Cutoff B)) ,
        ((shift_B_ZZ (weakenCutoffB c26 k26) (shift_A_ZZ (weakenCutoffA C0 k26) zz17)) =
        (shift_A_ZZ (weakenCutoffA C0 k26) (shift_B_ZZ (weakenCutoffB c26 k26) zz17))))) (fun (bb8 : BB) =>
      (forall (k26 : Hvl) (c26 : (Cutoff B)) ,
        ((shift_B_BB (weakenCutoffB c26 k26) (shift_A_BB (weakenCutoffA C0 k26) bb8)) =
        (shift_A_BB (weakenCutoffA C0 k26) (shift_B_BB (weakenCutoffB c26 k26) bb8))))) (fun (a26 : (Index A)) =>
      (fun (k26 : Hvl) (c26 : (Cutoff B)) =>
        eq_refl)) (fun (zz : ZZ) IHzz17 (k26 : Hvl) (c26 : (Cutoff B)) =>
      (f_equal arec (IHzz17 k26 c26))) (fun (z17 : (Index Z)) =>
      (fun (k26 : Hvl) (c26 : (Cutoff B)) =>
        eq_refl)) (fun (aa : AA) IHaa8 (k26 : Hvl) (c26 : (Cutoff B)) =>
      (f_equal aabs (IHaa8 (HS A k26) c26))) (fun (bb : BB) IHbb8 (k26 : Hvl) (c26 : (Cutoff B)) =>
      (f_equal babs (IHbb8 (HS B k26) c26))) (fun (b26 : (Index B)) =>
      (fun (k26 : Hvl) (c26 : (Cutoff B)) =>
        eq_refl)) (fun (zz : ZZ) IHzz17 (k26 : Hvl) (c26 : (Cutoff B)) =>
      (f_equal brec (IHzz17 k26 c26)))).
    Lemma shift_B__shift_A_0_AA_comm_ind  :
      (forall (aa8 : AA) ,
        (forall (k26 : Hvl) (c26 : (Cutoff B)) ,
          ((shift_B_AA (weakenCutoffB c26 k26) (shift_A_AA (weakenCutoffA C0 k26) aa8)) =
          (shift_A_AA (weakenCutoffA C0 k26) (shift_B_AA (weakenCutoffB c26 k26) aa8))))).
    Proof.
      pose proof (shift_B__shift_A_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_B__shift_A_0_ZZ_comm_ind  :
      (forall (zz17 : ZZ) ,
        (forall (k26 : Hvl) (c26 : (Cutoff B)) ,
          ((shift_B_ZZ (weakenCutoffB c26 k26) (shift_A_ZZ (weakenCutoffA C0 k26) zz17)) =
          (shift_A_ZZ (weakenCutoffA C0 k26) (shift_B_ZZ (weakenCutoffB c26 k26) zz17))))).
    Proof.
      pose proof (shift_B__shift_A_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_B__shift_A_0_BB_comm_ind  :
      (forall (bb8 : BB) ,
        (forall (k26 : Hvl) (c26 : (Cutoff B)) ,
          ((shift_B_BB (weakenCutoffB c26 k26) (shift_A_BB (weakenCutoffA C0 k26) bb8)) =
          (shift_A_BB (weakenCutoffA C0 k26) (shift_B_BB (weakenCutoffB c26 k26) bb8))))).
    Proof.
      pose proof (shift_B__shift_A_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_B__shift_B_0_AA_ZZ_BB_comm_ind := (ind_AA_ZZ_BB (fun (aa8 : AA) =>
      (forall (k26 : Hvl) (c26 : (Cutoff B)) ,
        ((shift_B_AA (weakenCutoffB (CS c26) k26) (shift_B_AA (weakenCutoffB C0 k26) aa8)) =
        (shift_B_AA (weakenCutoffB C0 k26) (shift_B_AA (weakenCutoffB c26 k26) aa8))))) (fun (zz17 : ZZ) =>
      (forall (k26 : Hvl) (c26 : (Cutoff B)) ,
        ((shift_B_ZZ (weakenCutoffB (CS c26) k26) (shift_B_ZZ (weakenCutoffB C0 k26) zz17)) =
        (shift_B_ZZ (weakenCutoffB C0 k26) (shift_B_ZZ (weakenCutoffB c26 k26) zz17))))) (fun (bb8 : BB) =>
      (forall (k26 : Hvl) (c26 : (Cutoff B)) ,
        ((shift_B_BB (weakenCutoffB (CS c26) k26) (shift_B_BB (weakenCutoffB C0 k26) bb8)) =
        (shift_B_BB (weakenCutoffB C0 k26) (shift_B_BB (weakenCutoffB c26 k26) bb8))))) (fun (a26 : (Index A)) =>
      (fun (k26 : Hvl) (c26 : (Cutoff B)) =>
        eq_refl)) (fun (zz : ZZ) IHzz17 (k26 : Hvl) (c26 : (Cutoff B)) =>
      (f_equal arec (IHzz17 k26 c26))) (fun (z17 : (Index Z)) =>
      (fun (k26 : Hvl) (c26 : (Cutoff B)) =>
        eq_refl)) (fun (aa : AA) IHaa8 (k26 : Hvl) (c26 : (Cutoff B)) =>
      (f_equal aabs (IHaa8 (HS A k26) c26))) (fun (bb : BB) IHbb8 (k26 : Hvl) (c26 : (Cutoff B)) =>
      (f_equal babs (IHbb8 (HS B k26) c26))) (fun (b26 : (Index B)) =>
      (fun (k26 : Hvl) (c26 : (Cutoff B)) =>
        (f_equal bvar (shift_B_Index_shift_B_Index0_comm_ind k26 c26 b26)))) (fun (zz : ZZ) IHzz17 (k26 : Hvl) (c26 : (Cutoff B)) =>
      (f_equal brec (IHzz17 k26 c26)))).
    Lemma shift_B__shift_B_0_AA_comm_ind  :
      (forall (aa8 : AA) ,
        (forall (k26 : Hvl) (c26 : (Cutoff B)) ,
          ((shift_B_AA (weakenCutoffB (CS c26) k26) (shift_B_AA (weakenCutoffB C0 k26) aa8)) =
          (shift_B_AA (weakenCutoffB C0 k26) (shift_B_AA (weakenCutoffB c26 k26) aa8))))).
    Proof.
      pose proof (shift_B__shift_B_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_B__shift_B_0_ZZ_comm_ind  :
      (forall (zz17 : ZZ) ,
        (forall (k26 : Hvl) (c26 : (Cutoff B)) ,
          ((shift_B_ZZ (weakenCutoffB (CS c26) k26) (shift_B_ZZ (weakenCutoffB C0 k26) zz17)) =
          (shift_B_ZZ (weakenCutoffB C0 k26) (shift_B_ZZ (weakenCutoffB c26 k26) zz17))))).
    Proof.
      pose proof (shift_B__shift_B_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_B__shift_B_0_BB_comm_ind  :
      (forall (bb8 : BB) ,
        (forall (k26 : Hvl) (c26 : (Cutoff B)) ,
          ((shift_B_BB (weakenCutoffB (CS c26) k26) (shift_B_BB (weakenCutoffB C0 k26) bb8)) =
          (shift_B_BB (weakenCutoffB C0 k26) (shift_B_BB (weakenCutoffB c26 k26) bb8))))).
    Proof.
      pose proof (shift_B__shift_B_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_B__shift_Z_0_AA_ZZ_BB_comm_ind := (ind_AA_ZZ_BB (fun (aa8 : AA) =>
      (forall (k26 : Hvl) (c26 : (Cutoff B)) ,
        ((shift_B_AA (weakenCutoffB c26 k26) (shift_Z_AA (weakenCutoffZ C0 k26) aa8)) =
        (shift_Z_AA (weakenCutoffZ C0 k26) (shift_B_AA (weakenCutoffB c26 k26) aa8))))) (fun (zz17 : ZZ) =>
      (forall (k26 : Hvl) (c26 : (Cutoff B)) ,
        ((shift_B_ZZ (weakenCutoffB c26 k26) (shift_Z_ZZ (weakenCutoffZ C0 k26) zz17)) =
        (shift_Z_ZZ (weakenCutoffZ C0 k26) (shift_B_ZZ (weakenCutoffB c26 k26) zz17))))) (fun (bb8 : BB) =>
      (forall (k26 : Hvl) (c26 : (Cutoff B)) ,
        ((shift_B_BB (weakenCutoffB c26 k26) (shift_Z_BB (weakenCutoffZ C0 k26) bb8)) =
        (shift_Z_BB (weakenCutoffZ C0 k26) (shift_B_BB (weakenCutoffB c26 k26) bb8))))) (fun (a26 : (Index A)) =>
      (fun (k26 : Hvl) (c26 : (Cutoff B)) =>
        eq_refl)) (fun (zz : ZZ) IHzz17 (k26 : Hvl) (c26 : (Cutoff B)) =>
      (f_equal arec (IHzz17 k26 c26))) (fun (z17 : (Index Z)) =>
      (fun (k26 : Hvl) (c26 : (Cutoff B)) =>
        eq_refl)) (fun (aa : AA) IHaa8 (k26 : Hvl) (c26 : (Cutoff B)) =>
      (f_equal aabs (IHaa8 (HS A k26) c26))) (fun (bb : BB) IHbb8 (k26 : Hvl) (c26 : (Cutoff B)) =>
      (f_equal babs (IHbb8 (HS B k26) c26))) (fun (b26 : (Index B)) =>
      (fun (k26 : Hvl) (c26 : (Cutoff B)) =>
        eq_refl)) (fun (zz : ZZ) IHzz17 (k26 : Hvl) (c26 : (Cutoff B)) =>
      (f_equal brec (IHzz17 k26 c26)))).
    Lemma shift_B__shift_Z_0_AA_comm_ind  :
      (forall (aa8 : AA) ,
        (forall (k26 : Hvl) (c26 : (Cutoff B)) ,
          ((shift_B_AA (weakenCutoffB c26 k26) (shift_Z_AA (weakenCutoffZ C0 k26) aa8)) =
          (shift_Z_AA (weakenCutoffZ C0 k26) (shift_B_AA (weakenCutoffB c26 k26) aa8))))).
    Proof.
      pose proof (shift_B__shift_Z_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_B__shift_Z_0_ZZ_comm_ind  :
      (forall (zz17 : ZZ) ,
        (forall (k26 : Hvl) (c26 : (Cutoff B)) ,
          ((shift_B_ZZ (weakenCutoffB c26 k26) (shift_Z_ZZ (weakenCutoffZ C0 k26) zz17)) =
          (shift_Z_ZZ (weakenCutoffZ C0 k26) (shift_B_ZZ (weakenCutoffB c26 k26) zz17))))).
    Proof.
      pose proof (shift_B__shift_Z_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_B__shift_Z_0_BB_comm_ind  :
      (forall (bb8 : BB) ,
        (forall (k26 : Hvl) (c26 : (Cutoff B)) ,
          ((shift_B_BB (weakenCutoffB c26 k26) (shift_Z_BB (weakenCutoffZ C0 k26) bb8)) =
          (shift_Z_BB (weakenCutoffZ C0 k26) (shift_B_BB (weakenCutoffB c26 k26) bb8))))).
    Proof.
      pose proof (shift_B__shift_Z_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_Z__shift_A_0_AA_ZZ_BB_comm_ind := (ind_AA_ZZ_BB (fun (aa8 : AA) =>
      (forall (k26 : Hvl) (c26 : (Cutoff Z)) ,
        ((shift_Z_AA (weakenCutoffZ c26 k26) (shift_A_AA (weakenCutoffA C0 k26) aa8)) =
        (shift_A_AA (weakenCutoffA C0 k26) (shift_Z_AA (weakenCutoffZ c26 k26) aa8))))) (fun (zz17 : ZZ) =>
      (forall (k26 : Hvl) (c26 : (Cutoff Z)) ,
        ((shift_Z_ZZ (weakenCutoffZ c26 k26) (shift_A_ZZ (weakenCutoffA C0 k26) zz17)) =
        (shift_A_ZZ (weakenCutoffA C0 k26) (shift_Z_ZZ (weakenCutoffZ c26 k26) zz17))))) (fun (bb8 : BB) =>
      (forall (k26 : Hvl) (c26 : (Cutoff Z)) ,
        ((shift_Z_BB (weakenCutoffZ c26 k26) (shift_A_BB (weakenCutoffA C0 k26) bb8)) =
        (shift_A_BB (weakenCutoffA C0 k26) (shift_Z_BB (weakenCutoffZ c26 k26) bb8))))) (fun (a26 : (Index A)) =>
      (fun (k26 : Hvl) (c26 : (Cutoff Z)) =>
        eq_refl)) (fun (zz : ZZ) IHzz17 (k26 : Hvl) (c26 : (Cutoff Z)) =>
      (f_equal arec (IHzz17 k26 c26))) (fun (z17 : (Index Z)) =>
      (fun (k26 : Hvl) (c26 : (Cutoff Z)) =>
        eq_refl)) (fun (aa : AA) IHaa8 (k26 : Hvl) (c26 : (Cutoff Z)) =>
      (f_equal aabs (IHaa8 (HS A k26) c26))) (fun (bb : BB) IHbb8 (k26 : Hvl) (c26 : (Cutoff Z)) =>
      (f_equal babs (IHbb8 (HS B k26) c26))) (fun (b26 : (Index B)) =>
      (fun (k26 : Hvl) (c26 : (Cutoff Z)) =>
        eq_refl)) (fun (zz : ZZ) IHzz17 (k26 : Hvl) (c26 : (Cutoff Z)) =>
      (f_equal brec (IHzz17 k26 c26)))).
    Lemma shift_Z__shift_A_0_AA_comm_ind  :
      (forall (aa8 : AA) ,
        (forall (k26 : Hvl) (c26 : (Cutoff Z)) ,
          ((shift_Z_AA (weakenCutoffZ c26 k26) (shift_A_AA (weakenCutoffA C0 k26) aa8)) =
          (shift_A_AA (weakenCutoffA C0 k26) (shift_Z_AA (weakenCutoffZ c26 k26) aa8))))).
    Proof.
      pose proof (shift_Z__shift_A_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_Z__shift_A_0_ZZ_comm_ind  :
      (forall (zz17 : ZZ) ,
        (forall (k26 : Hvl) (c26 : (Cutoff Z)) ,
          ((shift_Z_ZZ (weakenCutoffZ c26 k26) (shift_A_ZZ (weakenCutoffA C0 k26) zz17)) =
          (shift_A_ZZ (weakenCutoffA C0 k26) (shift_Z_ZZ (weakenCutoffZ c26 k26) zz17))))).
    Proof.
      pose proof (shift_Z__shift_A_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_Z__shift_A_0_BB_comm_ind  :
      (forall (bb8 : BB) ,
        (forall (k26 : Hvl) (c26 : (Cutoff Z)) ,
          ((shift_Z_BB (weakenCutoffZ c26 k26) (shift_A_BB (weakenCutoffA C0 k26) bb8)) =
          (shift_A_BB (weakenCutoffA C0 k26) (shift_Z_BB (weakenCutoffZ c26 k26) bb8))))).
    Proof.
      pose proof (shift_Z__shift_A_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_Z__shift_B_0_AA_ZZ_BB_comm_ind := (ind_AA_ZZ_BB (fun (aa8 : AA) =>
      (forall (k26 : Hvl) (c26 : (Cutoff Z)) ,
        ((shift_Z_AA (weakenCutoffZ c26 k26) (shift_B_AA (weakenCutoffB C0 k26) aa8)) =
        (shift_B_AA (weakenCutoffB C0 k26) (shift_Z_AA (weakenCutoffZ c26 k26) aa8))))) (fun (zz17 : ZZ) =>
      (forall (k26 : Hvl) (c26 : (Cutoff Z)) ,
        ((shift_Z_ZZ (weakenCutoffZ c26 k26) (shift_B_ZZ (weakenCutoffB C0 k26) zz17)) =
        (shift_B_ZZ (weakenCutoffB C0 k26) (shift_Z_ZZ (weakenCutoffZ c26 k26) zz17))))) (fun (bb8 : BB) =>
      (forall (k26 : Hvl) (c26 : (Cutoff Z)) ,
        ((shift_Z_BB (weakenCutoffZ c26 k26) (shift_B_BB (weakenCutoffB C0 k26) bb8)) =
        (shift_B_BB (weakenCutoffB C0 k26) (shift_Z_BB (weakenCutoffZ c26 k26) bb8))))) (fun (a26 : (Index A)) =>
      (fun (k26 : Hvl) (c26 : (Cutoff Z)) =>
        eq_refl)) (fun (zz : ZZ) IHzz17 (k26 : Hvl) (c26 : (Cutoff Z)) =>
      (f_equal arec (IHzz17 k26 c26))) (fun (z17 : (Index Z)) =>
      (fun (k26 : Hvl) (c26 : (Cutoff Z)) =>
        eq_refl)) (fun (aa : AA) IHaa8 (k26 : Hvl) (c26 : (Cutoff Z)) =>
      (f_equal aabs (IHaa8 (HS A k26) c26))) (fun (bb : BB) IHbb8 (k26 : Hvl) (c26 : (Cutoff Z)) =>
      (f_equal babs (IHbb8 (HS B k26) c26))) (fun (b26 : (Index B)) =>
      (fun (k26 : Hvl) (c26 : (Cutoff Z)) =>
        eq_refl)) (fun (zz : ZZ) IHzz17 (k26 : Hvl) (c26 : (Cutoff Z)) =>
      (f_equal brec (IHzz17 k26 c26)))).
    Lemma shift_Z__shift_B_0_AA_comm_ind  :
      (forall (aa8 : AA) ,
        (forall (k26 : Hvl) (c26 : (Cutoff Z)) ,
          ((shift_Z_AA (weakenCutoffZ c26 k26) (shift_B_AA (weakenCutoffB C0 k26) aa8)) =
          (shift_B_AA (weakenCutoffB C0 k26) (shift_Z_AA (weakenCutoffZ c26 k26) aa8))))).
    Proof.
      pose proof (shift_Z__shift_B_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_Z__shift_B_0_ZZ_comm_ind  :
      (forall (zz17 : ZZ) ,
        (forall (k26 : Hvl) (c26 : (Cutoff Z)) ,
          ((shift_Z_ZZ (weakenCutoffZ c26 k26) (shift_B_ZZ (weakenCutoffB C0 k26) zz17)) =
          (shift_B_ZZ (weakenCutoffB C0 k26) (shift_Z_ZZ (weakenCutoffZ c26 k26) zz17))))).
    Proof.
      pose proof (shift_Z__shift_B_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_Z__shift_B_0_BB_comm_ind  :
      (forall (bb8 : BB) ,
        (forall (k26 : Hvl) (c26 : (Cutoff Z)) ,
          ((shift_Z_BB (weakenCutoffZ c26 k26) (shift_B_BB (weakenCutoffB C0 k26) bb8)) =
          (shift_B_BB (weakenCutoffB C0 k26) (shift_Z_BB (weakenCutoffZ c26 k26) bb8))))).
    Proof.
      pose proof (shift_Z__shift_B_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_Z__shift_Z_0_AA_ZZ_BB_comm_ind := (ind_AA_ZZ_BB (fun (aa8 : AA) =>
      (forall (k26 : Hvl) (c26 : (Cutoff Z)) ,
        ((shift_Z_AA (weakenCutoffZ (CS c26) k26) (shift_Z_AA (weakenCutoffZ C0 k26) aa8)) =
        (shift_Z_AA (weakenCutoffZ C0 k26) (shift_Z_AA (weakenCutoffZ c26 k26) aa8))))) (fun (zz17 : ZZ) =>
      (forall (k26 : Hvl) (c26 : (Cutoff Z)) ,
        ((shift_Z_ZZ (weakenCutoffZ (CS c26) k26) (shift_Z_ZZ (weakenCutoffZ C0 k26) zz17)) =
        (shift_Z_ZZ (weakenCutoffZ C0 k26) (shift_Z_ZZ (weakenCutoffZ c26 k26) zz17))))) (fun (bb8 : BB) =>
      (forall (k26 : Hvl) (c26 : (Cutoff Z)) ,
        ((shift_Z_BB (weakenCutoffZ (CS c26) k26) (shift_Z_BB (weakenCutoffZ C0 k26) bb8)) =
        (shift_Z_BB (weakenCutoffZ C0 k26) (shift_Z_BB (weakenCutoffZ c26 k26) bb8))))) (fun (a26 : (Index A)) =>
      (fun (k26 : Hvl) (c26 : (Cutoff Z)) =>
        eq_refl)) (fun (zz : ZZ) IHzz17 (k26 : Hvl) (c26 : (Cutoff Z)) =>
      (f_equal arec (IHzz17 k26 c26))) (fun (z17 : (Index Z)) =>
      (fun (k26 : Hvl) (c26 : (Cutoff Z)) =>
        (f_equal zvar (shift_Z_Index_shift_Z_Index0_comm_ind k26 c26 z17)))) (fun (aa : AA) IHaa8 (k26 : Hvl) (c26 : (Cutoff Z)) =>
      (f_equal aabs (IHaa8 (HS A k26) c26))) (fun (bb : BB) IHbb8 (k26 : Hvl) (c26 : (Cutoff Z)) =>
      (f_equal babs (IHbb8 (HS B k26) c26))) (fun (b26 : (Index B)) =>
      (fun (k26 : Hvl) (c26 : (Cutoff Z)) =>
        eq_refl)) (fun (zz : ZZ) IHzz17 (k26 : Hvl) (c26 : (Cutoff Z)) =>
      (f_equal brec (IHzz17 k26 c26)))).
    Lemma shift_Z__shift_Z_0_AA_comm_ind  :
      (forall (aa8 : AA) ,
        (forall (k26 : Hvl) (c26 : (Cutoff Z)) ,
          ((shift_Z_AA (weakenCutoffZ (CS c26) k26) (shift_Z_AA (weakenCutoffZ C0 k26) aa8)) =
          (shift_Z_AA (weakenCutoffZ C0 k26) (shift_Z_AA (weakenCutoffZ c26 k26) aa8))))).
    Proof.
      pose proof (shift_Z__shift_Z_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_Z__shift_Z_0_ZZ_comm_ind  :
      (forall (zz17 : ZZ) ,
        (forall (k26 : Hvl) (c26 : (Cutoff Z)) ,
          ((shift_Z_ZZ (weakenCutoffZ (CS c26) k26) (shift_Z_ZZ (weakenCutoffZ C0 k26) zz17)) =
          (shift_Z_ZZ (weakenCutoffZ C0 k26) (shift_Z_ZZ (weakenCutoffZ c26 k26) zz17))))).
    Proof.
      pose proof (shift_Z__shift_Z_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_Z__shift_Z_0_BB_comm_ind  :
      (forall (bb8 : BB) ,
        (forall (k26 : Hvl) (c26 : (Cutoff Z)) ,
          ((shift_Z_BB (weakenCutoffZ (CS c26) k26) (shift_Z_BB (weakenCutoffZ C0 k26) bb8)) =
          (shift_Z_BB (weakenCutoffZ C0 k26) (shift_Z_BB (weakenCutoffZ c26 k26) bb8))))).
    Proof.
      pose proof (shift_Z__shift_Z_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
  End ShiftCommInd.
  Section ShiftComm.
    Definition shift_A__shift_A_0_AA_comm (aa8 : AA) : (forall (c26 : (Cutoff A)) ,
      ((shift_A_AA (CS c26) (shift_A_AA C0 aa8)) =
      (shift_A_AA C0 (shift_A_AA c26 aa8)))) := (shift_A__shift_A_0_AA_comm_ind aa8 H0).
    Definition shift_A__shift_A_0_ZZ_comm (zz17 : ZZ) : (forall (c26 : (Cutoff A)) ,
      ((shift_A_ZZ (CS c26) (shift_A_ZZ C0 zz17)) =
      (shift_A_ZZ C0 (shift_A_ZZ c26 zz17)))) := (shift_A__shift_A_0_ZZ_comm_ind zz17 H0).
    Definition shift_A__shift_A_0_BB_comm (bb8 : BB) : (forall (c26 : (Cutoff A)) ,
      ((shift_A_BB (CS c26) (shift_A_BB C0 bb8)) =
      (shift_A_BB C0 (shift_A_BB c26 bb8)))) := (shift_A__shift_A_0_BB_comm_ind bb8 H0).
    Definition shift_A__shift_B_0_AA_comm (aa8 : AA) : (forall (c26 : (Cutoff A)) ,
      ((shift_A_AA c26 (shift_B_AA C0 aa8)) =
      (shift_B_AA C0 (shift_A_AA c26 aa8)))) := (shift_A__shift_B_0_AA_comm_ind aa8 H0).
    Definition shift_A__shift_B_0_ZZ_comm (zz17 : ZZ) : (forall (c26 : (Cutoff A)) ,
      ((shift_A_ZZ c26 (shift_B_ZZ C0 zz17)) =
      (shift_B_ZZ C0 (shift_A_ZZ c26 zz17)))) := (shift_A__shift_B_0_ZZ_comm_ind zz17 H0).
    Definition shift_A__shift_B_0_BB_comm (bb8 : BB) : (forall (c26 : (Cutoff A)) ,
      ((shift_A_BB c26 (shift_B_BB C0 bb8)) =
      (shift_B_BB C0 (shift_A_BB c26 bb8)))) := (shift_A__shift_B_0_BB_comm_ind bb8 H0).
    Definition shift_A__shift_Z_0_AA_comm (aa8 : AA) : (forall (c26 : (Cutoff A)) ,
      ((shift_A_AA c26 (shift_Z_AA C0 aa8)) =
      (shift_Z_AA C0 (shift_A_AA c26 aa8)))) := (shift_A__shift_Z_0_AA_comm_ind aa8 H0).
    Definition shift_A__shift_Z_0_ZZ_comm (zz17 : ZZ) : (forall (c26 : (Cutoff A)) ,
      ((shift_A_ZZ c26 (shift_Z_ZZ C0 zz17)) =
      (shift_Z_ZZ C0 (shift_A_ZZ c26 zz17)))) := (shift_A__shift_Z_0_ZZ_comm_ind zz17 H0).
    Definition shift_A__shift_Z_0_BB_comm (bb8 : BB) : (forall (c26 : (Cutoff A)) ,
      ((shift_A_BB c26 (shift_Z_BB C0 bb8)) =
      (shift_Z_BB C0 (shift_A_BB c26 bb8)))) := (shift_A__shift_Z_0_BB_comm_ind bb8 H0).
    Definition shift_B__shift_A_0_AA_comm (aa8 : AA) : (forall (c26 : (Cutoff B)) ,
      ((shift_B_AA c26 (shift_A_AA C0 aa8)) =
      (shift_A_AA C0 (shift_B_AA c26 aa8)))) := (shift_B__shift_A_0_AA_comm_ind aa8 H0).
    Definition shift_B__shift_A_0_ZZ_comm (zz17 : ZZ) : (forall (c26 : (Cutoff B)) ,
      ((shift_B_ZZ c26 (shift_A_ZZ C0 zz17)) =
      (shift_A_ZZ C0 (shift_B_ZZ c26 zz17)))) := (shift_B__shift_A_0_ZZ_comm_ind zz17 H0).
    Definition shift_B__shift_A_0_BB_comm (bb8 : BB) : (forall (c26 : (Cutoff B)) ,
      ((shift_B_BB c26 (shift_A_BB C0 bb8)) =
      (shift_A_BB C0 (shift_B_BB c26 bb8)))) := (shift_B__shift_A_0_BB_comm_ind bb8 H0).
    Definition shift_B__shift_B_0_AA_comm (aa8 : AA) : (forall (c26 : (Cutoff B)) ,
      ((shift_B_AA (CS c26) (shift_B_AA C0 aa8)) =
      (shift_B_AA C0 (shift_B_AA c26 aa8)))) := (shift_B__shift_B_0_AA_comm_ind aa8 H0).
    Definition shift_B__shift_B_0_ZZ_comm (zz17 : ZZ) : (forall (c26 : (Cutoff B)) ,
      ((shift_B_ZZ (CS c26) (shift_B_ZZ C0 zz17)) =
      (shift_B_ZZ C0 (shift_B_ZZ c26 zz17)))) := (shift_B__shift_B_0_ZZ_comm_ind zz17 H0).
    Definition shift_B__shift_B_0_BB_comm (bb8 : BB) : (forall (c26 : (Cutoff B)) ,
      ((shift_B_BB (CS c26) (shift_B_BB C0 bb8)) =
      (shift_B_BB C0 (shift_B_BB c26 bb8)))) := (shift_B__shift_B_0_BB_comm_ind bb8 H0).
    Definition shift_B__shift_Z_0_AA_comm (aa8 : AA) : (forall (c26 : (Cutoff B)) ,
      ((shift_B_AA c26 (shift_Z_AA C0 aa8)) =
      (shift_Z_AA C0 (shift_B_AA c26 aa8)))) := (shift_B__shift_Z_0_AA_comm_ind aa8 H0).
    Definition shift_B__shift_Z_0_ZZ_comm (zz17 : ZZ) : (forall (c26 : (Cutoff B)) ,
      ((shift_B_ZZ c26 (shift_Z_ZZ C0 zz17)) =
      (shift_Z_ZZ C0 (shift_B_ZZ c26 zz17)))) := (shift_B__shift_Z_0_ZZ_comm_ind zz17 H0).
    Definition shift_B__shift_Z_0_BB_comm (bb8 : BB) : (forall (c26 : (Cutoff B)) ,
      ((shift_B_BB c26 (shift_Z_BB C0 bb8)) =
      (shift_Z_BB C0 (shift_B_BB c26 bb8)))) := (shift_B__shift_Z_0_BB_comm_ind bb8 H0).
    Definition shift_Z__shift_A_0_AA_comm (aa8 : AA) : (forall (c26 : (Cutoff Z)) ,
      ((shift_Z_AA c26 (shift_A_AA C0 aa8)) =
      (shift_A_AA C0 (shift_Z_AA c26 aa8)))) := (shift_Z__shift_A_0_AA_comm_ind aa8 H0).
    Definition shift_Z__shift_A_0_ZZ_comm (zz17 : ZZ) : (forall (c26 : (Cutoff Z)) ,
      ((shift_Z_ZZ c26 (shift_A_ZZ C0 zz17)) =
      (shift_A_ZZ C0 (shift_Z_ZZ c26 zz17)))) := (shift_Z__shift_A_0_ZZ_comm_ind zz17 H0).
    Definition shift_Z__shift_A_0_BB_comm (bb8 : BB) : (forall (c26 : (Cutoff Z)) ,
      ((shift_Z_BB c26 (shift_A_BB C0 bb8)) =
      (shift_A_BB C0 (shift_Z_BB c26 bb8)))) := (shift_Z__shift_A_0_BB_comm_ind bb8 H0).
    Definition shift_Z__shift_B_0_AA_comm (aa8 : AA) : (forall (c26 : (Cutoff Z)) ,
      ((shift_Z_AA c26 (shift_B_AA C0 aa8)) =
      (shift_B_AA C0 (shift_Z_AA c26 aa8)))) := (shift_Z__shift_B_0_AA_comm_ind aa8 H0).
    Definition shift_Z__shift_B_0_ZZ_comm (zz17 : ZZ) : (forall (c26 : (Cutoff Z)) ,
      ((shift_Z_ZZ c26 (shift_B_ZZ C0 zz17)) =
      (shift_B_ZZ C0 (shift_Z_ZZ c26 zz17)))) := (shift_Z__shift_B_0_ZZ_comm_ind zz17 H0).
    Definition shift_Z__shift_B_0_BB_comm (bb8 : BB) : (forall (c26 : (Cutoff Z)) ,
      ((shift_Z_BB c26 (shift_B_BB C0 bb8)) =
      (shift_B_BB C0 (shift_Z_BB c26 bb8)))) := (shift_Z__shift_B_0_BB_comm_ind bb8 H0).
    Definition shift_Z__shift_Z_0_AA_comm (aa8 : AA) : (forall (c26 : (Cutoff Z)) ,
      ((shift_Z_AA (CS c26) (shift_Z_AA C0 aa8)) =
      (shift_Z_AA C0 (shift_Z_AA c26 aa8)))) := (shift_Z__shift_Z_0_AA_comm_ind aa8 H0).
    Definition shift_Z__shift_Z_0_ZZ_comm (zz17 : ZZ) : (forall (c26 : (Cutoff Z)) ,
      ((shift_Z_ZZ (CS c26) (shift_Z_ZZ C0 zz17)) =
      (shift_Z_ZZ C0 (shift_Z_ZZ c26 zz17)))) := (shift_Z__shift_Z_0_ZZ_comm_ind zz17 H0).
    Definition shift_Z__shift_Z_0_BB_comm (bb8 : BB) : (forall (c26 : (Cutoff Z)) ,
      ((shift_Z_BB (CS c26) (shift_Z_BB C0 bb8)) =
      (shift_Z_BB C0 (shift_Z_BB c26 bb8)))) := (shift_Z__shift_Z_0_BB_comm_ind bb8 H0).
  End ShiftComm.
End ShiftInteraction.
 Hint Rewrite shift_A__shift_A_0_AA_comm shift_A__shift_B_0_AA_comm shift_A__shift_Z_0_AA_comm shift_B__shift_A_0_AA_comm shift_B__shift_B_0_AA_comm shift_B__shift_Z_0_AA_comm shift_Z__shift_A_0_AA_comm shift_Z__shift_B_0_AA_comm shift_Z__shift_Z_0_AA_comm shift_A__shift_A_0_BB_comm shift_A__shift_B_0_BB_comm shift_A__shift_Z_0_BB_comm shift_B__shift_A_0_BB_comm shift_B__shift_B_0_BB_comm shift_B__shift_Z_0_BB_comm shift_Z__shift_A_0_BB_comm shift_Z__shift_B_0_BB_comm shift_Z__shift_Z_0_BB_comm shift_A__shift_A_0_ZZ_comm shift_A__shift_B_0_ZZ_comm shift_A__shift_Z_0_ZZ_comm shift_B__shift_A_0_ZZ_comm shift_B__shift_B_0_ZZ_comm shift_B__shift_Z_0_ZZ_comm shift_Z__shift_A_0_ZZ_comm shift_Z__shift_B_0_ZZ_comm shift_Z__shift_Z_0_ZZ_comm : interaction.
 Hint Rewrite shift_A__shift_A_0_AA_comm shift_A__shift_B_0_AA_comm shift_A__shift_Z_0_AA_comm shift_B__shift_A_0_AA_comm shift_B__shift_B_0_AA_comm shift_B__shift_Z_0_AA_comm shift_Z__shift_A_0_AA_comm shift_Z__shift_B_0_AA_comm shift_Z__shift_Z_0_AA_comm shift_A__shift_A_0_BB_comm shift_A__shift_B_0_BB_comm shift_A__shift_Z_0_BB_comm shift_B__shift_A_0_BB_comm shift_B__shift_B_0_BB_comm shift_B__shift_Z_0_BB_comm shift_Z__shift_A_0_BB_comm shift_Z__shift_B_0_BB_comm shift_Z__shift_Z_0_BB_comm shift_A__shift_A_0_ZZ_comm shift_A__shift_B_0_ZZ_comm shift_A__shift_Z_0_ZZ_comm shift_B__shift_A_0_ZZ_comm shift_B__shift_B_0_ZZ_comm shift_B__shift_Z_0_ZZ_comm shift_Z__shift_A_0_ZZ_comm shift_Z__shift_B_0_ZZ_comm shift_Z__shift_Z_0_ZZ_comm : shift_shift0.
Section WeakenShift.
  Lemma weakenAA_shift_A_AA  :
    (forall (k134 : Hvl) (c53 : (Cutoff A)) (aa71 : AA) ,
      ((weakenAA (shift_A_AA c53 aa71) k134) =
      (shift_A_AA (weakenCutoffA c53 k134) (weakenAA aa71 k134)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenZZ_shift_A_ZZ  :
    (forall (k134 : Hvl) (c53 : (Cutoff A)) (zz104 : ZZ) ,
      ((weakenZZ (shift_A_ZZ c53 zz104) k134) =
      (shift_A_ZZ (weakenCutoffA c53 k134) (weakenZZ zz104 k134)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenBB_shift_A_BB  :
    (forall (k134 : Hvl) (c53 : (Cutoff A)) (bb71 : BB) ,
      ((weakenBB (shift_A_BB c53 bb71) k134) =
      (shift_A_BB (weakenCutoffA c53 k134) (weakenBB bb71 k134)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenAA_shift_B_AA  :
    (forall (k134 : Hvl) (c53 : (Cutoff B)) (aa71 : AA) ,
      ((weakenAA (shift_B_AA c53 aa71) k134) =
      (shift_B_AA (weakenCutoffB c53 k134) (weakenAA aa71 k134)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenZZ_shift_B_ZZ  :
    (forall (k134 : Hvl) (c53 : (Cutoff B)) (zz104 : ZZ) ,
      ((weakenZZ (shift_B_ZZ c53 zz104) k134) =
      (shift_B_ZZ (weakenCutoffB c53 k134) (weakenZZ zz104 k134)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenBB_shift_B_BB  :
    (forall (k134 : Hvl) (c53 : (Cutoff B)) (bb71 : BB) ,
      ((weakenBB (shift_B_BB c53 bb71) k134) =
      (shift_B_BB (weakenCutoffB c53 k134) (weakenBB bb71 k134)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenAA_shift_Z_AA  :
    (forall (k134 : Hvl) (c53 : (Cutoff Z)) (aa71 : AA) ,
      ((weakenAA (shift_Z_AA c53 aa71) k134) =
      (shift_Z_AA (weakenCutoffZ c53 k134) (weakenAA aa71 k134)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenZZ_shift_Z_ZZ  :
    (forall (k134 : Hvl) (c53 : (Cutoff Z)) (zz104 : ZZ) ,
      ((weakenZZ (shift_Z_ZZ c53 zz104) k134) =
      (shift_Z_ZZ (weakenCutoffZ c53 k134) (weakenZZ zz104 k134)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenBB_shift_Z_BB  :
    (forall (k134 : Hvl) (c53 : (Cutoff Z)) (bb71 : BB) ,
      ((weakenBB (shift_Z_BB c53 bb71) k134) =
      (shift_Z_BB (weakenCutoffZ c53 k134) (weakenBB bb71 k134)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
End WeakenShift.
Section ShiftSubstInteraction.
  Section ShiftSubstIndexCommInd.
    Lemma shift_A_AA_subst_A_Index0_comm_ind (c26 : (Cutoff A)) (aa14 : AA) :
      (forall (k35 : Hvl) (a35 : (Index A)) ,
        ((shift_A_AA (weakenCutoffA c26 k35) (subst_A_Index (weakenTrace X0 k35) aa14 a35)) =
        (subst_A_Index (weakenTrace X0 k35) (shift_A_AA c26 aa14) (shift_A_Index (weakenCutoffA (CS c26) k35) a35)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma shift_B_AA_subst_A_Index0_comm_ind (c26 : (Cutoff B)) (aa14 : AA) :
      (forall (k35 : Hvl) (a35 : (Index A)) ,
        ((shift_B_AA (weakenCutoffB c26 k35) (subst_A_Index (weakenTrace X0 k35) aa14 a35)) =
        (subst_A_Index (weakenTrace X0 k35) (shift_B_AA c26 aa14) a35))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma shift_Z_AA_subst_A_Index0_comm_ind (c26 : (Cutoff Z)) (aa14 : AA) :
      (forall (k35 : Hvl) (a35 : (Index A)) ,
        ((shift_Z_AA (weakenCutoffZ c26 k35) (subst_A_Index (weakenTrace X0 k35) aa14 a35)) =
        (subst_A_Index (weakenTrace X0 k35) (shift_Z_AA c26 aa14) a35))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma shift_A_BB_subst_B_Index0_comm_ind (c26 : (Cutoff A)) (bb14 : BB) :
      (forall (k35 : Hvl) (b35 : (Index B)) ,
        ((shift_A_BB (weakenCutoffA c26 k35) (subst_B_Index (weakenTrace X0 k35) bb14 b35)) =
        (subst_B_Index (weakenTrace X0 k35) (shift_A_BB c26 bb14) b35))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma shift_B_BB_subst_B_Index0_comm_ind (c26 : (Cutoff B)) (bb14 : BB) :
      (forall (k35 : Hvl) (b35 : (Index B)) ,
        ((shift_B_BB (weakenCutoffB c26 k35) (subst_B_Index (weakenTrace X0 k35) bb14 b35)) =
        (subst_B_Index (weakenTrace X0 k35) (shift_B_BB c26 bb14) (shift_B_Index (weakenCutoffB (CS c26) k35) b35)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma shift_Z_BB_subst_B_Index0_comm_ind (c26 : (Cutoff Z)) (bb14 : BB) :
      (forall (k35 : Hvl) (b35 : (Index B)) ,
        ((shift_Z_BB (weakenCutoffZ c26 k35) (subst_B_Index (weakenTrace X0 k35) bb14 b35)) =
        (subst_B_Index (weakenTrace X0 k35) (shift_Z_BB c26 bb14) b35))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma shift_A_ZZ_subst_Z_Index0_comm_ind (c26 : (Cutoff A)) (zz26 : ZZ) :
      (forall (k35 : Hvl) (z23 : (Index Z)) ,
        ((shift_A_ZZ (weakenCutoffA c26 k35) (subst_Z_Index (weakenTrace X0 k35) zz26 z23)) =
        (subst_Z_Index (weakenTrace X0 k35) (shift_A_ZZ c26 zz26) z23))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma shift_B_ZZ_subst_Z_Index0_comm_ind (c26 : (Cutoff B)) (zz26 : ZZ) :
      (forall (k35 : Hvl) (z23 : (Index Z)) ,
        ((shift_B_ZZ (weakenCutoffB c26 k35) (subst_Z_Index (weakenTrace X0 k35) zz26 z23)) =
        (subst_Z_Index (weakenTrace X0 k35) (shift_B_ZZ c26 zz26) z23))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma shift_Z_ZZ_subst_Z_Index0_comm_ind (c26 : (Cutoff Z)) (zz26 : ZZ) :
      (forall (k35 : Hvl) (z23 : (Index Z)) ,
        ((shift_Z_ZZ (weakenCutoffZ c26 k35) (subst_Z_Index (weakenTrace X0 k35) zz26 z23)) =
        (subst_Z_Index (weakenTrace X0 k35) (shift_Z_ZZ c26 zz26) (shift_Z_Index (weakenCutoffZ (CS c26) k35) z23)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
  End ShiftSubstIndexCommInd.
  Section ShiftSubstCommInd.
    Definition shift_A__subst_A_0_AA_ZZ_BB_comm_ind := (ind_AA_ZZ_BB (fun (aa32 : AA) =>
      (forall (k62 : Hvl) (c53 : (Cutoff A)) (aa33 : AA) ,
        ((shift_A_AA (weakenCutoffA c53 k62) (subst_A_AA (weakenTrace X0 k62) aa33 aa32)) =
        (subst_A_AA (weakenTrace X0 k62) (shift_A_AA c53 aa33) (shift_A_AA (weakenCutoffA (CS c53) k62) aa32))))) (fun (zz53 : ZZ) =>
      (forall (k62 : Hvl) (c53 : (Cutoff A)) (aa32 : AA) ,
        ((shift_A_ZZ (weakenCutoffA c53 k62) (subst_A_ZZ (weakenTrace X0 k62) aa32 zz53)) =
        (subst_A_ZZ (weakenTrace X0 k62) (shift_A_AA c53 aa32) (shift_A_ZZ (weakenCutoffA (CS c53) k62) zz53))))) (fun (bb32 : BB) =>
      (forall (k62 : Hvl) (c53 : (Cutoff A)) (aa32 : AA) ,
        ((shift_A_BB (weakenCutoffA c53 k62) (subst_A_BB (weakenTrace X0 k62) aa32 bb32)) =
        (subst_A_BB (weakenTrace X0 k62) (shift_A_AA c53 aa32) (shift_A_BB (weakenCutoffA (CS c53) k62) bb32))))) (fun (a62 : (Index A)) =>
      (fun (k62 : Hvl) (c53 : (Cutoff A)) (aa32 : AA) =>
        (shift_A_AA_subst_A_Index0_comm_ind c53 aa32 k62 a62))) (fun (zz : ZZ) IHzz21 (k62 : Hvl) (c53 : (Cutoff A)) (aa32 : AA) =>
      (f_equal arec (IHzz21 k62 c53 aa32))) (fun (z41 : (Index Z)) =>
      (fun (k62 : Hvl) (c53 : (Cutoff A)) (aa32 : AA) =>
        eq_refl)) (fun (aa : AA) IHaa8 (k62 : Hvl) (c53 : (Cutoff A)) (aa32 : AA) =>
      (f_equal aabs (eq_trans (f_equal2 shift_A_AA eq_refl (f_equal3 subst_A_AA (weakenTrace_append A X0 k62 (HS A H0)) eq_refl eq_refl)) (eq_trans (IHaa8 (HS A k62) c53 aa32) (f_equal3 subst_A_AA (eq_sym (weakenTrace_append A X0 k62 (HS A H0))) eq_refl eq_refl))))) (fun (bb : BB) IHbb9 (k62 : Hvl) (c53 : (Cutoff A)) (aa32 : AA) =>
      (f_equal babs (eq_trans (f_equal2 shift_A_BB eq_refl (f_equal3 subst_A_BB (weakenTrace_append A X0 k62 (HS B H0)) eq_refl eq_refl)) (eq_trans (IHbb9 (HS B k62) c53 aa32) (f_equal3 subst_A_BB (eq_sym (weakenTrace_append A X0 k62 (HS B H0))) eq_refl eq_refl))))) (fun (b62 : (Index B)) =>
      (fun (k62 : Hvl) (c53 : (Cutoff A)) (aa32 : AA) =>
        eq_refl)) (fun (zz : ZZ) IHzz21 (k62 : Hvl) (c53 : (Cutoff A)) (aa32 : AA) =>
      (f_equal brec (IHzz21 k62 c53 aa32)))).
    Lemma shift_A__subst_A_0_AA_comm_ind  :
      (forall (aa32 : AA) ,
        (forall (k62 : Hvl) (c53 : (Cutoff A)) (aa33 : AA) ,
          ((shift_A_AA (weakenCutoffA c53 k62) (subst_A_AA (weakenTrace X0 k62) aa33 aa32)) =
          (subst_A_AA (weakenTrace X0 k62) (shift_A_AA c53 aa33) (shift_A_AA (weakenCutoffA (CS c53) k62) aa32))))).
    Proof.
      pose proof (shift_A__subst_A_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_A__subst_A_0_ZZ_comm_ind  :
      (forall (zz53 : ZZ) ,
        (forall (k62 : Hvl) (c53 : (Cutoff A)) (aa33 : AA) ,
          ((shift_A_ZZ (weakenCutoffA c53 k62) (subst_A_ZZ (weakenTrace X0 k62) aa33 zz53)) =
          (subst_A_ZZ (weakenTrace X0 k62) (shift_A_AA c53 aa33) (shift_A_ZZ (weakenCutoffA (CS c53) k62) zz53))))).
    Proof.
      pose proof (shift_A__subst_A_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_A__subst_A_0_BB_comm_ind  :
      (forall (bb32 : BB) ,
        (forall (k62 : Hvl) (c53 : (Cutoff A)) (aa33 : AA) ,
          ((shift_A_BB (weakenCutoffA c53 k62) (subst_A_BB (weakenTrace X0 k62) aa33 bb32)) =
          (subst_A_BB (weakenTrace X0 k62) (shift_A_AA c53 aa33) (shift_A_BB (weakenCutoffA (CS c53) k62) bb32))))).
    Proof.
      pose proof (shift_A__subst_A_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_A__subst_B_0_AA_ZZ_BB_comm_ind := (ind_AA_ZZ_BB (fun (aa32 : AA) =>
      (forall (k62 : Hvl) (c53 : (Cutoff A)) (bb32 : BB) ,
        ((shift_A_AA (weakenCutoffA c53 k62) (subst_B_AA (weakenTrace X0 k62) bb32 aa32)) =
        (subst_B_AA (weakenTrace X0 k62) (shift_A_BB c53 bb32) (shift_A_AA (weakenCutoffA c53 k62) aa32))))) (fun (zz53 : ZZ) =>
      (forall (k62 : Hvl) (c53 : (Cutoff A)) (bb32 : BB) ,
        ((shift_A_ZZ (weakenCutoffA c53 k62) (subst_B_ZZ (weakenTrace X0 k62) bb32 zz53)) =
        (subst_B_ZZ (weakenTrace X0 k62) (shift_A_BB c53 bb32) (shift_A_ZZ (weakenCutoffA c53 k62) zz53))))) (fun (bb32 : BB) =>
      (forall (k62 : Hvl) (c53 : (Cutoff A)) (bb33 : BB) ,
        ((shift_A_BB (weakenCutoffA c53 k62) (subst_B_BB (weakenTrace X0 k62) bb33 bb32)) =
        (subst_B_BB (weakenTrace X0 k62) (shift_A_BB c53 bb33) (shift_A_BB (weakenCutoffA c53 k62) bb32))))) (fun (a62 : (Index A)) =>
      (fun (k62 : Hvl) (c53 : (Cutoff A)) (bb32 : BB) =>
        eq_refl)) (fun (zz : ZZ) IHzz21 (k62 : Hvl) (c53 : (Cutoff A)) (bb32 : BB) =>
      (f_equal arec (IHzz21 k62 c53 bb32))) (fun (z41 : (Index Z)) =>
      (fun (k62 : Hvl) (c53 : (Cutoff A)) (bb32 : BB) =>
        eq_refl)) (fun (aa : AA) IHaa8 (k62 : Hvl) (c53 : (Cutoff A)) (bb32 : BB) =>
      (f_equal aabs (eq_trans (f_equal2 shift_A_AA eq_refl (f_equal3 subst_B_AA (weakenTrace_append B X0 k62 (HS A H0)) eq_refl eq_refl)) (eq_trans (IHaa8 (HS A k62) c53 bb32) (f_equal3 subst_B_AA (eq_sym (weakenTrace_append B X0 k62 (HS A H0))) eq_refl eq_refl))))) (fun (bb : BB) IHbb9 (k62 : Hvl) (c53 : (Cutoff A)) (bb32 : BB) =>
      (f_equal babs (eq_trans (f_equal2 shift_A_BB eq_refl (f_equal3 subst_B_BB (weakenTrace_append B X0 k62 (HS B H0)) eq_refl eq_refl)) (eq_trans (IHbb9 (HS B k62) c53 bb32) (f_equal3 subst_B_BB (eq_sym (weakenTrace_append B X0 k62 (HS B H0))) eq_refl eq_refl))))) (fun (b62 : (Index B)) =>
      (fun (k62 : Hvl) (c53 : (Cutoff A)) (bb32 : BB) =>
        (shift_A_BB_subst_B_Index0_comm_ind c53 bb32 k62 b62))) (fun (zz : ZZ) IHzz21 (k62 : Hvl) (c53 : (Cutoff A)) (bb32 : BB) =>
      (f_equal brec (IHzz21 k62 c53 bb32)))).
    Lemma shift_A__subst_B_0_AA_comm_ind  :
      (forall (aa32 : AA) ,
        (forall (k62 : Hvl) (c53 : (Cutoff A)) (bb32 : BB) ,
          ((shift_A_AA (weakenCutoffA c53 k62) (subst_B_AA (weakenTrace X0 k62) bb32 aa32)) =
          (subst_B_AA (weakenTrace X0 k62) (shift_A_BB c53 bb32) (shift_A_AA (weakenCutoffA c53 k62) aa32))))).
    Proof.
      pose proof (shift_A__subst_B_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_A__subst_B_0_ZZ_comm_ind  :
      (forall (zz53 : ZZ) ,
        (forall (k62 : Hvl) (c53 : (Cutoff A)) (bb32 : BB) ,
          ((shift_A_ZZ (weakenCutoffA c53 k62) (subst_B_ZZ (weakenTrace X0 k62) bb32 zz53)) =
          (subst_B_ZZ (weakenTrace X0 k62) (shift_A_BB c53 bb32) (shift_A_ZZ (weakenCutoffA c53 k62) zz53))))).
    Proof.
      pose proof (shift_A__subst_B_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_A__subst_B_0_BB_comm_ind  :
      (forall (bb32 : BB) ,
        (forall (k62 : Hvl) (c53 : (Cutoff A)) (bb33 : BB) ,
          ((shift_A_BB (weakenCutoffA c53 k62) (subst_B_BB (weakenTrace X0 k62) bb33 bb32)) =
          (subst_B_BB (weakenTrace X0 k62) (shift_A_BB c53 bb33) (shift_A_BB (weakenCutoffA c53 k62) bb32))))).
    Proof.
      pose proof (shift_A__subst_B_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_A__subst_Z_0_AA_ZZ_BB_comm_ind := (ind_AA_ZZ_BB (fun (aa32 : AA) =>
      (forall (k62 : Hvl) (c53 : (Cutoff A)) (zz53 : ZZ) ,
        ((shift_A_AA (weakenCutoffA c53 k62) (subst_Z_AA (weakenTrace X0 k62) zz53 aa32)) =
        (subst_Z_AA (weakenTrace X0 k62) (shift_A_ZZ c53 zz53) (shift_A_AA (weakenCutoffA c53 k62) aa32))))) (fun (zz53 : ZZ) =>
      (forall (k62 : Hvl) (c53 : (Cutoff A)) (zz54 : ZZ) ,
        ((shift_A_ZZ (weakenCutoffA c53 k62) (subst_Z_ZZ (weakenTrace X0 k62) zz54 zz53)) =
        (subst_Z_ZZ (weakenTrace X0 k62) (shift_A_ZZ c53 zz54) (shift_A_ZZ (weakenCutoffA c53 k62) zz53))))) (fun (bb32 : BB) =>
      (forall (k62 : Hvl) (c53 : (Cutoff A)) (zz53 : ZZ) ,
        ((shift_A_BB (weakenCutoffA c53 k62) (subst_Z_BB (weakenTrace X0 k62) zz53 bb32)) =
        (subst_Z_BB (weakenTrace X0 k62) (shift_A_ZZ c53 zz53) (shift_A_BB (weakenCutoffA c53 k62) bb32))))) (fun (a62 : (Index A)) =>
      (fun (k62 : Hvl) (c53 : (Cutoff A)) (zz53 : ZZ) =>
        eq_refl)) (fun (zz : ZZ) IHzz21 (k62 : Hvl) (c53 : (Cutoff A)) (zz53 : ZZ) =>
      (f_equal arec (IHzz21 k62 c53 zz53))) (fun (z41 : (Index Z)) =>
      (fun (k62 : Hvl) (c53 : (Cutoff A)) (zz53 : ZZ) =>
        (shift_A_ZZ_subst_Z_Index0_comm_ind c53 zz53 k62 z41))) (fun (aa : AA) IHaa8 (k62 : Hvl) (c53 : (Cutoff A)) (zz53 : ZZ) =>
      (f_equal aabs (eq_trans (f_equal2 shift_A_AA eq_refl (f_equal3 subst_Z_AA (weakenTrace_append Z X0 k62 (HS A H0)) eq_refl eq_refl)) (eq_trans (IHaa8 (HS A k62) c53 zz53) (f_equal3 subst_Z_AA (eq_sym (weakenTrace_append Z X0 k62 (HS A H0))) eq_refl eq_refl))))) (fun (bb : BB) IHbb9 (k62 : Hvl) (c53 : (Cutoff A)) (zz53 : ZZ) =>
      (f_equal babs (eq_trans (f_equal2 shift_A_BB eq_refl (f_equal3 subst_Z_BB (weakenTrace_append Z X0 k62 (HS B H0)) eq_refl eq_refl)) (eq_trans (IHbb9 (HS B k62) c53 zz53) (f_equal3 subst_Z_BB (eq_sym (weakenTrace_append Z X0 k62 (HS B H0))) eq_refl eq_refl))))) (fun (b62 : (Index B)) =>
      (fun (k62 : Hvl) (c53 : (Cutoff A)) (zz53 : ZZ) =>
        eq_refl)) (fun (zz : ZZ) IHzz21 (k62 : Hvl) (c53 : (Cutoff A)) (zz53 : ZZ) =>
      (f_equal brec (IHzz21 k62 c53 zz53)))).
    Lemma shift_A__subst_Z_0_AA_comm_ind  :
      (forall (aa32 : AA) ,
        (forall (k62 : Hvl) (c53 : (Cutoff A)) (zz53 : ZZ) ,
          ((shift_A_AA (weakenCutoffA c53 k62) (subst_Z_AA (weakenTrace X0 k62) zz53 aa32)) =
          (subst_Z_AA (weakenTrace X0 k62) (shift_A_ZZ c53 zz53) (shift_A_AA (weakenCutoffA c53 k62) aa32))))).
    Proof.
      pose proof (shift_A__subst_Z_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_A__subst_Z_0_ZZ_comm_ind  :
      (forall (zz53 : ZZ) ,
        (forall (k62 : Hvl) (c53 : (Cutoff A)) (zz54 : ZZ) ,
          ((shift_A_ZZ (weakenCutoffA c53 k62) (subst_Z_ZZ (weakenTrace X0 k62) zz54 zz53)) =
          (subst_Z_ZZ (weakenTrace X0 k62) (shift_A_ZZ c53 zz54) (shift_A_ZZ (weakenCutoffA c53 k62) zz53))))).
    Proof.
      pose proof (shift_A__subst_Z_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_A__subst_Z_0_BB_comm_ind  :
      (forall (bb32 : BB) ,
        (forall (k62 : Hvl) (c53 : (Cutoff A)) (zz54 : ZZ) ,
          ((shift_A_BB (weakenCutoffA c53 k62) (subst_Z_BB (weakenTrace X0 k62) zz54 bb32)) =
          (subst_Z_BB (weakenTrace X0 k62) (shift_A_ZZ c53 zz54) (shift_A_BB (weakenCutoffA c53 k62) bb32))))).
    Proof.
      pose proof (shift_A__subst_Z_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_B__subst_A_0_AA_ZZ_BB_comm_ind := (ind_AA_ZZ_BB (fun (aa32 : AA) =>
      (forall (k62 : Hvl) (c53 : (Cutoff B)) (aa33 : AA) ,
        ((shift_B_AA (weakenCutoffB c53 k62) (subst_A_AA (weakenTrace X0 k62) aa33 aa32)) =
        (subst_A_AA (weakenTrace X0 k62) (shift_B_AA c53 aa33) (shift_B_AA (weakenCutoffB c53 k62) aa32))))) (fun (zz53 : ZZ) =>
      (forall (k62 : Hvl) (c53 : (Cutoff B)) (aa32 : AA) ,
        ((shift_B_ZZ (weakenCutoffB c53 k62) (subst_A_ZZ (weakenTrace X0 k62) aa32 zz53)) =
        (subst_A_ZZ (weakenTrace X0 k62) (shift_B_AA c53 aa32) (shift_B_ZZ (weakenCutoffB c53 k62) zz53))))) (fun (bb32 : BB) =>
      (forall (k62 : Hvl) (c53 : (Cutoff B)) (aa32 : AA) ,
        ((shift_B_BB (weakenCutoffB c53 k62) (subst_A_BB (weakenTrace X0 k62) aa32 bb32)) =
        (subst_A_BB (weakenTrace X0 k62) (shift_B_AA c53 aa32) (shift_B_BB (weakenCutoffB c53 k62) bb32))))) (fun (a62 : (Index A)) =>
      (fun (k62 : Hvl) (c53 : (Cutoff B)) (aa32 : AA) =>
        (shift_B_AA_subst_A_Index0_comm_ind c53 aa32 k62 a62))) (fun (zz : ZZ) IHzz21 (k62 : Hvl) (c53 : (Cutoff B)) (aa32 : AA) =>
      (f_equal arec (IHzz21 k62 c53 aa32))) (fun (z41 : (Index Z)) =>
      (fun (k62 : Hvl) (c53 : (Cutoff B)) (aa32 : AA) =>
        eq_refl)) (fun (aa : AA) IHaa8 (k62 : Hvl) (c53 : (Cutoff B)) (aa32 : AA) =>
      (f_equal aabs (eq_trans (f_equal2 shift_B_AA eq_refl (f_equal3 subst_A_AA (weakenTrace_append A X0 k62 (HS A H0)) eq_refl eq_refl)) (eq_trans (IHaa8 (HS A k62) c53 aa32) (f_equal3 subst_A_AA (eq_sym (weakenTrace_append A X0 k62 (HS A H0))) eq_refl eq_refl))))) (fun (bb : BB) IHbb9 (k62 : Hvl) (c53 : (Cutoff B)) (aa32 : AA) =>
      (f_equal babs (eq_trans (f_equal2 shift_B_BB eq_refl (f_equal3 subst_A_BB (weakenTrace_append A X0 k62 (HS B H0)) eq_refl eq_refl)) (eq_trans (IHbb9 (HS B k62) c53 aa32) (f_equal3 subst_A_BB (eq_sym (weakenTrace_append A X0 k62 (HS B H0))) eq_refl eq_refl))))) (fun (b62 : (Index B)) =>
      (fun (k62 : Hvl) (c53 : (Cutoff B)) (aa32 : AA) =>
        eq_refl)) (fun (zz : ZZ) IHzz21 (k62 : Hvl) (c53 : (Cutoff B)) (aa32 : AA) =>
      (f_equal brec (IHzz21 k62 c53 aa32)))).
    Lemma shift_B__subst_A_0_AA_comm_ind  :
      (forall (aa32 : AA) ,
        (forall (k62 : Hvl) (c53 : (Cutoff B)) (aa33 : AA) ,
          ((shift_B_AA (weakenCutoffB c53 k62) (subst_A_AA (weakenTrace X0 k62) aa33 aa32)) =
          (subst_A_AA (weakenTrace X0 k62) (shift_B_AA c53 aa33) (shift_B_AA (weakenCutoffB c53 k62) aa32))))).
    Proof.
      pose proof (shift_B__subst_A_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_B__subst_A_0_ZZ_comm_ind  :
      (forall (zz53 : ZZ) ,
        (forall (k62 : Hvl) (c53 : (Cutoff B)) (aa33 : AA) ,
          ((shift_B_ZZ (weakenCutoffB c53 k62) (subst_A_ZZ (weakenTrace X0 k62) aa33 zz53)) =
          (subst_A_ZZ (weakenTrace X0 k62) (shift_B_AA c53 aa33) (shift_B_ZZ (weakenCutoffB c53 k62) zz53))))).
    Proof.
      pose proof (shift_B__subst_A_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_B__subst_A_0_BB_comm_ind  :
      (forall (bb32 : BB) ,
        (forall (k62 : Hvl) (c53 : (Cutoff B)) (aa33 : AA) ,
          ((shift_B_BB (weakenCutoffB c53 k62) (subst_A_BB (weakenTrace X0 k62) aa33 bb32)) =
          (subst_A_BB (weakenTrace X0 k62) (shift_B_AA c53 aa33) (shift_B_BB (weakenCutoffB c53 k62) bb32))))).
    Proof.
      pose proof (shift_B__subst_A_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_B__subst_B_0_AA_ZZ_BB_comm_ind := (ind_AA_ZZ_BB (fun (aa32 : AA) =>
      (forall (k62 : Hvl) (c53 : (Cutoff B)) (bb32 : BB) ,
        ((shift_B_AA (weakenCutoffB c53 k62) (subst_B_AA (weakenTrace X0 k62) bb32 aa32)) =
        (subst_B_AA (weakenTrace X0 k62) (shift_B_BB c53 bb32) (shift_B_AA (weakenCutoffB (CS c53) k62) aa32))))) (fun (zz53 : ZZ) =>
      (forall (k62 : Hvl) (c53 : (Cutoff B)) (bb32 : BB) ,
        ((shift_B_ZZ (weakenCutoffB c53 k62) (subst_B_ZZ (weakenTrace X0 k62) bb32 zz53)) =
        (subst_B_ZZ (weakenTrace X0 k62) (shift_B_BB c53 bb32) (shift_B_ZZ (weakenCutoffB (CS c53) k62) zz53))))) (fun (bb32 : BB) =>
      (forall (k62 : Hvl) (c53 : (Cutoff B)) (bb33 : BB) ,
        ((shift_B_BB (weakenCutoffB c53 k62) (subst_B_BB (weakenTrace X0 k62) bb33 bb32)) =
        (subst_B_BB (weakenTrace X0 k62) (shift_B_BB c53 bb33) (shift_B_BB (weakenCutoffB (CS c53) k62) bb32))))) (fun (a62 : (Index A)) =>
      (fun (k62 : Hvl) (c53 : (Cutoff B)) (bb32 : BB) =>
        eq_refl)) (fun (zz : ZZ) IHzz21 (k62 : Hvl) (c53 : (Cutoff B)) (bb32 : BB) =>
      (f_equal arec (IHzz21 k62 c53 bb32))) (fun (z41 : (Index Z)) =>
      (fun (k62 : Hvl) (c53 : (Cutoff B)) (bb32 : BB) =>
        eq_refl)) (fun (aa : AA) IHaa8 (k62 : Hvl) (c53 : (Cutoff B)) (bb32 : BB) =>
      (f_equal aabs (eq_trans (f_equal2 shift_B_AA eq_refl (f_equal3 subst_B_AA (weakenTrace_append B X0 k62 (HS A H0)) eq_refl eq_refl)) (eq_trans (IHaa8 (HS A k62) c53 bb32) (f_equal3 subst_B_AA (eq_sym (weakenTrace_append B X0 k62 (HS A H0))) eq_refl eq_refl))))) (fun (bb : BB) IHbb9 (k62 : Hvl) (c53 : (Cutoff B)) (bb32 : BB) =>
      (f_equal babs (eq_trans (f_equal2 shift_B_BB eq_refl (f_equal3 subst_B_BB (weakenTrace_append B X0 k62 (HS B H0)) eq_refl eq_refl)) (eq_trans (IHbb9 (HS B k62) c53 bb32) (f_equal3 subst_B_BB (eq_sym (weakenTrace_append B X0 k62 (HS B H0))) eq_refl eq_refl))))) (fun (b62 : (Index B)) =>
      (fun (k62 : Hvl) (c53 : (Cutoff B)) (bb32 : BB) =>
        (shift_B_BB_subst_B_Index0_comm_ind c53 bb32 k62 b62))) (fun (zz : ZZ) IHzz21 (k62 : Hvl) (c53 : (Cutoff B)) (bb32 : BB) =>
      (f_equal brec (IHzz21 k62 c53 bb32)))).
    Lemma shift_B__subst_B_0_AA_comm_ind  :
      (forall (aa32 : AA) ,
        (forall (k62 : Hvl) (c53 : (Cutoff B)) (bb32 : BB) ,
          ((shift_B_AA (weakenCutoffB c53 k62) (subst_B_AA (weakenTrace X0 k62) bb32 aa32)) =
          (subst_B_AA (weakenTrace X0 k62) (shift_B_BB c53 bb32) (shift_B_AA (weakenCutoffB (CS c53) k62) aa32))))).
    Proof.
      pose proof (shift_B__subst_B_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_B__subst_B_0_ZZ_comm_ind  :
      (forall (zz53 : ZZ) ,
        (forall (k62 : Hvl) (c53 : (Cutoff B)) (bb32 : BB) ,
          ((shift_B_ZZ (weakenCutoffB c53 k62) (subst_B_ZZ (weakenTrace X0 k62) bb32 zz53)) =
          (subst_B_ZZ (weakenTrace X0 k62) (shift_B_BB c53 bb32) (shift_B_ZZ (weakenCutoffB (CS c53) k62) zz53))))).
    Proof.
      pose proof (shift_B__subst_B_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_B__subst_B_0_BB_comm_ind  :
      (forall (bb32 : BB) ,
        (forall (k62 : Hvl) (c53 : (Cutoff B)) (bb33 : BB) ,
          ((shift_B_BB (weakenCutoffB c53 k62) (subst_B_BB (weakenTrace X0 k62) bb33 bb32)) =
          (subst_B_BB (weakenTrace X0 k62) (shift_B_BB c53 bb33) (shift_B_BB (weakenCutoffB (CS c53) k62) bb32))))).
    Proof.
      pose proof (shift_B__subst_B_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_B__subst_Z_0_AA_ZZ_BB_comm_ind := (ind_AA_ZZ_BB (fun (aa32 : AA) =>
      (forall (k62 : Hvl) (c53 : (Cutoff B)) (zz53 : ZZ) ,
        ((shift_B_AA (weakenCutoffB c53 k62) (subst_Z_AA (weakenTrace X0 k62) zz53 aa32)) =
        (subst_Z_AA (weakenTrace X0 k62) (shift_B_ZZ c53 zz53) (shift_B_AA (weakenCutoffB c53 k62) aa32))))) (fun (zz53 : ZZ) =>
      (forall (k62 : Hvl) (c53 : (Cutoff B)) (zz54 : ZZ) ,
        ((shift_B_ZZ (weakenCutoffB c53 k62) (subst_Z_ZZ (weakenTrace X0 k62) zz54 zz53)) =
        (subst_Z_ZZ (weakenTrace X0 k62) (shift_B_ZZ c53 zz54) (shift_B_ZZ (weakenCutoffB c53 k62) zz53))))) (fun (bb32 : BB) =>
      (forall (k62 : Hvl) (c53 : (Cutoff B)) (zz53 : ZZ) ,
        ((shift_B_BB (weakenCutoffB c53 k62) (subst_Z_BB (weakenTrace X0 k62) zz53 bb32)) =
        (subst_Z_BB (weakenTrace X0 k62) (shift_B_ZZ c53 zz53) (shift_B_BB (weakenCutoffB c53 k62) bb32))))) (fun (a62 : (Index A)) =>
      (fun (k62 : Hvl) (c53 : (Cutoff B)) (zz53 : ZZ) =>
        eq_refl)) (fun (zz : ZZ) IHzz21 (k62 : Hvl) (c53 : (Cutoff B)) (zz53 : ZZ) =>
      (f_equal arec (IHzz21 k62 c53 zz53))) (fun (z41 : (Index Z)) =>
      (fun (k62 : Hvl) (c53 : (Cutoff B)) (zz53 : ZZ) =>
        (shift_B_ZZ_subst_Z_Index0_comm_ind c53 zz53 k62 z41))) (fun (aa : AA) IHaa8 (k62 : Hvl) (c53 : (Cutoff B)) (zz53 : ZZ) =>
      (f_equal aabs (eq_trans (f_equal2 shift_B_AA eq_refl (f_equal3 subst_Z_AA (weakenTrace_append Z X0 k62 (HS A H0)) eq_refl eq_refl)) (eq_trans (IHaa8 (HS A k62) c53 zz53) (f_equal3 subst_Z_AA (eq_sym (weakenTrace_append Z X0 k62 (HS A H0))) eq_refl eq_refl))))) (fun (bb : BB) IHbb9 (k62 : Hvl) (c53 : (Cutoff B)) (zz53 : ZZ) =>
      (f_equal babs (eq_trans (f_equal2 shift_B_BB eq_refl (f_equal3 subst_Z_BB (weakenTrace_append Z X0 k62 (HS B H0)) eq_refl eq_refl)) (eq_trans (IHbb9 (HS B k62) c53 zz53) (f_equal3 subst_Z_BB (eq_sym (weakenTrace_append Z X0 k62 (HS B H0))) eq_refl eq_refl))))) (fun (b62 : (Index B)) =>
      (fun (k62 : Hvl) (c53 : (Cutoff B)) (zz53 : ZZ) =>
        eq_refl)) (fun (zz : ZZ) IHzz21 (k62 : Hvl) (c53 : (Cutoff B)) (zz53 : ZZ) =>
      (f_equal brec (IHzz21 k62 c53 zz53)))).
    Lemma shift_B__subst_Z_0_AA_comm_ind  :
      (forall (aa32 : AA) ,
        (forall (k62 : Hvl) (c53 : (Cutoff B)) (zz53 : ZZ) ,
          ((shift_B_AA (weakenCutoffB c53 k62) (subst_Z_AA (weakenTrace X0 k62) zz53 aa32)) =
          (subst_Z_AA (weakenTrace X0 k62) (shift_B_ZZ c53 zz53) (shift_B_AA (weakenCutoffB c53 k62) aa32))))).
    Proof.
      pose proof (shift_B__subst_Z_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_B__subst_Z_0_ZZ_comm_ind  :
      (forall (zz53 : ZZ) ,
        (forall (k62 : Hvl) (c53 : (Cutoff B)) (zz54 : ZZ) ,
          ((shift_B_ZZ (weakenCutoffB c53 k62) (subst_Z_ZZ (weakenTrace X0 k62) zz54 zz53)) =
          (subst_Z_ZZ (weakenTrace X0 k62) (shift_B_ZZ c53 zz54) (shift_B_ZZ (weakenCutoffB c53 k62) zz53))))).
    Proof.
      pose proof (shift_B__subst_Z_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_B__subst_Z_0_BB_comm_ind  :
      (forall (bb32 : BB) ,
        (forall (k62 : Hvl) (c53 : (Cutoff B)) (zz54 : ZZ) ,
          ((shift_B_BB (weakenCutoffB c53 k62) (subst_Z_BB (weakenTrace X0 k62) zz54 bb32)) =
          (subst_Z_BB (weakenTrace X0 k62) (shift_B_ZZ c53 zz54) (shift_B_BB (weakenCutoffB c53 k62) bb32))))).
    Proof.
      pose proof (shift_B__subst_Z_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_Z__subst_A_0_AA_ZZ_BB_comm_ind := (ind_AA_ZZ_BB (fun (aa32 : AA) =>
      (forall (k62 : Hvl) (c53 : (Cutoff Z)) (aa33 : AA) ,
        ((shift_Z_AA (weakenCutoffZ c53 k62) (subst_A_AA (weakenTrace X0 k62) aa33 aa32)) =
        (subst_A_AA (weakenTrace X0 k62) (shift_Z_AA c53 aa33) (shift_Z_AA (weakenCutoffZ c53 k62) aa32))))) (fun (zz53 : ZZ) =>
      (forall (k62 : Hvl) (c53 : (Cutoff Z)) (aa32 : AA) ,
        ((shift_Z_ZZ (weakenCutoffZ c53 k62) (subst_A_ZZ (weakenTrace X0 k62) aa32 zz53)) =
        (subst_A_ZZ (weakenTrace X0 k62) (shift_Z_AA c53 aa32) (shift_Z_ZZ (weakenCutoffZ c53 k62) zz53))))) (fun (bb32 : BB) =>
      (forall (k62 : Hvl) (c53 : (Cutoff Z)) (aa32 : AA) ,
        ((shift_Z_BB (weakenCutoffZ c53 k62) (subst_A_BB (weakenTrace X0 k62) aa32 bb32)) =
        (subst_A_BB (weakenTrace X0 k62) (shift_Z_AA c53 aa32) (shift_Z_BB (weakenCutoffZ c53 k62) bb32))))) (fun (a62 : (Index A)) =>
      (fun (k62 : Hvl) (c53 : (Cutoff Z)) (aa32 : AA) =>
        (shift_Z_AA_subst_A_Index0_comm_ind c53 aa32 k62 a62))) (fun (zz : ZZ) IHzz21 (k62 : Hvl) (c53 : (Cutoff Z)) (aa32 : AA) =>
      (f_equal arec (IHzz21 k62 c53 aa32))) (fun (z41 : (Index Z)) =>
      (fun (k62 : Hvl) (c53 : (Cutoff Z)) (aa32 : AA) =>
        eq_refl)) (fun (aa : AA) IHaa8 (k62 : Hvl) (c53 : (Cutoff Z)) (aa32 : AA) =>
      (f_equal aabs (eq_trans (f_equal2 shift_Z_AA eq_refl (f_equal3 subst_A_AA (weakenTrace_append A X0 k62 (HS A H0)) eq_refl eq_refl)) (eq_trans (IHaa8 (HS A k62) c53 aa32) (f_equal3 subst_A_AA (eq_sym (weakenTrace_append A X0 k62 (HS A H0))) eq_refl eq_refl))))) (fun (bb : BB) IHbb9 (k62 : Hvl) (c53 : (Cutoff Z)) (aa32 : AA) =>
      (f_equal babs (eq_trans (f_equal2 shift_Z_BB eq_refl (f_equal3 subst_A_BB (weakenTrace_append A X0 k62 (HS B H0)) eq_refl eq_refl)) (eq_trans (IHbb9 (HS B k62) c53 aa32) (f_equal3 subst_A_BB (eq_sym (weakenTrace_append A X0 k62 (HS B H0))) eq_refl eq_refl))))) (fun (b62 : (Index B)) =>
      (fun (k62 : Hvl) (c53 : (Cutoff Z)) (aa32 : AA) =>
        eq_refl)) (fun (zz : ZZ) IHzz21 (k62 : Hvl) (c53 : (Cutoff Z)) (aa32 : AA) =>
      (f_equal brec (IHzz21 k62 c53 aa32)))).
    Lemma shift_Z__subst_A_0_AA_comm_ind  :
      (forall (aa32 : AA) ,
        (forall (k62 : Hvl) (c53 : (Cutoff Z)) (aa33 : AA) ,
          ((shift_Z_AA (weakenCutoffZ c53 k62) (subst_A_AA (weakenTrace X0 k62) aa33 aa32)) =
          (subst_A_AA (weakenTrace X0 k62) (shift_Z_AA c53 aa33) (shift_Z_AA (weakenCutoffZ c53 k62) aa32))))).
    Proof.
      pose proof (shift_Z__subst_A_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_Z__subst_A_0_ZZ_comm_ind  :
      (forall (zz53 : ZZ) ,
        (forall (k62 : Hvl) (c53 : (Cutoff Z)) (aa33 : AA) ,
          ((shift_Z_ZZ (weakenCutoffZ c53 k62) (subst_A_ZZ (weakenTrace X0 k62) aa33 zz53)) =
          (subst_A_ZZ (weakenTrace X0 k62) (shift_Z_AA c53 aa33) (shift_Z_ZZ (weakenCutoffZ c53 k62) zz53))))).
    Proof.
      pose proof (shift_Z__subst_A_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_Z__subst_A_0_BB_comm_ind  :
      (forall (bb32 : BB) ,
        (forall (k62 : Hvl) (c53 : (Cutoff Z)) (aa33 : AA) ,
          ((shift_Z_BB (weakenCutoffZ c53 k62) (subst_A_BB (weakenTrace X0 k62) aa33 bb32)) =
          (subst_A_BB (weakenTrace X0 k62) (shift_Z_AA c53 aa33) (shift_Z_BB (weakenCutoffZ c53 k62) bb32))))).
    Proof.
      pose proof (shift_Z__subst_A_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_Z__subst_B_0_AA_ZZ_BB_comm_ind := (ind_AA_ZZ_BB (fun (aa32 : AA) =>
      (forall (k62 : Hvl) (c53 : (Cutoff Z)) (bb32 : BB) ,
        ((shift_Z_AA (weakenCutoffZ c53 k62) (subst_B_AA (weakenTrace X0 k62) bb32 aa32)) =
        (subst_B_AA (weakenTrace X0 k62) (shift_Z_BB c53 bb32) (shift_Z_AA (weakenCutoffZ c53 k62) aa32))))) (fun (zz53 : ZZ) =>
      (forall (k62 : Hvl) (c53 : (Cutoff Z)) (bb32 : BB) ,
        ((shift_Z_ZZ (weakenCutoffZ c53 k62) (subst_B_ZZ (weakenTrace X0 k62) bb32 zz53)) =
        (subst_B_ZZ (weakenTrace X0 k62) (shift_Z_BB c53 bb32) (shift_Z_ZZ (weakenCutoffZ c53 k62) zz53))))) (fun (bb32 : BB) =>
      (forall (k62 : Hvl) (c53 : (Cutoff Z)) (bb33 : BB) ,
        ((shift_Z_BB (weakenCutoffZ c53 k62) (subst_B_BB (weakenTrace X0 k62) bb33 bb32)) =
        (subst_B_BB (weakenTrace X0 k62) (shift_Z_BB c53 bb33) (shift_Z_BB (weakenCutoffZ c53 k62) bb32))))) (fun (a62 : (Index A)) =>
      (fun (k62 : Hvl) (c53 : (Cutoff Z)) (bb32 : BB) =>
        eq_refl)) (fun (zz : ZZ) IHzz21 (k62 : Hvl) (c53 : (Cutoff Z)) (bb32 : BB) =>
      (f_equal arec (IHzz21 k62 c53 bb32))) (fun (z41 : (Index Z)) =>
      (fun (k62 : Hvl) (c53 : (Cutoff Z)) (bb32 : BB) =>
        eq_refl)) (fun (aa : AA) IHaa8 (k62 : Hvl) (c53 : (Cutoff Z)) (bb32 : BB) =>
      (f_equal aabs (eq_trans (f_equal2 shift_Z_AA eq_refl (f_equal3 subst_B_AA (weakenTrace_append B X0 k62 (HS A H0)) eq_refl eq_refl)) (eq_trans (IHaa8 (HS A k62) c53 bb32) (f_equal3 subst_B_AA (eq_sym (weakenTrace_append B X0 k62 (HS A H0))) eq_refl eq_refl))))) (fun (bb : BB) IHbb9 (k62 : Hvl) (c53 : (Cutoff Z)) (bb32 : BB) =>
      (f_equal babs (eq_trans (f_equal2 shift_Z_BB eq_refl (f_equal3 subst_B_BB (weakenTrace_append B X0 k62 (HS B H0)) eq_refl eq_refl)) (eq_trans (IHbb9 (HS B k62) c53 bb32) (f_equal3 subst_B_BB (eq_sym (weakenTrace_append B X0 k62 (HS B H0))) eq_refl eq_refl))))) (fun (b62 : (Index B)) =>
      (fun (k62 : Hvl) (c53 : (Cutoff Z)) (bb32 : BB) =>
        (shift_Z_BB_subst_B_Index0_comm_ind c53 bb32 k62 b62))) (fun (zz : ZZ) IHzz21 (k62 : Hvl) (c53 : (Cutoff Z)) (bb32 : BB) =>
      (f_equal brec (IHzz21 k62 c53 bb32)))).
    Lemma shift_Z__subst_B_0_AA_comm_ind  :
      (forall (aa32 : AA) ,
        (forall (k62 : Hvl) (c53 : (Cutoff Z)) (bb32 : BB) ,
          ((shift_Z_AA (weakenCutoffZ c53 k62) (subst_B_AA (weakenTrace X0 k62) bb32 aa32)) =
          (subst_B_AA (weakenTrace X0 k62) (shift_Z_BB c53 bb32) (shift_Z_AA (weakenCutoffZ c53 k62) aa32))))).
    Proof.
      pose proof (shift_Z__subst_B_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_Z__subst_B_0_ZZ_comm_ind  :
      (forall (zz53 : ZZ) ,
        (forall (k62 : Hvl) (c53 : (Cutoff Z)) (bb32 : BB) ,
          ((shift_Z_ZZ (weakenCutoffZ c53 k62) (subst_B_ZZ (weakenTrace X0 k62) bb32 zz53)) =
          (subst_B_ZZ (weakenTrace X0 k62) (shift_Z_BB c53 bb32) (shift_Z_ZZ (weakenCutoffZ c53 k62) zz53))))).
    Proof.
      pose proof (shift_Z__subst_B_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_Z__subst_B_0_BB_comm_ind  :
      (forall (bb32 : BB) ,
        (forall (k62 : Hvl) (c53 : (Cutoff Z)) (bb33 : BB) ,
          ((shift_Z_BB (weakenCutoffZ c53 k62) (subst_B_BB (weakenTrace X0 k62) bb33 bb32)) =
          (subst_B_BB (weakenTrace X0 k62) (shift_Z_BB c53 bb33) (shift_Z_BB (weakenCutoffZ c53 k62) bb32))))).
    Proof.
      pose proof (shift_Z__subst_B_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_Z__subst_Z_0_AA_ZZ_BB_comm_ind := (ind_AA_ZZ_BB (fun (aa32 : AA) =>
      (forall (k62 : Hvl) (c53 : (Cutoff Z)) (zz53 : ZZ) ,
        ((shift_Z_AA (weakenCutoffZ c53 k62) (subst_Z_AA (weakenTrace X0 k62) zz53 aa32)) =
        (subst_Z_AA (weakenTrace X0 k62) (shift_Z_ZZ c53 zz53) (shift_Z_AA (weakenCutoffZ (CS c53) k62) aa32))))) (fun (zz53 : ZZ) =>
      (forall (k62 : Hvl) (c53 : (Cutoff Z)) (zz54 : ZZ) ,
        ((shift_Z_ZZ (weakenCutoffZ c53 k62) (subst_Z_ZZ (weakenTrace X0 k62) zz54 zz53)) =
        (subst_Z_ZZ (weakenTrace X0 k62) (shift_Z_ZZ c53 zz54) (shift_Z_ZZ (weakenCutoffZ (CS c53) k62) zz53))))) (fun (bb32 : BB) =>
      (forall (k62 : Hvl) (c53 : (Cutoff Z)) (zz53 : ZZ) ,
        ((shift_Z_BB (weakenCutoffZ c53 k62) (subst_Z_BB (weakenTrace X0 k62) zz53 bb32)) =
        (subst_Z_BB (weakenTrace X0 k62) (shift_Z_ZZ c53 zz53) (shift_Z_BB (weakenCutoffZ (CS c53) k62) bb32))))) (fun (a62 : (Index A)) =>
      (fun (k62 : Hvl) (c53 : (Cutoff Z)) (zz53 : ZZ) =>
        eq_refl)) (fun (zz : ZZ) IHzz21 (k62 : Hvl) (c53 : (Cutoff Z)) (zz53 : ZZ) =>
      (f_equal arec (IHzz21 k62 c53 zz53))) (fun (z41 : (Index Z)) =>
      (fun (k62 : Hvl) (c53 : (Cutoff Z)) (zz53 : ZZ) =>
        (shift_Z_ZZ_subst_Z_Index0_comm_ind c53 zz53 k62 z41))) (fun (aa : AA) IHaa8 (k62 : Hvl) (c53 : (Cutoff Z)) (zz53 : ZZ) =>
      (f_equal aabs (eq_trans (f_equal2 shift_Z_AA eq_refl (f_equal3 subst_Z_AA (weakenTrace_append Z X0 k62 (HS A H0)) eq_refl eq_refl)) (eq_trans (IHaa8 (HS A k62) c53 zz53) (f_equal3 subst_Z_AA (eq_sym (weakenTrace_append Z X0 k62 (HS A H0))) eq_refl eq_refl))))) (fun (bb : BB) IHbb9 (k62 : Hvl) (c53 : (Cutoff Z)) (zz53 : ZZ) =>
      (f_equal babs (eq_trans (f_equal2 shift_Z_BB eq_refl (f_equal3 subst_Z_BB (weakenTrace_append Z X0 k62 (HS B H0)) eq_refl eq_refl)) (eq_trans (IHbb9 (HS B k62) c53 zz53) (f_equal3 subst_Z_BB (eq_sym (weakenTrace_append Z X0 k62 (HS B H0))) eq_refl eq_refl))))) (fun (b62 : (Index B)) =>
      (fun (k62 : Hvl) (c53 : (Cutoff Z)) (zz53 : ZZ) =>
        eq_refl)) (fun (zz : ZZ) IHzz21 (k62 : Hvl) (c53 : (Cutoff Z)) (zz53 : ZZ) =>
      (f_equal brec (IHzz21 k62 c53 zz53)))).
    Lemma shift_Z__subst_Z_0_AA_comm_ind  :
      (forall (aa32 : AA) ,
        (forall (k62 : Hvl) (c53 : (Cutoff Z)) (zz53 : ZZ) ,
          ((shift_Z_AA (weakenCutoffZ c53 k62) (subst_Z_AA (weakenTrace X0 k62) zz53 aa32)) =
          (subst_Z_AA (weakenTrace X0 k62) (shift_Z_ZZ c53 zz53) (shift_Z_AA (weakenCutoffZ (CS c53) k62) aa32))))).
    Proof.
      pose proof (shift_Z__subst_Z_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_Z__subst_Z_0_ZZ_comm_ind  :
      (forall (zz53 : ZZ) ,
        (forall (k62 : Hvl) (c53 : (Cutoff Z)) (zz54 : ZZ) ,
          ((shift_Z_ZZ (weakenCutoffZ c53 k62) (subst_Z_ZZ (weakenTrace X0 k62) zz54 zz53)) =
          (subst_Z_ZZ (weakenTrace X0 k62) (shift_Z_ZZ c53 zz54) (shift_Z_ZZ (weakenCutoffZ (CS c53) k62) zz53))))).
    Proof.
      pose proof (shift_Z__subst_Z_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_Z__subst_Z_0_BB_comm_ind  :
      (forall (bb32 : BB) ,
        (forall (k62 : Hvl) (c53 : (Cutoff Z)) (zz54 : ZZ) ,
          ((shift_Z_BB (weakenCutoffZ c53 k62) (subst_Z_BB (weakenTrace X0 k62) zz54 bb32)) =
          (subst_Z_BB (weakenTrace X0 k62) (shift_Z_ZZ c53 zz54) (shift_Z_BB (weakenCutoffZ (CS c53) k62) bb32))))).
    Proof.
      pose proof (shift_Z__subst_Z_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
  End ShiftSubstCommInd.
  Section ShiftSubstComm.
    Definition shift_A_AA_subst_A_AA0_comm (aa33 : AA) : (forall (c53 : (Cutoff A)) (aa32 : AA) ,
      ((shift_A_AA c53 (subst_A_AA X0 aa32 aa33)) =
      (subst_A_AA X0 (shift_A_AA c53 aa32) (shift_A_AA (CS c53) aa33)))) := (shift_A__subst_A_0_AA_comm_ind aa33 H0).
    Definition shift_A_ZZ_subst_A_ZZ0_comm (zz53 : ZZ) : (forall (c53 : (Cutoff A)) (aa32 : AA) ,
      ((shift_A_ZZ c53 (subst_A_ZZ X0 aa32 zz53)) =
      (subst_A_ZZ X0 (shift_A_AA c53 aa32) (shift_A_ZZ (CS c53) zz53)))) := (shift_A__subst_A_0_ZZ_comm_ind zz53 H0).
    Definition shift_A_BB_subst_A_BB0_comm (bb32 : BB) : (forall (c53 : (Cutoff A)) (aa32 : AA) ,
      ((shift_A_BB c53 (subst_A_BB X0 aa32 bb32)) =
      (subst_A_BB X0 (shift_A_AA c53 aa32) (shift_A_BB (CS c53) bb32)))) := (shift_A__subst_A_0_BB_comm_ind bb32 H0).
    Definition shift_A_AA_subst_B_AA0_comm (aa32 : AA) : (forall (c53 : (Cutoff A)) (bb32 : BB) ,
      ((shift_A_AA c53 (subst_B_AA X0 bb32 aa32)) =
      (subst_B_AA X0 (shift_A_BB c53 bb32) (shift_A_AA c53 aa32)))) := (shift_A__subst_B_0_AA_comm_ind aa32 H0).
    Definition shift_A_ZZ_subst_B_ZZ0_comm (zz53 : ZZ) : (forall (c53 : (Cutoff A)) (bb32 : BB) ,
      ((shift_A_ZZ c53 (subst_B_ZZ X0 bb32 zz53)) =
      (subst_B_ZZ X0 (shift_A_BB c53 bb32) (shift_A_ZZ c53 zz53)))) := (shift_A__subst_B_0_ZZ_comm_ind zz53 H0).
    Definition shift_A_BB_subst_B_BB0_comm (bb33 : BB) : (forall (c53 : (Cutoff A)) (bb32 : BB) ,
      ((shift_A_BB c53 (subst_B_BB X0 bb32 bb33)) =
      (subst_B_BB X0 (shift_A_BB c53 bb32) (shift_A_BB c53 bb33)))) := (shift_A__subst_B_0_BB_comm_ind bb33 H0).
    Definition shift_A_AA_subst_Z_AA0_comm (aa32 : AA) : (forall (c53 : (Cutoff A)) (zz53 : ZZ) ,
      ((shift_A_AA c53 (subst_Z_AA X0 zz53 aa32)) =
      (subst_Z_AA X0 (shift_A_ZZ c53 zz53) (shift_A_AA c53 aa32)))) := (shift_A__subst_Z_0_AA_comm_ind aa32 H0).
    Definition shift_A_ZZ_subst_Z_ZZ0_comm (zz54 : ZZ) : (forall (c53 : (Cutoff A)) (zz53 : ZZ) ,
      ((shift_A_ZZ c53 (subst_Z_ZZ X0 zz53 zz54)) =
      (subst_Z_ZZ X0 (shift_A_ZZ c53 zz53) (shift_A_ZZ c53 zz54)))) := (shift_A__subst_Z_0_ZZ_comm_ind zz54 H0).
    Definition shift_A_BB_subst_Z_BB0_comm (bb32 : BB) : (forall (c53 : (Cutoff A)) (zz53 : ZZ) ,
      ((shift_A_BB c53 (subst_Z_BB X0 zz53 bb32)) =
      (subst_Z_BB X0 (shift_A_ZZ c53 zz53) (shift_A_BB c53 bb32)))) := (shift_A__subst_Z_0_BB_comm_ind bb32 H0).
    Definition shift_B_AA_subst_A_AA0_comm (aa33 : AA) : (forall (c53 : (Cutoff B)) (aa32 : AA) ,
      ((shift_B_AA c53 (subst_A_AA X0 aa32 aa33)) =
      (subst_A_AA X0 (shift_B_AA c53 aa32) (shift_B_AA c53 aa33)))) := (shift_B__subst_A_0_AA_comm_ind aa33 H0).
    Definition shift_B_ZZ_subst_A_ZZ0_comm (zz53 : ZZ) : (forall (c53 : (Cutoff B)) (aa32 : AA) ,
      ((shift_B_ZZ c53 (subst_A_ZZ X0 aa32 zz53)) =
      (subst_A_ZZ X0 (shift_B_AA c53 aa32) (shift_B_ZZ c53 zz53)))) := (shift_B__subst_A_0_ZZ_comm_ind zz53 H0).
    Definition shift_B_BB_subst_A_BB0_comm (bb32 : BB) : (forall (c53 : (Cutoff B)) (aa32 : AA) ,
      ((shift_B_BB c53 (subst_A_BB X0 aa32 bb32)) =
      (subst_A_BB X0 (shift_B_AA c53 aa32) (shift_B_BB c53 bb32)))) := (shift_B__subst_A_0_BB_comm_ind bb32 H0).
    Definition shift_B_AA_subst_B_AA0_comm (aa32 : AA) : (forall (c53 : (Cutoff B)) (bb32 : BB) ,
      ((shift_B_AA c53 (subst_B_AA X0 bb32 aa32)) =
      (subst_B_AA X0 (shift_B_BB c53 bb32) (shift_B_AA (CS c53) aa32)))) := (shift_B__subst_B_0_AA_comm_ind aa32 H0).
    Definition shift_B_ZZ_subst_B_ZZ0_comm (zz53 : ZZ) : (forall (c53 : (Cutoff B)) (bb32 : BB) ,
      ((shift_B_ZZ c53 (subst_B_ZZ X0 bb32 zz53)) =
      (subst_B_ZZ X0 (shift_B_BB c53 bb32) (shift_B_ZZ (CS c53) zz53)))) := (shift_B__subst_B_0_ZZ_comm_ind zz53 H0).
    Definition shift_B_BB_subst_B_BB0_comm (bb33 : BB) : (forall (c53 : (Cutoff B)) (bb32 : BB) ,
      ((shift_B_BB c53 (subst_B_BB X0 bb32 bb33)) =
      (subst_B_BB X0 (shift_B_BB c53 bb32) (shift_B_BB (CS c53) bb33)))) := (shift_B__subst_B_0_BB_comm_ind bb33 H0).
    Definition shift_B_AA_subst_Z_AA0_comm (aa32 : AA) : (forall (c53 : (Cutoff B)) (zz53 : ZZ) ,
      ((shift_B_AA c53 (subst_Z_AA X0 zz53 aa32)) =
      (subst_Z_AA X0 (shift_B_ZZ c53 zz53) (shift_B_AA c53 aa32)))) := (shift_B__subst_Z_0_AA_comm_ind aa32 H0).
    Definition shift_B_ZZ_subst_Z_ZZ0_comm (zz54 : ZZ) : (forall (c53 : (Cutoff B)) (zz53 : ZZ) ,
      ((shift_B_ZZ c53 (subst_Z_ZZ X0 zz53 zz54)) =
      (subst_Z_ZZ X0 (shift_B_ZZ c53 zz53) (shift_B_ZZ c53 zz54)))) := (shift_B__subst_Z_0_ZZ_comm_ind zz54 H0).
    Definition shift_B_BB_subst_Z_BB0_comm (bb32 : BB) : (forall (c53 : (Cutoff B)) (zz53 : ZZ) ,
      ((shift_B_BB c53 (subst_Z_BB X0 zz53 bb32)) =
      (subst_Z_BB X0 (shift_B_ZZ c53 zz53) (shift_B_BB c53 bb32)))) := (shift_B__subst_Z_0_BB_comm_ind bb32 H0).
    Definition shift_Z_AA_subst_A_AA0_comm (aa33 : AA) : (forall (c53 : (Cutoff Z)) (aa32 : AA) ,
      ((shift_Z_AA c53 (subst_A_AA X0 aa32 aa33)) =
      (subst_A_AA X0 (shift_Z_AA c53 aa32) (shift_Z_AA c53 aa33)))) := (shift_Z__subst_A_0_AA_comm_ind aa33 H0).
    Definition shift_Z_ZZ_subst_A_ZZ0_comm (zz53 : ZZ) : (forall (c53 : (Cutoff Z)) (aa32 : AA) ,
      ((shift_Z_ZZ c53 (subst_A_ZZ X0 aa32 zz53)) =
      (subst_A_ZZ X0 (shift_Z_AA c53 aa32) (shift_Z_ZZ c53 zz53)))) := (shift_Z__subst_A_0_ZZ_comm_ind zz53 H0).
    Definition shift_Z_BB_subst_A_BB0_comm (bb32 : BB) : (forall (c53 : (Cutoff Z)) (aa32 : AA) ,
      ((shift_Z_BB c53 (subst_A_BB X0 aa32 bb32)) =
      (subst_A_BB X0 (shift_Z_AA c53 aa32) (shift_Z_BB c53 bb32)))) := (shift_Z__subst_A_0_BB_comm_ind bb32 H0).
    Definition shift_Z_AA_subst_B_AA0_comm (aa32 : AA) : (forall (c53 : (Cutoff Z)) (bb32 : BB) ,
      ((shift_Z_AA c53 (subst_B_AA X0 bb32 aa32)) =
      (subst_B_AA X0 (shift_Z_BB c53 bb32) (shift_Z_AA c53 aa32)))) := (shift_Z__subst_B_0_AA_comm_ind aa32 H0).
    Definition shift_Z_ZZ_subst_B_ZZ0_comm (zz53 : ZZ) : (forall (c53 : (Cutoff Z)) (bb32 : BB) ,
      ((shift_Z_ZZ c53 (subst_B_ZZ X0 bb32 zz53)) =
      (subst_B_ZZ X0 (shift_Z_BB c53 bb32) (shift_Z_ZZ c53 zz53)))) := (shift_Z__subst_B_0_ZZ_comm_ind zz53 H0).
    Definition shift_Z_BB_subst_B_BB0_comm (bb33 : BB) : (forall (c53 : (Cutoff Z)) (bb32 : BB) ,
      ((shift_Z_BB c53 (subst_B_BB X0 bb32 bb33)) =
      (subst_B_BB X0 (shift_Z_BB c53 bb32) (shift_Z_BB c53 bb33)))) := (shift_Z__subst_B_0_BB_comm_ind bb33 H0).
    Definition shift_Z_AA_subst_Z_AA0_comm (aa32 : AA) : (forall (c53 : (Cutoff Z)) (zz53 : ZZ) ,
      ((shift_Z_AA c53 (subst_Z_AA X0 zz53 aa32)) =
      (subst_Z_AA X0 (shift_Z_ZZ c53 zz53) (shift_Z_AA (CS c53) aa32)))) := (shift_Z__subst_Z_0_AA_comm_ind aa32 H0).
    Definition shift_Z_ZZ_subst_Z_ZZ0_comm (zz54 : ZZ) : (forall (c53 : (Cutoff Z)) (zz53 : ZZ) ,
      ((shift_Z_ZZ c53 (subst_Z_ZZ X0 zz53 zz54)) =
      (subst_Z_ZZ X0 (shift_Z_ZZ c53 zz53) (shift_Z_ZZ (CS c53) zz54)))) := (shift_Z__subst_Z_0_ZZ_comm_ind zz54 H0).
    Definition shift_Z_BB_subst_Z_BB0_comm (bb32 : BB) : (forall (c53 : (Cutoff Z)) (zz53 : ZZ) ,
      ((shift_Z_BB c53 (subst_Z_BB X0 zz53 bb32)) =
      (subst_Z_BB X0 (shift_Z_ZZ c53 zz53) (shift_Z_BB (CS c53) bb32)))) := (shift_Z__subst_Z_0_BB_comm_ind bb32 H0).
  End ShiftSubstComm.
  Section SubstShiftIndexCommInd.
    Lemma subst_A_Index_shift_A_AA0_comm_ind (d : (Trace A)) (aa32 : AA) :
      (forall (k62 : Hvl) (a62 : (Index A)) ,
        ((subst_A_Index (weakenTrace (XS A d) k62) aa32 (shift_A_Index (weakenCutoffA C0 k62) a62)) =
        (shift_A_AA (weakenCutoffA C0 k62) (subst_A_Index (weakenTrace d k62) aa32 a62)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma subst_A_Index_shift_B_AA0_comm_ind (d : (Trace A)) (aa32 : AA) :
      (forall (k62 : Hvl) (a62 : (Index A)) ,
        ((subst_A_Index (weakenTrace (XS B d) k62) aa32 a62) =
        (shift_B_AA (weakenCutoffB C0 k62) (subst_A_Index (weakenTrace d k62) aa32 a62)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma subst_A_Index_shift_Z_AA0_comm_ind (d : (Trace A)) (aa32 : AA) :
      (forall (k62 : Hvl) (a62 : (Index A)) ,
        ((subst_A_Index (weakenTrace (XS Z d) k62) aa32 a62) =
        (shift_Z_AA (weakenCutoffZ C0 k62) (subst_A_Index (weakenTrace d k62) aa32 a62)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma subst_B_Index_shift_A_BB0_comm_ind (d : (Trace B)) (bb32 : BB) :
      (forall (k62 : Hvl) (b62 : (Index B)) ,
        ((subst_B_Index (weakenTrace (XS A d) k62) bb32 b62) =
        (shift_A_BB (weakenCutoffA C0 k62) (subst_B_Index (weakenTrace d k62) bb32 b62)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma subst_B_Index_shift_B_BB0_comm_ind (d : (Trace B)) (bb32 : BB) :
      (forall (k62 : Hvl) (b62 : (Index B)) ,
        ((subst_B_Index (weakenTrace (XS B d) k62) bb32 (shift_B_Index (weakenCutoffB C0 k62) b62)) =
        (shift_B_BB (weakenCutoffB C0 k62) (subst_B_Index (weakenTrace d k62) bb32 b62)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma subst_B_Index_shift_Z_BB0_comm_ind (d : (Trace B)) (bb32 : BB) :
      (forall (k62 : Hvl) (b62 : (Index B)) ,
        ((subst_B_Index (weakenTrace (XS Z d) k62) bb32 b62) =
        (shift_Z_BB (weakenCutoffZ C0 k62) (subst_B_Index (weakenTrace d k62) bb32 b62)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma subst_Z_Index_shift_A_ZZ0_comm_ind (d : (Trace Z)) (zz53 : ZZ) :
      (forall (k62 : Hvl) (z41 : (Index Z)) ,
        ((subst_Z_Index (weakenTrace (XS A d) k62) zz53 z41) =
        (shift_A_ZZ (weakenCutoffA C0 k62) (subst_Z_Index (weakenTrace d k62) zz53 z41)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma subst_Z_Index_shift_B_ZZ0_comm_ind (d : (Trace Z)) (zz53 : ZZ) :
      (forall (k62 : Hvl) (z41 : (Index Z)) ,
        ((subst_Z_Index (weakenTrace (XS B d) k62) zz53 z41) =
        (shift_B_ZZ (weakenCutoffB C0 k62) (subst_Z_Index (weakenTrace d k62) zz53 z41)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma subst_Z_Index_shift_Z_ZZ0_comm_ind (d : (Trace Z)) (zz53 : ZZ) :
      (forall (k62 : Hvl) (z41 : (Index Z)) ,
        ((subst_Z_Index (weakenTrace (XS Z d) k62) zz53 (shift_Z_Index (weakenCutoffZ C0 k62) z41)) =
        (shift_Z_ZZ (weakenCutoffZ C0 k62) (subst_Z_Index (weakenTrace d k62) zz53 z41)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
  End SubstShiftIndexCommInd.
  Section SubstShiftCommInd.
    Definition subst_A__shift_A_0_AA_ZZ_BB_comm_ind := (ind_AA_ZZ_BB (fun (aa62 : AA) =>
      (forall (k125 : Hvl) (d62 : (Trace A)) (aa63 : AA) ,
        ((subst_A_AA (weakenTrace (XS A d62) k125) aa63 (shift_A_AA (weakenCutoffA C0 k125) aa62)) =
        (shift_A_AA (weakenCutoffA C0 k125) (subst_A_AA (weakenTrace d62 k125) aa63 aa62))))) (fun (zz92 : ZZ) =>
      (forall (k125 : Hvl) (d62 : (Trace A)) (aa62 : AA) ,
        ((subst_A_ZZ (weakenTrace (XS A d62) k125) aa62 (shift_A_ZZ (weakenCutoffA C0 k125) zz92)) =
        (shift_A_ZZ (weakenCutoffA C0 k125) (subst_A_ZZ (weakenTrace d62 k125) aa62 zz92))))) (fun (bb62 : BB) =>
      (forall (k125 : Hvl) (d62 : (Trace A)) (aa62 : AA) ,
        ((subst_A_BB (weakenTrace (XS A d62) k125) aa62 (shift_A_BB (weakenCutoffA C0 k125) bb62)) =
        (shift_A_BB (weakenCutoffA C0 k125) (subst_A_BB (weakenTrace d62 k125) aa62 bb62))))) (fun (a89 : (Index A)) =>
      (fun (k125 : Hvl) (d62 : (Trace A)) (aa62 : AA) =>
        (subst_A_Index_shift_A_AA0_comm_ind d62 aa62 k125 a89))) (fun (zz : ZZ) IHzz21 (k125 : Hvl) (d62 : (Trace A)) (aa62 : AA) =>
      (f_equal arec (IHzz21 k125 d62 aa62))) (fun (z59 : (Index Z)) =>
      (fun (k125 : Hvl) (d62 : (Trace A)) (aa62 : AA) =>
        eq_refl)) (fun (aa : AA) IHaa8 (k125 : Hvl) (d62 : (Trace A)) (aa62 : AA) =>
      (f_equal aabs (eq_trans (f_equal3 subst_A_AA (weakenTrace_append A (XS A d62) k125 (HS A H0)) eq_refl eq_refl) (eq_trans (IHaa8 (HS A k125) d62 aa62) (f_equal2 shift_A_AA eq_refl (f_equal3 subst_A_AA (eq_sym (weakenTrace_append A d62 k125 (HS A H0))) eq_refl eq_refl)))))) (fun (bb : BB) IHbb9 (k125 : Hvl) (d62 : (Trace A)) (aa62 : AA) =>
      (f_equal babs (eq_trans (f_equal3 subst_A_BB (weakenTrace_append A (XS A d62) k125 (HS B H0)) eq_refl eq_refl) (eq_trans (IHbb9 (HS B k125) d62 aa62) (f_equal2 shift_A_BB eq_refl (f_equal3 subst_A_BB (eq_sym (weakenTrace_append A d62 k125 (HS B H0))) eq_refl eq_refl)))))) (fun (b89 : (Index B)) =>
      (fun (k125 : Hvl) (d62 : (Trace A)) (aa62 : AA) =>
        eq_refl)) (fun (zz : ZZ) IHzz21 (k125 : Hvl) (d62 : (Trace A)) (aa62 : AA) =>
      (f_equal brec (IHzz21 k125 d62 aa62)))).
    Lemma subst_A__shift_A_0_AA_comm_ind  :
      (forall (aa62 : AA) ,
        (forall (k125 : Hvl) (d62 : (Trace A)) (aa63 : AA) ,
          ((subst_A_AA (weakenTrace (XS A d62) k125) aa63 (shift_A_AA (weakenCutoffA C0 k125) aa62)) =
          (shift_A_AA (weakenCutoffA C0 k125) (subst_A_AA (weakenTrace d62 k125) aa63 aa62))))).
    Proof.
      pose proof (subst_A__shift_A_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_A__shift_A_0_ZZ_comm_ind  :
      (forall (zz92 : ZZ) ,
        (forall (k125 : Hvl) (d62 : (Trace A)) (aa63 : AA) ,
          ((subst_A_ZZ (weakenTrace (XS A d62) k125) aa63 (shift_A_ZZ (weakenCutoffA C0 k125) zz92)) =
          (shift_A_ZZ (weakenCutoffA C0 k125) (subst_A_ZZ (weakenTrace d62 k125) aa63 zz92))))).
    Proof.
      pose proof (subst_A__shift_A_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_A__shift_A_0_BB_comm_ind  :
      (forall (bb62 : BB) ,
        (forall (k125 : Hvl) (d62 : (Trace A)) (aa63 : AA) ,
          ((subst_A_BB (weakenTrace (XS A d62) k125) aa63 (shift_A_BB (weakenCutoffA C0 k125) bb62)) =
          (shift_A_BB (weakenCutoffA C0 k125) (subst_A_BB (weakenTrace d62 k125) aa63 bb62))))).
    Proof.
      pose proof (subst_A__shift_A_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_A__shift_B_0_AA_ZZ_BB_comm_ind := (ind_AA_ZZ_BB (fun (aa62 : AA) =>
      (forall (k125 : Hvl) (d62 : (Trace A)) (aa63 : AA) ,
        ((subst_A_AA (weakenTrace (XS B d62) k125) aa63 (shift_B_AA (weakenCutoffB C0 k125) aa62)) =
        (shift_B_AA (weakenCutoffB C0 k125) (subst_A_AA (weakenTrace d62 k125) aa63 aa62))))) (fun (zz92 : ZZ) =>
      (forall (k125 : Hvl) (d62 : (Trace A)) (aa62 : AA) ,
        ((subst_A_ZZ (weakenTrace (XS B d62) k125) aa62 (shift_B_ZZ (weakenCutoffB C0 k125) zz92)) =
        (shift_B_ZZ (weakenCutoffB C0 k125) (subst_A_ZZ (weakenTrace d62 k125) aa62 zz92))))) (fun (bb62 : BB) =>
      (forall (k125 : Hvl) (d62 : (Trace A)) (aa62 : AA) ,
        ((subst_A_BB (weakenTrace (XS B d62) k125) aa62 (shift_B_BB (weakenCutoffB C0 k125) bb62)) =
        (shift_B_BB (weakenCutoffB C0 k125) (subst_A_BB (weakenTrace d62 k125) aa62 bb62))))) (fun (a89 : (Index A)) =>
      (fun (k125 : Hvl) (d62 : (Trace A)) (aa62 : AA) =>
        (subst_A_Index_shift_B_AA0_comm_ind d62 aa62 k125 a89))) (fun (zz : ZZ) IHzz21 (k125 : Hvl) (d62 : (Trace A)) (aa62 : AA) =>
      (f_equal arec (IHzz21 k125 d62 aa62))) (fun (z59 : (Index Z)) =>
      (fun (k125 : Hvl) (d62 : (Trace A)) (aa62 : AA) =>
        eq_refl)) (fun (aa : AA) IHaa8 (k125 : Hvl) (d62 : (Trace A)) (aa62 : AA) =>
      (f_equal aabs (eq_trans (f_equal3 subst_A_AA (weakenTrace_append A (XS B d62) k125 (HS A H0)) eq_refl eq_refl) (eq_trans (IHaa8 (HS A k125) d62 aa62) (f_equal2 shift_B_AA eq_refl (f_equal3 subst_A_AA (eq_sym (weakenTrace_append A d62 k125 (HS A H0))) eq_refl eq_refl)))))) (fun (bb : BB) IHbb9 (k125 : Hvl) (d62 : (Trace A)) (aa62 : AA) =>
      (f_equal babs (eq_trans (f_equal3 subst_A_BB (weakenTrace_append A (XS B d62) k125 (HS B H0)) eq_refl eq_refl) (eq_trans (IHbb9 (HS B k125) d62 aa62) (f_equal2 shift_B_BB eq_refl (f_equal3 subst_A_BB (eq_sym (weakenTrace_append A d62 k125 (HS B H0))) eq_refl eq_refl)))))) (fun (b89 : (Index B)) =>
      (fun (k125 : Hvl) (d62 : (Trace A)) (aa62 : AA) =>
        eq_refl)) (fun (zz : ZZ) IHzz21 (k125 : Hvl) (d62 : (Trace A)) (aa62 : AA) =>
      (f_equal brec (IHzz21 k125 d62 aa62)))).
    Lemma subst_A__shift_B_0_AA_comm_ind  :
      (forall (aa62 : AA) ,
        (forall (k125 : Hvl) (d62 : (Trace A)) (aa63 : AA) ,
          ((subst_A_AA (weakenTrace (XS B d62) k125) aa63 (shift_B_AA (weakenCutoffB C0 k125) aa62)) =
          (shift_B_AA (weakenCutoffB C0 k125) (subst_A_AA (weakenTrace d62 k125) aa63 aa62))))).
    Proof.
      pose proof (subst_A__shift_B_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_A__shift_B_0_ZZ_comm_ind  :
      (forall (zz92 : ZZ) ,
        (forall (k125 : Hvl) (d62 : (Trace A)) (aa63 : AA) ,
          ((subst_A_ZZ (weakenTrace (XS B d62) k125) aa63 (shift_B_ZZ (weakenCutoffB C0 k125) zz92)) =
          (shift_B_ZZ (weakenCutoffB C0 k125) (subst_A_ZZ (weakenTrace d62 k125) aa63 zz92))))).
    Proof.
      pose proof (subst_A__shift_B_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_A__shift_B_0_BB_comm_ind  :
      (forall (bb62 : BB) ,
        (forall (k125 : Hvl) (d62 : (Trace A)) (aa63 : AA) ,
          ((subst_A_BB (weakenTrace (XS B d62) k125) aa63 (shift_B_BB (weakenCutoffB C0 k125) bb62)) =
          (shift_B_BB (weakenCutoffB C0 k125) (subst_A_BB (weakenTrace d62 k125) aa63 bb62))))).
    Proof.
      pose proof (subst_A__shift_B_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_A__shift_Z_0_AA_ZZ_BB_comm_ind := (ind_AA_ZZ_BB (fun (aa62 : AA) =>
      (forall (k125 : Hvl) (d62 : (Trace A)) (aa63 : AA) ,
        ((subst_A_AA (weakenTrace (XS Z d62) k125) aa63 (shift_Z_AA (weakenCutoffZ C0 k125) aa62)) =
        (shift_Z_AA (weakenCutoffZ C0 k125) (subst_A_AA (weakenTrace d62 k125) aa63 aa62))))) (fun (zz92 : ZZ) =>
      (forall (k125 : Hvl) (d62 : (Trace A)) (aa62 : AA) ,
        ((subst_A_ZZ (weakenTrace (XS Z d62) k125) aa62 (shift_Z_ZZ (weakenCutoffZ C0 k125) zz92)) =
        (shift_Z_ZZ (weakenCutoffZ C0 k125) (subst_A_ZZ (weakenTrace d62 k125) aa62 zz92))))) (fun (bb62 : BB) =>
      (forall (k125 : Hvl) (d62 : (Trace A)) (aa62 : AA) ,
        ((subst_A_BB (weakenTrace (XS Z d62) k125) aa62 (shift_Z_BB (weakenCutoffZ C0 k125) bb62)) =
        (shift_Z_BB (weakenCutoffZ C0 k125) (subst_A_BB (weakenTrace d62 k125) aa62 bb62))))) (fun (a89 : (Index A)) =>
      (fun (k125 : Hvl) (d62 : (Trace A)) (aa62 : AA) =>
        (subst_A_Index_shift_Z_AA0_comm_ind d62 aa62 k125 a89))) (fun (zz : ZZ) IHzz21 (k125 : Hvl) (d62 : (Trace A)) (aa62 : AA) =>
      (f_equal arec (IHzz21 k125 d62 aa62))) (fun (z59 : (Index Z)) =>
      (fun (k125 : Hvl) (d62 : (Trace A)) (aa62 : AA) =>
        eq_refl)) (fun (aa : AA) IHaa8 (k125 : Hvl) (d62 : (Trace A)) (aa62 : AA) =>
      (f_equal aabs (eq_trans (f_equal3 subst_A_AA (weakenTrace_append A (XS Z d62) k125 (HS A H0)) eq_refl eq_refl) (eq_trans (IHaa8 (HS A k125) d62 aa62) (f_equal2 shift_Z_AA eq_refl (f_equal3 subst_A_AA (eq_sym (weakenTrace_append A d62 k125 (HS A H0))) eq_refl eq_refl)))))) (fun (bb : BB) IHbb9 (k125 : Hvl) (d62 : (Trace A)) (aa62 : AA) =>
      (f_equal babs (eq_trans (f_equal3 subst_A_BB (weakenTrace_append A (XS Z d62) k125 (HS B H0)) eq_refl eq_refl) (eq_trans (IHbb9 (HS B k125) d62 aa62) (f_equal2 shift_Z_BB eq_refl (f_equal3 subst_A_BB (eq_sym (weakenTrace_append A d62 k125 (HS B H0))) eq_refl eq_refl)))))) (fun (b89 : (Index B)) =>
      (fun (k125 : Hvl) (d62 : (Trace A)) (aa62 : AA) =>
        eq_refl)) (fun (zz : ZZ) IHzz21 (k125 : Hvl) (d62 : (Trace A)) (aa62 : AA) =>
      (f_equal brec (IHzz21 k125 d62 aa62)))).
    Lemma subst_A__shift_Z_0_AA_comm_ind  :
      (forall (aa62 : AA) ,
        (forall (k125 : Hvl) (d62 : (Trace A)) (aa63 : AA) ,
          ((subst_A_AA (weakenTrace (XS Z d62) k125) aa63 (shift_Z_AA (weakenCutoffZ C0 k125) aa62)) =
          (shift_Z_AA (weakenCutoffZ C0 k125) (subst_A_AA (weakenTrace d62 k125) aa63 aa62))))).
    Proof.
      pose proof (subst_A__shift_Z_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_A__shift_Z_0_ZZ_comm_ind  :
      (forall (zz92 : ZZ) ,
        (forall (k125 : Hvl) (d62 : (Trace A)) (aa63 : AA) ,
          ((subst_A_ZZ (weakenTrace (XS Z d62) k125) aa63 (shift_Z_ZZ (weakenCutoffZ C0 k125) zz92)) =
          (shift_Z_ZZ (weakenCutoffZ C0 k125) (subst_A_ZZ (weakenTrace d62 k125) aa63 zz92))))).
    Proof.
      pose proof (subst_A__shift_Z_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_A__shift_Z_0_BB_comm_ind  :
      (forall (bb62 : BB) ,
        (forall (k125 : Hvl) (d62 : (Trace A)) (aa63 : AA) ,
          ((subst_A_BB (weakenTrace (XS Z d62) k125) aa63 (shift_Z_BB (weakenCutoffZ C0 k125) bb62)) =
          (shift_Z_BB (weakenCutoffZ C0 k125) (subst_A_BB (weakenTrace d62 k125) aa63 bb62))))).
    Proof.
      pose proof (subst_A__shift_Z_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_B__shift_A_0_AA_ZZ_BB_comm_ind := (ind_AA_ZZ_BB (fun (aa62 : AA) =>
      (forall (k125 : Hvl) (d62 : (Trace B)) (bb62 : BB) ,
        ((subst_B_AA (weakenTrace (XS A d62) k125) bb62 (shift_A_AA (weakenCutoffA C0 k125) aa62)) =
        (shift_A_AA (weakenCutoffA C0 k125) (subst_B_AA (weakenTrace d62 k125) bb62 aa62))))) (fun (zz92 : ZZ) =>
      (forall (k125 : Hvl) (d62 : (Trace B)) (bb62 : BB) ,
        ((subst_B_ZZ (weakenTrace (XS A d62) k125) bb62 (shift_A_ZZ (weakenCutoffA C0 k125) zz92)) =
        (shift_A_ZZ (weakenCutoffA C0 k125) (subst_B_ZZ (weakenTrace d62 k125) bb62 zz92))))) (fun (bb62 : BB) =>
      (forall (k125 : Hvl) (d62 : (Trace B)) (bb63 : BB) ,
        ((subst_B_BB (weakenTrace (XS A d62) k125) bb63 (shift_A_BB (weakenCutoffA C0 k125) bb62)) =
        (shift_A_BB (weakenCutoffA C0 k125) (subst_B_BB (weakenTrace d62 k125) bb63 bb62))))) (fun (a89 : (Index A)) =>
      (fun (k125 : Hvl) (d62 : (Trace B)) (bb62 : BB) =>
        eq_refl)) (fun (zz : ZZ) IHzz21 (k125 : Hvl) (d62 : (Trace B)) (bb62 : BB) =>
      (f_equal arec (IHzz21 k125 d62 bb62))) (fun (z59 : (Index Z)) =>
      (fun (k125 : Hvl) (d62 : (Trace B)) (bb62 : BB) =>
        eq_refl)) (fun (aa : AA) IHaa8 (k125 : Hvl) (d62 : (Trace B)) (bb62 : BB) =>
      (f_equal aabs (eq_trans (f_equal3 subst_B_AA (weakenTrace_append B (XS A d62) k125 (HS A H0)) eq_refl eq_refl) (eq_trans (IHaa8 (HS A k125) d62 bb62) (f_equal2 shift_A_AA eq_refl (f_equal3 subst_B_AA (eq_sym (weakenTrace_append B d62 k125 (HS A H0))) eq_refl eq_refl)))))) (fun (bb : BB) IHbb9 (k125 : Hvl) (d62 : (Trace B)) (bb62 : BB) =>
      (f_equal babs (eq_trans (f_equal3 subst_B_BB (weakenTrace_append B (XS A d62) k125 (HS B H0)) eq_refl eq_refl) (eq_trans (IHbb9 (HS B k125) d62 bb62) (f_equal2 shift_A_BB eq_refl (f_equal3 subst_B_BB (eq_sym (weakenTrace_append B d62 k125 (HS B H0))) eq_refl eq_refl)))))) (fun (b89 : (Index B)) =>
      (fun (k125 : Hvl) (d62 : (Trace B)) (bb62 : BB) =>
        (subst_B_Index_shift_A_BB0_comm_ind d62 bb62 k125 b89))) (fun (zz : ZZ) IHzz21 (k125 : Hvl) (d62 : (Trace B)) (bb62 : BB) =>
      (f_equal brec (IHzz21 k125 d62 bb62)))).
    Lemma subst_B__shift_A_0_AA_comm_ind  :
      (forall (aa62 : AA) ,
        (forall (k125 : Hvl) (d62 : (Trace B)) (bb62 : BB) ,
          ((subst_B_AA (weakenTrace (XS A d62) k125) bb62 (shift_A_AA (weakenCutoffA C0 k125) aa62)) =
          (shift_A_AA (weakenCutoffA C0 k125) (subst_B_AA (weakenTrace d62 k125) bb62 aa62))))).
    Proof.
      pose proof (subst_B__shift_A_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_B__shift_A_0_ZZ_comm_ind  :
      (forall (zz92 : ZZ) ,
        (forall (k125 : Hvl) (d62 : (Trace B)) (bb62 : BB) ,
          ((subst_B_ZZ (weakenTrace (XS A d62) k125) bb62 (shift_A_ZZ (weakenCutoffA C0 k125) zz92)) =
          (shift_A_ZZ (weakenCutoffA C0 k125) (subst_B_ZZ (weakenTrace d62 k125) bb62 zz92))))).
    Proof.
      pose proof (subst_B__shift_A_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_B__shift_A_0_BB_comm_ind  :
      (forall (bb62 : BB) ,
        (forall (k125 : Hvl) (d62 : (Trace B)) (bb63 : BB) ,
          ((subst_B_BB (weakenTrace (XS A d62) k125) bb63 (shift_A_BB (weakenCutoffA C0 k125) bb62)) =
          (shift_A_BB (weakenCutoffA C0 k125) (subst_B_BB (weakenTrace d62 k125) bb63 bb62))))).
    Proof.
      pose proof (subst_B__shift_A_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_B__shift_B_0_AA_ZZ_BB_comm_ind := (ind_AA_ZZ_BB (fun (aa62 : AA) =>
      (forall (k125 : Hvl) (d62 : (Trace B)) (bb62 : BB) ,
        ((subst_B_AA (weakenTrace (XS B d62) k125) bb62 (shift_B_AA (weakenCutoffB C0 k125) aa62)) =
        (shift_B_AA (weakenCutoffB C0 k125) (subst_B_AA (weakenTrace d62 k125) bb62 aa62))))) (fun (zz92 : ZZ) =>
      (forall (k125 : Hvl) (d62 : (Trace B)) (bb62 : BB) ,
        ((subst_B_ZZ (weakenTrace (XS B d62) k125) bb62 (shift_B_ZZ (weakenCutoffB C0 k125) zz92)) =
        (shift_B_ZZ (weakenCutoffB C0 k125) (subst_B_ZZ (weakenTrace d62 k125) bb62 zz92))))) (fun (bb62 : BB) =>
      (forall (k125 : Hvl) (d62 : (Trace B)) (bb63 : BB) ,
        ((subst_B_BB (weakenTrace (XS B d62) k125) bb63 (shift_B_BB (weakenCutoffB C0 k125) bb62)) =
        (shift_B_BB (weakenCutoffB C0 k125) (subst_B_BB (weakenTrace d62 k125) bb63 bb62))))) (fun (a89 : (Index A)) =>
      (fun (k125 : Hvl) (d62 : (Trace B)) (bb62 : BB) =>
        eq_refl)) (fun (zz : ZZ) IHzz21 (k125 : Hvl) (d62 : (Trace B)) (bb62 : BB) =>
      (f_equal arec (IHzz21 k125 d62 bb62))) (fun (z59 : (Index Z)) =>
      (fun (k125 : Hvl) (d62 : (Trace B)) (bb62 : BB) =>
        eq_refl)) (fun (aa : AA) IHaa8 (k125 : Hvl) (d62 : (Trace B)) (bb62 : BB) =>
      (f_equal aabs (eq_trans (f_equal3 subst_B_AA (weakenTrace_append B (XS B d62) k125 (HS A H0)) eq_refl eq_refl) (eq_trans (IHaa8 (HS A k125) d62 bb62) (f_equal2 shift_B_AA eq_refl (f_equal3 subst_B_AA (eq_sym (weakenTrace_append B d62 k125 (HS A H0))) eq_refl eq_refl)))))) (fun (bb : BB) IHbb9 (k125 : Hvl) (d62 : (Trace B)) (bb62 : BB) =>
      (f_equal babs (eq_trans (f_equal3 subst_B_BB (weakenTrace_append B (XS B d62) k125 (HS B H0)) eq_refl eq_refl) (eq_trans (IHbb9 (HS B k125) d62 bb62) (f_equal2 shift_B_BB eq_refl (f_equal3 subst_B_BB (eq_sym (weakenTrace_append B d62 k125 (HS B H0))) eq_refl eq_refl)))))) (fun (b89 : (Index B)) =>
      (fun (k125 : Hvl) (d62 : (Trace B)) (bb62 : BB) =>
        (subst_B_Index_shift_B_BB0_comm_ind d62 bb62 k125 b89))) (fun (zz : ZZ) IHzz21 (k125 : Hvl) (d62 : (Trace B)) (bb62 : BB) =>
      (f_equal brec (IHzz21 k125 d62 bb62)))).
    Lemma subst_B__shift_B_0_AA_comm_ind  :
      (forall (aa62 : AA) ,
        (forall (k125 : Hvl) (d62 : (Trace B)) (bb62 : BB) ,
          ((subst_B_AA (weakenTrace (XS B d62) k125) bb62 (shift_B_AA (weakenCutoffB C0 k125) aa62)) =
          (shift_B_AA (weakenCutoffB C0 k125) (subst_B_AA (weakenTrace d62 k125) bb62 aa62))))).
    Proof.
      pose proof (subst_B__shift_B_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_B__shift_B_0_ZZ_comm_ind  :
      (forall (zz92 : ZZ) ,
        (forall (k125 : Hvl) (d62 : (Trace B)) (bb62 : BB) ,
          ((subst_B_ZZ (weakenTrace (XS B d62) k125) bb62 (shift_B_ZZ (weakenCutoffB C0 k125) zz92)) =
          (shift_B_ZZ (weakenCutoffB C0 k125) (subst_B_ZZ (weakenTrace d62 k125) bb62 zz92))))).
    Proof.
      pose proof (subst_B__shift_B_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_B__shift_B_0_BB_comm_ind  :
      (forall (bb62 : BB) ,
        (forall (k125 : Hvl) (d62 : (Trace B)) (bb63 : BB) ,
          ((subst_B_BB (weakenTrace (XS B d62) k125) bb63 (shift_B_BB (weakenCutoffB C0 k125) bb62)) =
          (shift_B_BB (weakenCutoffB C0 k125) (subst_B_BB (weakenTrace d62 k125) bb63 bb62))))).
    Proof.
      pose proof (subst_B__shift_B_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_B__shift_Z_0_AA_ZZ_BB_comm_ind := (ind_AA_ZZ_BB (fun (aa62 : AA) =>
      (forall (k125 : Hvl) (d62 : (Trace B)) (bb62 : BB) ,
        ((subst_B_AA (weakenTrace (XS Z d62) k125) bb62 (shift_Z_AA (weakenCutoffZ C0 k125) aa62)) =
        (shift_Z_AA (weakenCutoffZ C0 k125) (subst_B_AA (weakenTrace d62 k125) bb62 aa62))))) (fun (zz92 : ZZ) =>
      (forall (k125 : Hvl) (d62 : (Trace B)) (bb62 : BB) ,
        ((subst_B_ZZ (weakenTrace (XS Z d62) k125) bb62 (shift_Z_ZZ (weakenCutoffZ C0 k125) zz92)) =
        (shift_Z_ZZ (weakenCutoffZ C0 k125) (subst_B_ZZ (weakenTrace d62 k125) bb62 zz92))))) (fun (bb62 : BB) =>
      (forall (k125 : Hvl) (d62 : (Trace B)) (bb63 : BB) ,
        ((subst_B_BB (weakenTrace (XS Z d62) k125) bb63 (shift_Z_BB (weakenCutoffZ C0 k125) bb62)) =
        (shift_Z_BB (weakenCutoffZ C0 k125) (subst_B_BB (weakenTrace d62 k125) bb63 bb62))))) (fun (a89 : (Index A)) =>
      (fun (k125 : Hvl) (d62 : (Trace B)) (bb62 : BB) =>
        eq_refl)) (fun (zz : ZZ) IHzz21 (k125 : Hvl) (d62 : (Trace B)) (bb62 : BB) =>
      (f_equal arec (IHzz21 k125 d62 bb62))) (fun (z59 : (Index Z)) =>
      (fun (k125 : Hvl) (d62 : (Trace B)) (bb62 : BB) =>
        eq_refl)) (fun (aa : AA) IHaa8 (k125 : Hvl) (d62 : (Trace B)) (bb62 : BB) =>
      (f_equal aabs (eq_trans (f_equal3 subst_B_AA (weakenTrace_append B (XS Z d62) k125 (HS A H0)) eq_refl eq_refl) (eq_trans (IHaa8 (HS A k125) d62 bb62) (f_equal2 shift_Z_AA eq_refl (f_equal3 subst_B_AA (eq_sym (weakenTrace_append B d62 k125 (HS A H0))) eq_refl eq_refl)))))) (fun (bb : BB) IHbb9 (k125 : Hvl) (d62 : (Trace B)) (bb62 : BB) =>
      (f_equal babs (eq_trans (f_equal3 subst_B_BB (weakenTrace_append B (XS Z d62) k125 (HS B H0)) eq_refl eq_refl) (eq_trans (IHbb9 (HS B k125) d62 bb62) (f_equal2 shift_Z_BB eq_refl (f_equal3 subst_B_BB (eq_sym (weakenTrace_append B d62 k125 (HS B H0))) eq_refl eq_refl)))))) (fun (b89 : (Index B)) =>
      (fun (k125 : Hvl) (d62 : (Trace B)) (bb62 : BB) =>
        (subst_B_Index_shift_Z_BB0_comm_ind d62 bb62 k125 b89))) (fun (zz : ZZ) IHzz21 (k125 : Hvl) (d62 : (Trace B)) (bb62 : BB) =>
      (f_equal brec (IHzz21 k125 d62 bb62)))).
    Lemma subst_B__shift_Z_0_AA_comm_ind  :
      (forall (aa62 : AA) ,
        (forall (k125 : Hvl) (d62 : (Trace B)) (bb62 : BB) ,
          ((subst_B_AA (weakenTrace (XS Z d62) k125) bb62 (shift_Z_AA (weakenCutoffZ C0 k125) aa62)) =
          (shift_Z_AA (weakenCutoffZ C0 k125) (subst_B_AA (weakenTrace d62 k125) bb62 aa62))))).
    Proof.
      pose proof (subst_B__shift_Z_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_B__shift_Z_0_ZZ_comm_ind  :
      (forall (zz92 : ZZ) ,
        (forall (k125 : Hvl) (d62 : (Trace B)) (bb62 : BB) ,
          ((subst_B_ZZ (weakenTrace (XS Z d62) k125) bb62 (shift_Z_ZZ (weakenCutoffZ C0 k125) zz92)) =
          (shift_Z_ZZ (weakenCutoffZ C0 k125) (subst_B_ZZ (weakenTrace d62 k125) bb62 zz92))))).
    Proof.
      pose proof (subst_B__shift_Z_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_B__shift_Z_0_BB_comm_ind  :
      (forall (bb62 : BB) ,
        (forall (k125 : Hvl) (d62 : (Trace B)) (bb63 : BB) ,
          ((subst_B_BB (weakenTrace (XS Z d62) k125) bb63 (shift_Z_BB (weakenCutoffZ C0 k125) bb62)) =
          (shift_Z_BB (weakenCutoffZ C0 k125) (subst_B_BB (weakenTrace d62 k125) bb63 bb62))))).
    Proof.
      pose proof (subst_B__shift_Z_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_Z__shift_A_0_AA_ZZ_BB_comm_ind := (ind_AA_ZZ_BB (fun (aa62 : AA) =>
      (forall (k125 : Hvl) (d62 : (Trace Z)) (zz92 : ZZ) ,
        ((subst_Z_AA (weakenTrace (XS A d62) k125) zz92 (shift_A_AA (weakenCutoffA C0 k125) aa62)) =
        (shift_A_AA (weakenCutoffA C0 k125) (subst_Z_AA (weakenTrace d62 k125) zz92 aa62))))) (fun (zz92 : ZZ) =>
      (forall (k125 : Hvl) (d62 : (Trace Z)) (zz93 : ZZ) ,
        ((subst_Z_ZZ (weakenTrace (XS A d62) k125) zz93 (shift_A_ZZ (weakenCutoffA C0 k125) zz92)) =
        (shift_A_ZZ (weakenCutoffA C0 k125) (subst_Z_ZZ (weakenTrace d62 k125) zz93 zz92))))) (fun (bb62 : BB) =>
      (forall (k125 : Hvl) (d62 : (Trace Z)) (zz92 : ZZ) ,
        ((subst_Z_BB (weakenTrace (XS A d62) k125) zz92 (shift_A_BB (weakenCutoffA C0 k125) bb62)) =
        (shift_A_BB (weakenCutoffA C0 k125) (subst_Z_BB (weakenTrace d62 k125) zz92 bb62))))) (fun (a89 : (Index A)) =>
      (fun (k125 : Hvl) (d62 : (Trace Z)) (zz92 : ZZ) =>
        eq_refl)) (fun (zz : ZZ) IHzz21 (k125 : Hvl) (d62 : (Trace Z)) (zz92 : ZZ) =>
      (f_equal arec (IHzz21 k125 d62 zz92))) (fun (z59 : (Index Z)) =>
      (fun (k125 : Hvl) (d62 : (Trace Z)) (zz92 : ZZ) =>
        (subst_Z_Index_shift_A_ZZ0_comm_ind d62 zz92 k125 z59))) (fun (aa : AA) IHaa8 (k125 : Hvl) (d62 : (Trace Z)) (zz92 : ZZ) =>
      (f_equal aabs (eq_trans (f_equal3 subst_Z_AA (weakenTrace_append Z (XS A d62) k125 (HS A H0)) eq_refl eq_refl) (eq_trans (IHaa8 (HS A k125) d62 zz92) (f_equal2 shift_A_AA eq_refl (f_equal3 subst_Z_AA (eq_sym (weakenTrace_append Z d62 k125 (HS A H0))) eq_refl eq_refl)))))) (fun (bb : BB) IHbb9 (k125 : Hvl) (d62 : (Trace Z)) (zz92 : ZZ) =>
      (f_equal babs (eq_trans (f_equal3 subst_Z_BB (weakenTrace_append Z (XS A d62) k125 (HS B H0)) eq_refl eq_refl) (eq_trans (IHbb9 (HS B k125) d62 zz92) (f_equal2 shift_A_BB eq_refl (f_equal3 subst_Z_BB (eq_sym (weakenTrace_append Z d62 k125 (HS B H0))) eq_refl eq_refl)))))) (fun (b89 : (Index B)) =>
      (fun (k125 : Hvl) (d62 : (Trace Z)) (zz92 : ZZ) =>
        eq_refl)) (fun (zz : ZZ) IHzz21 (k125 : Hvl) (d62 : (Trace Z)) (zz92 : ZZ) =>
      (f_equal brec (IHzz21 k125 d62 zz92)))).
    Lemma subst_Z__shift_A_0_AA_comm_ind  :
      (forall (aa62 : AA) ,
        (forall (k125 : Hvl) (d62 : (Trace Z)) (zz92 : ZZ) ,
          ((subst_Z_AA (weakenTrace (XS A d62) k125) zz92 (shift_A_AA (weakenCutoffA C0 k125) aa62)) =
          (shift_A_AA (weakenCutoffA C0 k125) (subst_Z_AA (weakenTrace d62 k125) zz92 aa62))))).
    Proof.
      pose proof (subst_Z__shift_A_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_Z__shift_A_0_ZZ_comm_ind  :
      (forall (zz92 : ZZ) ,
        (forall (k125 : Hvl) (d62 : (Trace Z)) (zz93 : ZZ) ,
          ((subst_Z_ZZ (weakenTrace (XS A d62) k125) zz93 (shift_A_ZZ (weakenCutoffA C0 k125) zz92)) =
          (shift_A_ZZ (weakenCutoffA C0 k125) (subst_Z_ZZ (weakenTrace d62 k125) zz93 zz92))))).
    Proof.
      pose proof (subst_Z__shift_A_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_Z__shift_A_0_BB_comm_ind  :
      (forall (bb62 : BB) ,
        (forall (k125 : Hvl) (d62 : (Trace Z)) (zz93 : ZZ) ,
          ((subst_Z_BB (weakenTrace (XS A d62) k125) zz93 (shift_A_BB (weakenCutoffA C0 k125) bb62)) =
          (shift_A_BB (weakenCutoffA C0 k125) (subst_Z_BB (weakenTrace d62 k125) zz93 bb62))))).
    Proof.
      pose proof (subst_Z__shift_A_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_Z__shift_B_0_AA_ZZ_BB_comm_ind := (ind_AA_ZZ_BB (fun (aa62 : AA) =>
      (forall (k125 : Hvl) (d62 : (Trace Z)) (zz92 : ZZ) ,
        ((subst_Z_AA (weakenTrace (XS B d62) k125) zz92 (shift_B_AA (weakenCutoffB C0 k125) aa62)) =
        (shift_B_AA (weakenCutoffB C0 k125) (subst_Z_AA (weakenTrace d62 k125) zz92 aa62))))) (fun (zz92 : ZZ) =>
      (forall (k125 : Hvl) (d62 : (Trace Z)) (zz93 : ZZ) ,
        ((subst_Z_ZZ (weakenTrace (XS B d62) k125) zz93 (shift_B_ZZ (weakenCutoffB C0 k125) zz92)) =
        (shift_B_ZZ (weakenCutoffB C0 k125) (subst_Z_ZZ (weakenTrace d62 k125) zz93 zz92))))) (fun (bb62 : BB) =>
      (forall (k125 : Hvl) (d62 : (Trace Z)) (zz92 : ZZ) ,
        ((subst_Z_BB (weakenTrace (XS B d62) k125) zz92 (shift_B_BB (weakenCutoffB C0 k125) bb62)) =
        (shift_B_BB (weakenCutoffB C0 k125) (subst_Z_BB (weakenTrace d62 k125) zz92 bb62))))) (fun (a89 : (Index A)) =>
      (fun (k125 : Hvl) (d62 : (Trace Z)) (zz92 : ZZ) =>
        eq_refl)) (fun (zz : ZZ) IHzz21 (k125 : Hvl) (d62 : (Trace Z)) (zz92 : ZZ) =>
      (f_equal arec (IHzz21 k125 d62 zz92))) (fun (z59 : (Index Z)) =>
      (fun (k125 : Hvl) (d62 : (Trace Z)) (zz92 : ZZ) =>
        (subst_Z_Index_shift_B_ZZ0_comm_ind d62 zz92 k125 z59))) (fun (aa : AA) IHaa8 (k125 : Hvl) (d62 : (Trace Z)) (zz92 : ZZ) =>
      (f_equal aabs (eq_trans (f_equal3 subst_Z_AA (weakenTrace_append Z (XS B d62) k125 (HS A H0)) eq_refl eq_refl) (eq_trans (IHaa8 (HS A k125) d62 zz92) (f_equal2 shift_B_AA eq_refl (f_equal3 subst_Z_AA (eq_sym (weakenTrace_append Z d62 k125 (HS A H0))) eq_refl eq_refl)))))) (fun (bb : BB) IHbb9 (k125 : Hvl) (d62 : (Trace Z)) (zz92 : ZZ) =>
      (f_equal babs (eq_trans (f_equal3 subst_Z_BB (weakenTrace_append Z (XS B d62) k125 (HS B H0)) eq_refl eq_refl) (eq_trans (IHbb9 (HS B k125) d62 zz92) (f_equal2 shift_B_BB eq_refl (f_equal3 subst_Z_BB (eq_sym (weakenTrace_append Z d62 k125 (HS B H0))) eq_refl eq_refl)))))) (fun (b89 : (Index B)) =>
      (fun (k125 : Hvl) (d62 : (Trace Z)) (zz92 : ZZ) =>
        eq_refl)) (fun (zz : ZZ) IHzz21 (k125 : Hvl) (d62 : (Trace Z)) (zz92 : ZZ) =>
      (f_equal brec (IHzz21 k125 d62 zz92)))).
    Lemma subst_Z__shift_B_0_AA_comm_ind  :
      (forall (aa62 : AA) ,
        (forall (k125 : Hvl) (d62 : (Trace Z)) (zz92 : ZZ) ,
          ((subst_Z_AA (weakenTrace (XS B d62) k125) zz92 (shift_B_AA (weakenCutoffB C0 k125) aa62)) =
          (shift_B_AA (weakenCutoffB C0 k125) (subst_Z_AA (weakenTrace d62 k125) zz92 aa62))))).
    Proof.
      pose proof (subst_Z__shift_B_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_Z__shift_B_0_ZZ_comm_ind  :
      (forall (zz92 : ZZ) ,
        (forall (k125 : Hvl) (d62 : (Trace Z)) (zz93 : ZZ) ,
          ((subst_Z_ZZ (weakenTrace (XS B d62) k125) zz93 (shift_B_ZZ (weakenCutoffB C0 k125) zz92)) =
          (shift_B_ZZ (weakenCutoffB C0 k125) (subst_Z_ZZ (weakenTrace d62 k125) zz93 zz92))))).
    Proof.
      pose proof (subst_Z__shift_B_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_Z__shift_B_0_BB_comm_ind  :
      (forall (bb62 : BB) ,
        (forall (k125 : Hvl) (d62 : (Trace Z)) (zz93 : ZZ) ,
          ((subst_Z_BB (weakenTrace (XS B d62) k125) zz93 (shift_B_BB (weakenCutoffB C0 k125) bb62)) =
          (shift_B_BB (weakenCutoffB C0 k125) (subst_Z_BB (weakenTrace d62 k125) zz93 bb62))))).
    Proof.
      pose proof (subst_Z__shift_B_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_Z__shift_Z_0_AA_ZZ_BB_comm_ind := (ind_AA_ZZ_BB (fun (aa62 : AA) =>
      (forall (k125 : Hvl) (d62 : (Trace Z)) (zz92 : ZZ) ,
        ((subst_Z_AA (weakenTrace (XS Z d62) k125) zz92 (shift_Z_AA (weakenCutoffZ C0 k125) aa62)) =
        (shift_Z_AA (weakenCutoffZ C0 k125) (subst_Z_AA (weakenTrace d62 k125) zz92 aa62))))) (fun (zz92 : ZZ) =>
      (forall (k125 : Hvl) (d62 : (Trace Z)) (zz93 : ZZ) ,
        ((subst_Z_ZZ (weakenTrace (XS Z d62) k125) zz93 (shift_Z_ZZ (weakenCutoffZ C0 k125) zz92)) =
        (shift_Z_ZZ (weakenCutoffZ C0 k125) (subst_Z_ZZ (weakenTrace d62 k125) zz93 zz92))))) (fun (bb62 : BB) =>
      (forall (k125 : Hvl) (d62 : (Trace Z)) (zz92 : ZZ) ,
        ((subst_Z_BB (weakenTrace (XS Z d62) k125) zz92 (shift_Z_BB (weakenCutoffZ C0 k125) bb62)) =
        (shift_Z_BB (weakenCutoffZ C0 k125) (subst_Z_BB (weakenTrace d62 k125) zz92 bb62))))) (fun (a89 : (Index A)) =>
      (fun (k125 : Hvl) (d62 : (Trace Z)) (zz92 : ZZ) =>
        eq_refl)) (fun (zz : ZZ) IHzz21 (k125 : Hvl) (d62 : (Trace Z)) (zz92 : ZZ) =>
      (f_equal arec (IHzz21 k125 d62 zz92))) (fun (z59 : (Index Z)) =>
      (fun (k125 : Hvl) (d62 : (Trace Z)) (zz92 : ZZ) =>
        (subst_Z_Index_shift_Z_ZZ0_comm_ind d62 zz92 k125 z59))) (fun (aa : AA) IHaa8 (k125 : Hvl) (d62 : (Trace Z)) (zz92 : ZZ) =>
      (f_equal aabs (eq_trans (f_equal3 subst_Z_AA (weakenTrace_append Z (XS Z d62) k125 (HS A H0)) eq_refl eq_refl) (eq_trans (IHaa8 (HS A k125) d62 zz92) (f_equal2 shift_Z_AA eq_refl (f_equal3 subst_Z_AA (eq_sym (weakenTrace_append Z d62 k125 (HS A H0))) eq_refl eq_refl)))))) (fun (bb : BB) IHbb9 (k125 : Hvl) (d62 : (Trace Z)) (zz92 : ZZ) =>
      (f_equal babs (eq_trans (f_equal3 subst_Z_BB (weakenTrace_append Z (XS Z d62) k125 (HS B H0)) eq_refl eq_refl) (eq_trans (IHbb9 (HS B k125) d62 zz92) (f_equal2 shift_Z_BB eq_refl (f_equal3 subst_Z_BB (eq_sym (weakenTrace_append Z d62 k125 (HS B H0))) eq_refl eq_refl)))))) (fun (b89 : (Index B)) =>
      (fun (k125 : Hvl) (d62 : (Trace Z)) (zz92 : ZZ) =>
        eq_refl)) (fun (zz : ZZ) IHzz21 (k125 : Hvl) (d62 : (Trace Z)) (zz92 : ZZ) =>
      (f_equal brec (IHzz21 k125 d62 zz92)))).
    Lemma subst_Z__shift_Z_0_AA_comm_ind  :
      (forall (aa62 : AA) ,
        (forall (k125 : Hvl) (d62 : (Trace Z)) (zz92 : ZZ) ,
          ((subst_Z_AA (weakenTrace (XS Z d62) k125) zz92 (shift_Z_AA (weakenCutoffZ C0 k125) aa62)) =
          (shift_Z_AA (weakenCutoffZ C0 k125) (subst_Z_AA (weakenTrace d62 k125) zz92 aa62))))).
    Proof.
      pose proof (subst_Z__shift_Z_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_Z__shift_Z_0_ZZ_comm_ind  :
      (forall (zz92 : ZZ) ,
        (forall (k125 : Hvl) (d62 : (Trace Z)) (zz93 : ZZ) ,
          ((subst_Z_ZZ (weakenTrace (XS Z d62) k125) zz93 (shift_Z_ZZ (weakenCutoffZ C0 k125) zz92)) =
          (shift_Z_ZZ (weakenCutoffZ C0 k125) (subst_Z_ZZ (weakenTrace d62 k125) zz93 zz92))))).
    Proof.
      pose proof (subst_Z__shift_Z_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_Z__shift_Z_0_BB_comm_ind  :
      (forall (bb62 : BB) ,
        (forall (k125 : Hvl) (d62 : (Trace Z)) (zz93 : ZZ) ,
          ((subst_Z_BB (weakenTrace (XS Z d62) k125) zz93 (shift_Z_BB (weakenCutoffZ C0 k125) bb62)) =
          (shift_Z_BB (weakenCutoffZ C0 k125) (subst_Z_BB (weakenTrace d62 k125) zz93 bb62))))).
    Proof.
      pose proof (subst_Z__shift_Z_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
  End SubstShiftCommInd.
  Section SubstShiftComm.
    Definition subst_A_AA_shift_A_AA0_comm (aa63 : AA) : (forall (d62 : (Trace A)) (aa62 : AA) ,
      ((subst_A_AA (XS A d62) aa62 (shift_A_AA C0 aa63)) =
      (shift_A_AA C0 (subst_A_AA d62 aa62 aa63)))) := (subst_A__shift_A_0_AA_comm_ind aa63 H0).
    Definition subst_A_ZZ_shift_A_ZZ0_comm (zz92 : ZZ) : (forall (d62 : (Trace A)) (aa62 : AA) ,
      ((subst_A_ZZ (XS A d62) aa62 (shift_A_ZZ C0 zz92)) =
      (shift_A_ZZ C0 (subst_A_ZZ d62 aa62 zz92)))) := (subst_A__shift_A_0_ZZ_comm_ind zz92 H0).
    Definition subst_A_BB_shift_A_BB0_comm (bb62 : BB) : (forall (d62 : (Trace A)) (aa62 : AA) ,
      ((subst_A_BB (XS A d62) aa62 (shift_A_BB C0 bb62)) =
      (shift_A_BB C0 (subst_A_BB d62 aa62 bb62)))) := (subst_A__shift_A_0_BB_comm_ind bb62 H0).
    Definition subst_A_AA_shift_B_AA0_comm (aa63 : AA) : (forall (d62 : (Trace A)) (aa62 : AA) ,
      ((subst_A_AA (XS B d62) aa62 (shift_B_AA C0 aa63)) =
      (shift_B_AA C0 (subst_A_AA d62 aa62 aa63)))) := (subst_A__shift_B_0_AA_comm_ind aa63 H0).
    Definition subst_A_ZZ_shift_B_ZZ0_comm (zz92 : ZZ) : (forall (d62 : (Trace A)) (aa62 : AA) ,
      ((subst_A_ZZ (XS B d62) aa62 (shift_B_ZZ C0 zz92)) =
      (shift_B_ZZ C0 (subst_A_ZZ d62 aa62 zz92)))) := (subst_A__shift_B_0_ZZ_comm_ind zz92 H0).
    Definition subst_A_BB_shift_B_BB0_comm (bb62 : BB) : (forall (d62 : (Trace A)) (aa62 : AA) ,
      ((subst_A_BB (XS B d62) aa62 (shift_B_BB C0 bb62)) =
      (shift_B_BB C0 (subst_A_BB d62 aa62 bb62)))) := (subst_A__shift_B_0_BB_comm_ind bb62 H0).
    Definition subst_A_AA_shift_Z_AA0_comm (aa63 : AA) : (forall (d62 : (Trace A)) (aa62 : AA) ,
      ((subst_A_AA (XS Z d62) aa62 (shift_Z_AA C0 aa63)) =
      (shift_Z_AA C0 (subst_A_AA d62 aa62 aa63)))) := (subst_A__shift_Z_0_AA_comm_ind aa63 H0).
    Definition subst_A_ZZ_shift_Z_ZZ0_comm (zz92 : ZZ) : (forall (d62 : (Trace A)) (aa62 : AA) ,
      ((subst_A_ZZ (XS Z d62) aa62 (shift_Z_ZZ C0 zz92)) =
      (shift_Z_ZZ C0 (subst_A_ZZ d62 aa62 zz92)))) := (subst_A__shift_Z_0_ZZ_comm_ind zz92 H0).
    Definition subst_A_BB_shift_Z_BB0_comm (bb62 : BB) : (forall (d62 : (Trace A)) (aa62 : AA) ,
      ((subst_A_BB (XS Z d62) aa62 (shift_Z_BB C0 bb62)) =
      (shift_Z_BB C0 (subst_A_BB d62 aa62 bb62)))) := (subst_A__shift_Z_0_BB_comm_ind bb62 H0).
    Definition subst_B_AA_shift_A_AA0_comm (aa62 : AA) : (forall (d62 : (Trace B)) (bb62 : BB) ,
      ((subst_B_AA (XS A d62) bb62 (shift_A_AA C0 aa62)) =
      (shift_A_AA C0 (subst_B_AA d62 bb62 aa62)))) := (subst_B__shift_A_0_AA_comm_ind aa62 H0).
    Definition subst_B_ZZ_shift_A_ZZ0_comm (zz92 : ZZ) : (forall (d62 : (Trace B)) (bb62 : BB) ,
      ((subst_B_ZZ (XS A d62) bb62 (shift_A_ZZ C0 zz92)) =
      (shift_A_ZZ C0 (subst_B_ZZ d62 bb62 zz92)))) := (subst_B__shift_A_0_ZZ_comm_ind zz92 H0).
    Definition subst_B_BB_shift_A_BB0_comm (bb63 : BB) : (forall (d62 : (Trace B)) (bb62 : BB) ,
      ((subst_B_BB (XS A d62) bb62 (shift_A_BB C0 bb63)) =
      (shift_A_BB C0 (subst_B_BB d62 bb62 bb63)))) := (subst_B__shift_A_0_BB_comm_ind bb63 H0).
    Definition subst_B_AA_shift_B_AA0_comm (aa62 : AA) : (forall (d62 : (Trace B)) (bb62 : BB) ,
      ((subst_B_AA (XS B d62) bb62 (shift_B_AA C0 aa62)) =
      (shift_B_AA C0 (subst_B_AA d62 bb62 aa62)))) := (subst_B__shift_B_0_AA_comm_ind aa62 H0).
    Definition subst_B_ZZ_shift_B_ZZ0_comm (zz92 : ZZ) : (forall (d62 : (Trace B)) (bb62 : BB) ,
      ((subst_B_ZZ (XS B d62) bb62 (shift_B_ZZ C0 zz92)) =
      (shift_B_ZZ C0 (subst_B_ZZ d62 bb62 zz92)))) := (subst_B__shift_B_0_ZZ_comm_ind zz92 H0).
    Definition subst_B_BB_shift_B_BB0_comm (bb63 : BB) : (forall (d62 : (Trace B)) (bb62 : BB) ,
      ((subst_B_BB (XS B d62) bb62 (shift_B_BB C0 bb63)) =
      (shift_B_BB C0 (subst_B_BB d62 bb62 bb63)))) := (subst_B__shift_B_0_BB_comm_ind bb63 H0).
    Definition subst_B_AA_shift_Z_AA0_comm (aa62 : AA) : (forall (d62 : (Trace B)) (bb62 : BB) ,
      ((subst_B_AA (XS Z d62) bb62 (shift_Z_AA C0 aa62)) =
      (shift_Z_AA C0 (subst_B_AA d62 bb62 aa62)))) := (subst_B__shift_Z_0_AA_comm_ind aa62 H0).
    Definition subst_B_ZZ_shift_Z_ZZ0_comm (zz92 : ZZ) : (forall (d62 : (Trace B)) (bb62 : BB) ,
      ((subst_B_ZZ (XS Z d62) bb62 (shift_Z_ZZ C0 zz92)) =
      (shift_Z_ZZ C0 (subst_B_ZZ d62 bb62 zz92)))) := (subst_B__shift_Z_0_ZZ_comm_ind zz92 H0).
    Definition subst_B_BB_shift_Z_BB0_comm (bb63 : BB) : (forall (d62 : (Trace B)) (bb62 : BB) ,
      ((subst_B_BB (XS Z d62) bb62 (shift_Z_BB C0 bb63)) =
      (shift_Z_BB C0 (subst_B_BB d62 bb62 bb63)))) := (subst_B__shift_Z_0_BB_comm_ind bb63 H0).
    Definition subst_Z_AA_shift_A_AA0_comm (aa62 : AA) : (forall (d62 : (Trace Z)) (zz92 : ZZ) ,
      ((subst_Z_AA (XS A d62) zz92 (shift_A_AA C0 aa62)) =
      (shift_A_AA C0 (subst_Z_AA d62 zz92 aa62)))) := (subst_Z__shift_A_0_AA_comm_ind aa62 H0).
    Definition subst_Z_ZZ_shift_A_ZZ0_comm (zz93 : ZZ) : (forall (d62 : (Trace Z)) (zz92 : ZZ) ,
      ((subst_Z_ZZ (XS A d62) zz92 (shift_A_ZZ C0 zz93)) =
      (shift_A_ZZ C0 (subst_Z_ZZ d62 zz92 zz93)))) := (subst_Z__shift_A_0_ZZ_comm_ind zz93 H0).
    Definition subst_Z_BB_shift_A_BB0_comm (bb62 : BB) : (forall (d62 : (Trace Z)) (zz92 : ZZ) ,
      ((subst_Z_BB (XS A d62) zz92 (shift_A_BB C0 bb62)) =
      (shift_A_BB C0 (subst_Z_BB d62 zz92 bb62)))) := (subst_Z__shift_A_0_BB_comm_ind bb62 H0).
    Definition subst_Z_AA_shift_B_AA0_comm (aa62 : AA) : (forall (d62 : (Trace Z)) (zz92 : ZZ) ,
      ((subst_Z_AA (XS B d62) zz92 (shift_B_AA C0 aa62)) =
      (shift_B_AA C0 (subst_Z_AA d62 zz92 aa62)))) := (subst_Z__shift_B_0_AA_comm_ind aa62 H0).
    Definition subst_Z_ZZ_shift_B_ZZ0_comm (zz93 : ZZ) : (forall (d62 : (Trace Z)) (zz92 : ZZ) ,
      ((subst_Z_ZZ (XS B d62) zz92 (shift_B_ZZ C0 zz93)) =
      (shift_B_ZZ C0 (subst_Z_ZZ d62 zz92 zz93)))) := (subst_Z__shift_B_0_ZZ_comm_ind zz93 H0).
    Definition subst_Z_BB_shift_B_BB0_comm (bb62 : BB) : (forall (d62 : (Trace Z)) (zz92 : ZZ) ,
      ((subst_Z_BB (XS B d62) zz92 (shift_B_BB C0 bb62)) =
      (shift_B_BB C0 (subst_Z_BB d62 zz92 bb62)))) := (subst_Z__shift_B_0_BB_comm_ind bb62 H0).
    Definition subst_Z_AA_shift_Z_AA0_comm (aa62 : AA) : (forall (d62 : (Trace Z)) (zz92 : ZZ) ,
      ((subst_Z_AA (XS Z d62) zz92 (shift_Z_AA C0 aa62)) =
      (shift_Z_AA C0 (subst_Z_AA d62 zz92 aa62)))) := (subst_Z__shift_Z_0_AA_comm_ind aa62 H0).
    Definition subst_Z_ZZ_shift_Z_ZZ0_comm (zz93 : ZZ) : (forall (d62 : (Trace Z)) (zz92 : ZZ) ,
      ((subst_Z_ZZ (XS Z d62) zz92 (shift_Z_ZZ C0 zz93)) =
      (shift_Z_ZZ C0 (subst_Z_ZZ d62 zz92 zz93)))) := (subst_Z__shift_Z_0_ZZ_comm_ind zz93 H0).
    Definition subst_Z_BB_shift_Z_BB0_comm (bb62 : BB) : (forall (d62 : (Trace Z)) (zz92 : ZZ) ,
      ((subst_Z_BB (XS Z d62) zz92 (shift_Z_BB C0 bb62)) =
      (shift_Z_BB C0 (subst_Z_BB d62 zz92 bb62)))) := (subst_Z__shift_Z_0_BB_comm_ind bb62 H0).
  End SubstShiftComm.
  Section SubstSubordInd.
    
  End SubstSubordInd.
  Section SubstSubord.
    
  End SubstSubord.
End ShiftSubstInteraction.
 Hint Rewrite subst_A_AA0_shift_A_AA0_reflection subst_B_AA0_shift_B_AA0_reflection subst_Z_AA0_shift_Z_AA0_reflection subst_A_BB0_shift_A_BB0_reflection subst_B_BB0_shift_B_BB0_reflection subst_Z_BB0_shift_Z_BB0_reflection subst_A_ZZ0_shift_A_ZZ0_reflection subst_B_ZZ0_shift_B_ZZ0_reflection subst_Z_ZZ0_shift_Z_ZZ0_reflection : interaction.
 Hint Rewrite subst_A_AA0_shift_A_AA0_reflection subst_B_AA0_shift_B_AA0_reflection subst_Z_AA0_shift_Z_AA0_reflection subst_A_BB0_shift_A_BB0_reflection subst_B_BB0_shift_B_BB0_reflection subst_Z_BB0_shift_Z_BB0_reflection subst_A_ZZ0_shift_A_ZZ0_reflection subst_B_ZZ0_shift_B_ZZ0_reflection subst_Z_ZZ0_shift_Z_ZZ0_reflection : reflection.
 Hint Rewrite subst_A_AA_shift_A_AA0_comm subst_A_AA_shift_B_AA0_comm subst_A_AA_shift_Z_AA0_comm subst_B_AA_shift_A_AA0_comm subst_B_AA_shift_B_AA0_comm subst_B_AA_shift_Z_AA0_comm subst_Z_AA_shift_A_AA0_comm subst_Z_AA_shift_B_AA0_comm subst_Z_AA_shift_Z_AA0_comm subst_A_BB_shift_A_BB0_comm subst_A_BB_shift_B_BB0_comm subst_A_BB_shift_Z_BB0_comm subst_B_BB_shift_A_BB0_comm subst_B_BB_shift_B_BB0_comm subst_B_BB_shift_Z_BB0_comm subst_Z_BB_shift_A_BB0_comm subst_Z_BB_shift_B_BB0_comm subst_Z_BB_shift_Z_BB0_comm subst_A_ZZ_shift_A_ZZ0_comm subst_A_ZZ_shift_B_ZZ0_comm subst_A_ZZ_shift_Z_ZZ0_comm subst_B_ZZ_shift_A_ZZ0_comm subst_B_ZZ_shift_B_ZZ0_comm subst_B_ZZ_shift_Z_ZZ0_comm subst_Z_ZZ_shift_A_ZZ0_comm subst_Z_ZZ_shift_B_ZZ0_comm subst_Z_ZZ_shift_Z_ZZ0_comm : interaction.
 Hint Rewrite subst_A_AA_shift_A_AA0_comm subst_A_AA_shift_B_AA0_comm subst_A_AA_shift_Z_AA0_comm subst_B_AA_shift_A_AA0_comm subst_B_AA_shift_B_AA0_comm subst_B_AA_shift_Z_AA0_comm subst_Z_AA_shift_A_AA0_comm subst_Z_AA_shift_B_AA0_comm subst_Z_AA_shift_Z_AA0_comm subst_A_BB_shift_A_BB0_comm subst_A_BB_shift_B_BB0_comm subst_A_BB_shift_Z_BB0_comm subst_B_BB_shift_A_BB0_comm subst_B_BB_shift_B_BB0_comm subst_B_BB_shift_Z_BB0_comm subst_Z_BB_shift_A_BB0_comm subst_Z_BB_shift_B_BB0_comm subst_Z_BB_shift_Z_BB0_comm subst_A_ZZ_shift_A_ZZ0_comm subst_A_ZZ_shift_B_ZZ0_comm subst_A_ZZ_shift_Z_ZZ0_comm subst_B_ZZ_shift_A_ZZ0_comm subst_B_ZZ_shift_B_ZZ0_comm subst_B_ZZ_shift_Z_ZZ0_comm subst_Z_ZZ_shift_A_ZZ0_comm subst_Z_ZZ_shift_B_ZZ0_comm subst_Z_ZZ_shift_Z_ZZ0_comm : subst_shift0.
 Hint Rewrite shift_A_AA_subst_A_AA0_comm shift_A_AA_subst_B_AA0_comm shift_A_AA_subst_Z_AA0_comm shift_B_AA_subst_A_AA0_comm shift_B_AA_subst_B_AA0_comm shift_B_AA_subst_Z_AA0_comm shift_Z_AA_subst_A_AA0_comm shift_Z_AA_subst_B_AA0_comm shift_Z_AA_subst_Z_AA0_comm shift_A_BB_subst_A_BB0_comm shift_A_BB_subst_B_BB0_comm shift_A_BB_subst_Z_BB0_comm shift_B_BB_subst_A_BB0_comm shift_B_BB_subst_B_BB0_comm shift_B_BB_subst_Z_BB0_comm shift_Z_BB_subst_A_BB0_comm shift_Z_BB_subst_B_BB0_comm shift_Z_BB_subst_Z_BB0_comm shift_A_ZZ_subst_A_ZZ0_comm shift_A_ZZ_subst_B_ZZ0_comm shift_A_ZZ_subst_Z_ZZ0_comm shift_B_ZZ_subst_A_ZZ0_comm shift_B_ZZ_subst_B_ZZ0_comm shift_B_ZZ_subst_Z_ZZ0_comm shift_Z_ZZ_subst_A_ZZ0_comm shift_Z_ZZ_subst_B_ZZ0_comm shift_Z_ZZ_subst_Z_ZZ0_comm : interaction.
 Hint Rewrite shift_A_AA_subst_A_AA0_comm shift_A_AA_subst_B_AA0_comm shift_A_AA_subst_Z_AA0_comm shift_B_AA_subst_A_AA0_comm shift_B_AA_subst_B_AA0_comm shift_B_AA_subst_Z_AA0_comm shift_Z_AA_subst_A_AA0_comm shift_Z_AA_subst_B_AA0_comm shift_Z_AA_subst_Z_AA0_comm shift_A_BB_subst_A_BB0_comm shift_A_BB_subst_B_BB0_comm shift_A_BB_subst_Z_BB0_comm shift_B_BB_subst_A_BB0_comm shift_B_BB_subst_B_BB0_comm shift_B_BB_subst_Z_BB0_comm shift_Z_BB_subst_A_BB0_comm shift_Z_BB_subst_B_BB0_comm shift_Z_BB_subst_Z_BB0_comm shift_A_ZZ_subst_A_ZZ0_comm shift_A_ZZ_subst_B_ZZ0_comm shift_A_ZZ_subst_Z_ZZ0_comm shift_B_ZZ_subst_A_ZZ0_comm shift_B_ZZ_subst_B_ZZ0_comm shift_B_ZZ_subst_Z_ZZ0_comm shift_Z_ZZ_subst_A_ZZ0_comm shift_Z_ZZ_subst_B_ZZ0_comm shift_Z_ZZ_subst_Z_ZZ0_comm : shift_subst0.
Section SubstSubstInteraction.
  Section SubstSubstIndexCommInd.
    Lemma subst_A_AA_subst_A_Index0_commright_ind (d62 : (Trace A)) (aa62 : AA) (aa63 : AA) :
      (forall (k125 : Hvl) (a89 : (Index A)) ,
        ((subst_A_AA (weakenTrace d62 k125) aa62 (subst_A_Index (weakenTrace X0 k125) aa63 a89)) =
        (subst_A_AA (weakenTrace X0 k125) (subst_A_AA d62 aa62 aa63) (subst_A_Index (weakenTrace (XS A d62) k125) aa62 a89)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_B_AA_subst_A_Index0_commright_ind (d62 : (Trace B)) (bb62 : BB) (aa62 : AA) :
      (forall (k125 : Hvl) (a89 : (Index A)) ,
        ((subst_B_AA (weakenTrace d62 k125) bb62 (subst_A_Index (weakenTrace X0 k125) aa62 a89)) =
        (subst_A_Index (weakenTrace X0 k125) (subst_B_AA d62 bb62 aa62) a89))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_Z_AA_subst_A_Index0_commright_ind (d62 : (Trace Z)) (zz92 : ZZ) (aa62 : AA) :
      (forall (k125 : Hvl) (a89 : (Index A)) ,
        ((subst_Z_AA (weakenTrace d62 k125) zz92 (subst_A_Index (weakenTrace X0 k125) aa62 a89)) =
        (subst_A_Index (weakenTrace X0 k125) (subst_Z_AA d62 zz92 aa62) a89))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_A_BB_subst_B_Index0_commright_ind (d62 : (Trace A)) (aa62 : AA) (bb62 : BB) :
      (forall (k125 : Hvl) (b89 : (Index B)) ,
        ((subst_A_BB (weakenTrace d62 k125) aa62 (subst_B_Index (weakenTrace X0 k125) bb62 b89)) =
        (subst_B_Index (weakenTrace X0 k125) (subst_A_BB d62 aa62 bb62) b89))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_B_BB_subst_B_Index0_commright_ind (d62 : (Trace B)) (bb62 : BB) (bb63 : BB) :
      (forall (k125 : Hvl) (b89 : (Index B)) ,
        ((subst_B_BB (weakenTrace d62 k125) bb62 (subst_B_Index (weakenTrace X0 k125) bb63 b89)) =
        (subst_B_BB (weakenTrace X0 k125) (subst_B_BB d62 bb62 bb63) (subst_B_Index (weakenTrace (XS B d62) k125) bb62 b89)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_Z_BB_subst_B_Index0_commright_ind (d62 : (Trace Z)) (zz92 : ZZ) (bb62 : BB) :
      (forall (k125 : Hvl) (b89 : (Index B)) ,
        ((subst_Z_BB (weakenTrace d62 k125) zz92 (subst_B_Index (weakenTrace X0 k125) bb62 b89)) =
        (subst_B_Index (weakenTrace X0 k125) (subst_Z_BB d62 zz92 bb62) b89))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_A_ZZ_subst_Z_Index0_commright_ind (d62 : (Trace A)) (aa62 : AA) (zz92 : ZZ) :
      (forall (k125 : Hvl) (z59 : (Index Z)) ,
        ((subst_A_ZZ (weakenTrace d62 k125) aa62 (subst_Z_Index (weakenTrace X0 k125) zz92 z59)) =
        (subst_Z_Index (weakenTrace X0 k125) (subst_A_ZZ d62 aa62 zz92) z59))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_B_ZZ_subst_Z_Index0_commright_ind (d62 : (Trace B)) (bb62 : BB) (zz92 : ZZ) :
      (forall (k125 : Hvl) (z59 : (Index Z)) ,
        ((subst_B_ZZ (weakenTrace d62 k125) bb62 (subst_Z_Index (weakenTrace X0 k125) zz92 z59)) =
        (subst_Z_Index (weakenTrace X0 k125) (subst_B_ZZ d62 bb62 zz92) z59))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_Z_ZZ_subst_Z_Index0_commright_ind (d62 : (Trace Z)) (zz92 : ZZ) (zz93 : ZZ) :
      (forall (k125 : Hvl) (z59 : (Index Z)) ,
        ((subst_Z_ZZ (weakenTrace d62 k125) zz92 (subst_Z_Index (weakenTrace X0 k125) zz93 z59)) =
        (subst_Z_ZZ (weakenTrace X0 k125) (subst_Z_ZZ d62 zz92 zz93) (subst_Z_Index (weakenTrace (XS Z d62) k125) zz92 z59)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_A_BB_subst_B_Index0_commleft_ind (d62 : (Trace A)) (aa62 : AA) (bb62 : BB) :
      (forall (k125 : Hvl) (a89 : (Index A)) ,
        ((subst_A_Index (weakenTrace d62 k125) aa62 a89) =
        (subst_B_AA (weakenTrace X0 k125) (subst_A_BB d62 aa62 bb62) (subst_A_Index (weakenTrace (XS B d62) k125) aa62 a89)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_A_ZZ_subst_Z_Index0_commleft_ind (d62 : (Trace A)) (aa62 : AA) (zz92 : ZZ) :
      (forall (k125 : Hvl) (a89 : (Index A)) ,
        ((subst_A_Index (weakenTrace d62 k125) aa62 a89) =
        (subst_Z_AA (weakenTrace X0 k125) (subst_A_ZZ d62 aa62 zz92) (subst_A_Index (weakenTrace (XS Z d62) k125) aa62 a89)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_B_AA_subst_A_Index0_commleft_ind (d62 : (Trace B)) (bb62 : BB) (aa62 : AA) :
      (forall (k125 : Hvl) (b89 : (Index B)) ,
        ((subst_B_Index (weakenTrace d62 k125) bb62 b89) =
        (subst_A_BB (weakenTrace X0 k125) (subst_B_AA d62 bb62 aa62) (subst_B_Index (weakenTrace (XS A d62) k125) bb62 b89)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_B_ZZ_subst_Z_Index0_commleft_ind (d62 : (Trace B)) (bb62 : BB) (zz92 : ZZ) :
      (forall (k125 : Hvl) (b89 : (Index B)) ,
        ((subst_B_Index (weakenTrace d62 k125) bb62 b89) =
        (subst_Z_BB (weakenTrace X0 k125) (subst_B_ZZ d62 bb62 zz92) (subst_B_Index (weakenTrace (XS Z d62) k125) bb62 b89)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_Z_AA_subst_A_Index0_commleft_ind (d62 : (Trace Z)) (zz92 : ZZ) (aa62 : AA) :
      (forall (k125 : Hvl) (z59 : (Index Z)) ,
        ((subst_Z_Index (weakenTrace d62 k125) zz92 z59) =
        (subst_A_ZZ (weakenTrace X0 k125) (subst_Z_AA d62 zz92 aa62) (subst_Z_Index (weakenTrace (XS A d62) k125) zz92 z59)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_Z_BB_subst_B_Index0_commleft_ind (d62 : (Trace Z)) (zz92 : ZZ) (bb62 : BB) :
      (forall (k125 : Hvl) (z59 : (Index Z)) ,
        ((subst_Z_Index (weakenTrace d62 k125) zz92 z59) =
        (subst_B_ZZ (weakenTrace X0 k125) (subst_Z_BB d62 zz92 bb62) (subst_Z_Index (weakenTrace (XS B d62) k125) zz92 z59)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
  End SubstSubstIndexCommInd.
  Section SubstSubstCommInd.
    Definition subst_A__subst_A_0_AA_ZZ_BB_comm_ind := (ind_AA_ZZ_BB (fun (aa71 : AA) =>
      (forall (k134 : Hvl) (d71 : (Trace A)) (aa72 : AA) (aa73 : AA) ,
        ((subst_A_AA (weakenTrace d71 k134) aa72 (subst_A_AA (weakenTrace X0 k134) aa73 aa71)) =
        (subst_A_AA (weakenTrace X0 k134) (subst_A_AA d71 aa72 aa73) (subst_A_AA (weakenTrace (XS A d71) k134) aa72 aa71))))) (fun (zz104 : ZZ) =>
      (forall (k134 : Hvl) (d71 : (Trace A)) (aa71 : AA) (aa72 : AA) ,
        ((subst_A_ZZ (weakenTrace d71 k134) aa71 (subst_A_ZZ (weakenTrace X0 k134) aa72 zz104)) =
        (subst_A_ZZ (weakenTrace X0 k134) (subst_A_AA d71 aa71 aa72) (subst_A_ZZ (weakenTrace (XS A d71) k134) aa71 zz104))))) (fun (bb71 : BB) =>
      (forall (k134 : Hvl) (d71 : (Trace A)) (aa71 : AA) (aa72 : AA) ,
        ((subst_A_BB (weakenTrace d71 k134) aa71 (subst_A_BB (weakenTrace X0 k134) aa72 bb71)) =
        (subst_A_BB (weakenTrace X0 k134) (subst_A_AA d71 aa71 aa72) (subst_A_BB (weakenTrace (XS A d71) k134) aa71 bb71))))) (fun (a98 : (Index A)) =>
      (fun (k134 : Hvl) (d71 : (Trace A)) (aa71 : AA) (aa72 : AA) =>
        (subst_A_AA_subst_A_Index0_commright_ind d71 aa71 aa72 k134 a98))) (fun (zz : ZZ) IHzz21 (k134 : Hvl) (d71 : (Trace A)) (aa71 : AA) (aa72 : AA) =>
      (f_equal arec (IHzz21 k134 d71 aa71 aa72))) (fun (z65 : (Index Z)) =>
      (fun (k134 : Hvl) (d71 : (Trace A)) (aa71 : AA) (aa72 : AA) =>
        eq_refl)) (fun (aa : AA) IHaa8 (k134 : Hvl) (d71 : (Trace A)) (aa71 : AA) (aa72 : AA) =>
      (f_equal aabs (eq_trans (f_equal3 subst_A_AA (weakenTrace_append A d71 k134 (HS A H0)) eq_refl (f_equal3 subst_A_AA (weakenTrace_append A X0 k134 (HS A H0)) eq_refl eq_refl)) (eq_trans (IHaa8 (HS A k134) d71 aa71 aa72) (f_equal3 subst_A_AA (eq_sym (weakenTrace_append A X0 k134 (HS A H0))) eq_refl (f_equal3 subst_A_AA (eq_sym (weakenTrace_append A (XS A d71) k134 (HS A H0))) eq_refl eq_refl)))))) (fun (bb : BB) IHbb9 (k134 : Hvl) (d71 : (Trace A)) (aa71 : AA) (aa72 : AA) =>
      (f_equal babs (eq_trans (f_equal3 subst_A_BB (weakenTrace_append A d71 k134 (HS B H0)) eq_refl (f_equal3 subst_A_BB (weakenTrace_append A X0 k134 (HS B H0)) eq_refl eq_refl)) (eq_trans (IHbb9 (HS B k134) d71 aa71 aa72) (f_equal3 subst_A_BB (eq_sym (weakenTrace_append A X0 k134 (HS B H0))) eq_refl (f_equal3 subst_A_BB (eq_sym (weakenTrace_append A (XS A d71) k134 (HS B H0))) eq_refl eq_refl)))))) (fun (b98 : (Index B)) =>
      (fun (k134 : Hvl) (d71 : (Trace A)) (aa71 : AA) (aa72 : AA) =>
        eq_refl)) (fun (zz : ZZ) IHzz21 (k134 : Hvl) (d71 : (Trace A)) (aa71 : AA) (aa72 : AA) =>
      (f_equal brec (IHzz21 k134 d71 aa71 aa72)))).
    Lemma subst_A__subst_A_0_AA_comm_ind  :
      (forall (aa71 : AA) ,
        (forall (k134 : Hvl) (d71 : (Trace A)) (aa72 : AA) (aa73 : AA) ,
          ((subst_A_AA (weakenTrace d71 k134) aa72 (subst_A_AA (weakenTrace X0 k134) aa73 aa71)) =
          (subst_A_AA (weakenTrace X0 k134) (subst_A_AA d71 aa72 aa73) (subst_A_AA (weakenTrace (XS A d71) k134) aa72 aa71))))).
    Proof.
      pose proof (subst_A__subst_A_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_A__subst_A_0_ZZ_comm_ind  :
      (forall (zz104 : ZZ) ,
        (forall (k134 : Hvl) (d71 : (Trace A)) (aa72 : AA) (aa73 : AA) ,
          ((subst_A_ZZ (weakenTrace d71 k134) aa72 (subst_A_ZZ (weakenTrace X0 k134) aa73 zz104)) =
          (subst_A_ZZ (weakenTrace X0 k134) (subst_A_AA d71 aa72 aa73) (subst_A_ZZ (weakenTrace (XS A d71) k134) aa72 zz104))))).
    Proof.
      pose proof (subst_A__subst_A_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_A__subst_A_0_BB_comm_ind  :
      (forall (bb71 : BB) ,
        (forall (k134 : Hvl) (d71 : (Trace A)) (aa72 : AA) (aa73 : AA) ,
          ((subst_A_BB (weakenTrace d71 k134) aa72 (subst_A_BB (weakenTrace X0 k134) aa73 bb71)) =
          (subst_A_BB (weakenTrace X0 k134) (subst_A_AA d71 aa72 aa73) (subst_A_BB (weakenTrace (XS A d71) k134) aa72 bb71))))).
    Proof.
      pose proof (subst_A__subst_A_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_A__subst_B_0_AA_ZZ_BB_comm_ind := (ind_AA_ZZ_BB (fun (aa71 : AA) =>
      (forall (k134 : Hvl) (d71 : (Trace A)) (aa72 : AA) (bb71 : BB) ,
        ((subst_A_AA (weakenTrace d71 k134) aa72 (subst_B_AA (weakenTrace X0 k134) bb71 aa71)) =
        (subst_B_AA (weakenTrace X0 k134) (subst_A_BB d71 aa72 bb71) (subst_A_AA (weakenTrace (XS B d71) k134) aa72 aa71))))) (fun (zz104 : ZZ) =>
      (forall (k134 : Hvl) (d71 : (Trace A)) (aa71 : AA) (bb71 : BB) ,
        ((subst_A_ZZ (weakenTrace d71 k134) aa71 (subst_B_ZZ (weakenTrace X0 k134) bb71 zz104)) =
        (subst_B_ZZ (weakenTrace X0 k134) (subst_A_BB d71 aa71 bb71) (subst_A_ZZ (weakenTrace (XS B d71) k134) aa71 zz104))))) (fun (bb71 : BB) =>
      (forall (k134 : Hvl) (d71 : (Trace A)) (aa71 : AA) (bb72 : BB) ,
        ((subst_A_BB (weakenTrace d71 k134) aa71 (subst_B_BB (weakenTrace X0 k134) bb72 bb71)) =
        (subst_B_BB (weakenTrace X0 k134) (subst_A_BB d71 aa71 bb72) (subst_A_BB (weakenTrace (XS B d71) k134) aa71 bb71))))) (fun (a98 : (Index A)) =>
      (fun (k134 : Hvl) (d71 : (Trace A)) (aa71 : AA) (bb71 : BB) =>
        (subst_A_BB_subst_B_Index0_commleft_ind d71 aa71 bb71 k134 a98))) (fun (zz : ZZ) IHzz21 (k134 : Hvl) (d71 : (Trace A)) (aa71 : AA) (bb71 : BB) =>
      (f_equal arec (IHzz21 k134 d71 aa71 bb71))) (fun (z65 : (Index Z)) =>
      (fun (k134 : Hvl) (d71 : (Trace A)) (aa71 : AA) (bb71 : BB) =>
        eq_refl)) (fun (aa : AA) IHaa8 (k134 : Hvl) (d71 : (Trace A)) (aa71 : AA) (bb71 : BB) =>
      (f_equal aabs (eq_trans (f_equal3 subst_A_AA (weakenTrace_append A d71 k134 (HS A H0)) eq_refl (f_equal3 subst_B_AA (weakenTrace_append B X0 k134 (HS A H0)) eq_refl eq_refl)) (eq_trans (IHaa8 (HS A k134) d71 aa71 bb71) (f_equal3 subst_B_AA (eq_sym (weakenTrace_append B X0 k134 (HS A H0))) eq_refl (f_equal3 subst_A_AA (eq_sym (weakenTrace_append A (XS B d71) k134 (HS A H0))) eq_refl eq_refl)))))) (fun (bb : BB) IHbb9 (k134 : Hvl) (d71 : (Trace A)) (aa71 : AA) (bb71 : BB) =>
      (f_equal babs (eq_trans (f_equal3 subst_A_BB (weakenTrace_append A d71 k134 (HS B H0)) eq_refl (f_equal3 subst_B_BB (weakenTrace_append B X0 k134 (HS B H0)) eq_refl eq_refl)) (eq_trans (IHbb9 (HS B k134) d71 aa71 bb71) (f_equal3 subst_B_BB (eq_sym (weakenTrace_append B X0 k134 (HS B H0))) eq_refl (f_equal3 subst_A_BB (eq_sym (weakenTrace_append A (XS B d71) k134 (HS B H0))) eq_refl eq_refl)))))) (fun (b98 : (Index B)) =>
      (fun (k134 : Hvl) (d71 : (Trace A)) (aa71 : AA) (bb71 : BB) =>
        (subst_A_BB_subst_B_Index0_commright_ind d71 aa71 bb71 k134 b98))) (fun (zz : ZZ) IHzz21 (k134 : Hvl) (d71 : (Trace A)) (aa71 : AA) (bb71 : BB) =>
      (f_equal brec (IHzz21 k134 d71 aa71 bb71)))).
    Lemma subst_A__subst_B_0_AA_comm_ind  :
      (forall (aa71 : AA) ,
        (forall (k134 : Hvl) (d71 : (Trace A)) (aa72 : AA) (bb71 : BB) ,
          ((subst_A_AA (weakenTrace d71 k134) aa72 (subst_B_AA (weakenTrace X0 k134) bb71 aa71)) =
          (subst_B_AA (weakenTrace X0 k134) (subst_A_BB d71 aa72 bb71) (subst_A_AA (weakenTrace (XS B d71) k134) aa72 aa71))))).
    Proof.
      pose proof (subst_A__subst_B_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_A__subst_B_0_ZZ_comm_ind  :
      (forall (zz104 : ZZ) ,
        (forall (k134 : Hvl) (d71 : (Trace A)) (aa72 : AA) (bb71 : BB) ,
          ((subst_A_ZZ (weakenTrace d71 k134) aa72 (subst_B_ZZ (weakenTrace X0 k134) bb71 zz104)) =
          (subst_B_ZZ (weakenTrace X0 k134) (subst_A_BB d71 aa72 bb71) (subst_A_ZZ (weakenTrace (XS B d71) k134) aa72 zz104))))).
    Proof.
      pose proof (subst_A__subst_B_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_A__subst_B_0_BB_comm_ind  :
      (forall (bb71 : BB) ,
        (forall (k134 : Hvl) (d71 : (Trace A)) (aa72 : AA) (bb72 : BB) ,
          ((subst_A_BB (weakenTrace d71 k134) aa72 (subst_B_BB (weakenTrace X0 k134) bb72 bb71)) =
          (subst_B_BB (weakenTrace X0 k134) (subst_A_BB d71 aa72 bb72) (subst_A_BB (weakenTrace (XS B d71) k134) aa72 bb71))))).
    Proof.
      pose proof (subst_A__subst_B_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_A__subst_Z_0_AA_ZZ_BB_comm_ind := (ind_AA_ZZ_BB (fun (aa71 : AA) =>
      (forall (k134 : Hvl) (d71 : (Trace A)) (aa72 : AA) (zz104 : ZZ) ,
        ((subst_A_AA (weakenTrace d71 k134) aa72 (subst_Z_AA (weakenTrace X0 k134) zz104 aa71)) =
        (subst_Z_AA (weakenTrace X0 k134) (subst_A_ZZ d71 aa72 zz104) (subst_A_AA (weakenTrace (XS Z d71) k134) aa72 aa71))))) (fun (zz104 : ZZ) =>
      (forall (k134 : Hvl) (d71 : (Trace A)) (aa71 : AA) (zz105 : ZZ) ,
        ((subst_A_ZZ (weakenTrace d71 k134) aa71 (subst_Z_ZZ (weakenTrace X0 k134) zz105 zz104)) =
        (subst_Z_ZZ (weakenTrace X0 k134) (subst_A_ZZ d71 aa71 zz105) (subst_A_ZZ (weakenTrace (XS Z d71) k134) aa71 zz104))))) (fun (bb71 : BB) =>
      (forall (k134 : Hvl) (d71 : (Trace A)) (aa71 : AA) (zz104 : ZZ) ,
        ((subst_A_BB (weakenTrace d71 k134) aa71 (subst_Z_BB (weakenTrace X0 k134) zz104 bb71)) =
        (subst_Z_BB (weakenTrace X0 k134) (subst_A_ZZ d71 aa71 zz104) (subst_A_BB (weakenTrace (XS Z d71) k134) aa71 bb71))))) (fun (a98 : (Index A)) =>
      (fun (k134 : Hvl) (d71 : (Trace A)) (aa71 : AA) (zz104 : ZZ) =>
        (subst_A_ZZ_subst_Z_Index0_commleft_ind d71 aa71 zz104 k134 a98))) (fun (zz : ZZ) IHzz21 (k134 : Hvl) (d71 : (Trace A)) (aa71 : AA) (zz104 : ZZ) =>
      (f_equal arec (IHzz21 k134 d71 aa71 zz104))) (fun (z65 : (Index Z)) =>
      (fun (k134 : Hvl) (d71 : (Trace A)) (aa71 : AA) (zz104 : ZZ) =>
        (subst_A_ZZ_subst_Z_Index0_commright_ind d71 aa71 zz104 k134 z65))) (fun (aa : AA) IHaa8 (k134 : Hvl) (d71 : (Trace A)) (aa71 : AA) (zz104 : ZZ) =>
      (f_equal aabs (eq_trans (f_equal3 subst_A_AA (weakenTrace_append A d71 k134 (HS A H0)) eq_refl (f_equal3 subst_Z_AA (weakenTrace_append Z X0 k134 (HS A H0)) eq_refl eq_refl)) (eq_trans (IHaa8 (HS A k134) d71 aa71 zz104) (f_equal3 subst_Z_AA (eq_sym (weakenTrace_append Z X0 k134 (HS A H0))) eq_refl (f_equal3 subst_A_AA (eq_sym (weakenTrace_append A (XS Z d71) k134 (HS A H0))) eq_refl eq_refl)))))) (fun (bb : BB) IHbb9 (k134 : Hvl) (d71 : (Trace A)) (aa71 : AA) (zz104 : ZZ) =>
      (f_equal babs (eq_trans (f_equal3 subst_A_BB (weakenTrace_append A d71 k134 (HS B H0)) eq_refl (f_equal3 subst_Z_BB (weakenTrace_append Z X0 k134 (HS B H0)) eq_refl eq_refl)) (eq_trans (IHbb9 (HS B k134) d71 aa71 zz104) (f_equal3 subst_Z_BB (eq_sym (weakenTrace_append Z X0 k134 (HS B H0))) eq_refl (f_equal3 subst_A_BB (eq_sym (weakenTrace_append A (XS Z d71) k134 (HS B H0))) eq_refl eq_refl)))))) (fun (b98 : (Index B)) =>
      (fun (k134 : Hvl) (d71 : (Trace A)) (aa71 : AA) (zz104 : ZZ) =>
        eq_refl)) (fun (zz : ZZ) IHzz21 (k134 : Hvl) (d71 : (Trace A)) (aa71 : AA) (zz104 : ZZ) =>
      (f_equal brec (IHzz21 k134 d71 aa71 zz104)))).
    Lemma subst_A__subst_Z_0_AA_comm_ind  :
      (forall (aa71 : AA) ,
        (forall (k134 : Hvl) (d71 : (Trace A)) (aa72 : AA) (zz104 : ZZ) ,
          ((subst_A_AA (weakenTrace d71 k134) aa72 (subst_Z_AA (weakenTrace X0 k134) zz104 aa71)) =
          (subst_Z_AA (weakenTrace X0 k134) (subst_A_ZZ d71 aa72 zz104) (subst_A_AA (weakenTrace (XS Z d71) k134) aa72 aa71))))).
    Proof.
      pose proof (subst_A__subst_Z_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_A__subst_Z_0_ZZ_comm_ind  :
      (forall (zz104 : ZZ) ,
        (forall (k134 : Hvl) (d71 : (Trace A)) (aa72 : AA) (zz105 : ZZ) ,
          ((subst_A_ZZ (weakenTrace d71 k134) aa72 (subst_Z_ZZ (weakenTrace X0 k134) zz105 zz104)) =
          (subst_Z_ZZ (weakenTrace X0 k134) (subst_A_ZZ d71 aa72 zz105) (subst_A_ZZ (weakenTrace (XS Z d71) k134) aa72 zz104))))).
    Proof.
      pose proof (subst_A__subst_Z_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_A__subst_Z_0_BB_comm_ind  :
      (forall (bb71 : BB) ,
        (forall (k134 : Hvl) (d71 : (Trace A)) (aa72 : AA) (zz105 : ZZ) ,
          ((subst_A_BB (weakenTrace d71 k134) aa72 (subst_Z_BB (weakenTrace X0 k134) zz105 bb71)) =
          (subst_Z_BB (weakenTrace X0 k134) (subst_A_ZZ d71 aa72 zz105) (subst_A_BB (weakenTrace (XS Z d71) k134) aa72 bb71))))).
    Proof.
      pose proof (subst_A__subst_Z_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_B__subst_A_0_AA_ZZ_BB_comm_ind := (ind_AA_ZZ_BB (fun (aa71 : AA) =>
      (forall (k134 : Hvl) (d71 : (Trace B)) (bb71 : BB) (aa72 : AA) ,
        ((subst_B_AA (weakenTrace d71 k134) bb71 (subst_A_AA (weakenTrace X0 k134) aa72 aa71)) =
        (subst_A_AA (weakenTrace X0 k134) (subst_B_AA d71 bb71 aa72) (subst_B_AA (weakenTrace (XS A d71) k134) bb71 aa71))))) (fun (zz104 : ZZ) =>
      (forall (k134 : Hvl) (d71 : (Trace B)) (bb71 : BB) (aa71 : AA) ,
        ((subst_B_ZZ (weakenTrace d71 k134) bb71 (subst_A_ZZ (weakenTrace X0 k134) aa71 zz104)) =
        (subst_A_ZZ (weakenTrace X0 k134) (subst_B_AA d71 bb71 aa71) (subst_B_ZZ (weakenTrace (XS A d71) k134) bb71 zz104))))) (fun (bb71 : BB) =>
      (forall (k134 : Hvl) (d71 : (Trace B)) (bb72 : BB) (aa71 : AA) ,
        ((subst_B_BB (weakenTrace d71 k134) bb72 (subst_A_BB (weakenTrace X0 k134) aa71 bb71)) =
        (subst_A_BB (weakenTrace X0 k134) (subst_B_AA d71 bb72 aa71) (subst_B_BB (weakenTrace (XS A d71) k134) bb72 bb71))))) (fun (a98 : (Index A)) =>
      (fun (k134 : Hvl) (d71 : (Trace B)) (bb71 : BB) (aa71 : AA) =>
        (subst_B_AA_subst_A_Index0_commright_ind d71 bb71 aa71 k134 a98))) (fun (zz : ZZ) IHzz21 (k134 : Hvl) (d71 : (Trace B)) (bb71 : BB) (aa71 : AA) =>
      (f_equal arec (IHzz21 k134 d71 bb71 aa71))) (fun (z65 : (Index Z)) =>
      (fun (k134 : Hvl) (d71 : (Trace B)) (bb71 : BB) (aa71 : AA) =>
        eq_refl)) (fun (aa : AA) IHaa8 (k134 : Hvl) (d71 : (Trace B)) (bb71 : BB) (aa71 : AA) =>
      (f_equal aabs (eq_trans (f_equal3 subst_B_AA (weakenTrace_append B d71 k134 (HS A H0)) eq_refl (f_equal3 subst_A_AA (weakenTrace_append A X0 k134 (HS A H0)) eq_refl eq_refl)) (eq_trans (IHaa8 (HS A k134) d71 bb71 aa71) (f_equal3 subst_A_AA (eq_sym (weakenTrace_append A X0 k134 (HS A H0))) eq_refl (f_equal3 subst_B_AA (eq_sym (weakenTrace_append B (XS A d71) k134 (HS A H0))) eq_refl eq_refl)))))) (fun (bb : BB) IHbb9 (k134 : Hvl) (d71 : (Trace B)) (bb71 : BB) (aa71 : AA) =>
      (f_equal babs (eq_trans (f_equal3 subst_B_BB (weakenTrace_append B d71 k134 (HS B H0)) eq_refl (f_equal3 subst_A_BB (weakenTrace_append A X0 k134 (HS B H0)) eq_refl eq_refl)) (eq_trans (IHbb9 (HS B k134) d71 bb71 aa71) (f_equal3 subst_A_BB (eq_sym (weakenTrace_append A X0 k134 (HS B H0))) eq_refl (f_equal3 subst_B_BB (eq_sym (weakenTrace_append B (XS A d71) k134 (HS B H0))) eq_refl eq_refl)))))) (fun (b98 : (Index B)) =>
      (fun (k134 : Hvl) (d71 : (Trace B)) (bb71 : BB) (aa71 : AA) =>
        (subst_B_AA_subst_A_Index0_commleft_ind d71 bb71 aa71 k134 b98))) (fun (zz : ZZ) IHzz21 (k134 : Hvl) (d71 : (Trace B)) (bb71 : BB) (aa71 : AA) =>
      (f_equal brec (IHzz21 k134 d71 bb71 aa71)))).
    Lemma subst_B__subst_A_0_AA_comm_ind  :
      (forall (aa71 : AA) ,
        (forall (k134 : Hvl) (d71 : (Trace B)) (bb71 : BB) (aa72 : AA) ,
          ((subst_B_AA (weakenTrace d71 k134) bb71 (subst_A_AA (weakenTrace X0 k134) aa72 aa71)) =
          (subst_A_AA (weakenTrace X0 k134) (subst_B_AA d71 bb71 aa72) (subst_B_AA (weakenTrace (XS A d71) k134) bb71 aa71))))).
    Proof.
      pose proof (subst_B__subst_A_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_B__subst_A_0_ZZ_comm_ind  :
      (forall (zz104 : ZZ) ,
        (forall (k134 : Hvl) (d71 : (Trace B)) (bb71 : BB) (aa72 : AA) ,
          ((subst_B_ZZ (weakenTrace d71 k134) bb71 (subst_A_ZZ (weakenTrace X0 k134) aa72 zz104)) =
          (subst_A_ZZ (weakenTrace X0 k134) (subst_B_AA d71 bb71 aa72) (subst_B_ZZ (weakenTrace (XS A d71) k134) bb71 zz104))))).
    Proof.
      pose proof (subst_B__subst_A_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_B__subst_A_0_BB_comm_ind  :
      (forall (bb71 : BB) ,
        (forall (k134 : Hvl) (d71 : (Trace B)) (bb72 : BB) (aa72 : AA) ,
          ((subst_B_BB (weakenTrace d71 k134) bb72 (subst_A_BB (weakenTrace X0 k134) aa72 bb71)) =
          (subst_A_BB (weakenTrace X0 k134) (subst_B_AA d71 bb72 aa72) (subst_B_BB (weakenTrace (XS A d71) k134) bb72 bb71))))).
    Proof.
      pose proof (subst_B__subst_A_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_B__subst_B_0_AA_ZZ_BB_comm_ind := (ind_AA_ZZ_BB (fun (aa71 : AA) =>
      (forall (k134 : Hvl) (d71 : (Trace B)) (bb71 : BB) (bb72 : BB) ,
        ((subst_B_AA (weakenTrace d71 k134) bb71 (subst_B_AA (weakenTrace X0 k134) bb72 aa71)) =
        (subst_B_AA (weakenTrace X0 k134) (subst_B_BB d71 bb71 bb72) (subst_B_AA (weakenTrace (XS B d71) k134) bb71 aa71))))) (fun (zz104 : ZZ) =>
      (forall (k134 : Hvl) (d71 : (Trace B)) (bb71 : BB) (bb72 : BB) ,
        ((subst_B_ZZ (weakenTrace d71 k134) bb71 (subst_B_ZZ (weakenTrace X0 k134) bb72 zz104)) =
        (subst_B_ZZ (weakenTrace X0 k134) (subst_B_BB d71 bb71 bb72) (subst_B_ZZ (weakenTrace (XS B d71) k134) bb71 zz104))))) (fun (bb71 : BB) =>
      (forall (k134 : Hvl) (d71 : (Trace B)) (bb72 : BB) (bb73 : BB) ,
        ((subst_B_BB (weakenTrace d71 k134) bb72 (subst_B_BB (weakenTrace X0 k134) bb73 bb71)) =
        (subst_B_BB (weakenTrace X0 k134) (subst_B_BB d71 bb72 bb73) (subst_B_BB (weakenTrace (XS B d71) k134) bb72 bb71))))) (fun (a98 : (Index A)) =>
      (fun (k134 : Hvl) (d71 : (Trace B)) (bb71 : BB) (bb72 : BB) =>
        eq_refl)) (fun (zz : ZZ) IHzz21 (k134 : Hvl) (d71 : (Trace B)) (bb71 : BB) (bb72 : BB) =>
      (f_equal arec (IHzz21 k134 d71 bb71 bb72))) (fun (z65 : (Index Z)) =>
      (fun (k134 : Hvl) (d71 : (Trace B)) (bb71 : BB) (bb72 : BB) =>
        eq_refl)) (fun (aa : AA) IHaa8 (k134 : Hvl) (d71 : (Trace B)) (bb71 : BB) (bb72 : BB) =>
      (f_equal aabs (eq_trans (f_equal3 subst_B_AA (weakenTrace_append B d71 k134 (HS A H0)) eq_refl (f_equal3 subst_B_AA (weakenTrace_append B X0 k134 (HS A H0)) eq_refl eq_refl)) (eq_trans (IHaa8 (HS A k134) d71 bb71 bb72) (f_equal3 subst_B_AA (eq_sym (weakenTrace_append B X0 k134 (HS A H0))) eq_refl (f_equal3 subst_B_AA (eq_sym (weakenTrace_append B (XS B d71) k134 (HS A H0))) eq_refl eq_refl)))))) (fun (bb : BB) IHbb9 (k134 : Hvl) (d71 : (Trace B)) (bb71 : BB) (bb72 : BB) =>
      (f_equal babs (eq_trans (f_equal3 subst_B_BB (weakenTrace_append B d71 k134 (HS B H0)) eq_refl (f_equal3 subst_B_BB (weakenTrace_append B X0 k134 (HS B H0)) eq_refl eq_refl)) (eq_trans (IHbb9 (HS B k134) d71 bb71 bb72) (f_equal3 subst_B_BB (eq_sym (weakenTrace_append B X0 k134 (HS B H0))) eq_refl (f_equal3 subst_B_BB (eq_sym (weakenTrace_append B (XS B d71) k134 (HS B H0))) eq_refl eq_refl)))))) (fun (b98 : (Index B)) =>
      (fun (k134 : Hvl) (d71 : (Trace B)) (bb71 : BB) (bb72 : BB) =>
        (subst_B_BB_subst_B_Index0_commright_ind d71 bb71 bb72 k134 b98))) (fun (zz : ZZ) IHzz21 (k134 : Hvl) (d71 : (Trace B)) (bb71 : BB) (bb72 : BB) =>
      (f_equal brec (IHzz21 k134 d71 bb71 bb72)))).
    Lemma subst_B__subst_B_0_AA_comm_ind  :
      (forall (aa71 : AA) ,
        (forall (k134 : Hvl) (d71 : (Trace B)) (bb71 : BB) (bb72 : BB) ,
          ((subst_B_AA (weakenTrace d71 k134) bb71 (subst_B_AA (weakenTrace X0 k134) bb72 aa71)) =
          (subst_B_AA (weakenTrace X0 k134) (subst_B_BB d71 bb71 bb72) (subst_B_AA (weakenTrace (XS B d71) k134) bb71 aa71))))).
    Proof.
      pose proof (subst_B__subst_B_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_B__subst_B_0_ZZ_comm_ind  :
      (forall (zz104 : ZZ) ,
        (forall (k134 : Hvl) (d71 : (Trace B)) (bb71 : BB) (bb72 : BB) ,
          ((subst_B_ZZ (weakenTrace d71 k134) bb71 (subst_B_ZZ (weakenTrace X0 k134) bb72 zz104)) =
          (subst_B_ZZ (weakenTrace X0 k134) (subst_B_BB d71 bb71 bb72) (subst_B_ZZ (weakenTrace (XS B d71) k134) bb71 zz104))))).
    Proof.
      pose proof (subst_B__subst_B_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_B__subst_B_0_BB_comm_ind  :
      (forall (bb71 : BB) ,
        (forall (k134 : Hvl) (d71 : (Trace B)) (bb72 : BB) (bb73 : BB) ,
          ((subst_B_BB (weakenTrace d71 k134) bb72 (subst_B_BB (weakenTrace X0 k134) bb73 bb71)) =
          (subst_B_BB (weakenTrace X0 k134) (subst_B_BB d71 bb72 bb73) (subst_B_BB (weakenTrace (XS B d71) k134) bb72 bb71))))).
    Proof.
      pose proof (subst_B__subst_B_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_B__subst_Z_0_AA_ZZ_BB_comm_ind := (ind_AA_ZZ_BB (fun (aa71 : AA) =>
      (forall (k134 : Hvl) (d71 : (Trace B)) (bb71 : BB) (zz104 : ZZ) ,
        ((subst_B_AA (weakenTrace d71 k134) bb71 (subst_Z_AA (weakenTrace X0 k134) zz104 aa71)) =
        (subst_Z_AA (weakenTrace X0 k134) (subst_B_ZZ d71 bb71 zz104) (subst_B_AA (weakenTrace (XS Z d71) k134) bb71 aa71))))) (fun (zz104 : ZZ) =>
      (forall (k134 : Hvl) (d71 : (Trace B)) (bb71 : BB) (zz105 : ZZ) ,
        ((subst_B_ZZ (weakenTrace d71 k134) bb71 (subst_Z_ZZ (weakenTrace X0 k134) zz105 zz104)) =
        (subst_Z_ZZ (weakenTrace X0 k134) (subst_B_ZZ d71 bb71 zz105) (subst_B_ZZ (weakenTrace (XS Z d71) k134) bb71 zz104))))) (fun (bb71 : BB) =>
      (forall (k134 : Hvl) (d71 : (Trace B)) (bb72 : BB) (zz104 : ZZ) ,
        ((subst_B_BB (weakenTrace d71 k134) bb72 (subst_Z_BB (weakenTrace X0 k134) zz104 bb71)) =
        (subst_Z_BB (weakenTrace X0 k134) (subst_B_ZZ d71 bb72 zz104) (subst_B_BB (weakenTrace (XS Z d71) k134) bb72 bb71))))) (fun (a98 : (Index A)) =>
      (fun (k134 : Hvl) (d71 : (Trace B)) (bb71 : BB) (zz104 : ZZ) =>
        eq_refl)) (fun (zz : ZZ) IHzz21 (k134 : Hvl) (d71 : (Trace B)) (bb71 : BB) (zz104 : ZZ) =>
      (f_equal arec (IHzz21 k134 d71 bb71 zz104))) (fun (z65 : (Index Z)) =>
      (fun (k134 : Hvl) (d71 : (Trace B)) (bb71 : BB) (zz104 : ZZ) =>
        (subst_B_ZZ_subst_Z_Index0_commright_ind d71 bb71 zz104 k134 z65))) (fun (aa : AA) IHaa8 (k134 : Hvl) (d71 : (Trace B)) (bb71 : BB) (zz104 : ZZ) =>
      (f_equal aabs (eq_trans (f_equal3 subst_B_AA (weakenTrace_append B d71 k134 (HS A H0)) eq_refl (f_equal3 subst_Z_AA (weakenTrace_append Z X0 k134 (HS A H0)) eq_refl eq_refl)) (eq_trans (IHaa8 (HS A k134) d71 bb71 zz104) (f_equal3 subst_Z_AA (eq_sym (weakenTrace_append Z X0 k134 (HS A H0))) eq_refl (f_equal3 subst_B_AA (eq_sym (weakenTrace_append B (XS Z d71) k134 (HS A H0))) eq_refl eq_refl)))))) (fun (bb : BB) IHbb9 (k134 : Hvl) (d71 : (Trace B)) (bb71 : BB) (zz104 : ZZ) =>
      (f_equal babs (eq_trans (f_equal3 subst_B_BB (weakenTrace_append B d71 k134 (HS B H0)) eq_refl (f_equal3 subst_Z_BB (weakenTrace_append Z X0 k134 (HS B H0)) eq_refl eq_refl)) (eq_trans (IHbb9 (HS B k134) d71 bb71 zz104) (f_equal3 subst_Z_BB (eq_sym (weakenTrace_append Z X0 k134 (HS B H0))) eq_refl (f_equal3 subst_B_BB (eq_sym (weakenTrace_append B (XS Z d71) k134 (HS B H0))) eq_refl eq_refl)))))) (fun (b98 : (Index B)) =>
      (fun (k134 : Hvl) (d71 : (Trace B)) (bb71 : BB) (zz104 : ZZ) =>
        (subst_B_ZZ_subst_Z_Index0_commleft_ind d71 bb71 zz104 k134 b98))) (fun (zz : ZZ) IHzz21 (k134 : Hvl) (d71 : (Trace B)) (bb71 : BB) (zz104 : ZZ) =>
      (f_equal brec (IHzz21 k134 d71 bb71 zz104)))).
    Lemma subst_B__subst_Z_0_AA_comm_ind  :
      (forall (aa71 : AA) ,
        (forall (k134 : Hvl) (d71 : (Trace B)) (bb71 : BB) (zz104 : ZZ) ,
          ((subst_B_AA (weakenTrace d71 k134) bb71 (subst_Z_AA (weakenTrace X0 k134) zz104 aa71)) =
          (subst_Z_AA (weakenTrace X0 k134) (subst_B_ZZ d71 bb71 zz104) (subst_B_AA (weakenTrace (XS Z d71) k134) bb71 aa71))))).
    Proof.
      pose proof (subst_B__subst_Z_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_B__subst_Z_0_ZZ_comm_ind  :
      (forall (zz104 : ZZ) ,
        (forall (k134 : Hvl) (d71 : (Trace B)) (bb71 : BB) (zz105 : ZZ) ,
          ((subst_B_ZZ (weakenTrace d71 k134) bb71 (subst_Z_ZZ (weakenTrace X0 k134) zz105 zz104)) =
          (subst_Z_ZZ (weakenTrace X0 k134) (subst_B_ZZ d71 bb71 zz105) (subst_B_ZZ (weakenTrace (XS Z d71) k134) bb71 zz104))))).
    Proof.
      pose proof (subst_B__subst_Z_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_B__subst_Z_0_BB_comm_ind  :
      (forall (bb71 : BB) ,
        (forall (k134 : Hvl) (d71 : (Trace B)) (bb72 : BB) (zz105 : ZZ) ,
          ((subst_B_BB (weakenTrace d71 k134) bb72 (subst_Z_BB (weakenTrace X0 k134) zz105 bb71)) =
          (subst_Z_BB (weakenTrace X0 k134) (subst_B_ZZ d71 bb72 zz105) (subst_B_BB (weakenTrace (XS Z d71) k134) bb72 bb71))))).
    Proof.
      pose proof (subst_B__subst_Z_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_Z__subst_A_0_AA_ZZ_BB_comm_ind := (ind_AA_ZZ_BB (fun (aa71 : AA) =>
      (forall (k134 : Hvl) (d71 : (Trace Z)) (zz104 : ZZ) (aa72 : AA) ,
        ((subst_Z_AA (weakenTrace d71 k134) zz104 (subst_A_AA (weakenTrace X0 k134) aa72 aa71)) =
        (subst_A_AA (weakenTrace X0 k134) (subst_Z_AA d71 zz104 aa72) (subst_Z_AA (weakenTrace (XS A d71) k134) zz104 aa71))))) (fun (zz104 : ZZ) =>
      (forall (k134 : Hvl) (d71 : (Trace Z)) (zz105 : ZZ) (aa71 : AA) ,
        ((subst_Z_ZZ (weakenTrace d71 k134) zz105 (subst_A_ZZ (weakenTrace X0 k134) aa71 zz104)) =
        (subst_A_ZZ (weakenTrace X0 k134) (subst_Z_AA d71 zz105 aa71) (subst_Z_ZZ (weakenTrace (XS A d71) k134) zz105 zz104))))) (fun (bb71 : BB) =>
      (forall (k134 : Hvl) (d71 : (Trace Z)) (zz104 : ZZ) (aa71 : AA) ,
        ((subst_Z_BB (weakenTrace d71 k134) zz104 (subst_A_BB (weakenTrace X0 k134) aa71 bb71)) =
        (subst_A_BB (weakenTrace X0 k134) (subst_Z_AA d71 zz104 aa71) (subst_Z_BB (weakenTrace (XS A d71) k134) zz104 bb71))))) (fun (a98 : (Index A)) =>
      (fun (k134 : Hvl) (d71 : (Trace Z)) (zz104 : ZZ) (aa71 : AA) =>
        (subst_Z_AA_subst_A_Index0_commright_ind d71 zz104 aa71 k134 a98))) (fun (zz : ZZ) IHzz21 (k134 : Hvl) (d71 : (Trace Z)) (zz104 : ZZ) (aa71 : AA) =>
      (f_equal arec (IHzz21 k134 d71 zz104 aa71))) (fun (z65 : (Index Z)) =>
      (fun (k134 : Hvl) (d71 : (Trace Z)) (zz104 : ZZ) (aa71 : AA) =>
        (subst_Z_AA_subst_A_Index0_commleft_ind d71 zz104 aa71 k134 z65))) (fun (aa : AA) IHaa8 (k134 : Hvl) (d71 : (Trace Z)) (zz104 : ZZ) (aa71 : AA) =>
      (f_equal aabs (eq_trans (f_equal3 subst_Z_AA (weakenTrace_append Z d71 k134 (HS A H0)) eq_refl (f_equal3 subst_A_AA (weakenTrace_append A X0 k134 (HS A H0)) eq_refl eq_refl)) (eq_trans (IHaa8 (HS A k134) d71 zz104 aa71) (f_equal3 subst_A_AA (eq_sym (weakenTrace_append A X0 k134 (HS A H0))) eq_refl (f_equal3 subst_Z_AA (eq_sym (weakenTrace_append Z (XS A d71) k134 (HS A H0))) eq_refl eq_refl)))))) (fun (bb : BB) IHbb9 (k134 : Hvl) (d71 : (Trace Z)) (zz104 : ZZ) (aa71 : AA) =>
      (f_equal babs (eq_trans (f_equal3 subst_Z_BB (weakenTrace_append Z d71 k134 (HS B H0)) eq_refl (f_equal3 subst_A_BB (weakenTrace_append A X0 k134 (HS B H0)) eq_refl eq_refl)) (eq_trans (IHbb9 (HS B k134) d71 zz104 aa71) (f_equal3 subst_A_BB (eq_sym (weakenTrace_append A X0 k134 (HS B H0))) eq_refl (f_equal3 subst_Z_BB (eq_sym (weakenTrace_append Z (XS A d71) k134 (HS B H0))) eq_refl eq_refl)))))) (fun (b98 : (Index B)) =>
      (fun (k134 : Hvl) (d71 : (Trace Z)) (zz104 : ZZ) (aa71 : AA) =>
        eq_refl)) (fun (zz : ZZ) IHzz21 (k134 : Hvl) (d71 : (Trace Z)) (zz104 : ZZ) (aa71 : AA) =>
      (f_equal brec (IHzz21 k134 d71 zz104 aa71)))).
    Lemma subst_Z__subst_A_0_AA_comm_ind  :
      (forall (aa71 : AA) ,
        (forall (k134 : Hvl) (d71 : (Trace Z)) (zz104 : ZZ) (aa72 : AA) ,
          ((subst_Z_AA (weakenTrace d71 k134) zz104 (subst_A_AA (weakenTrace X0 k134) aa72 aa71)) =
          (subst_A_AA (weakenTrace X0 k134) (subst_Z_AA d71 zz104 aa72) (subst_Z_AA (weakenTrace (XS A d71) k134) zz104 aa71))))).
    Proof.
      pose proof (subst_Z__subst_A_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_Z__subst_A_0_ZZ_comm_ind  :
      (forall (zz104 : ZZ) ,
        (forall (k134 : Hvl) (d71 : (Trace Z)) (zz105 : ZZ) (aa72 : AA) ,
          ((subst_Z_ZZ (weakenTrace d71 k134) zz105 (subst_A_ZZ (weakenTrace X0 k134) aa72 zz104)) =
          (subst_A_ZZ (weakenTrace X0 k134) (subst_Z_AA d71 zz105 aa72) (subst_Z_ZZ (weakenTrace (XS A d71) k134) zz105 zz104))))).
    Proof.
      pose proof (subst_Z__subst_A_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_Z__subst_A_0_BB_comm_ind  :
      (forall (bb71 : BB) ,
        (forall (k134 : Hvl) (d71 : (Trace Z)) (zz105 : ZZ) (aa72 : AA) ,
          ((subst_Z_BB (weakenTrace d71 k134) zz105 (subst_A_BB (weakenTrace X0 k134) aa72 bb71)) =
          (subst_A_BB (weakenTrace X0 k134) (subst_Z_AA d71 zz105 aa72) (subst_Z_BB (weakenTrace (XS A d71) k134) zz105 bb71))))).
    Proof.
      pose proof (subst_Z__subst_A_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_Z__subst_B_0_AA_ZZ_BB_comm_ind := (ind_AA_ZZ_BB (fun (aa71 : AA) =>
      (forall (k134 : Hvl) (d71 : (Trace Z)) (zz104 : ZZ) (bb71 : BB) ,
        ((subst_Z_AA (weakenTrace d71 k134) zz104 (subst_B_AA (weakenTrace X0 k134) bb71 aa71)) =
        (subst_B_AA (weakenTrace X0 k134) (subst_Z_BB d71 zz104 bb71) (subst_Z_AA (weakenTrace (XS B d71) k134) zz104 aa71))))) (fun (zz104 : ZZ) =>
      (forall (k134 : Hvl) (d71 : (Trace Z)) (zz105 : ZZ) (bb71 : BB) ,
        ((subst_Z_ZZ (weakenTrace d71 k134) zz105 (subst_B_ZZ (weakenTrace X0 k134) bb71 zz104)) =
        (subst_B_ZZ (weakenTrace X0 k134) (subst_Z_BB d71 zz105 bb71) (subst_Z_ZZ (weakenTrace (XS B d71) k134) zz105 zz104))))) (fun (bb71 : BB) =>
      (forall (k134 : Hvl) (d71 : (Trace Z)) (zz104 : ZZ) (bb72 : BB) ,
        ((subst_Z_BB (weakenTrace d71 k134) zz104 (subst_B_BB (weakenTrace X0 k134) bb72 bb71)) =
        (subst_B_BB (weakenTrace X0 k134) (subst_Z_BB d71 zz104 bb72) (subst_Z_BB (weakenTrace (XS B d71) k134) zz104 bb71))))) (fun (a98 : (Index A)) =>
      (fun (k134 : Hvl) (d71 : (Trace Z)) (zz104 : ZZ) (bb71 : BB) =>
        eq_refl)) (fun (zz : ZZ) IHzz21 (k134 : Hvl) (d71 : (Trace Z)) (zz104 : ZZ) (bb71 : BB) =>
      (f_equal arec (IHzz21 k134 d71 zz104 bb71))) (fun (z65 : (Index Z)) =>
      (fun (k134 : Hvl) (d71 : (Trace Z)) (zz104 : ZZ) (bb71 : BB) =>
        (subst_Z_BB_subst_B_Index0_commleft_ind d71 zz104 bb71 k134 z65))) (fun (aa : AA) IHaa8 (k134 : Hvl) (d71 : (Trace Z)) (zz104 : ZZ) (bb71 : BB) =>
      (f_equal aabs (eq_trans (f_equal3 subst_Z_AA (weakenTrace_append Z d71 k134 (HS A H0)) eq_refl (f_equal3 subst_B_AA (weakenTrace_append B X0 k134 (HS A H0)) eq_refl eq_refl)) (eq_trans (IHaa8 (HS A k134) d71 zz104 bb71) (f_equal3 subst_B_AA (eq_sym (weakenTrace_append B X0 k134 (HS A H0))) eq_refl (f_equal3 subst_Z_AA (eq_sym (weakenTrace_append Z (XS B d71) k134 (HS A H0))) eq_refl eq_refl)))))) (fun (bb : BB) IHbb9 (k134 : Hvl) (d71 : (Trace Z)) (zz104 : ZZ) (bb71 : BB) =>
      (f_equal babs (eq_trans (f_equal3 subst_Z_BB (weakenTrace_append Z d71 k134 (HS B H0)) eq_refl (f_equal3 subst_B_BB (weakenTrace_append B X0 k134 (HS B H0)) eq_refl eq_refl)) (eq_trans (IHbb9 (HS B k134) d71 zz104 bb71) (f_equal3 subst_B_BB (eq_sym (weakenTrace_append B X0 k134 (HS B H0))) eq_refl (f_equal3 subst_Z_BB (eq_sym (weakenTrace_append Z (XS B d71) k134 (HS B H0))) eq_refl eq_refl)))))) (fun (b98 : (Index B)) =>
      (fun (k134 : Hvl) (d71 : (Trace Z)) (zz104 : ZZ) (bb71 : BB) =>
        (subst_Z_BB_subst_B_Index0_commright_ind d71 zz104 bb71 k134 b98))) (fun (zz : ZZ) IHzz21 (k134 : Hvl) (d71 : (Trace Z)) (zz104 : ZZ) (bb71 : BB) =>
      (f_equal brec (IHzz21 k134 d71 zz104 bb71)))).
    Lemma subst_Z__subst_B_0_AA_comm_ind  :
      (forall (aa71 : AA) ,
        (forall (k134 : Hvl) (d71 : (Trace Z)) (zz104 : ZZ) (bb71 : BB) ,
          ((subst_Z_AA (weakenTrace d71 k134) zz104 (subst_B_AA (weakenTrace X0 k134) bb71 aa71)) =
          (subst_B_AA (weakenTrace X0 k134) (subst_Z_BB d71 zz104 bb71) (subst_Z_AA (weakenTrace (XS B d71) k134) zz104 aa71))))).
    Proof.
      pose proof (subst_Z__subst_B_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_Z__subst_B_0_ZZ_comm_ind  :
      (forall (zz104 : ZZ) ,
        (forall (k134 : Hvl) (d71 : (Trace Z)) (zz105 : ZZ) (bb71 : BB) ,
          ((subst_Z_ZZ (weakenTrace d71 k134) zz105 (subst_B_ZZ (weakenTrace X0 k134) bb71 zz104)) =
          (subst_B_ZZ (weakenTrace X0 k134) (subst_Z_BB d71 zz105 bb71) (subst_Z_ZZ (weakenTrace (XS B d71) k134) zz105 zz104))))).
    Proof.
      pose proof (subst_Z__subst_B_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_Z__subst_B_0_BB_comm_ind  :
      (forall (bb71 : BB) ,
        (forall (k134 : Hvl) (d71 : (Trace Z)) (zz105 : ZZ) (bb72 : BB) ,
          ((subst_Z_BB (weakenTrace d71 k134) zz105 (subst_B_BB (weakenTrace X0 k134) bb72 bb71)) =
          (subst_B_BB (weakenTrace X0 k134) (subst_Z_BB d71 zz105 bb72) (subst_Z_BB (weakenTrace (XS B d71) k134) zz105 bb71))))).
    Proof.
      pose proof (subst_Z__subst_B_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_Z__subst_Z_0_AA_ZZ_BB_comm_ind := (ind_AA_ZZ_BB (fun (aa71 : AA) =>
      (forall (k134 : Hvl) (d71 : (Trace Z)) (zz104 : ZZ) (zz105 : ZZ) ,
        ((subst_Z_AA (weakenTrace d71 k134) zz104 (subst_Z_AA (weakenTrace X0 k134) zz105 aa71)) =
        (subst_Z_AA (weakenTrace X0 k134) (subst_Z_ZZ d71 zz104 zz105) (subst_Z_AA (weakenTrace (XS Z d71) k134) zz104 aa71))))) (fun (zz104 : ZZ) =>
      (forall (k134 : Hvl) (d71 : (Trace Z)) (zz105 : ZZ) (zz106 : ZZ) ,
        ((subst_Z_ZZ (weakenTrace d71 k134) zz105 (subst_Z_ZZ (weakenTrace X0 k134) zz106 zz104)) =
        (subst_Z_ZZ (weakenTrace X0 k134) (subst_Z_ZZ d71 zz105 zz106) (subst_Z_ZZ (weakenTrace (XS Z d71) k134) zz105 zz104))))) (fun (bb71 : BB) =>
      (forall (k134 : Hvl) (d71 : (Trace Z)) (zz104 : ZZ) (zz105 : ZZ) ,
        ((subst_Z_BB (weakenTrace d71 k134) zz104 (subst_Z_BB (weakenTrace X0 k134) zz105 bb71)) =
        (subst_Z_BB (weakenTrace X0 k134) (subst_Z_ZZ d71 zz104 zz105) (subst_Z_BB (weakenTrace (XS Z d71) k134) zz104 bb71))))) (fun (a98 : (Index A)) =>
      (fun (k134 : Hvl) (d71 : (Trace Z)) (zz104 : ZZ) (zz105 : ZZ) =>
        eq_refl)) (fun (zz : ZZ) IHzz21 (k134 : Hvl) (d71 : (Trace Z)) (zz104 : ZZ) (zz105 : ZZ) =>
      (f_equal arec (IHzz21 k134 d71 zz104 zz105))) (fun (z65 : (Index Z)) =>
      (fun (k134 : Hvl) (d71 : (Trace Z)) (zz104 : ZZ) (zz105 : ZZ) =>
        (subst_Z_ZZ_subst_Z_Index0_commright_ind d71 zz104 zz105 k134 z65))) (fun (aa : AA) IHaa8 (k134 : Hvl) (d71 : (Trace Z)) (zz104 : ZZ) (zz105 : ZZ) =>
      (f_equal aabs (eq_trans (f_equal3 subst_Z_AA (weakenTrace_append Z d71 k134 (HS A H0)) eq_refl (f_equal3 subst_Z_AA (weakenTrace_append Z X0 k134 (HS A H0)) eq_refl eq_refl)) (eq_trans (IHaa8 (HS A k134) d71 zz104 zz105) (f_equal3 subst_Z_AA (eq_sym (weakenTrace_append Z X0 k134 (HS A H0))) eq_refl (f_equal3 subst_Z_AA (eq_sym (weakenTrace_append Z (XS Z d71) k134 (HS A H0))) eq_refl eq_refl)))))) (fun (bb : BB) IHbb9 (k134 : Hvl) (d71 : (Trace Z)) (zz104 : ZZ) (zz105 : ZZ) =>
      (f_equal babs (eq_trans (f_equal3 subst_Z_BB (weakenTrace_append Z d71 k134 (HS B H0)) eq_refl (f_equal3 subst_Z_BB (weakenTrace_append Z X0 k134 (HS B H0)) eq_refl eq_refl)) (eq_trans (IHbb9 (HS B k134) d71 zz104 zz105) (f_equal3 subst_Z_BB (eq_sym (weakenTrace_append Z X0 k134 (HS B H0))) eq_refl (f_equal3 subst_Z_BB (eq_sym (weakenTrace_append Z (XS Z d71) k134 (HS B H0))) eq_refl eq_refl)))))) (fun (b98 : (Index B)) =>
      (fun (k134 : Hvl) (d71 : (Trace Z)) (zz104 : ZZ) (zz105 : ZZ) =>
        eq_refl)) (fun (zz : ZZ) IHzz21 (k134 : Hvl) (d71 : (Trace Z)) (zz104 : ZZ) (zz105 : ZZ) =>
      (f_equal brec (IHzz21 k134 d71 zz104 zz105)))).
    Lemma subst_Z__subst_Z_0_AA_comm_ind  :
      (forall (aa71 : AA) ,
        (forall (k134 : Hvl) (d71 : (Trace Z)) (zz104 : ZZ) (zz105 : ZZ) ,
          ((subst_Z_AA (weakenTrace d71 k134) zz104 (subst_Z_AA (weakenTrace X0 k134) zz105 aa71)) =
          (subst_Z_AA (weakenTrace X0 k134) (subst_Z_ZZ d71 zz104 zz105) (subst_Z_AA (weakenTrace (XS Z d71) k134) zz104 aa71))))).
    Proof.
      pose proof (subst_Z__subst_Z_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_Z__subst_Z_0_ZZ_comm_ind  :
      (forall (zz104 : ZZ) ,
        (forall (k134 : Hvl) (d71 : (Trace Z)) (zz105 : ZZ) (zz106 : ZZ) ,
          ((subst_Z_ZZ (weakenTrace d71 k134) zz105 (subst_Z_ZZ (weakenTrace X0 k134) zz106 zz104)) =
          (subst_Z_ZZ (weakenTrace X0 k134) (subst_Z_ZZ d71 zz105 zz106) (subst_Z_ZZ (weakenTrace (XS Z d71) k134) zz105 zz104))))).
    Proof.
      pose proof (subst_Z__subst_Z_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_Z__subst_Z_0_BB_comm_ind  :
      (forall (bb71 : BB) ,
        (forall (k134 : Hvl) (d71 : (Trace Z)) (zz105 : ZZ) (zz106 : ZZ) ,
          ((subst_Z_BB (weakenTrace d71 k134) zz105 (subst_Z_BB (weakenTrace X0 k134) zz106 bb71)) =
          (subst_Z_BB (weakenTrace X0 k134) (subst_Z_ZZ d71 zz105 zz106) (subst_Z_BB (weakenTrace (XS Z d71) k134) zz105 bb71))))).
    Proof.
      pose proof (subst_Z__subst_Z_0_AA_ZZ_BB_comm_ind).
      destruct_conjs .
      easy .
    Qed.
  End SubstSubstCommInd.
  Section SubstSubstComm.
    Definition subst_A_AA_subst_A_AA0_comm (aa73 : AA) : (forall (d71 : (Trace A)) (aa71 : AA) (aa72 : AA) ,
      ((subst_A_AA d71 aa71 (subst_A_AA X0 aa72 aa73)) =
      (subst_A_AA X0 (subst_A_AA d71 aa71 aa72) (subst_A_AA (XS A d71) aa71 aa73)))) := (subst_A__subst_A_0_AA_comm_ind aa73 H0).
    Definition subst_A_ZZ_subst_A_ZZ0_comm (zz104 : ZZ) : (forall (d71 : (Trace A)) (aa71 : AA) (aa72 : AA) ,
      ((subst_A_ZZ d71 aa71 (subst_A_ZZ X0 aa72 zz104)) =
      (subst_A_ZZ X0 (subst_A_AA d71 aa71 aa72) (subst_A_ZZ (XS A d71) aa71 zz104)))) := (subst_A__subst_A_0_ZZ_comm_ind zz104 H0).
    Definition subst_A_BB_subst_A_BB0_comm (bb71 : BB) : (forall (d71 : (Trace A)) (aa71 : AA) (aa72 : AA) ,
      ((subst_A_BB d71 aa71 (subst_A_BB X0 aa72 bb71)) =
      (subst_A_BB X0 (subst_A_AA d71 aa71 aa72) (subst_A_BB (XS A d71) aa71 bb71)))) := (subst_A__subst_A_0_BB_comm_ind bb71 H0).
    Definition subst_A_AA_subst_B_AA0_comm (aa72 : AA) : (forall (d71 : (Trace A)) (aa71 : AA) (bb71 : BB) ,
      ((subst_A_AA d71 aa71 (subst_B_AA X0 bb71 aa72)) =
      (subst_B_AA X0 (subst_A_BB d71 aa71 bb71) (subst_A_AA (XS B d71) aa71 aa72)))) := (subst_A__subst_B_0_AA_comm_ind aa72 H0).
    Definition subst_A_ZZ_subst_B_ZZ0_comm (zz104 : ZZ) : (forall (d71 : (Trace A)) (aa71 : AA) (bb71 : BB) ,
      ((subst_A_ZZ d71 aa71 (subst_B_ZZ X0 bb71 zz104)) =
      (subst_B_ZZ X0 (subst_A_BB d71 aa71 bb71) (subst_A_ZZ (XS B d71) aa71 zz104)))) := (subst_A__subst_B_0_ZZ_comm_ind zz104 H0).
    Definition subst_A_BB_subst_B_BB0_comm (bb72 : BB) : (forall (d71 : (Trace A)) (aa71 : AA) (bb71 : BB) ,
      ((subst_A_BB d71 aa71 (subst_B_BB X0 bb71 bb72)) =
      (subst_B_BB X0 (subst_A_BB d71 aa71 bb71) (subst_A_BB (XS B d71) aa71 bb72)))) := (subst_A__subst_B_0_BB_comm_ind bb72 H0).
    Definition subst_A_AA_subst_Z_AA0_comm (aa72 : AA) : (forall (d71 : (Trace A)) (aa71 : AA) (zz104 : ZZ) ,
      ((subst_A_AA d71 aa71 (subst_Z_AA X0 zz104 aa72)) =
      (subst_Z_AA X0 (subst_A_ZZ d71 aa71 zz104) (subst_A_AA (XS Z d71) aa71 aa72)))) := (subst_A__subst_Z_0_AA_comm_ind aa72 H0).
    Definition subst_A_ZZ_subst_Z_ZZ0_comm (zz105 : ZZ) : (forall (d71 : (Trace A)) (aa71 : AA) (zz104 : ZZ) ,
      ((subst_A_ZZ d71 aa71 (subst_Z_ZZ X0 zz104 zz105)) =
      (subst_Z_ZZ X0 (subst_A_ZZ d71 aa71 zz104) (subst_A_ZZ (XS Z d71) aa71 zz105)))) := (subst_A__subst_Z_0_ZZ_comm_ind zz105 H0).
    Definition subst_A_BB_subst_Z_BB0_comm (bb71 : BB) : (forall (d71 : (Trace A)) (aa71 : AA) (zz104 : ZZ) ,
      ((subst_A_BB d71 aa71 (subst_Z_BB X0 zz104 bb71)) =
      (subst_Z_BB X0 (subst_A_ZZ d71 aa71 zz104) (subst_A_BB (XS Z d71) aa71 bb71)))) := (subst_A__subst_Z_0_BB_comm_ind bb71 H0).
    Definition subst_B_AA_subst_A_AA0_comm (aa72 : AA) : (forall (d71 : (Trace B)) (bb71 : BB) (aa71 : AA) ,
      ((subst_B_AA d71 bb71 (subst_A_AA X0 aa71 aa72)) =
      (subst_A_AA X0 (subst_B_AA d71 bb71 aa71) (subst_B_AA (XS A d71) bb71 aa72)))) := (subst_B__subst_A_0_AA_comm_ind aa72 H0).
    Definition subst_B_ZZ_subst_A_ZZ0_comm (zz104 : ZZ) : (forall (d71 : (Trace B)) (bb71 : BB) (aa71 : AA) ,
      ((subst_B_ZZ d71 bb71 (subst_A_ZZ X0 aa71 zz104)) =
      (subst_A_ZZ X0 (subst_B_AA d71 bb71 aa71) (subst_B_ZZ (XS A d71) bb71 zz104)))) := (subst_B__subst_A_0_ZZ_comm_ind zz104 H0).
    Definition subst_B_BB_subst_A_BB0_comm (bb72 : BB) : (forall (d71 : (Trace B)) (bb71 : BB) (aa71 : AA) ,
      ((subst_B_BB d71 bb71 (subst_A_BB X0 aa71 bb72)) =
      (subst_A_BB X0 (subst_B_AA d71 bb71 aa71) (subst_B_BB (XS A d71) bb71 bb72)))) := (subst_B__subst_A_0_BB_comm_ind bb72 H0).
    Definition subst_B_AA_subst_B_AA0_comm (aa71 : AA) : (forall (d71 : (Trace B)) (bb71 : BB) (bb72 : BB) ,
      ((subst_B_AA d71 bb71 (subst_B_AA X0 bb72 aa71)) =
      (subst_B_AA X0 (subst_B_BB d71 bb71 bb72) (subst_B_AA (XS B d71) bb71 aa71)))) := (subst_B__subst_B_0_AA_comm_ind aa71 H0).
    Definition subst_B_ZZ_subst_B_ZZ0_comm (zz104 : ZZ) : (forall (d71 : (Trace B)) (bb71 : BB) (bb72 : BB) ,
      ((subst_B_ZZ d71 bb71 (subst_B_ZZ X0 bb72 zz104)) =
      (subst_B_ZZ X0 (subst_B_BB d71 bb71 bb72) (subst_B_ZZ (XS B d71) bb71 zz104)))) := (subst_B__subst_B_0_ZZ_comm_ind zz104 H0).
    Definition subst_B_BB_subst_B_BB0_comm (bb73 : BB) : (forall (d71 : (Trace B)) (bb71 : BB) (bb72 : BB) ,
      ((subst_B_BB d71 bb71 (subst_B_BB X0 bb72 bb73)) =
      (subst_B_BB X0 (subst_B_BB d71 bb71 bb72) (subst_B_BB (XS B d71) bb71 bb73)))) := (subst_B__subst_B_0_BB_comm_ind bb73 H0).
    Definition subst_B_AA_subst_Z_AA0_comm (aa71 : AA) : (forall (d71 : (Trace B)) (bb71 : BB) (zz104 : ZZ) ,
      ((subst_B_AA d71 bb71 (subst_Z_AA X0 zz104 aa71)) =
      (subst_Z_AA X0 (subst_B_ZZ d71 bb71 zz104) (subst_B_AA (XS Z d71) bb71 aa71)))) := (subst_B__subst_Z_0_AA_comm_ind aa71 H0).
    Definition subst_B_ZZ_subst_Z_ZZ0_comm (zz105 : ZZ) : (forall (d71 : (Trace B)) (bb71 : BB) (zz104 : ZZ) ,
      ((subst_B_ZZ d71 bb71 (subst_Z_ZZ X0 zz104 zz105)) =
      (subst_Z_ZZ X0 (subst_B_ZZ d71 bb71 zz104) (subst_B_ZZ (XS Z d71) bb71 zz105)))) := (subst_B__subst_Z_0_ZZ_comm_ind zz105 H0).
    Definition subst_B_BB_subst_Z_BB0_comm (bb72 : BB) : (forall (d71 : (Trace B)) (bb71 : BB) (zz104 : ZZ) ,
      ((subst_B_BB d71 bb71 (subst_Z_BB X0 zz104 bb72)) =
      (subst_Z_BB X0 (subst_B_ZZ d71 bb71 zz104) (subst_B_BB (XS Z d71) bb71 bb72)))) := (subst_B__subst_Z_0_BB_comm_ind bb72 H0).
    Definition subst_Z_AA_subst_A_AA0_comm (aa72 : AA) : (forall (d71 : (Trace Z)) (zz104 : ZZ) (aa71 : AA) ,
      ((subst_Z_AA d71 zz104 (subst_A_AA X0 aa71 aa72)) =
      (subst_A_AA X0 (subst_Z_AA d71 zz104 aa71) (subst_Z_AA (XS A d71) zz104 aa72)))) := (subst_Z__subst_A_0_AA_comm_ind aa72 H0).
    Definition subst_Z_ZZ_subst_A_ZZ0_comm (zz105 : ZZ) : (forall (d71 : (Trace Z)) (zz104 : ZZ) (aa71 : AA) ,
      ((subst_Z_ZZ d71 zz104 (subst_A_ZZ X0 aa71 zz105)) =
      (subst_A_ZZ X0 (subst_Z_AA d71 zz104 aa71) (subst_Z_ZZ (XS A d71) zz104 zz105)))) := (subst_Z__subst_A_0_ZZ_comm_ind zz105 H0).
    Definition subst_Z_BB_subst_A_BB0_comm (bb71 : BB) : (forall (d71 : (Trace Z)) (zz104 : ZZ) (aa71 : AA) ,
      ((subst_Z_BB d71 zz104 (subst_A_BB X0 aa71 bb71)) =
      (subst_A_BB X0 (subst_Z_AA d71 zz104 aa71) (subst_Z_BB (XS A d71) zz104 bb71)))) := (subst_Z__subst_A_0_BB_comm_ind bb71 H0).
    Definition subst_Z_AA_subst_B_AA0_comm (aa71 : AA) : (forall (d71 : (Trace Z)) (zz104 : ZZ) (bb71 : BB) ,
      ((subst_Z_AA d71 zz104 (subst_B_AA X0 bb71 aa71)) =
      (subst_B_AA X0 (subst_Z_BB d71 zz104 bb71) (subst_Z_AA (XS B d71) zz104 aa71)))) := (subst_Z__subst_B_0_AA_comm_ind aa71 H0).
    Definition subst_Z_ZZ_subst_B_ZZ0_comm (zz105 : ZZ) : (forall (d71 : (Trace Z)) (zz104 : ZZ) (bb71 : BB) ,
      ((subst_Z_ZZ d71 zz104 (subst_B_ZZ X0 bb71 zz105)) =
      (subst_B_ZZ X0 (subst_Z_BB d71 zz104 bb71) (subst_Z_ZZ (XS B d71) zz104 zz105)))) := (subst_Z__subst_B_0_ZZ_comm_ind zz105 H0).
    Definition subst_Z_BB_subst_B_BB0_comm (bb72 : BB) : (forall (d71 : (Trace Z)) (zz104 : ZZ) (bb71 : BB) ,
      ((subst_Z_BB d71 zz104 (subst_B_BB X0 bb71 bb72)) =
      (subst_B_BB X0 (subst_Z_BB d71 zz104 bb71) (subst_Z_BB (XS B d71) zz104 bb72)))) := (subst_Z__subst_B_0_BB_comm_ind bb72 H0).
    Definition subst_Z_AA_subst_Z_AA0_comm (aa71 : AA) : (forall (d71 : (Trace Z)) (zz104 : ZZ) (zz105 : ZZ) ,
      ((subst_Z_AA d71 zz104 (subst_Z_AA X0 zz105 aa71)) =
      (subst_Z_AA X0 (subst_Z_ZZ d71 zz104 zz105) (subst_Z_AA (XS Z d71) zz104 aa71)))) := (subst_Z__subst_Z_0_AA_comm_ind aa71 H0).
    Definition subst_Z_ZZ_subst_Z_ZZ0_comm (zz106 : ZZ) : (forall (d71 : (Trace Z)) (zz104 : ZZ) (zz105 : ZZ) ,
      ((subst_Z_ZZ d71 zz104 (subst_Z_ZZ X0 zz105 zz106)) =
      (subst_Z_ZZ X0 (subst_Z_ZZ d71 zz104 zz105) (subst_Z_ZZ (XS Z d71) zz104 zz106)))) := (subst_Z__subst_Z_0_ZZ_comm_ind zz106 H0).
    Definition subst_Z_BB_subst_Z_BB0_comm (bb71 : BB) : (forall (d71 : (Trace Z)) (zz104 : ZZ) (zz105 : ZZ) ,
      ((subst_Z_BB d71 zz104 (subst_Z_BB X0 zz105 bb71)) =
      (subst_Z_BB X0 (subst_Z_ZZ d71 zz104 zz105) (subst_Z_BB (XS Z d71) zz104 bb71)))) := (subst_Z__subst_Z_0_BB_comm_ind bb71 H0).
  End SubstSubstComm.
  Section WeakenSubst.
    Lemma weakenAA_subst_A_AA  :
      (forall (k134 : Hvl) (d71 : (Trace A)) (aa71 : AA) (aa72 : AA) ,
        ((weakenAA (subst_A_AA d71 aa71 aa72) k134) =
        (subst_A_AA (weakenTrace d71 k134) aa71 (weakenAA aa72 k134)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenZZ_subst_A_ZZ  :
      (forall (k134 : Hvl) (d71 : (Trace A)) (aa71 : AA) (zz104 : ZZ) ,
        ((weakenZZ (subst_A_ZZ d71 aa71 zz104) k134) =
        (subst_A_ZZ (weakenTrace d71 k134) aa71 (weakenZZ zz104 k134)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenBB_subst_A_BB  :
      (forall (k134 : Hvl) (d71 : (Trace A)) (aa71 : AA) (bb71 : BB) ,
        ((weakenBB (subst_A_BB d71 aa71 bb71) k134) =
        (subst_A_BB (weakenTrace d71 k134) aa71 (weakenBB bb71 k134)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenAA_subst_B_AA  :
      (forall (k134 : Hvl) (d71 : (Trace B)) (bb71 : BB) (aa71 : AA) ,
        ((weakenAA (subst_B_AA d71 bb71 aa71) k134) =
        (subst_B_AA (weakenTrace d71 k134) bb71 (weakenAA aa71 k134)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenZZ_subst_B_ZZ  :
      (forall (k134 : Hvl) (d71 : (Trace B)) (bb71 : BB) (zz104 : ZZ) ,
        ((weakenZZ (subst_B_ZZ d71 bb71 zz104) k134) =
        (subst_B_ZZ (weakenTrace d71 k134) bb71 (weakenZZ zz104 k134)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenBB_subst_B_BB  :
      (forall (k134 : Hvl) (d71 : (Trace B)) (bb71 : BB) (bb72 : BB) ,
        ((weakenBB (subst_B_BB d71 bb71 bb72) k134) =
        (subst_B_BB (weakenTrace d71 k134) bb71 (weakenBB bb72 k134)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenAA_subst_Z_AA  :
      (forall (k134 : Hvl) (d71 : (Trace Z)) (zz104 : ZZ) (aa71 : AA) ,
        ((weakenAA (subst_Z_AA d71 zz104 aa71) k134) =
        (subst_Z_AA (weakenTrace d71 k134) zz104 (weakenAA aa71 k134)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenZZ_subst_Z_ZZ  :
      (forall (k134 : Hvl) (d71 : (Trace Z)) (zz104 : ZZ) (zz105 : ZZ) ,
        ((weakenZZ (subst_Z_ZZ d71 zz104 zz105) k134) =
        (subst_Z_ZZ (weakenTrace d71 k134) zz104 (weakenZZ zz105 k134)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenBB_subst_Z_BB  :
      (forall (k134 : Hvl) (d71 : (Trace Z)) (zz104 : ZZ) (bb71 : BB) ,
        ((weakenBB (subst_Z_BB d71 zz104 bb71) k134) =
        (subst_Z_BB (weakenTrace d71 k134) zz104 (weakenBB bb71 k134)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
  End WeakenSubst.
  Section WeakenAppend.
    Lemma weakenAA_append  :
      (forall (aa71 : AA) (k134 : Hvl) (k135 : Hvl) ,
        ((weakenAA (weakenAA aa71 k134) k135) =
        (weakenAA aa71 (appendHvl k134 k135)))).
    Proof.
      needleGenericWeakenAppend .
    Qed.
    Lemma weakenZZ_append  :
      (forall (zz104 : ZZ) (k134 : Hvl) (k135 : Hvl) ,
        ((weakenZZ (weakenZZ zz104 k134) k135) =
        (weakenZZ zz104 (appendHvl k134 k135)))).
    Proof.
      needleGenericWeakenAppend .
    Qed.
    Lemma weakenBB_append  :
      (forall (bb71 : BB) (k134 : Hvl) (k135 : Hvl) ,
        ((weakenBB (weakenBB bb71 k134) k135) =
        (weakenBB bb71 (appendHvl k134 k135)))).
    Proof.
      needleGenericWeakenAppend .
    Qed.
  End WeakenAppend.
End SubstSubstInteraction.
 Hint Rewrite subst_A_AA_subst_A_AA0_comm subst_B_AA_subst_B_AA0_comm subst_Z_AA_subst_Z_AA0_comm subst_A_BB_subst_A_BB0_comm subst_B_BB_subst_B_BB0_comm subst_Z_BB_subst_Z_BB0_comm subst_A_ZZ_subst_A_ZZ0_comm subst_B_ZZ_subst_B_ZZ0_comm subst_Z_ZZ_subst_Z_ZZ0_comm : interaction.
 Hint Rewrite subst_A_AA_subst_A_AA0_comm subst_B_AA_subst_B_AA0_comm subst_Z_AA_subst_Z_AA0_comm subst_A_BB_subst_A_BB0_comm subst_B_BB_subst_B_BB0_comm subst_Z_BB_subst_Z_BB0_comm subst_A_ZZ_subst_A_ZZ0_comm subst_B_ZZ_subst_B_ZZ0_comm subst_Z_ZZ_subst_Z_ZZ0_comm : subst_subst0.
 Hint Rewrite weakenAA_shift_A_AA weakenAA_shift_B_AA weakenAA_shift_Z_AA weakenBB_shift_A_BB weakenBB_shift_B_BB weakenBB_shift_Z_BB weakenZZ_shift_A_ZZ weakenZZ_shift_B_ZZ weakenZZ_shift_Z_ZZ : weaken_shift.
 Hint Rewrite weakenAA_subst_A_AA weakenAA_subst_B_AA weakenAA_subst_Z_AA weakenBB_subst_A_BB weakenBB_subst_B_BB weakenBB_subst_Z_BB weakenZZ_subst_A_ZZ weakenZZ_subst_B_ZZ weakenZZ_subst_Z_ZZ : weaken_subst.
Section WellFormed.
  Fixpoint wfindex {a107 : Namespace} (k134 : Hvl) (i : (Index a107)) {struct k134} : Prop :=
    match k134 with
      | (H0) => False
      | (HS b107 k134) => match (eq_namespace_dec a107 b107) with
        | (left e) => match i with
          | (I0) => True
          | (IS i) => (wfindex k134 i)
        end
        | (right n) => (wfindex k134 i)
      end
    end.
  Inductive wfAA (k134 : Hvl) : AA -> Prop :=
    | wfAA_avar {a107 : (Index A)}
        (wfi : (wfindex k134 a107)) :
        (wfAA k134 (avar a107))
    | wfAA_arec {zz110 : ZZ}
        (wf : (wfZZ k134 zz110)) :
        (wfAA k134 (arec zz110))
  with wfZZ (k134 : Hvl) : ZZ -> Prop :=
    | wfZZ_zvar {z71 : (Index Z)}
        (wfi : (wfindex k134 z71)) :
        (wfZZ k134 (zvar z71))
    | wfZZ_aabs {aa74 : AA}
        (wf : (wfAA (HS A k134) aa74)) :
        (wfZZ k134 (aabs aa74))
    | wfZZ_babs {bb74 : BB}
        (wf : (wfBB (HS B k134) bb74)) :
        (wfZZ k134 (babs bb74))
  with wfBB (k134 : Hvl) : BB -> Prop :=
    | wfBB_bvar {b107 : (Index B)}
        (wfi : (wfindex k134 b107)) :
        (wfBB k134 (bvar b107))
    | wfBB_brec {zz110 : ZZ}
        (wf : (wfZZ k134 zz110)) :
        (wfBB k134 (brec zz110)).
  Scheme ind_wfAA := Induction for wfAA Sort Prop
  with ind_wfZZ := Induction for wfZZ Sort Prop
  with ind_wfBB := Induction for wfBB Sort Prop.
  Combined Scheme ind_wfAA_ZZ_BB from ind_wfAA, ind_wfZZ, ind_wfBB.
End WellFormed.
Section ShiftWellFormed.
  Inductive shifthvl_A : (forall (c62 : (Cutoff A)) (k134 : Hvl) (k135 : Hvl) ,
    Prop) :=
    | shifthvl_A_here (k134 : Hvl)
        :
        (shifthvl_A C0 k134 (HS A k134))
    | shifthvl_A_there_A
        (c62 : (Cutoff A)) (k134 : Hvl)
        (k135 : Hvl) :
        (shifthvl_A c62 k134 k135) -> (shifthvl_A (CS c62) (HS A k134) (HS A k135))
    | shifthvl_A_there_B
        (c62 : (Cutoff A)) (k134 : Hvl)
        (k135 : Hvl) :
        (shifthvl_A c62 k134 k135) -> (shifthvl_A c62 (HS B k134) (HS B k135))
    | shifthvl_A_there_Z
        (c62 : (Cutoff A)) (k134 : Hvl)
        (k135 : Hvl) :
        (shifthvl_A c62 k134 k135) -> (shifthvl_A c62 (HS Z k134) (HS Z k135)).
  Inductive shifthvl_B : (forall (c62 : (Cutoff B)) (k134 : Hvl) (k135 : Hvl) ,
    Prop) :=
    | shifthvl_B_here (k134 : Hvl)
        :
        (shifthvl_B C0 k134 (HS B k134))
    | shifthvl_B_there_A
        (c62 : (Cutoff B)) (k134 : Hvl)
        (k135 : Hvl) :
        (shifthvl_B c62 k134 k135) -> (shifthvl_B c62 (HS A k134) (HS A k135))
    | shifthvl_B_there_B
        (c62 : (Cutoff B)) (k134 : Hvl)
        (k135 : Hvl) :
        (shifthvl_B c62 k134 k135) -> (shifthvl_B (CS c62) (HS B k134) (HS B k135))
    | shifthvl_B_there_Z
        (c62 : (Cutoff B)) (k134 : Hvl)
        (k135 : Hvl) :
        (shifthvl_B c62 k134 k135) -> (shifthvl_B c62 (HS Z k134) (HS Z k135)).
  Inductive shifthvl_Z : (forall (c62 : (Cutoff Z)) (k134 : Hvl) (k135 : Hvl) ,
    Prop) :=
    | shifthvl_Z_here (k134 : Hvl)
        :
        (shifthvl_Z C0 k134 (HS Z k134))
    | shifthvl_Z_there_A
        (c62 : (Cutoff Z)) (k134 : Hvl)
        (k135 : Hvl) :
        (shifthvl_Z c62 k134 k135) -> (shifthvl_Z c62 (HS A k134) (HS A k135))
    | shifthvl_Z_there_B
        (c62 : (Cutoff Z)) (k134 : Hvl)
        (k135 : Hvl) :
        (shifthvl_Z c62 k134 k135) -> (shifthvl_Z c62 (HS B k134) (HS B k135))
    | shifthvl_Z_there_Z
        (c62 : (Cutoff Z)) (k134 : Hvl)
        (k135 : Hvl) :
        (shifthvl_Z c62 k134 k135) -> (shifthvl_Z (CS c62) (HS Z k134) (HS Z k135)).
  Lemma weaken_shifthvl_A  :
    (forall (k134 : Hvl) {c62 : (Cutoff A)} {k135 : Hvl} {k136 : Hvl} ,
      (shifthvl_A c62 k135 k136) -> (shifthvl_A (weakenCutoffA c62 k134) (appendHvl k135 k134) (appendHvl k136 k134))).
  Proof.
    needleGenericWeakenHVarlistInsert .
  Qed.
  Lemma weaken_shifthvl_B  :
    (forall (k134 : Hvl) {c62 : (Cutoff B)} {k135 : Hvl} {k136 : Hvl} ,
      (shifthvl_B c62 k135 k136) -> (shifthvl_B (weakenCutoffB c62 k134) (appendHvl k135 k134) (appendHvl k136 k134))).
  Proof.
    needleGenericWeakenHVarlistInsert .
  Qed.
  Lemma weaken_shifthvl_Z  :
    (forall (k134 : Hvl) {c62 : (Cutoff Z)} {k135 : Hvl} {k136 : Hvl} ,
      (shifthvl_Z c62 k135 k136) -> (shifthvl_Z (weakenCutoffZ c62 k134) (appendHvl k135 k134) (appendHvl k136 k134))).
  Proof.
    needleGenericWeakenHVarlistInsert .
  Qed.
  Lemma shift_A__wfindex_A  :
    (forall (c62 : (Cutoff A)) (k134 : Hvl) (k135 : Hvl) (ins : (shifthvl_A c62 k134 k135)) (a107 : (Index A)) ,
      (wfindex k134 a107) -> (wfindex k135 (shift_A_Index c62 a107))).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma shift_A__wfindex_B  :
    (forall (c62 : (Cutoff A)) (k134 : Hvl) (k135 : Hvl) (ins : (shifthvl_A c62 k134 k135)) (b107 : (Index B)) ,
      (wfindex k134 b107) -> (wfindex k135 b107)).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma shift_A__wfindex_Z  :
    (forall (c62 : (Cutoff A)) (k134 : Hvl) (k135 : Hvl) (ins : (shifthvl_A c62 k134 k135)) (z71 : (Index Z)) ,
      (wfindex k134 z71) -> (wfindex k135 z71)).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma shift_B__wfindex_A  :
    (forall (c62 : (Cutoff B)) (k134 : Hvl) (k135 : Hvl) (ins : (shifthvl_B c62 k134 k135)) (a107 : (Index A)) ,
      (wfindex k134 a107) -> (wfindex k135 a107)).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma shift_B__wfindex_B  :
    (forall (c62 : (Cutoff B)) (k134 : Hvl) (k135 : Hvl) (ins : (shifthvl_B c62 k134 k135)) (b107 : (Index B)) ,
      (wfindex k134 b107) -> (wfindex k135 (shift_B_Index c62 b107))).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma shift_B__wfindex_Z  :
    (forall (c62 : (Cutoff B)) (k134 : Hvl) (k135 : Hvl) (ins : (shifthvl_B c62 k134 k135)) (z71 : (Index Z)) ,
      (wfindex k134 z71) -> (wfindex k135 z71)).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma shift_Z__wfindex_A  :
    (forall (c62 : (Cutoff Z)) (k134 : Hvl) (k135 : Hvl) (ins : (shifthvl_Z c62 k134 k135)) (a107 : (Index A)) ,
      (wfindex k134 a107) -> (wfindex k135 a107)).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma shift_Z__wfindex_B  :
    (forall (c62 : (Cutoff Z)) (k134 : Hvl) (k135 : Hvl) (ins : (shifthvl_Z c62 k134 k135)) (b107 : (Index B)) ,
      (wfindex k134 b107) -> (wfindex k135 b107)).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma shift_Z__wfindex_Z  :
    (forall (c62 : (Cutoff Z)) (k134 : Hvl) (k135 : Hvl) (ins : (shifthvl_Z c62 k134 k135)) (z71 : (Index Z)) ,
      (wfindex k134 z71) -> (wfindex k135 (shift_Z_Index c62 z71))).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Definition shift_A__wfAA_ZZ_BB : (forall (k134 : Hvl) ,
    (forall (aa74 : AA) (wf : (wfAA k134 aa74)) ,
      (forall (c62 : (Cutoff A)) (k135 : Hvl) ,
        (shifthvl_A c62 k134 k135) -> (wfAA k135 (shift_A_AA c62 aa74)))) /\
    (forall (zz110 : ZZ) (wf : (wfZZ k134 zz110)) ,
      (forall (c62 : (Cutoff A)) (k135 : Hvl) ,
        (shifthvl_A c62 k134 k135) -> (wfZZ k135 (shift_A_ZZ c62 zz110)))) /\
    (forall (bb74 : BB) (wf : (wfBB k134 bb74)) ,
      (forall (c62 : (Cutoff A)) (k135 : Hvl) ,
        (shifthvl_A c62 k134 k135) -> (wfBB k135 (shift_A_BB c62 bb74))))) := (ind_wfAA_ZZ_BB (fun (k134 : Hvl) (aa74 : AA) (wf : (wfAA k134 aa74)) =>
    (forall (c62 : (Cutoff A)) (k135 : Hvl) ,
      (shifthvl_A c62 k134 k135) -> (wfAA k135 (shift_A_AA c62 aa74)))) (fun (k134 : Hvl) (zz110 : ZZ) (wf : (wfZZ k134 zz110)) =>
    (forall (c62 : (Cutoff A)) (k135 : Hvl) ,
      (shifthvl_A c62 k134 k135) -> (wfZZ k135 (shift_A_ZZ c62 zz110)))) (fun (k134 : Hvl) (bb74 : BB) (wf : (wfBB k134 bb74)) =>
    (forall (c62 : (Cutoff A)) (k135 : Hvl) ,
      (shifthvl_A c62 k134 k135) -> (wfBB k135 (shift_A_BB c62 bb74)))) (fun (k134 : Hvl) (a107 : (Index A)) (wfi : (wfindex k134 a107)) (c62 : (Cutoff A)) (k135 : Hvl) (ins : (shifthvl_A c62 k134 k135)) =>
    (wfAA_avar k135 (shift_A__wfindex_A c62 k134 k135 ins a107 wfi))) (fun (k134 : Hvl) (zz : ZZ) (wf : (wfZZ k134 zz)) IHzz21 (c62 : (Cutoff A)) (k135 : Hvl) (ins : (shifthvl_A c62 k134 k135)) =>
    (wfAA_arec k135 (IHzz21 c62 k135 (weaken_shifthvl_A H0 ins)))) (fun (k134 : Hvl) (z71 : (Index Z)) (wfi : (wfindex k134 z71)) (c62 : (Cutoff A)) (k135 : Hvl) (ins : (shifthvl_A c62 k134 k135)) =>
    (wfZZ_zvar k135 (shift_A__wfindex_Z c62 k134 k135 ins z71 wfi))) (fun (k134 : Hvl) (aa : AA) (wf : (wfAA (HS A k134) aa)) IHaa8 (c62 : (Cutoff A)) (k135 : Hvl) (ins : (shifthvl_A c62 k134 k135)) =>
    (wfZZ_aabs k135 (IHaa8 (CS c62) (HS A k135) (weaken_shifthvl_A (HS A H0) ins)))) (fun (k134 : Hvl) (bb : BB) (wf : (wfBB (HS B k134) bb)) IHbb9 (c62 : (Cutoff A)) (k135 : Hvl) (ins : (shifthvl_A c62 k134 k135)) =>
    (wfZZ_babs k135 (IHbb9 c62 (HS B k135) (weaken_shifthvl_A (HS B H0) ins)))) (fun (k134 : Hvl) (b107 : (Index B)) (wfi : (wfindex k134 b107)) (c62 : (Cutoff A)) (k135 : Hvl) (ins : (shifthvl_A c62 k134 k135)) =>
    (wfBB_bvar k135 (shift_A__wfindex_B c62 k134 k135 ins b107 wfi))) (fun (k134 : Hvl) (zz : ZZ) (wf : (wfZZ k134 zz)) IHzz21 (c62 : (Cutoff A)) (k135 : Hvl) (ins : (shifthvl_A c62 k134 k135)) =>
    (wfBB_brec k135 (IHzz21 c62 k135 (weaken_shifthvl_A H0 ins))))).
  Lemma shift_A__wfAA (k134 : Hvl) :
    (forall (aa74 : AA) (wf : (wfAA k134 aa74)) ,
      (forall (c62 : (Cutoff A)) (k135 : Hvl) ,
        (shifthvl_A c62 k134 k135) -> (wfAA k135 (shift_A_AA c62 aa74)))).
  Proof.
    pose proof ((shift_A__wfAA_ZZ_BB k134)).
    destruct_conjs .
    easy .
  Qed.
  Lemma shift_A__wfZZ (k134 : Hvl) :
    (forall (zz110 : ZZ) (wf0 : (wfZZ k134 zz110)) ,
      (forall (c63 : (Cutoff A)) (k136 : Hvl) ,
        (shifthvl_A c63 k134 k136) -> (wfZZ k136 (shift_A_ZZ c63 zz110)))).
  Proof.
    pose proof ((shift_A__wfAA_ZZ_BB k134)).
    destruct_conjs .
    easy .
  Qed.
  Lemma shift_A__wfBB (k134 : Hvl) :
    (forall (bb74 : BB) (wf1 : (wfBB k134 bb74)) ,
      (forall (c64 : (Cutoff A)) (k137 : Hvl) ,
        (shifthvl_A c64 k134 k137) -> (wfBB k137 (shift_A_BB c64 bb74)))).
  Proof.
    pose proof ((shift_A__wfAA_ZZ_BB k134)).
    destruct_conjs .
    easy .
  Qed.
  Definition shift_B__wfAA_ZZ_BB : (forall (k134 : Hvl) ,
    (forall (aa74 : AA) (wf : (wfAA k134 aa74)) ,
      (forall (c62 : (Cutoff B)) (k135 : Hvl) ,
        (shifthvl_B c62 k134 k135) -> (wfAA k135 (shift_B_AA c62 aa74)))) /\
    (forall (zz110 : ZZ) (wf : (wfZZ k134 zz110)) ,
      (forall (c62 : (Cutoff B)) (k135 : Hvl) ,
        (shifthvl_B c62 k134 k135) -> (wfZZ k135 (shift_B_ZZ c62 zz110)))) /\
    (forall (bb74 : BB) (wf : (wfBB k134 bb74)) ,
      (forall (c62 : (Cutoff B)) (k135 : Hvl) ,
        (shifthvl_B c62 k134 k135) -> (wfBB k135 (shift_B_BB c62 bb74))))) := (ind_wfAA_ZZ_BB (fun (k134 : Hvl) (aa74 : AA) (wf : (wfAA k134 aa74)) =>
    (forall (c62 : (Cutoff B)) (k135 : Hvl) ,
      (shifthvl_B c62 k134 k135) -> (wfAA k135 (shift_B_AA c62 aa74)))) (fun (k134 : Hvl) (zz110 : ZZ) (wf : (wfZZ k134 zz110)) =>
    (forall (c62 : (Cutoff B)) (k135 : Hvl) ,
      (shifthvl_B c62 k134 k135) -> (wfZZ k135 (shift_B_ZZ c62 zz110)))) (fun (k134 : Hvl) (bb74 : BB) (wf : (wfBB k134 bb74)) =>
    (forall (c62 : (Cutoff B)) (k135 : Hvl) ,
      (shifthvl_B c62 k134 k135) -> (wfBB k135 (shift_B_BB c62 bb74)))) (fun (k134 : Hvl) (a107 : (Index A)) (wfi : (wfindex k134 a107)) (c62 : (Cutoff B)) (k135 : Hvl) (ins : (shifthvl_B c62 k134 k135)) =>
    (wfAA_avar k135 (shift_B__wfindex_A c62 k134 k135 ins a107 wfi))) (fun (k134 : Hvl) (zz : ZZ) (wf : (wfZZ k134 zz)) IHzz21 (c62 : (Cutoff B)) (k135 : Hvl) (ins : (shifthvl_B c62 k134 k135)) =>
    (wfAA_arec k135 (IHzz21 c62 k135 (weaken_shifthvl_B H0 ins)))) (fun (k134 : Hvl) (z71 : (Index Z)) (wfi : (wfindex k134 z71)) (c62 : (Cutoff B)) (k135 : Hvl) (ins : (shifthvl_B c62 k134 k135)) =>
    (wfZZ_zvar k135 (shift_B__wfindex_Z c62 k134 k135 ins z71 wfi))) (fun (k134 : Hvl) (aa : AA) (wf : (wfAA (HS A k134) aa)) IHaa8 (c62 : (Cutoff B)) (k135 : Hvl) (ins : (shifthvl_B c62 k134 k135)) =>
    (wfZZ_aabs k135 (IHaa8 c62 (HS A k135) (weaken_shifthvl_B (HS A H0) ins)))) (fun (k134 : Hvl) (bb : BB) (wf : (wfBB (HS B k134) bb)) IHbb9 (c62 : (Cutoff B)) (k135 : Hvl) (ins : (shifthvl_B c62 k134 k135)) =>
    (wfZZ_babs k135 (IHbb9 (CS c62) (HS B k135) (weaken_shifthvl_B (HS B H0) ins)))) (fun (k134 : Hvl) (b107 : (Index B)) (wfi : (wfindex k134 b107)) (c62 : (Cutoff B)) (k135 : Hvl) (ins : (shifthvl_B c62 k134 k135)) =>
    (wfBB_bvar k135 (shift_B__wfindex_B c62 k134 k135 ins b107 wfi))) (fun (k134 : Hvl) (zz : ZZ) (wf : (wfZZ k134 zz)) IHzz21 (c62 : (Cutoff B)) (k135 : Hvl) (ins : (shifthvl_B c62 k134 k135)) =>
    (wfBB_brec k135 (IHzz21 c62 k135 (weaken_shifthvl_B H0 ins))))).
  Lemma shift_B__wfAA (k134 : Hvl) :
    (forall (aa74 : AA) (wf : (wfAA k134 aa74)) ,
      (forall (c62 : (Cutoff B)) (k135 : Hvl) ,
        (shifthvl_B c62 k134 k135) -> (wfAA k135 (shift_B_AA c62 aa74)))).
  Proof.
    pose proof ((shift_B__wfAA_ZZ_BB k134)).
    destruct_conjs .
    easy .
  Qed.
  Lemma shift_B__wfZZ (k134 : Hvl) :
    (forall (zz110 : ZZ) (wf0 : (wfZZ k134 zz110)) ,
      (forall (c63 : (Cutoff B)) (k136 : Hvl) ,
        (shifthvl_B c63 k134 k136) -> (wfZZ k136 (shift_B_ZZ c63 zz110)))).
  Proof.
    pose proof ((shift_B__wfAA_ZZ_BB k134)).
    destruct_conjs .
    easy .
  Qed.
  Lemma shift_B__wfBB (k134 : Hvl) :
    (forall (bb74 : BB) (wf1 : (wfBB k134 bb74)) ,
      (forall (c64 : (Cutoff B)) (k137 : Hvl) ,
        (shifthvl_B c64 k134 k137) -> (wfBB k137 (shift_B_BB c64 bb74)))).
  Proof.
    pose proof ((shift_B__wfAA_ZZ_BB k134)).
    destruct_conjs .
    easy .
  Qed.
  Definition shift_Z__wfAA_ZZ_BB : (forall (k134 : Hvl) ,
    (forall (aa74 : AA) (wf : (wfAA k134 aa74)) ,
      (forall (c62 : (Cutoff Z)) (k135 : Hvl) ,
        (shifthvl_Z c62 k134 k135) -> (wfAA k135 (shift_Z_AA c62 aa74)))) /\
    (forall (zz110 : ZZ) (wf : (wfZZ k134 zz110)) ,
      (forall (c62 : (Cutoff Z)) (k135 : Hvl) ,
        (shifthvl_Z c62 k134 k135) -> (wfZZ k135 (shift_Z_ZZ c62 zz110)))) /\
    (forall (bb74 : BB) (wf : (wfBB k134 bb74)) ,
      (forall (c62 : (Cutoff Z)) (k135 : Hvl) ,
        (shifthvl_Z c62 k134 k135) -> (wfBB k135 (shift_Z_BB c62 bb74))))) := (ind_wfAA_ZZ_BB (fun (k134 : Hvl) (aa74 : AA) (wf : (wfAA k134 aa74)) =>
    (forall (c62 : (Cutoff Z)) (k135 : Hvl) ,
      (shifthvl_Z c62 k134 k135) -> (wfAA k135 (shift_Z_AA c62 aa74)))) (fun (k134 : Hvl) (zz110 : ZZ) (wf : (wfZZ k134 zz110)) =>
    (forall (c62 : (Cutoff Z)) (k135 : Hvl) ,
      (shifthvl_Z c62 k134 k135) -> (wfZZ k135 (shift_Z_ZZ c62 zz110)))) (fun (k134 : Hvl) (bb74 : BB) (wf : (wfBB k134 bb74)) =>
    (forall (c62 : (Cutoff Z)) (k135 : Hvl) ,
      (shifthvl_Z c62 k134 k135) -> (wfBB k135 (shift_Z_BB c62 bb74)))) (fun (k134 : Hvl) (a107 : (Index A)) (wfi : (wfindex k134 a107)) (c62 : (Cutoff Z)) (k135 : Hvl) (ins : (shifthvl_Z c62 k134 k135)) =>
    (wfAA_avar k135 (shift_Z__wfindex_A c62 k134 k135 ins a107 wfi))) (fun (k134 : Hvl) (zz : ZZ) (wf : (wfZZ k134 zz)) IHzz21 (c62 : (Cutoff Z)) (k135 : Hvl) (ins : (shifthvl_Z c62 k134 k135)) =>
    (wfAA_arec k135 (IHzz21 c62 k135 (weaken_shifthvl_Z H0 ins)))) (fun (k134 : Hvl) (z71 : (Index Z)) (wfi : (wfindex k134 z71)) (c62 : (Cutoff Z)) (k135 : Hvl) (ins : (shifthvl_Z c62 k134 k135)) =>
    (wfZZ_zvar k135 (shift_Z__wfindex_Z c62 k134 k135 ins z71 wfi))) (fun (k134 : Hvl) (aa : AA) (wf : (wfAA (HS A k134) aa)) IHaa8 (c62 : (Cutoff Z)) (k135 : Hvl) (ins : (shifthvl_Z c62 k134 k135)) =>
    (wfZZ_aabs k135 (IHaa8 c62 (HS A k135) (weaken_shifthvl_Z (HS A H0) ins)))) (fun (k134 : Hvl) (bb : BB) (wf : (wfBB (HS B k134) bb)) IHbb9 (c62 : (Cutoff Z)) (k135 : Hvl) (ins : (shifthvl_Z c62 k134 k135)) =>
    (wfZZ_babs k135 (IHbb9 c62 (HS B k135) (weaken_shifthvl_Z (HS B H0) ins)))) (fun (k134 : Hvl) (b107 : (Index B)) (wfi : (wfindex k134 b107)) (c62 : (Cutoff Z)) (k135 : Hvl) (ins : (shifthvl_Z c62 k134 k135)) =>
    (wfBB_bvar k135 (shift_Z__wfindex_B c62 k134 k135 ins b107 wfi))) (fun (k134 : Hvl) (zz : ZZ) (wf : (wfZZ k134 zz)) IHzz21 (c62 : (Cutoff Z)) (k135 : Hvl) (ins : (shifthvl_Z c62 k134 k135)) =>
    (wfBB_brec k135 (IHzz21 c62 k135 (weaken_shifthvl_Z H0 ins))))).
  Lemma shift_Z__wfAA (k134 : Hvl) :
    (forall (aa74 : AA) (wf : (wfAA k134 aa74)) ,
      (forall (c62 : (Cutoff Z)) (k135 : Hvl) ,
        (shifthvl_Z c62 k134 k135) -> (wfAA k135 (shift_Z_AA c62 aa74)))).
  Proof.
    pose proof ((shift_Z__wfAA_ZZ_BB k134)).
    destruct_conjs .
    easy .
  Qed.
  Lemma shift_Z__wfZZ (k134 : Hvl) :
    (forall (zz110 : ZZ) (wf0 : (wfZZ k134 zz110)) ,
      (forall (c63 : (Cutoff Z)) (k136 : Hvl) ,
        (shifthvl_Z c63 k134 k136) -> (wfZZ k136 (shift_Z_ZZ c63 zz110)))).
  Proof.
    pose proof ((shift_Z__wfAA_ZZ_BB k134)).
    destruct_conjs .
    easy .
  Qed.
  Lemma shift_Z__wfBB (k134 : Hvl) :
    (forall (bb74 : BB) (wf1 : (wfBB k134 bb74)) ,
      (forall (c64 : (Cutoff Z)) (k137 : Hvl) ,
        (shifthvl_Z c64 k134 k137) -> (wfBB k137 (shift_Z_BB c64 bb74)))).
  Proof.
    pose proof ((shift_Z__wfAA_ZZ_BB k134)).
    destruct_conjs .
    easy .
  Qed.
End ShiftWellFormed.
 Hint Resolve shift_A__wfindex_A shift_A__wfindex_B shift_A__wfindex_Z shift_B__wfindex_A shift_B__wfindex_B shift_B__wfindex_Z shift_Z__wfindex_A shift_Z__wfindex_B shift_Z__wfindex_Z : infra.
 Hint Resolve shift_A__wfindex_A shift_A__wfindex_B shift_A__wfindex_Z shift_B__wfindex_A shift_B__wfindex_B shift_B__wfindex_Z shift_Z__wfindex_A shift_Z__wfindex_B shift_Z__wfindex_Z : shift.
 Hint Resolve shift_A__wfindex_A shift_A__wfindex_B shift_A__wfindex_Z shift_B__wfindex_A shift_B__wfindex_B shift_B__wfindex_Z shift_Z__wfindex_A shift_Z__wfindex_B shift_Z__wfindex_Z : shift_wf.
 Hint Resolve shift_A__wfindex_A shift_A__wfindex_B shift_A__wfindex_Z shift_B__wfindex_A shift_B__wfindex_B shift_B__wfindex_Z shift_Z__wfindex_A shift_Z__wfindex_B shift_Z__wfindex_Z : wf.
 Hint Resolve shift_A__wfAA shift_A__wfBB shift_A__wfZZ shift_B__wfAA shift_B__wfBB shift_B__wfZZ shift_Z__wfAA shift_Z__wfBB shift_Z__wfZZ : infra.
 Hint Resolve shift_A__wfAA shift_A__wfBB shift_A__wfZZ shift_B__wfAA shift_B__wfBB shift_B__wfZZ shift_Z__wfAA shift_Z__wfBB shift_Z__wfZZ : shift.
 Hint Resolve shift_A__wfAA shift_A__wfBB shift_A__wfZZ shift_B__wfAA shift_B__wfBB shift_B__wfZZ shift_Z__wfAA shift_Z__wfBB shift_Z__wfZZ : shift_wf.
 Hint Resolve shift_A__wfAA shift_A__wfBB shift_A__wfZZ shift_B__wfAA shift_B__wfBB shift_B__wfZZ shift_Z__wfAA shift_Z__wfBB shift_Z__wfZZ : wf.
 Hint Constructors shifthvl_A shifthvl_B shifthvl_Z : infra.
 Hint Constructors shifthvl_A shifthvl_B shifthvl_Z : shift.
 Hint Constructors shifthvl_A shifthvl_B shifthvl_Z : shift_wf.
 Hint Constructors shifthvl_A shifthvl_B shifthvl_Z : wf.
 Hint Resolve weaken_shifthvl_A weaken_shifthvl_B weaken_shifthvl_Z : infra.
 Hint Resolve weaken_shifthvl_A weaken_shifthvl_B weaken_shifthvl_Z : shift.
 Hint Resolve weaken_shifthvl_A weaken_shifthvl_B weaken_shifthvl_Z : shift_wf.
 Hint Resolve weaken_shifthvl_A weaken_shifthvl_B weaken_shifthvl_Z : weaken.
 Hint Resolve weaken_shifthvl_A weaken_shifthvl_B weaken_shifthvl_Z : wf.
Section SubstWellFormed.
  Inductive substhvl_A (k134 : Hvl) : (forall (d71 : (Trace A)) (k135 : Hvl) (k136 : Hvl) ,
    Prop) :=
    | substhvl_A_here :
        (substhvl_A k134 X0 (HS A k134) k134)
    | substhvl_A_there_A
        {d71 : (Trace A)} {k135 : Hvl}
        {k136 : Hvl} :
        (substhvl_A k134 d71 k135 k136) -> (substhvl_A k134 (XS A d71) (HS A k135) (HS A k136))
    | substhvl_A_there_B
        {d71 : (Trace A)} {k135 : Hvl}
        {k136 : Hvl} :
        (substhvl_A k134 d71 k135 k136) -> (substhvl_A k134 (XS B d71) (HS B k135) (HS B k136))
    | substhvl_A_there_Z
        {d71 : (Trace A)} {k135 : Hvl}
        {k136 : Hvl} :
        (substhvl_A k134 d71 k135 k136) -> (substhvl_A k134 (XS Z d71) (HS Z k135) (HS Z k136)).
  Inductive substhvl_B (k134 : Hvl) : (forall (d71 : (Trace B)) (k135 : Hvl) (k136 : Hvl) ,
    Prop) :=
    | substhvl_B_here :
        (substhvl_B k134 X0 (HS B k134) k134)
    | substhvl_B_there_A
        {d71 : (Trace B)} {k135 : Hvl}
        {k136 : Hvl} :
        (substhvl_B k134 d71 k135 k136) -> (substhvl_B k134 (XS A d71) (HS A k135) (HS A k136))
    | substhvl_B_there_B
        {d71 : (Trace B)} {k135 : Hvl}
        {k136 : Hvl} :
        (substhvl_B k134 d71 k135 k136) -> (substhvl_B k134 (XS B d71) (HS B k135) (HS B k136))
    | substhvl_B_there_Z
        {d71 : (Trace B)} {k135 : Hvl}
        {k136 : Hvl} :
        (substhvl_B k134 d71 k135 k136) -> (substhvl_B k134 (XS Z d71) (HS Z k135) (HS Z k136)).
  Inductive substhvl_Z (k134 : Hvl) : (forall (d71 : (Trace Z)) (k135 : Hvl) (k136 : Hvl) ,
    Prop) :=
    | substhvl_Z_here :
        (substhvl_Z k134 X0 (HS Z k134) k134)
    | substhvl_Z_there_A
        {d71 : (Trace Z)} {k135 : Hvl}
        {k136 : Hvl} :
        (substhvl_Z k134 d71 k135 k136) -> (substhvl_Z k134 (XS A d71) (HS A k135) (HS A k136))
    | substhvl_Z_there_B
        {d71 : (Trace Z)} {k135 : Hvl}
        {k136 : Hvl} :
        (substhvl_Z k134 d71 k135 k136) -> (substhvl_Z k134 (XS B d71) (HS B k135) (HS B k136))
    | substhvl_Z_there_Z
        {d71 : (Trace Z)} {k135 : Hvl}
        {k136 : Hvl} :
        (substhvl_Z k134 d71 k135 k136) -> (substhvl_Z k134 (XS Z d71) (HS Z k135) (HS Z k136)).
  Lemma weaken_substhvl_A  :
    (forall {k135 : Hvl} (k134 : Hvl) {d71 : (Trace A)} {k136 : Hvl} {k137 : Hvl} ,
      (substhvl_A k135 d71 k136 k137) -> (substhvl_A k135 (weakenTrace d71 k134) (appendHvl k136 k134) (appendHvl k137 k134))).
  Proof.
    needleGenericWeakenSubstHvl .
  Qed.
  Lemma weaken_substhvl_B  :
    (forall {k135 : Hvl} (k134 : Hvl) {d71 : (Trace B)} {k136 : Hvl} {k137 : Hvl} ,
      (substhvl_B k135 d71 k136 k137) -> (substhvl_B k135 (weakenTrace d71 k134) (appendHvl k136 k134) (appendHvl k137 k134))).
  Proof.
    needleGenericWeakenSubstHvl .
  Qed.
  Lemma weaken_substhvl_Z  :
    (forall {k135 : Hvl} (k134 : Hvl) {d71 : (Trace Z)} {k136 : Hvl} {k137 : Hvl} ,
      (substhvl_Z k135 d71 k136 k137) -> (substhvl_Z k135 (weakenTrace d71 k134) (appendHvl k136 k134) (appendHvl k137 k134))).
  Proof.
    needleGenericWeakenSubstHvl .
  Qed.
  Lemma substhvl_A_wfindex_A {k134 : Hvl} {aa74 : AA} (wft : (wfAA k134 aa74)) :
    (forall {d71 : (Trace A)} {k135 : Hvl} {k136 : Hvl} ,
      (substhvl_A k134 d71 k135 k136) -> (forall {a107 : (Index A)} ,
        (wfindex k135 a107) -> (wfAA k136 (subst_A_Index d71 aa74 a107)))).
  Proof.
    needleGenericSubstHvlWfIndexHom .
  Qed.
  Lemma substhvl_B_wfindex_B {k134 : Hvl} {bb74 : BB} (wft : (wfBB k134 bb74)) :
    (forall {d71 : (Trace B)} {k135 : Hvl} {k136 : Hvl} ,
      (substhvl_B k134 d71 k135 k136) -> (forall {b107 : (Index B)} ,
        (wfindex k135 b107) -> (wfBB k136 (subst_B_Index d71 bb74 b107)))).
  Proof.
    needleGenericSubstHvlWfIndexHom .
  Qed.
  Lemma substhvl_Z_wfindex_Z {k134 : Hvl} {zz110 : ZZ} (wft : (wfZZ k134 zz110)) :
    (forall {d71 : (Trace Z)} {k135 : Hvl} {k136 : Hvl} ,
      (substhvl_Z k134 d71 k135 k136) -> (forall {z71 : (Index Z)} ,
        (wfindex k135 z71) -> (wfZZ k136 (subst_Z_Index d71 zz110 z71)))).
  Proof.
    needleGenericSubstHvlWfIndexHom .
  Qed.
  Lemma substhvl_A_wfindex_B {k134 : Hvl} :
    (forall {d71 : (Trace A)} {k135 : Hvl} {k136 : Hvl} ,
      (substhvl_A k134 d71 k135 k136) -> (forall {b107 : (Index B)} ,
        (wfindex k135 b107) -> (wfindex k136 b107))).
  Proof.
    needleGenericSubstHvlWfIndexHet .
  Qed.
  Lemma substhvl_A_wfindex_Z {k134 : Hvl} :
    (forall {d71 : (Trace A)} {k135 : Hvl} {k136 : Hvl} ,
      (substhvl_A k134 d71 k135 k136) -> (forall {z71 : (Index Z)} ,
        (wfindex k135 z71) -> (wfindex k136 z71))).
  Proof.
    needleGenericSubstHvlWfIndexHet .
  Qed.
  Lemma substhvl_B_wfindex_A {k134 : Hvl} :
    (forall {d71 : (Trace B)} {k135 : Hvl} {k136 : Hvl} ,
      (substhvl_B k134 d71 k135 k136) -> (forall {a107 : (Index A)} ,
        (wfindex k135 a107) -> (wfindex k136 a107))).
  Proof.
    needleGenericSubstHvlWfIndexHet .
  Qed.
  Lemma substhvl_B_wfindex_Z {k134 : Hvl} :
    (forall {d71 : (Trace B)} {k135 : Hvl} {k136 : Hvl} ,
      (substhvl_B k134 d71 k135 k136) -> (forall {z71 : (Index Z)} ,
        (wfindex k135 z71) -> (wfindex k136 z71))).
  Proof.
    needleGenericSubstHvlWfIndexHet .
  Qed.
  Lemma substhvl_Z_wfindex_A {k134 : Hvl} :
    (forall {d71 : (Trace Z)} {k135 : Hvl} {k136 : Hvl} ,
      (substhvl_Z k134 d71 k135 k136) -> (forall {a107 : (Index A)} ,
        (wfindex k135 a107) -> (wfindex k136 a107))).
  Proof.
    needleGenericSubstHvlWfIndexHet .
  Qed.
  Lemma substhvl_Z_wfindex_B {k134 : Hvl} :
    (forall {d71 : (Trace Z)} {k135 : Hvl} {k136 : Hvl} ,
      (substhvl_Z k134 d71 k135 k136) -> (forall {b107 : (Index B)} ,
        (wfindex k135 b107) -> (wfindex k136 b107))).
  Proof.
    needleGenericSubstHvlWfIndexHet .
  Qed.
  Definition substhvl_A_wfAA_ZZ_BB {k134 : Hvl} {aa74 : AA} (wf : (wfAA k134 aa74)) : (forall (k135 : Hvl) ,
    (forall (aa75 : AA) (wf0 : (wfAA k135 aa75)) ,
      (forall {d71 : (Trace A)} {k136 : Hvl} ,
        (substhvl_A k134 d71 k135 k136) -> (wfAA k136 (subst_A_AA d71 aa74 aa75)))) /\
    (forall (zz110 : ZZ) (wf0 : (wfZZ k135 zz110)) ,
      (forall {d71 : (Trace A)} {k136 : Hvl} ,
        (substhvl_A k134 d71 k135 k136) -> (wfZZ k136 (subst_A_ZZ d71 aa74 zz110)))) /\
    (forall (bb74 : BB) (wf0 : (wfBB k135 bb74)) ,
      (forall {d71 : (Trace A)} {k136 : Hvl} ,
        (substhvl_A k134 d71 k135 k136) -> (wfBB k136 (subst_A_BB d71 aa74 bb74))))) := (ind_wfAA_ZZ_BB (fun (k135 : Hvl) (aa75 : AA) (wf0 : (wfAA k135 aa75)) =>
    (forall {d71 : (Trace A)} {k136 : Hvl} ,
      (substhvl_A k134 d71 k135 k136) -> (wfAA k136 (subst_A_AA d71 aa74 aa75)))) (fun (k135 : Hvl) (zz110 : ZZ) (wf0 : (wfZZ k135 zz110)) =>
    (forall {d71 : (Trace A)} {k136 : Hvl} ,
      (substhvl_A k134 d71 k135 k136) -> (wfZZ k136 (subst_A_ZZ d71 aa74 zz110)))) (fun (k135 : Hvl) (bb74 : BB) (wf0 : (wfBB k135 bb74)) =>
    (forall {d71 : (Trace A)} {k136 : Hvl} ,
      (substhvl_A k134 d71 k135 k136) -> (wfBB k136 (subst_A_BB d71 aa74 bb74)))) (fun (k135 : Hvl) {a107 : (Index A)} (wfi : (wfindex k135 a107)) {d71 : (Trace A)} {k136 : Hvl} (del : (substhvl_A k134 d71 k135 k136)) =>
    (substhvl_A_wfindex_A wf del wfi)) (fun (k135 : Hvl) (zz : ZZ) (wf0 : (wfZZ k135 zz)) IHzz21 {d71 : (Trace A)} {k136 : Hvl} (del : (substhvl_A k134 d71 k135 k136)) =>
    (wfAA_arec k136 (IHzz21 (weakenTrace d71 H0) k136 (weaken_substhvl_A H0 del)))) (fun (k135 : Hvl) {z71 : (Index Z)} (wfi : (wfindex k135 z71)) {d71 : (Trace A)} {k136 : Hvl} (del : (substhvl_A k134 d71 k135 k136)) =>
    (wfZZ_zvar k136 (substhvl_A_wfindex_Z del wfi))) (fun (k135 : Hvl) (aa : AA) (wf0 : (wfAA (HS A k135) aa)) IHaa8 {d71 : (Trace A)} {k136 : Hvl} (del : (substhvl_A k134 d71 k135 k136)) =>
    (wfZZ_aabs k136 (IHaa8 (weakenTrace d71 (HS A H0)) (HS A k136) (weaken_substhvl_A (HS A H0) del)))) (fun (k135 : Hvl) (bb : BB) (wf0 : (wfBB (HS B k135) bb)) IHbb9 {d71 : (Trace A)} {k136 : Hvl} (del : (substhvl_A k134 d71 k135 k136)) =>
    (wfZZ_babs k136 (IHbb9 (weakenTrace d71 (HS B H0)) (HS B k136) (weaken_substhvl_A (HS B H0) del)))) (fun (k135 : Hvl) {b107 : (Index B)} (wfi : (wfindex k135 b107)) {d71 : (Trace A)} {k136 : Hvl} (del : (substhvl_A k134 d71 k135 k136)) =>
    (wfBB_bvar k136 (substhvl_A_wfindex_B del wfi))) (fun (k135 : Hvl) (zz : ZZ) (wf0 : (wfZZ k135 zz)) IHzz21 {d71 : (Trace A)} {k136 : Hvl} (del : (substhvl_A k134 d71 k135 k136)) =>
    (wfBB_brec k136 (IHzz21 (weakenTrace d71 H0) k136 (weaken_substhvl_A H0 del))))).
  Lemma substhvl_A_wfAA {k134 : Hvl} {aa74 : AA} (wf : (wfAA k134 aa74)) (k135 : Hvl) (aa75 : AA) (wf0 : (wfAA k135 aa75)) :
    (forall {d71 : (Trace A)} {k136 : Hvl} ,
      (substhvl_A k134 d71 k135 k136) -> (wfAA k136 (subst_A_AA d71 aa74 aa75))).
  Proof.
    apply ((substhvl_A_wfAA_ZZ_BB wf k135)).
    auto .
  Qed.
  Lemma substhvl_A_wfZZ {k134 : Hvl} {aa74 : AA} (wf : (wfAA k134 aa74)) (k135 : Hvl) (zz110 : ZZ) (wf1 : (wfZZ k135 zz110)) :
    (forall {d72 : (Trace A)} {k137 : Hvl} ,
      (substhvl_A k134 d72 k135 k137) -> (wfZZ k137 (subst_A_ZZ d72 aa74 zz110))).
  Proof.
    apply ((substhvl_A_wfAA_ZZ_BB wf k135)).
    auto .
  Qed.
  Lemma substhvl_A_wfBB {k134 : Hvl} {aa74 : AA} (wf : (wfAA k134 aa74)) (k135 : Hvl) (bb74 : BB) (wf2 : (wfBB k135 bb74)) :
    (forall {d73 : (Trace A)} {k138 : Hvl} ,
      (substhvl_A k134 d73 k135 k138) -> (wfBB k138 (subst_A_BB d73 aa74 bb74))).
  Proof.
    apply ((substhvl_A_wfAA_ZZ_BB wf k135)).
    auto .
  Qed.
  Definition substhvl_B_wfAA_ZZ_BB {k134 : Hvl} {bb74 : BB} (wf : (wfBB k134 bb74)) : (forall (k135 : Hvl) ,
    (forall (aa74 : AA) (wf0 : (wfAA k135 aa74)) ,
      (forall {d71 : (Trace B)} {k136 : Hvl} ,
        (substhvl_B k134 d71 k135 k136) -> (wfAA k136 (subst_B_AA d71 bb74 aa74)))) /\
    (forall (zz110 : ZZ) (wf0 : (wfZZ k135 zz110)) ,
      (forall {d71 : (Trace B)} {k136 : Hvl} ,
        (substhvl_B k134 d71 k135 k136) -> (wfZZ k136 (subst_B_ZZ d71 bb74 zz110)))) /\
    (forall (bb75 : BB) (wf0 : (wfBB k135 bb75)) ,
      (forall {d71 : (Trace B)} {k136 : Hvl} ,
        (substhvl_B k134 d71 k135 k136) -> (wfBB k136 (subst_B_BB d71 bb74 bb75))))) := (ind_wfAA_ZZ_BB (fun (k135 : Hvl) (aa74 : AA) (wf0 : (wfAA k135 aa74)) =>
    (forall {d71 : (Trace B)} {k136 : Hvl} ,
      (substhvl_B k134 d71 k135 k136) -> (wfAA k136 (subst_B_AA d71 bb74 aa74)))) (fun (k135 : Hvl) (zz110 : ZZ) (wf0 : (wfZZ k135 zz110)) =>
    (forall {d71 : (Trace B)} {k136 : Hvl} ,
      (substhvl_B k134 d71 k135 k136) -> (wfZZ k136 (subst_B_ZZ d71 bb74 zz110)))) (fun (k135 : Hvl) (bb75 : BB) (wf0 : (wfBB k135 bb75)) =>
    (forall {d71 : (Trace B)} {k136 : Hvl} ,
      (substhvl_B k134 d71 k135 k136) -> (wfBB k136 (subst_B_BB d71 bb74 bb75)))) (fun (k135 : Hvl) {a107 : (Index A)} (wfi : (wfindex k135 a107)) {d71 : (Trace B)} {k136 : Hvl} (del : (substhvl_B k134 d71 k135 k136)) =>
    (wfAA_avar k136 (substhvl_B_wfindex_A del wfi))) (fun (k135 : Hvl) (zz : ZZ) (wf0 : (wfZZ k135 zz)) IHzz21 {d71 : (Trace B)} {k136 : Hvl} (del : (substhvl_B k134 d71 k135 k136)) =>
    (wfAA_arec k136 (IHzz21 (weakenTrace d71 H0) k136 (weaken_substhvl_B H0 del)))) (fun (k135 : Hvl) {z71 : (Index Z)} (wfi : (wfindex k135 z71)) {d71 : (Trace B)} {k136 : Hvl} (del : (substhvl_B k134 d71 k135 k136)) =>
    (wfZZ_zvar k136 (substhvl_B_wfindex_Z del wfi))) (fun (k135 : Hvl) (aa : AA) (wf0 : (wfAA (HS A k135) aa)) IHaa8 {d71 : (Trace B)} {k136 : Hvl} (del : (substhvl_B k134 d71 k135 k136)) =>
    (wfZZ_aabs k136 (IHaa8 (weakenTrace d71 (HS A H0)) (HS A k136) (weaken_substhvl_B (HS A H0) del)))) (fun (k135 : Hvl) (bb : BB) (wf0 : (wfBB (HS B k135) bb)) IHbb9 {d71 : (Trace B)} {k136 : Hvl} (del : (substhvl_B k134 d71 k135 k136)) =>
    (wfZZ_babs k136 (IHbb9 (weakenTrace d71 (HS B H0)) (HS B k136) (weaken_substhvl_B (HS B H0) del)))) (fun (k135 : Hvl) {b107 : (Index B)} (wfi : (wfindex k135 b107)) {d71 : (Trace B)} {k136 : Hvl} (del : (substhvl_B k134 d71 k135 k136)) =>
    (substhvl_B_wfindex_B wf del wfi)) (fun (k135 : Hvl) (zz : ZZ) (wf0 : (wfZZ k135 zz)) IHzz21 {d71 : (Trace B)} {k136 : Hvl} (del : (substhvl_B k134 d71 k135 k136)) =>
    (wfBB_brec k136 (IHzz21 (weakenTrace d71 H0) k136 (weaken_substhvl_B H0 del))))).
  Lemma substhvl_B_wfAA {k134 : Hvl} {bb74 : BB} (wf : (wfBB k134 bb74)) (k135 : Hvl) (aa74 : AA) (wf0 : (wfAA k135 aa74)) :
    (forall {d71 : (Trace B)} {k136 : Hvl} ,
      (substhvl_B k134 d71 k135 k136) -> (wfAA k136 (subst_B_AA d71 bb74 aa74))).
  Proof.
    apply ((substhvl_B_wfAA_ZZ_BB wf k135)).
    auto .
  Qed.
  Lemma substhvl_B_wfZZ {k134 : Hvl} {bb74 : BB} (wf : (wfBB k134 bb74)) (k135 : Hvl) (zz110 : ZZ) (wf1 : (wfZZ k135 zz110)) :
    (forall {d72 : (Trace B)} {k137 : Hvl} ,
      (substhvl_B k134 d72 k135 k137) -> (wfZZ k137 (subst_B_ZZ d72 bb74 zz110))).
  Proof.
    apply ((substhvl_B_wfAA_ZZ_BB wf k135)).
    auto .
  Qed.
  Lemma substhvl_B_wfBB {k134 : Hvl} {bb74 : BB} (wf : (wfBB k134 bb74)) (k135 : Hvl) (bb75 : BB) (wf2 : (wfBB k135 bb75)) :
    (forall {d73 : (Trace B)} {k138 : Hvl} ,
      (substhvl_B k134 d73 k135 k138) -> (wfBB k138 (subst_B_BB d73 bb74 bb75))).
  Proof.
    apply ((substhvl_B_wfAA_ZZ_BB wf k135)).
    auto .
  Qed.
  Definition substhvl_Z_wfAA_ZZ_BB {k134 : Hvl} {zz110 : ZZ} (wf : (wfZZ k134 zz110)) : (forall (k135 : Hvl) ,
    (forall (aa74 : AA) (wf0 : (wfAA k135 aa74)) ,
      (forall {d71 : (Trace Z)} {k136 : Hvl} ,
        (substhvl_Z k134 d71 k135 k136) -> (wfAA k136 (subst_Z_AA d71 zz110 aa74)))) /\
    (forall (zz111 : ZZ) (wf0 : (wfZZ k135 zz111)) ,
      (forall {d71 : (Trace Z)} {k136 : Hvl} ,
        (substhvl_Z k134 d71 k135 k136) -> (wfZZ k136 (subst_Z_ZZ d71 zz110 zz111)))) /\
    (forall (bb74 : BB) (wf0 : (wfBB k135 bb74)) ,
      (forall {d71 : (Trace Z)} {k136 : Hvl} ,
        (substhvl_Z k134 d71 k135 k136) -> (wfBB k136 (subst_Z_BB d71 zz110 bb74))))) := (ind_wfAA_ZZ_BB (fun (k135 : Hvl) (aa74 : AA) (wf0 : (wfAA k135 aa74)) =>
    (forall {d71 : (Trace Z)} {k136 : Hvl} ,
      (substhvl_Z k134 d71 k135 k136) -> (wfAA k136 (subst_Z_AA d71 zz110 aa74)))) (fun (k135 : Hvl) (zz111 : ZZ) (wf0 : (wfZZ k135 zz111)) =>
    (forall {d71 : (Trace Z)} {k136 : Hvl} ,
      (substhvl_Z k134 d71 k135 k136) -> (wfZZ k136 (subst_Z_ZZ d71 zz110 zz111)))) (fun (k135 : Hvl) (bb74 : BB) (wf0 : (wfBB k135 bb74)) =>
    (forall {d71 : (Trace Z)} {k136 : Hvl} ,
      (substhvl_Z k134 d71 k135 k136) -> (wfBB k136 (subst_Z_BB d71 zz110 bb74)))) (fun (k135 : Hvl) {a107 : (Index A)} (wfi : (wfindex k135 a107)) {d71 : (Trace Z)} {k136 : Hvl} (del : (substhvl_Z k134 d71 k135 k136)) =>
    (wfAA_avar k136 (substhvl_Z_wfindex_A del wfi))) (fun (k135 : Hvl) (zz : ZZ) (wf0 : (wfZZ k135 zz)) IHzz21 {d71 : (Trace Z)} {k136 : Hvl} (del : (substhvl_Z k134 d71 k135 k136)) =>
    (wfAA_arec k136 (IHzz21 (weakenTrace d71 H0) k136 (weaken_substhvl_Z H0 del)))) (fun (k135 : Hvl) {z71 : (Index Z)} (wfi : (wfindex k135 z71)) {d71 : (Trace Z)} {k136 : Hvl} (del : (substhvl_Z k134 d71 k135 k136)) =>
    (substhvl_Z_wfindex_Z wf del wfi)) (fun (k135 : Hvl) (aa : AA) (wf0 : (wfAA (HS A k135) aa)) IHaa8 {d71 : (Trace Z)} {k136 : Hvl} (del : (substhvl_Z k134 d71 k135 k136)) =>
    (wfZZ_aabs k136 (IHaa8 (weakenTrace d71 (HS A H0)) (HS A k136) (weaken_substhvl_Z (HS A H0) del)))) (fun (k135 : Hvl) (bb : BB) (wf0 : (wfBB (HS B k135) bb)) IHbb9 {d71 : (Trace Z)} {k136 : Hvl} (del : (substhvl_Z k134 d71 k135 k136)) =>
    (wfZZ_babs k136 (IHbb9 (weakenTrace d71 (HS B H0)) (HS B k136) (weaken_substhvl_Z (HS B H0) del)))) (fun (k135 : Hvl) {b107 : (Index B)} (wfi : (wfindex k135 b107)) {d71 : (Trace Z)} {k136 : Hvl} (del : (substhvl_Z k134 d71 k135 k136)) =>
    (wfBB_bvar k136 (substhvl_Z_wfindex_B del wfi))) (fun (k135 : Hvl) (zz : ZZ) (wf0 : (wfZZ k135 zz)) IHzz21 {d71 : (Trace Z)} {k136 : Hvl} (del : (substhvl_Z k134 d71 k135 k136)) =>
    (wfBB_brec k136 (IHzz21 (weakenTrace d71 H0) k136 (weaken_substhvl_Z H0 del))))).
  Lemma substhvl_Z_wfAA {k134 : Hvl} {zz110 : ZZ} (wf : (wfZZ k134 zz110)) (k135 : Hvl) (aa74 : AA) (wf0 : (wfAA k135 aa74)) :
    (forall {d71 : (Trace Z)} {k136 : Hvl} ,
      (substhvl_Z k134 d71 k135 k136) -> (wfAA k136 (subst_Z_AA d71 zz110 aa74))).
  Proof.
    apply ((substhvl_Z_wfAA_ZZ_BB wf k135)).
    auto .
  Qed.
  Lemma substhvl_Z_wfZZ {k134 : Hvl} {zz110 : ZZ} (wf : (wfZZ k134 zz110)) (k135 : Hvl) (zz111 : ZZ) (wf1 : (wfZZ k135 zz111)) :
    (forall {d72 : (Trace Z)} {k137 : Hvl} ,
      (substhvl_Z k134 d72 k135 k137) -> (wfZZ k137 (subst_Z_ZZ d72 zz110 zz111))).
  Proof.
    apply ((substhvl_Z_wfAA_ZZ_BB wf k135)).
    auto .
  Qed.
  Lemma substhvl_Z_wfBB {k134 : Hvl} {zz110 : ZZ} (wf : (wfZZ k134 zz110)) (k135 : Hvl) (bb74 : BB) (wf2 : (wfBB k135 bb74)) :
    (forall {d73 : (Trace Z)} {k138 : Hvl} ,
      (substhvl_Z k134 d73 k135 k138) -> (wfBB k138 (subst_Z_BB d73 zz110 bb74))).
  Proof.
    apply ((substhvl_Z_wfAA_ZZ_BB wf k135)).
    auto .
  Qed.
End SubstWellFormed.
 Hint Resolve substhvl_A_wfindex_A substhvl_A_wfindex_B substhvl_A_wfindex_Z substhvl_B_wfindex_A substhvl_B_wfindex_B substhvl_B_wfindex_Z substhvl_Z_wfindex_A substhvl_Z_wfindex_B substhvl_Z_wfindex_Z : infra.
 Hint Resolve substhvl_A_wfindex_A substhvl_A_wfindex_B substhvl_A_wfindex_Z substhvl_B_wfindex_A substhvl_B_wfindex_B substhvl_B_wfindex_Z substhvl_Z_wfindex_A substhvl_Z_wfindex_B substhvl_Z_wfindex_Z : subst.
 Hint Resolve substhvl_A_wfindex_A substhvl_A_wfindex_B substhvl_A_wfindex_Z substhvl_B_wfindex_A substhvl_B_wfindex_B substhvl_B_wfindex_Z substhvl_Z_wfindex_A substhvl_Z_wfindex_B substhvl_Z_wfindex_Z : subst_wf.
 Hint Resolve substhvl_A_wfindex_A substhvl_A_wfindex_B substhvl_A_wfindex_Z substhvl_B_wfindex_A substhvl_B_wfindex_B substhvl_B_wfindex_Z substhvl_Z_wfindex_A substhvl_Z_wfindex_B substhvl_Z_wfindex_Z : wf.
 Hint Resolve substhvl_A_wfAA substhvl_A_wfBB substhvl_A_wfZZ substhvl_B_wfAA substhvl_B_wfBB substhvl_B_wfZZ substhvl_Z_wfAA substhvl_Z_wfBB substhvl_Z_wfZZ : infra.
 Hint Resolve substhvl_A_wfAA substhvl_A_wfBB substhvl_A_wfZZ substhvl_B_wfAA substhvl_B_wfBB substhvl_B_wfZZ substhvl_Z_wfAA substhvl_Z_wfBB substhvl_Z_wfZZ : subst.
 Hint Resolve substhvl_A_wfAA substhvl_A_wfBB substhvl_A_wfZZ substhvl_B_wfAA substhvl_B_wfBB substhvl_B_wfZZ substhvl_Z_wfAA substhvl_Z_wfBB substhvl_Z_wfZZ : subst_wf.
 Hint Resolve substhvl_A_wfAA substhvl_A_wfBB substhvl_A_wfZZ substhvl_B_wfAA substhvl_B_wfBB substhvl_B_wfZZ substhvl_Z_wfAA substhvl_Z_wfBB substhvl_Z_wfZZ : wf.
 Hint Constructors substhvl_A substhvl_B substhvl_Z : infra.
 Hint Constructors substhvl_A substhvl_B substhvl_Z : subst.
 Hint Constructors substhvl_A substhvl_B substhvl_Z : subst_wf.
 Hint Constructors substhvl_A substhvl_B substhvl_Z : wf.
Section Context.
  Inductive Env : Type :=
    | empty 
    | ea (G : Env)
    | eb (G : Env).
  Fixpoint appendEnv (G : Env) (G0 : Env) : Env :=
    match G0 with
      | (empty) => G
      | (ea G1) => (ea (appendEnv G G1))
      | (eb G1) => (eb (appendEnv G G1))
    end.
  Fixpoint domainEnv (G : Env) : Hvl :=
    match G with
      | (empty) => H0
      | (ea G0) => (HS A (domainEnv G0))
      | (eb G0) => (HS B (domainEnv G0))
    end.
  Lemma appendEnv_assoc  :
    (forall (G : Env) (G0 : Env) (G1 : Env) ,
      ((appendEnv G (appendEnv G0 G1)) =
      (appendEnv (appendEnv G G0) G1))).
  Proof.
    needleGenericAppendEnvAssoc .
  Qed.
  Lemma domainEnv_appendEnv  :
    (forall (G : Env) (G0 : Env) ,
      ((domainEnv (appendEnv G G0)) =
      (appendHvl (domainEnv G) (domainEnv G0)))).
  Proof.
    needleGenericDomainEnvAppendEnv .
  Qed.
  Fixpoint shift_A_Env (c62 : (Cutoff A)) (G : Env) : Env :=
    match G with
      | (empty) => empty
      | (ea G0) => (ea (shift_A_Env c62 G0))
      | (eb G0) => (eb (shift_A_Env c62 G0))
    end.
  Fixpoint shift_B_Env (c62 : (Cutoff B)) (G : Env) : Env :=
    match G with
      | (empty) => empty
      | (ea G0) => (ea (shift_B_Env c62 G0))
      | (eb G0) => (eb (shift_B_Env c62 G0))
    end.
  Fixpoint subst_A_Env (d71 : (Trace A)) (aa74 : AA) (G : Env) : Env :=
    match G with
      | (empty) => empty
      | (ea G0) => (ea (subst_A_Env d71 aa74 G0))
      | (eb G0) => (eb (subst_A_Env d71 aa74 G0))
    end.
  Fixpoint subst_B_Env (d71 : (Trace B)) (bb74 : BB) (G : Env) : Env :=
    match G with
      | (empty) => empty
      | (ea G0) => (ea (subst_B_Env d71 bb74 G0))
      | (eb G0) => (eb (subst_B_Env d71 bb74 G0))
    end.
  Lemma domainEnv_shift_A_Env  :
    (forall (c62 : (Cutoff A)) (G : Env) ,
      ((domainEnv (shift_A_Env c62 G)) =
      (domainEnv G))).
  Proof.
    needleGenericDomainEnvShiftEnv .
  Qed.
  Lemma domainEnv_shift_B_Env  :
    (forall (c62 : (Cutoff B)) (G : Env) ,
      ((domainEnv (shift_B_Env c62 G)) =
      (domainEnv G))).
  Proof.
    needleGenericDomainEnvShiftEnv .
  Qed.
  Lemma domainEnv_subst_A_Env  :
    (forall (d71 : (Trace A)) (aa74 : AA) (G : Env) ,
      ((domainEnv (subst_A_Env d71 aa74 G)) =
      (domainEnv G))).
  Proof.
    needleGenericDomainEnvSubstEnv .
  Qed.
  Lemma domainEnv_subst_B_Env  :
    (forall (d71 : (Trace B)) (bb74 : BB) (G : Env) ,
      ((domainEnv (subst_B_Env d71 bb74 G)) =
      (domainEnv G))).
  Proof.
    needleGenericDomainEnvSubstEnv .
  Qed.
End Context.
 Hint Rewrite domainEnv_appendEnv : interaction.
 Hint Rewrite domainEnv_appendEnv : env_domain_append.
Section ContextStuff.
  Section ShiftEnvAppendEnv.
    Lemma shift_A_Env_appendEnv  :
      (forall (c62 : (Cutoff A)) (G : Env) (G0 : Env) ,
        ((shift_A_Env c62 (appendEnv G G0)) =
        (appendEnv (shift_A_Env c62 G) (shift_A_Env (weakenCutoffA c62 (domainEnv G)) G0)))).
    Proof.
      needleGenericShiftEnvAppendEnv .
    Qed.
    Lemma shift_B_Env_appendEnv  :
      (forall (c62 : (Cutoff B)) (G : Env) (G0 : Env) ,
        ((shift_B_Env c62 (appendEnv G G0)) =
        (appendEnv (shift_B_Env c62 G) (shift_B_Env (weakenCutoffB c62 (domainEnv G)) G0)))).
    Proof.
      needleGenericShiftEnvAppendEnv .
    Qed.
  End ShiftEnvAppendEnv.
  Section SubstEnvAppendEnv.
    Lemma subst_A_Env_appendEnv  :
      (forall (d71 : (Trace A)) (aa74 : AA) (G : Env) (G0 : Env) ,
        ((subst_A_Env d71 aa74 (appendEnv G G0)) =
        (appendEnv (subst_A_Env d71 aa74 G) (subst_A_Env (weakenTrace d71 (domainEnv G)) aa74 G0)))).
    Proof.
      needleGenericSubstEnvAppendEnv .
    Qed.
    Lemma subst_B_Env_appendEnv  :
      (forall (d71 : (Trace B)) (bb74 : BB) (G : Env) (G0 : Env) ,
        ((subst_B_Env d71 bb74 (appendEnv G G0)) =
        (appendEnv (subst_B_Env d71 bb74 G) (subst_B_Env (weakenTrace d71 (domainEnv G)) bb74 G0)))).
    Proof.
      needleGenericSubstEnvAppendEnv .
    Qed.
  End SubstEnvAppendEnv.
  Section Lookups.
    Inductive lookup_ea : Env -> (Index A) -> Prop :=
      | lookup_ea_here {G : Env} :
          (lookup_ea (ea G) I0)
      | lookup_ea_there_ea {G : Env}
          {a107 : (Index A)} :
          (lookup_ea G a107) -> (lookup_ea (ea G) (IS a107))
      | lookup_ea_there_eb {G : Env}
          {a107 : (Index A)} :
          (lookup_ea G a107) -> (lookup_ea (eb G) a107).
    Inductive lookup_eb : Env -> (Index B) -> Prop :=
      | lookup_eb_here {G : Env} :
          (lookup_eb (eb G) I0)
      | lookup_eb_there_ea {G : Env}
          {b107 : (Index B)} :
          (lookup_eb G b107) -> (lookup_eb (ea G) b107)
      | lookup_eb_there_eb {G : Env}
          {b107 : (Index B)} :
          (lookup_eb G b107) -> (lookup_eb (eb G) (IS b107)).
    Lemma weaken_lookup_ea  :
      (forall {G : Env} (G0 : Env) {a107 : (Index A)} ,
        (lookup_ea G a107) -> (lookup_ea (appendEnv G G0) (weakenIndexA a107 (domainEnv G0)))).
    Proof.
      needleGenericWeakenLookup .
    Qed.
    Lemma weaken_lookup_eb  :
      (forall {G : Env} (G0 : Env) {b107 : (Index B)} ,
        (lookup_eb G b107) -> (lookup_eb (appendEnv G G0) (weakenIndexB b107 (domainEnv G0)))).
    Proof.
      needleGenericWeakenLookup .
    Qed.
    Lemma lookup_ea_wfindex  :
      (forall {G : Env} {a107 : (Index A)} ,
        (lookup_ea G a107) -> (wfindex (domainEnv G) a107)).
    Proof.
      needleGenericLookupWellformedIndex .
    Qed.
    Lemma lookup_eb_wfindex  :
      (forall {G : Env} {b107 : (Index B)} ,
        (lookup_eb G b107) -> (wfindex (domainEnv G) b107)).
    Proof.
      needleGenericLookupWellformedIndex .
    Qed.
  End Lookups.
  Inductive shift_ea : (Cutoff A) -> Env -> Env -> Prop :=
    | shift_ea_here {G : Env} :
        (shift_ea C0 G (ea G))
    | shiftea_there_ea
        {c62 : (Cutoff A)} {G : Env}
        {G0 : Env} :
        (shift_ea c62 G G0) -> (shift_ea (CS c62) (ea G) (ea G0))
    | shiftea_there_eb
        {c62 : (Cutoff A)} {G : Env}
        {G0 : Env} :
        (shift_ea c62 G G0) -> (shift_ea c62 (eb G) (eb G0)).
  Lemma weaken_shift_ea  :
    (forall (G : Env) {c62 : (Cutoff A)} {G0 : Env} {G1 : Env} ,
      (shift_ea c62 G0 G1) -> (shift_ea (weakenCutoffA c62 (domainEnv G)) (appendEnv G0 G) (appendEnv G1 G))).
  Proof.
    needleGenericWeakenInsertEnv .
  Qed.
  Inductive shift_eb : (Cutoff B) -> Env -> Env -> Prop :=
    | shift_eb_here {G : Env} :
        (shift_eb C0 G (eb G))
    | shifteb_there_ea
        {c62 : (Cutoff B)} {G : Env}
        {G0 : Env} :
        (shift_eb c62 G G0) -> (shift_eb c62 (ea G) (ea G0))
    | shifteb_there_eb
        {c62 : (Cutoff B)} {G : Env}
        {G0 : Env} :
        (shift_eb c62 G G0) -> (shift_eb (CS c62) (eb G) (eb G0)).
  Lemma weaken_shift_eb  :
    (forall (G : Env) {c62 : (Cutoff B)} {G0 : Env} {G1 : Env} ,
      (shift_eb c62 G0 G1) -> (shift_eb (weakenCutoffB c62 (domainEnv G)) (appendEnv G0 G) (appendEnv G1 G))).
  Proof.
    needleGenericWeakenInsertEnv .
  Qed.
  Lemma shift_ea_shifthvl_A  :
    (forall {c62 : (Cutoff A)} {G : Env} {G0 : Env} ,
      (shift_ea c62 G G0) -> (shifthvl_A c62 (domainEnv G) (domainEnv G0))).
  Proof.
    needleGenericInsertEnvInsertHvl .
  Qed.
  Lemma shift_eb_shifthvl_B  :
    (forall {c62 : (Cutoff B)} {G : Env} {G0 : Env} ,
      (shift_eb c62 G G0) -> (shifthvl_B c62 (domainEnv G) (domainEnv G0))).
  Proof.
    needleGenericInsertEnvInsertHvl .
  Qed.
  Inductive subst_ea (G : Env) (aa74 : AA) : (Trace A) -> Env -> Env -> Prop :=
    | subst_ea_here :
        (subst_ea G aa74 X0 (ea G) G)
    | subst_ea_there_ea
        {d71 : (Trace A)} {G0 : Env}
        {G1 : Env} :
        (subst_ea G aa74 d71 G0 G1) -> (subst_ea G aa74 (XS A d71) (ea G0) (ea G1))
    | subst_ea_there_eb
        {d71 : (Trace A)} {G0 : Env}
        {G1 : Env} :
        (subst_ea G aa74 d71 G0 G1) -> (subst_ea G aa74 (XS B d71) (eb G0) (eb G1)).
  Lemma weaken_subst_ea {G : Env} {aa74 : AA} :
    (forall (G0 : Env) {d71 : (Trace A)} {G1 : Env} {G2 : Env} ,
      (subst_ea G aa74 d71 G1 G2) -> (subst_ea G aa74 (weakenTrace d71 (domainEnv G0)) (appendEnv G1 G0) (appendEnv G2 G0))).
  Proof.
    needleGenericWeakenSubstEnv .
  Qed.
  Inductive subst_eb (G : Env) (bb74 : BB) : (Trace B) -> Env -> Env -> Prop :=
    | subst_eb_here :
        (subst_eb G bb74 X0 (eb G) G)
    | subst_eb_there_ea
        {d71 : (Trace B)} {G0 : Env}
        {G1 : Env} :
        (subst_eb G bb74 d71 G0 G1) -> (subst_eb G bb74 (XS A d71) (ea G0) (ea G1))
    | subst_eb_there_eb
        {d71 : (Trace B)} {G0 : Env}
        {G1 : Env} :
        (subst_eb G bb74 d71 G0 G1) -> (subst_eb G bb74 (XS B d71) (eb G0) (eb G1)).
  Lemma weaken_subst_eb {G : Env} {bb74 : BB} :
    (forall (G0 : Env) {d71 : (Trace B)} {G1 : Env} {G2 : Env} ,
      (subst_eb G bb74 d71 G1 G2) -> (subst_eb G bb74 (weakenTrace d71 (domainEnv G0)) (appendEnv G1 G0) (appendEnv G2 G0))).
  Proof.
    needleGenericWeakenSubstEnv .
  Qed.
  Lemma subst_ea_substhvl_A {G : Env} {aa74 : AA} :
    (forall {d71 : (Trace A)} {G0 : Env} {G1 : Env} ,
      (subst_ea G aa74 d71 G0 G1) -> (substhvl_A (domainEnv G) d71 (domainEnv G0) (domainEnv G1))).
  Proof.
    needleGenericSubstEnvSubstHvl .
  Qed.
  Lemma subst_eb_substhvl_B {G : Env} {bb74 : BB} :
    (forall {d71 : (Trace B)} {G0 : Env} {G1 : Env} ,
      (subst_eb G bb74 d71 G0 G1) -> (substhvl_B (domainEnv G) d71 (domainEnv G0) (domainEnv G1))).
  Proof.
    needleGenericSubstEnvSubstHvl .
  Qed.
End ContextStuff.
 Hint Constructors lookup_ea lookup_eb : infra.
 Hint Constructors lookup_ea lookup_eb : lookup.
 Hint Constructors lookup_ea lookup_eb : shift.
 Hint Constructors lookup_ea lookup_eb : subst.
 Hint Resolve weaken_lookup_ea weaken_lookup_eb : infra.
 Hint Resolve weaken_lookup_ea weaken_lookup_eb : lookup.
 Hint Resolve weaken_lookup_ea weaken_lookup_eb : shift.
 Hint Resolve weaken_lookup_ea weaken_lookup_eb : weaken.
Lemma weaken_lookup_ea_here  :
  (forall {G : Env} (G0 : Env) ,
    (lookup_ea (appendEnv (ea G) G0) (weakenIndexA I0 (domainEnv G0)))).
Proof.
  eauto with lookup .
Qed.
Lemma weaken_lookup_eb_here  :
  (forall {G : Env} (G0 : Env) ,
    (lookup_eb (appendEnv (eb G) G0) (weakenIndexB I0 (domainEnv G0)))).
Proof.
  eauto with lookup .
Qed.
 Hint Constructors wfAA wfBB wfZZ : infra.
 Hint Constructors wfAA wfBB wfZZ : wf.
 Hint Extern 10 ((wfAA _ _)) => autorewrite with env_domain_append in *  : infra wf.
 Hint Extern 10 ((wfBB _ _)) => autorewrite with env_domain_append in *  : infra wf.
 Hint Extern 10 ((wfZZ _ _)) => autorewrite with env_domain_append in *  : infra wf.
 Hint Extern 2 ((wfAA _ _)) => match goal with
  | H : (wfAA _ (avar _)) |- _ => inversion H; subst; clear H
  | H : (wfAA _ (arec _)) |- _ => inversion H; subst; clear H
end : infra wf.
 Hint Extern 2 ((wfZZ _ _)) => match goal with
  | H : (wfZZ _ (zvar _)) |- _ => inversion H; subst; clear H
  | H : (wfZZ _ (aabs _)) |- _ => inversion H; subst; clear H
  | H : (wfZZ _ (babs _)) |- _ => inversion H; subst; clear H
end : infra wf.
 Hint Extern 2 ((wfBB _ _)) => match goal with
  | H : (wfBB _ (bvar _)) |- _ => inversion H; subst; clear H
  | H : (wfBB _ (brec _)) |- _ => inversion H; subst; clear H
end : infra wf.
 Hint Resolve lookup_ea_wfindex lookup_eb_wfindex : infra.
 Hint Resolve lookup_ea_wfindex lookup_eb_wfindex : lookup.
 Hint Resolve lookup_ea_wfindex lookup_eb_wfindex : wf.
 Hint Constructors shift_ea shift_eb : infra.
 Hint Constructors shift_ea shift_eb : shift.
 Hint Constructors shift_ea shift_eb : subst.
 Hint Resolve weaken_shift_ea weaken_shift_eb : infra.
 Hint Resolve weaken_shift_ea weaken_shift_eb : shift.
 Hint Resolve shift_ea_shifthvl_A shift_eb_shifthvl_B : infra.
 Hint Resolve shift_ea_shifthvl_A shift_eb_shifthvl_B : shift.
 Hint Resolve shift_ea_shifthvl_A shift_eb_shifthvl_B : shift_wf.
 Hint Resolve shift_ea_shifthvl_A shift_eb_shifthvl_B : wf.
 Hint Constructors subst_ea subst_eb : infra.
 Hint Constructors subst_ea subst_eb : subst.
 Hint Resolve weaken_subst_ea weaken_subst_eb : infra.
 Hint Resolve weaken_subst_ea weaken_subst_eb : subst.
 Hint Resolve subst_ea_substhvl_A subst_eb_substhvl_B : infra.
 Hint Resolve subst_ea_substhvl_A subst_eb_substhvl_B : subst.
 Hint Resolve subst_ea_substhvl_A subst_eb_substhvl_B : subst_wf.
 Hint Resolve subst_ea_substhvl_A subst_eb_substhvl_B : wf.
 Hint Resolve subst_ea_substhvl_A subst_eb_substhvl_B : substenv_substhvl.
Lemma shift_ea_lookup_ea  :
  (forall {c62 : (Cutoff A)} {G : Env} {G0 : Env} (ins : (shift_ea c62 G G0)) {a107 : (Index A)} ,
    (lookup_ea G a107) -> (lookup_ea G0 (shift_A_Index c62 a107))).
Proof.
  needleGenericInsertLookup .
Qed.
Lemma shift_ea_lookup_eb  :
  (forall {c62 : (Cutoff A)} {G : Env} {G0 : Env} (ins : (shift_ea c62 G G0)) {b107 : (Index B)} ,
    (lookup_eb G b107) -> (lookup_eb G0 b107)).
Proof.
  needleGenericInsertLookup .
Qed.
Lemma shift_eb_lookup_ea  :
  (forall {c62 : (Cutoff B)} {G : Env} {G0 : Env} (ins : (shift_eb c62 G G0)) {a107 : (Index A)} ,
    (lookup_ea G a107) -> (lookup_ea G0 a107)).
Proof.
  needleGenericInsertLookup .
Qed.
Lemma shift_eb_lookup_eb  :
  (forall {c62 : (Cutoff B)} {G : Env} {G0 : Env} (ins : (shift_eb c62 G G0)) {b107 : (Index B)} ,
    (lookup_eb G b107) -> (lookup_eb G0 (shift_B_Index c62 b107))).
Proof.
  needleGenericInsertLookup .
Qed.
 Hint Resolve shift_ea_lookup_ea shift_ea_lookup_eb shift_eb_lookup_ea shift_eb_lookup_eb : infra.
 Hint Resolve shift_ea_lookup_ea shift_ea_lookup_eb shift_eb_lookup_ea shift_eb_lookup_eb : lookup.
 Hint Resolve shift_ea_lookup_ea shift_ea_lookup_eb shift_eb_lookup_ea shift_eb_lookup_eb : shift.
Lemma subst_ea_lookup_eb {G : Env} {aa74 : AA} :
  (forall {d71 : (Trace A)} {G0 : Env} {G1 : Env} (sub : (subst_ea G aa74 d71 G0 G1)) {b107 : (Index B)} ,
    (lookup_eb G0 b107) -> (lookup_eb G1 b107)).
Proof.
  needleGenericSubstEnvLookup .
Qed.
Lemma subst_eb_lookup_ea {G : Env} {bb74 : BB} :
  (forall {d71 : (Trace B)} {G0 : Env} {G1 : Env} (sub : (subst_eb G bb74 d71 G0 G1)) {a107 : (Index A)} ,
    (lookup_ea G0 a107) -> (lookup_ea G1 a107)).
Proof.
  needleGenericSubstEnvLookup .
Qed.
 Hint Resolve subst_ea_lookup_eb subst_eb_lookup_ea : infra.
 Hint Resolve subst_ea_lookup_eb subst_eb_lookup_ea : lookup.
 Hint Resolve subst_ea_lookup_eb subst_eb_lookup_ea : subst.
Fixpoint size_AA (aa : AA) {struct aa} : nat :=
  match aa with
    | (avar a) => 1
    | (arec zz) => (plus 1 (size_ZZ zz))
  end
with size_ZZ (zz : ZZ) {struct zz} : nat :=
  match zz with
    | (zvar z) => 1
    | (aabs aa) => (plus 1 (size_AA aa))
    | (babs bb) => (plus 1 (size_BB bb))
  end
with size_BB (bb : BB) {struct bb} : nat :=
  match bb with
    | (bvar b) => 1
    | (brec zz) => (plus 1 (size_ZZ zz))
  end.
Lemma shift_A__size_AA_ZZ_BB  :
  (forall (aa71 : AA) (c53 : (Cutoff A)) ,
    ((size_AA (shift_A_AA c53 aa71)) =
    (size_AA aa71))) /\
  (forall (zz104 : ZZ) (c53 : (Cutoff A)) ,
    ((size_ZZ (shift_A_ZZ c53 zz104)) =
    (size_ZZ zz104))) /\
  (forall (bb71 : BB) (c53 : (Cutoff A)) ,
    ((size_BB (shift_A_BB c53 bb71)) =
    (size_BB bb71))).
Proof.
  apply_mutual_ind ind_AA_ZZ_BB.
  - intros a99 c53.
  reflexivity.
  - intros zz104 IHzz104.
  intros c54; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHzz104 c54)).
  - intros a100 c54.
  reflexivity.
  - intros aa71 IHaa71.
  intros c55; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHaa71 (CS c55))).
  - intros bb71 IHbb71.
  intros c55; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHbb71 c55)).
  - intros a102 c55.
  reflexivity.
  - intros zz105 IHzz105.
  intros c56; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHzz105 c56)).
Qed.

Lemma shift_A__size_AA  :
  (forall (aa72 : AA) (c56 : (Cutoff A)) ,
    ((size_AA (shift_A_AA c56 aa72)) =
    (size_AA aa72))).
Proof.
  pose proof (shift_A__size_AA_ZZ_BB).
  destruct_conjs .
  easy .
Qed.
Lemma shift_A__size_ZZ  :
  (forall (zz106 : ZZ) (c56 : (Cutoff A)) ,
    ((size_ZZ (shift_A_ZZ c56 zz106)) =
    (size_ZZ zz106))).
Proof.
  pose proof (shift_A__size_AA_ZZ_BB).
  destruct_conjs .
  easy .
Qed.
Lemma shift_A__size_BB  :
  (forall (bb72 : BB) (c56 : (Cutoff A)) ,
    ((size_BB (shift_A_BB c56 bb72)) =
    (size_BB bb72))).
Proof.
  pose proof (shift_A__size_AA_ZZ_BB).
  destruct_conjs .
  easy .
Qed.
Lemma shift_B__size_AA_ZZ_BB  :
  (forall (aa72 : AA) (c56 : (Cutoff B)) ,
    ((size_AA (shift_B_AA c56 aa72)) =
    (size_AA aa72))) /\
  (forall (zz106 : ZZ) (c56 : (Cutoff B)) ,
    ((size_ZZ (shift_B_ZZ c56 zz106)) =
    (size_ZZ zz106))) /\
  (forall (bb72 : BB) (c56 : (Cutoff B)) ,
    ((size_BB (shift_B_BB c56 bb72)) =
    (size_BB bb72))).
Proof.
  apply_mutual_ind ind_AA_ZZ_BB.
  - intros b100 c56.
  reflexivity.
  - intros zz106 IHzz106.
  intros c57; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHzz106 c57)).
  - intros b101 c57.
  reflexivity.
  - intros aa72 IHaa72.
  intros c58; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHaa72 c58)).
  - intros bb72 IHbb72.
  intros c58; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHbb72 (CS c58))).
  - intros b104 c58.
  reflexivity.
  - intros zz107 IHzz107.
  intros c59; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHzz107 c59)).
Qed.

Lemma shift_B__size_AA  :
  (forall (aa73 : AA) (c59 : (Cutoff B)) ,
    ((size_AA (shift_B_AA c59 aa73)) =
    (size_AA aa73))).
Proof.
  pose proof (shift_B__size_AA_ZZ_BB).
  destruct_conjs .
  easy .
Qed.
Lemma shift_B__size_ZZ  :
  (forall (zz108 : ZZ) (c59 : (Cutoff B)) ,
    ((size_ZZ (shift_B_ZZ c59 zz108)) =
    (size_ZZ zz108))).
Proof.
  pose proof (shift_B__size_AA_ZZ_BB).
  destruct_conjs .
  easy .
Qed.
Lemma shift_B__size_BB  :
  (forall (bb73 : BB) (c59 : (Cutoff B)) ,
    ((size_BB (shift_B_BB c59 bb73)) =
    (size_BB bb73))).
Proof.
  pose proof (shift_B__size_AA_ZZ_BB).
  destruct_conjs .
  easy .
Qed.
Lemma shift_Z__size_AA_ZZ_BB  :
  (forall (aa73 : AA) (c59 : (Cutoff Z)) ,
    ((size_AA (shift_Z_AA c59 aa73)) =
    (size_AA aa73))) /\
  (forall (zz108 : ZZ) (c59 : (Cutoff Z)) ,
    ((size_ZZ (shift_Z_ZZ c59 zz108)) =
    (size_ZZ zz108))) /\
  (forall (bb73 : BB) (c59 : (Cutoff Z)) ,
    ((size_BB (shift_Z_BB c59 bb73)) =
    (size_BB bb73))).
Proof.
  apply_mutual_ind ind_AA_ZZ_BB.
  - intros z67 c59.
  reflexivity.
  - intros zz108 IHzz108.
  intros c60; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHzz108 c60)).
  - intros z69 c60.
  reflexivity.
  - intros aa73 IHaa73.
  intros c61; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHaa73 c61)).
  - intros bb73 IHbb73.
  intros c61; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHbb73 c61)).
  - intros z70 c61.
  reflexivity.
  - intros zz109 IHzz109.
  intros c62; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHzz109 c62)).
Qed.

Lemma shift_Z__size_AA  :
  (forall (aa74 : AA) (c62 : (Cutoff Z)) ,
    ((size_AA (shift_Z_AA c62 aa74)) =
    (size_AA aa74))).
Proof.
  pose proof (shift_Z__size_AA_ZZ_BB).
  destruct_conjs .
  easy .
Qed.
Lemma shift_Z__size_ZZ  :
  (forall (zz110 : ZZ) (c62 : (Cutoff Z)) ,
    ((size_ZZ (shift_Z_ZZ c62 zz110)) =
    (size_ZZ zz110))).
Proof.
  pose proof (shift_Z__size_AA_ZZ_BB).
  destruct_conjs .
  easy .
Qed.
Lemma shift_Z__size_BB  :
  (forall (bb74 : BB) (c62 : (Cutoff Z)) ,
    ((size_BB (shift_Z_BB c62 bb74)) =
    (size_BB bb74))).
Proof.
  pose proof (shift_Z__size_AA_ZZ_BB).
  destruct_conjs .
  easy .
Qed.
 Hint Rewrite shift_A__size_AA shift_B__size_AA shift_Z__size_AA shift_A__size_BB shift_B__size_BB shift_Z__size_BB shift_A__size_ZZ shift_B__size_ZZ shift_Z__size_ZZ : interaction.
 Hint Rewrite shift_A__size_AA shift_B__size_AA shift_Z__size_AA shift_A__size_BB shift_B__size_BB shift_Z__size_BB shift_A__size_ZZ shift_B__size_ZZ shift_Z__size_ZZ : shift_size.
Lemma weaken_size_AA  :
  (forall (k134 : Hvl) (aa74 : AA) ,
    ((size_AA (weakenAA aa74 k134)) =
    (size_AA aa74))).
Proof.
  needleGenericWeakenSize .
Qed.
Lemma weaken_size_BB  :
  (forall (k134 : Hvl) (bb74 : BB) ,
    ((size_BB (weakenBB bb74 k134)) =
    (size_BB bb74))).
Proof.
  needleGenericWeakenSize .
Qed.
Lemma weaken_size_ZZ  :
  (forall (k134 : Hvl) (zz110 : ZZ) ,
    ((size_ZZ (weakenZZ zz110 k134)) =
    (size_ZZ zz110))).
Proof.
  needleGenericWeakenSize .
Qed.
 Hint Rewrite weaken_size_AA weaken_size_BB weaken_size_ZZ : interaction.
 Hint Rewrite weaken_size_AA weaken_size_BB weaken_size_ZZ : weaken_size.